/* Generated by Cython 3.2.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/_core/include/numpy/arrayobject.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/_core/include/numpy/arrayscalars.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/_core/include/numpy/ndarrayobject.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/_core/include/numpy/ndarraytypes.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/_core/include/numpy/ufuncobject.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/api.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/array/concatenate.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/builder.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/c/abi.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/c/bridge.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/c/dlpack.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/c/dlpack_abi.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/compute/api.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/compute/cast.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/compute/expression.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/config.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/csv/api.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/extension/bool8.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/extension/fixed_shape_tensor.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/extension/json.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/extension/opaque.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/extension/uuid.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/extension_type.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/io/api.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/ipc/api.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/json/options.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/json/reader.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/api.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/arrow_to_pandas.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/async.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/benchmark.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/common.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/config.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/csv.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/extension_type.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/gdb.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/helpers.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/inference.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/ipc.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/numpy_init.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/platform.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/pyarrow.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/python/udf.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/result.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/byte_size.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/cancel.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/compression.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/decimal.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/future.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/io_util.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/iterator.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/key_value_metadata.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/secure_string.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/thread_pool.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/arrow/util/value_parsing.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/parquet/api/reader.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/parquet/api/schema.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/parquet/api/writer.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/parquet/arrow/reader.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/parquet/arrow/schema.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/parquet/arrow/writer.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/parquet/encryption/encryption.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/parquet/geospatial/statistics.h",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include/parquet/properties.h"
        ],
        "extra_compile_args": [
            "-DWITH_IO_URING",
            "-std=c++20"
        ],
        "include_dirs": [
            ".",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow/include",
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/_core/include"
        ],
        "language": "c++",
        "libraries": [
            "arrow",
            "parquet",
            "uring"
        ],
        "library_dirs": [
            "/tmp/build-env-7lykdjnz/lib/python3.12/site-packages/pyarrow"
        ],
        "name": "jollyjack.jollyjack_cython",
        "sources": [
            "jollyjack/jollyjack_cython.pyx",
            "jollyjack/jollyjack.cc",
            "jollyjack/read_into_memory_io_uring.cc",
            "jollyjack/io_uring_reader_1.cc",
            "jollyjack/direct_reader.cc"
        ]
    },
    "module_name": "jollyjack.jollyjack_cython"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CppInitCode */
#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__jollyjack__jollyjack_cython
#define __PYX_HAVE_API__jollyjack__jollyjack_cython
/* Early includes */
#include <string.h>
#include <stdio.h>

    /* Using NumPy API declarations from "numpy/__init__.cython-30.pxd" */
    
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <vector>
#include <string_view>
#include <string>
#include <memory>
#include <stdint.h>
#include <functional>
#include <optional>
#include <utility>

    #if __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600)
    // move should be defined for these versions of MSVC, but __cplusplus isn't set usefully
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include <unordered_map>
#include <unordered_set>

    #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE PyObject *
    __Pyx_CAPI_PyList_GetItemRef(PyObject *list, Py_ssize_t index)
    {
        PyObject *item = PyList_GetItem(list, index);
        Py_XINCREF(item);
        return item;
    }
    #else
    #define __Pyx_CAPI_PyList_GetItemRef PyList_GetItemRef
    #endif

    #if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Extend(PyObject *list, PyObject *iterable)
    {
        return PyList_SetSlice(list, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, iterable);
    }

    static CYTHON_INLINE int
    __Pyx_CAPI_PyList_Clear(PyObject *list)
    {
        return PyList_SetSlice(list, 0, PY_SSIZE_T_MAX, NULL);
    }
    #else
    #define __Pyx_CAPI_PyList_Extend PyList_Extend
    #define __Pyx_CAPI_PyList_Clear PyList_Clear
    #endif
    

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    
#include <stddef.h>

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif

    #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int
    __Pyx_CAPI_PyDict_GetItemStringRef(PyObject *mp, const char *key, PyObject **result)
    {
        int res;
        PyObject *key_obj = PyUnicode_FromString(key);
        if (key_obj == NULL) {
            *result = NULL;
            return -1;
        }
        res = __Pyx_PyDict_GetItemRef(mp, key_obj, result);
        Py_DECREF(key_obj);
        return res;
    }
    #else
    #define __Pyx_CAPI_PyDict_GetItemStringRef PyDict_GetItemStringRef
    #endif
    #if PY_VERSION_HEX < 0x030d0000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030F0000)
    static CYTHON_INLINE int
    __Pyx_CAPI_PyDict_SetDefaultRef(PyObject *d, PyObject *key, PyObject *default_value,
                        PyObject **result)
    {
        PyObject *value;
        if (__Pyx_PyDict_GetItemRef(d, key, &value) < 0) {
            // get error
            if (result) {
                *result = NULL;
            }
            return -1;
        }
        if (value != NULL) {
            // present
            if (result) {
                *result = value;
            }
            else {
                Py_DECREF(value);
            }
            return 1;
        }

        // missing: set the item
        if (PyDict_SetItem(d, key, default_value) < 0) {
            // set error
            if (result) {
                *result = NULL;
            }
            return -1;
        }
        if (result) {
            Py_INCREF(default_value);
            *result = default_value;
        }
        return 0;
    }
    #else
    #define __Pyx_CAPI_PyDict_SetDefaultRef PyDict_SetDefaultRef
    #endif
    

    #if PY_VERSION_HEX < 0x030d0000
    static CYTHON_INLINE int __Pyx_PyWeakref_GetRef(PyObject *ref, PyObject **pobj)
    {
        PyObject *obj = PyWeakref_GetObject(ref);
        if (obj == NULL) {
            // SystemError if ref is NULL
            *pobj = NULL;
            return -1;
        }
        if (obj == Py_None) {
            *pobj = NULL;
            return 0;
        }
        Py_INCREF(obj);
        *pobj = obj;
        return 1;
    }
    #else
    #define __Pyx_PyWeakref_GetRef PyWeakref_GetRef
    #endif
    
#include "pythread.h"

    #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    

    #if CYTHON_COMPILING_IN_LIMITED_API
    #ifdef _MSC_VER
    #pragma message ("This module uses CPython specific internals of 'datetime.datetime', which are not available in the limited API.")
    #else
    #warning This module uses CPython specific internals of 'datetime.datetime', which are not available in the limited API.
    #endif
    #endif
    
#include "datetime.h"

    #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)         PyDateTimeAPI->DateTime_FromDateAndTimeAndFold(year, month, day, hour, minute, second,             microsecond, tz, fold, PyDateTimeAPI->DateTimeType)
    #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)         PyDateTimeAPI->Time_FromTimeAndFold(hour, minute, second, microsecond, tz, fold, PyDateTimeAPI->TimeType)

    #define __Pyx_TimeZone_UTC PyDateTime_TimeZone_UTC
    #define __Pyx_TimeZone_FromOffsetAndName(offset, name) PyTimeZone_FromOffsetAndName(offset, name)

    /* Backport for Python < 3.10 */
    #if PY_VERSION_HEX < 0x030a00a1
        #ifndef PyDateTime_TIME_GET_TZINFO
            #define PyDateTime_TIME_GET_TZINFO(o)                 ((((PyDateTime_Time*)o)->hastzinfo) ? ((PyDateTime_Time*)o)->tzinfo : Py_None)
        #endif
        #ifndef PyDateTime_DATE_GET_TZINFO
            #define PyDateTime_DATE_GET_TZINFO(o)                 ((((PyDateTime_DateTime*)o)->hastzinfo) ? ((PyDateTime_DateTime*)o)->tzinfo : Py_None)
        #endif
    #endif
    

    #include <memory>
    #include <string>
    #include <string_view>
    #include <utility>

    namespace arrow {
    namespace py {

    template <typename T>
    std::shared_ptr<T> to_shared(std::unique_ptr<T>& t) {
        return std::move(t);
    }
    template <typename T>
    std::shared_ptr<T> to_shared(std::unique_ptr<T>&& t) {
        return std::move(t);
    }

    // Needed until https://github.com/cython/cython/issues/6651 is fixed
    inline std::string to_string(std::string_view s) {
        return std::string(s);
    }

    }  // namespace py
    }  // namespace arrow
    
#include "arrow/python/platform.h"
#include <Python.h>
#include "arrow/api.h"
#include "arrow/result.h"
#include "arrow/util/future.h"
#include "arrow/python/async.h"
#include "arrow/python/common.h"
#include "arrow/util/key_value_metadata.h"
#include "arrow/util/decimal.h"
#include "arrow/config.h"
#include <variant>
#include "arrow/c/dlpack_abi.h"
#include "arrow/c/dlpack.h"
#include "arrow/builder.h"
#include "arrow/util/cancel.h"
#include "arrow/io/api.h"
#include "arrow/ipc/api.h"
#include "arrow/util/value_parsing.h"
#include "arrow/csv/api.h"
#include "arrow/json/options.h"
#include "arrow/json/reader.h"
#include "arrow/util/thread_pool.h"
#include "arrow/compute/api.h"

    #include "arrow/compute/function.h"
    namespace arrow {
    namespace compute {
    namespace internal {
    Result<std::unique_ptr<FunctionOptions>> DeserializeFunctionOptions(
        const Buffer& buffer);
    } //  namespace internal
    } //  namespace compute
    } //  namespace arrow
    

    #include <unordered_map>

    #include "arrow/type.h"
    #include "arrow/datum.h"

    namespace arrow {
    namespace compute {
    struct KnownFieldValues {
      std::unordered_map<FieldRef, Datum, FieldRef::Hash> map;
    };
    } //  namespace compute
    } //  namespace arrow
    
#include "arrow/compute/expression.h"
#include "arrow/extension_type.h"
#include "arrow/extension/json.h"
#include "arrow/extension/uuid.h"
#include "arrow/extension/fixed_shape_tensor.h"
#include "arrow/extension/opaque.h"
#include "arrow/extension/bool8.h"
#include "arrow/util/compression.h"
#include "arrow/util/io_util.h"
#include "arrow/util/iterator.h"
#include "arrow/util/secure_string.h"
#include "arrow/array/concatenate.h"
#include "arrow/c/abi.h"
#include "arrow/c/bridge.h"
#include "arrow/util/byte_size.h"
#include "arrow/python/udf.h"
#include "arrow/compute/cast.h"
#include "parquet/api/schema.h"
#include <array>
#include "parquet/geospatial/statistics.h"
#include "parquet/api/reader.h"
#include "parquet/api/writer.h"
#include "parquet/arrow/reader.h"
#include "parquet/arrow/schema.h"
#include "parquet/properties.h"
#include "parquet/arrow/writer.h"
#include "parquet/encryption/encryption.h"
#include "arrow/python/csv.h"
#include "arrow/python/api.h"
#include "arrow/python/arrow_to_pandas.h"
#include "arrow/python/numpy_init.h"
#include "arrow/python/pyarrow.h"
#include "arrow/python/inference.h"
#include "arrow/python/ipc.h"
#include "arrow/python/extension_type.h"
#include "arrow/python/benchmark.h"
#include "arrow/python/gdb.h"
#include "arrow/python/helpers.h"
#include "arrow/python/config.h"
#include "jollyjack.h"
#include "read_into_memory_io_uring.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif (defined(_Complex_I) && !defined(_MSC_VER)) || ((defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) && !defined(__STDC_NO_COMPLEX__) && !defined(_MSC_VER))
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "jollyjack/jollyjack_cython.pyx",
  "<stringsource>",
  "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd",
  "cpython/contextvars.pxd",
  "cpython/datetime.pxd",
  "pyarrow/_parquet.pxd",
  "cpython/type.pxd",
  "cpython/bool.pxd",
  "cpython/complex.pxd",
  "pyarrow/lib.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":744
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
*/

#line 744 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":745
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
*/

#line 745 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":746
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * 
*/

#line 746 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":747
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_uint8      uint8_t
*/

#line 747 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":749
 * ctypedef npy_int64      int64_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
*/

#line 749 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":750
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
*/

#line 750 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":751
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * 
*/

#line 751 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":752
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_float32    float32_t
*/

#line 752 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":754
 * ctypedef npy_uint64     uint64_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
*/

#line 754 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":755
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
*/

#line 755 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":762
 * ctypedef double complex complex128_t
 * 
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
*/

#line 762 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":763
 * 
 * ctypedef npy_longlong   longlong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
*/

#line 763 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":765
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
*/

#line 765 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":766
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
*/

#line 766 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":768
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
*/

#line 768 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":769
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
*/

#line 769 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":770
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef float complex       cfloat_t
*/

#line 770 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< long double > __pyx_t_long_double_complex;
  #else
    typedef long double _Complex __pyx_t_long_double_complex;
  #endif
#else
    typedef struct { long double real, imag; } __pyx_t_long_double_complex;
#endif
static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double, long double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7pyarrow_3lib__Weakrefable;
struct __pyx_obj_7pyarrow_3lib_IpcWriteOptions;
struct __pyx_obj_7pyarrow_3lib_IpcReadOptions;
struct __pyx_obj_7pyarrow_3lib_Message;
struct __pyx_obj_7pyarrow_3lib_MemoryPool;
struct __pyx_obj_7pyarrow_3lib_DataType;
struct __pyx_obj_7pyarrow_3lib_ListType;
struct __pyx_obj_7pyarrow_3lib_LargeListType;
struct __pyx_obj_7pyarrow_3lib_ListViewType;
struct __pyx_obj_7pyarrow_3lib_LargeListViewType;
struct __pyx_obj_7pyarrow_3lib_MapType;
struct __pyx_obj_7pyarrow_3lib_FixedSizeListType;
struct __pyx_obj_7pyarrow_3lib_StructType;
struct __pyx_obj_7pyarrow_3lib_DictionaryMemo;
struct __pyx_obj_7pyarrow_3lib_DictionaryType;
struct __pyx_obj_7pyarrow_3lib_TimestampType;
struct __pyx_obj_7pyarrow_3lib_Time32Type;
struct __pyx_obj_7pyarrow_3lib_Time64Type;
struct __pyx_obj_7pyarrow_3lib_DurationType;
struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType;
struct __pyx_obj_7pyarrow_3lib_Decimal32Type;
struct __pyx_obj_7pyarrow_3lib_Decimal64Type;
struct __pyx_obj_7pyarrow_3lib_Decimal128Type;
struct __pyx_obj_7pyarrow_3lib_Decimal256Type;
struct __pyx_obj_7pyarrow_3lib_RunEndEncodedType;
struct __pyx_obj_7pyarrow_3lib_BaseExtensionType;
struct __pyx_obj_7pyarrow_3lib_ExtensionType;
struct __pyx_obj_7pyarrow_3lib_FixedShapeTensorType;
struct __pyx_obj_7pyarrow_3lib_Bool8Type;
struct __pyx_obj_7pyarrow_3lib_OpaqueType;
struct __pyx_obj_7pyarrow_3lib_UuidType;
struct __pyx_obj_7pyarrow_3lib_JsonType;
struct __pyx_obj_7pyarrow_3lib__Metadata;
struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata;
struct __pyx_obj_7pyarrow_3lib_Field;
struct __pyx_obj_7pyarrow_3lib_Schema;
struct __pyx_obj_7pyarrow_3lib_Scalar;
struct __pyx_obj_7pyarrow_3lib_ArrayStatistics;
struct __pyx_obj_7pyarrow_3lib__PandasConvertible;
struct __pyx_obj_7pyarrow_3lib_Array;
struct __pyx_obj_7pyarrow_3lib_Tensor;
struct __pyx_obj_7pyarrow_3lib_SparseCSRMatrix;
struct __pyx_obj_7pyarrow_3lib_SparseCSCMatrix;
struct __pyx_obj_7pyarrow_3lib_SparseCOOTensor;
struct __pyx_obj_7pyarrow_3lib_SparseCSFTensor;
struct __pyx_obj_7pyarrow_3lib_NullArray;
struct __pyx_obj_7pyarrow_3lib_BooleanArray;
struct __pyx_obj_7pyarrow_3lib_NumericArray;
struct __pyx_obj_7pyarrow_3lib_IntegerArray;
struct __pyx_obj_7pyarrow_3lib_FloatingPointArray;
struct __pyx_obj_7pyarrow_3lib_Int8Array;
struct __pyx_obj_7pyarrow_3lib_UInt8Array;
struct __pyx_obj_7pyarrow_3lib_Int16Array;
struct __pyx_obj_7pyarrow_3lib_UInt16Array;
struct __pyx_obj_7pyarrow_3lib_Int32Array;
struct __pyx_obj_7pyarrow_3lib_UInt32Array;
struct __pyx_obj_7pyarrow_3lib_Int64Array;
struct __pyx_obj_7pyarrow_3lib_UInt64Array;
struct __pyx_obj_7pyarrow_3lib_HalfFloatArray;
struct __pyx_obj_7pyarrow_3lib_FloatArray;
struct __pyx_obj_7pyarrow_3lib_DoubleArray;
struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray;
struct __pyx_obj_7pyarrow_3lib_Decimal32Array;
struct __pyx_obj_7pyarrow_3lib_Decimal64Array;
struct __pyx_obj_7pyarrow_3lib_Decimal128Array;
struct __pyx_obj_7pyarrow_3lib_Decimal256Array;
struct __pyx_obj_7pyarrow_3lib_StructArray;
struct __pyx_obj_7pyarrow_3lib_BaseListArray;
struct __pyx_obj_7pyarrow_3lib_ListArray;
struct __pyx_obj_7pyarrow_3lib_LargeListArray;
struct __pyx_obj_7pyarrow_3lib_ListViewArray;
struct __pyx_obj_7pyarrow_3lib_LargeListViewArray;
struct __pyx_obj_7pyarrow_3lib_MapArray;
struct __pyx_obj_7pyarrow_3lib_FixedSizeListArray;
struct __pyx_obj_7pyarrow_3lib_UnionArray;
struct __pyx_obj_7pyarrow_3lib_StringArray;
struct __pyx_obj_7pyarrow_3lib_BinaryArray;
struct __pyx_obj_7pyarrow_3lib_StringViewArray;
struct __pyx_obj_7pyarrow_3lib_BinaryViewArray;
struct __pyx_obj_7pyarrow_3lib_DictionaryArray;
struct __pyx_obj_7pyarrow_3lib_ExtensionArray;
struct __pyx_obj_7pyarrow_3lib_MonthDayNanoIntervalArray;
struct __pyx_obj_7pyarrow_3lib_ChunkedArray;
struct __pyx_obj_7pyarrow_3lib__Tabular;
struct __pyx_obj_7pyarrow_3lib_Table;
struct __pyx_obj_7pyarrow_3lib_RecordBatch;
struct __pyx_obj_7pyarrow_3lib_Device;
struct __pyx_obj_7pyarrow_3lib_MemoryManager;
struct __pyx_obj_7pyarrow_3lib_Buffer;
struct __pyx_obj_7pyarrow_3lib_ResizableBuffer;
struct __pyx_obj_7pyarrow_3lib_NativeFile;
struct __pyx_obj_7pyarrow_3lib_BufferedInputStream;
struct __pyx_obj_7pyarrow_3lib_BufferedOutputStream;
struct __pyx_obj_7pyarrow_3lib_CompressedInputStream;
struct __pyx_obj_7pyarrow_3lib_CompressedOutputStream;
struct __pyx_obj_7pyarrow_3lib__CRecordBatchWriter;
struct __pyx_obj_7pyarrow_3lib_RecordBatchReader;
struct __pyx_obj_7pyarrow_3lib_CacheOptions;
struct __pyx_obj_7pyarrow_3lib_Codec;
struct __pyx_obj_7pyarrow_3lib_StopToken;
struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties;
struct __pyx_obj_7pyarrow_8_parquet_ParquetSchema;
struct __pyx_obj_7pyarrow_8_parquet_FileMetaData;
struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData;
struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData;
struct __pyx_obj_7pyarrow_8_parquet_Statistics;
struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics;
struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties;
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":116
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
*/

#line 116 "cpython/contextvars.pxd"
struct __pyx_opt_args_7cpython_11contextvars_get_value {

#line 116 "cpython/contextvars.pxd"
  int __pyx_n;

#line 116 "cpython/contextvars.pxd"
  PyObject *default_value;

#line 116 "cpython/contextvars.pxd"
};

/* "cpython/contextvars.pxd":134
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
*/

#line 134 "cpython/contextvars.pxd"
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {

#line 134 "cpython/contextvars.pxd"
  int __pyx_n;

#line 134 "cpython/contextvars.pxd"
  PyObject *default_value;

#line 134 "cpython/contextvars.pxd"
};

#line 134 "cpython/contextvars.pxd"
struct __pyx_opt_args_7cpython_8datetime_time_new;

#line 134 "cpython/contextvars.pxd"
struct __pyx_opt_args_7cpython_8datetime_datetime_new;

#line 134 "cpython/contextvars.pxd"
struct __pyx_opt_args_7cpython_8datetime_timezone_new;

#line 134 "cpython/contextvars.pxd"
struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp;

/* "cpython/datetime.pxd":265
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
*/

#line 265 "cpython/datetime.pxd"
struct __pyx_opt_args_7cpython_8datetime_time_new {

#line 265 "cpython/datetime.pxd"
  int __pyx_n;

#line 265 "cpython/datetime.pxd"
  int fold;

#line 265 "cpython/datetime.pxd"
};

/* "cpython/datetime.pxd":270
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
*/

#line 270 "cpython/datetime.pxd"
struct __pyx_opt_args_7cpython_8datetime_datetime_new {

#line 270 "cpython/datetime.pxd"
  int __pyx_n;

#line 270 "cpython/datetime.pxd"
  int fold;

#line 270 "cpython/datetime.pxd"
};

/* "cpython/datetime.pxd":279
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 * 
*/

#line 279 "cpython/datetime.pxd"
struct __pyx_opt_args_7cpython_8datetime_timezone_new {

#line 279 "cpython/datetime.pxd"
  int __pyx_n;

#line 279 "cpython/datetime.pxd"
  PyObject *name;

#line 279 "cpython/datetime.pxd"
};

/* "cpython/datetime.pxd":283
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
*/

#line 283 "cpython/datetime.pxd"
struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp {

#line 283 "cpython/datetime.pxd"
  int __pyx_n;

#line 283 "cpython/datetime.pxd"
  PyObject *tz;

#line 283 "cpython/datetime.pxd"
};

/* "pyarrow/includes/libarrow.pxd":1560
 * 
 * # Use typedef to emulate syntax for std::function<void(..)>
 * ctypedef void CallbackTransform(object, const shared_ptr[CBuffer]& src,             # <<<<<<<<<<<<<<
 *                                 shared_ptr[CBuffer]* dest)
 * 
*/

#line 1560 "pyarrow/includes/libarrow.pxd"
typedef void __pyx_t_7pyarrow_8includes_8libarrow_CallbackTransform(PyObject *, std::shared_ptr< arrow::Buffer>  const &, std::shared_ptr< arrow::Buffer>  *);

/* "pyarrow/includes/libarrow.pxd":1563
 *                                 shared_ptr[CBuffer]* dest)
 * 
 * ctypedef CResult[shared_ptr[CInputStream]] StreamWrapFunc(             # <<<<<<<<<<<<<<
 *     shared_ptr[CInputStream])
 * 
*/

#line 1563 "pyarrow/includes/libarrow.pxd"
typedef arrow::Result<std::shared_ptr< arrow::io::InputStream> >  __pyx_t_7pyarrow_8includes_8libarrow_StreamWrapFunc(std::shared_ptr< arrow::io::InputStream> );

/* "pyarrow/includes/libarrow.pxd":3239
 *     int64_t TotalBufferSize(const CTable& table)
 * 
 * ctypedef PyObject* CallbackUdf(object user_function, const CUdfContext& context, object inputs)             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 3239 "pyarrow/includes/libarrow.pxd"
typedef PyObject *__pyx_t_7pyarrow_8includes_8libarrow_CallbackUdf(PyObject *,  arrow::py::UdfContext const &, PyObject *);

/* "pyarrow/includes/libarrow_python.pxd":24
 * 
 * 
 * ctypedef CInvalidRowResult PyInvalidRowCallback(object,             # <<<<<<<<<<<<<<
 *                                                 const CCSVInvalidRow&)
 * 
*/

#line 24 "pyarrow/includes/libarrow_python.pxd"
typedef  arrow::csv::InvalidRowResult __pyx_t_7pyarrow_8includes_15libarrow_python_PyInvalidRowCallback(PyObject *,  arrow::csv::InvalidRow const &);

#line 24 "pyarrow/includes/libarrow_python.pxd"
struct __pyx_opt_args_7pyarrow_3lib_ensure_type;

/* "pyarrow/lib.pxd":676
 * 
 * # Default is allow_none=False
 * cpdef DataType ensure_type(object type, bint allow_none=*)             # <<<<<<<<<<<<<<
 * 
 * cdef DataType primitive_type(Type type)
*/

#line 676 "pyarrow/lib.pxd"
struct __pyx_opt_args_7pyarrow_3lib_ensure_type {

#line 676 "pyarrow/lib.pxd"
  int __pyx_n;

#line 676 "pyarrow/lib.pxd"
  int allow_none;

#line 676 "pyarrow/lib.pxd"
};

#line 676 "pyarrow/lib.pxd"
struct __pyx_opt_args_7pyarrow_8_parquet__create_writer_properties;

#line 676 "pyarrow/lib.pxd"
struct __pyx_opt_args_7pyarrow_8_parquet__create_arrow_writer_properties;

/* "pyarrow/_parquet.pxd":41
 *         return self.properties
 * 
 * cdef shared_ptr[WriterProperties] _create_writer_properties(             # <<<<<<<<<<<<<<
 *     use_dictionary=*,
 *     compression=*,
*/

#line 41 "pyarrow/_parquet.pxd"
struct __pyx_opt_args_7pyarrow_8_parquet__create_writer_properties {

#line 41 "pyarrow/_parquet.pxd"
  int __pyx_n;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *use_dictionary;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *compression;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *version;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *write_statistics;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *data_page_size;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *max_rows_per_page;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *compression_level;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *use_byte_stream_split;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *column_encoding;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *data_page_version;

#line 41 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *encryption_properties;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *write_batch_size;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *dictionary_pagesize_limit;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *write_page_index;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *write_page_checksum;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *sorting_columns;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *store_decimal_as_integer;

#line 41 "pyarrow/_parquet.pxd"
  PyObject *use_content_defined_chunking;

#line 41 "pyarrow/_parquet.pxd"
};

/* "pyarrow/_parquet.pxd":63
 * 
 * 
 * cdef shared_ptr[ArrowWriterProperties] _create_arrow_writer_properties(             # <<<<<<<<<<<<<<
 *     use_deprecated_int96_timestamps=*,
 *     coerce_timestamps=*,
*/

#line 63 "pyarrow/_parquet.pxd"
struct __pyx_opt_args_7pyarrow_8_parquet__create_arrow_writer_properties {

#line 63 "pyarrow/_parquet.pxd"
  int __pyx_n;

#line 63 "pyarrow/_parquet.pxd"
  PyObject *use_deprecated_int96_timestamps;

#line 63 "pyarrow/_parquet.pxd"
  PyObject *coerce_timestamps;

#line 63 "pyarrow/_parquet.pxd"
  PyObject *allow_truncated_timestamps;

#line 63 "pyarrow/_parquet.pxd"
  PyObject *writer_engine_version;

#line 63 "pyarrow/_parquet.pxd"
  PyObject *use_compliant_nested_type;

#line 63 "pyarrow/_parquet.pxd"
  PyObject *store_schema;

#line 63 "pyarrow/_parquet.pxd"
  PyObject *write_time_adjusted_to_utc;

#line 63 "pyarrow/_parquet.pxd"
};

#line 63 "pyarrow/_parquet.pxd"
struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_torch;

#line 63 "pyarrow/_parquet.pxd"
struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_numpy;

/* "jollyjack/jollyjack_cython.pyx":22
 *     return isinstance(obj, Iterable) and isinstance(obj[0], Iterable) and not isinstance(obj[0], str)
 * 
 * cpdef void read_into_torch (object source, FileMetaData metadata, tensor, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):             # <<<<<<<<<<<<<<
 * 
 *     import torch
*/

#line 22 "jollyjack/jollyjack_cython.pyx"
struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_torch {

#line 22 "jollyjack/jollyjack_cython.pyx"
  int __pyx_n;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *row_ranges;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *column_indices;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *column_names;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *pre_buffer;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *use_threads;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *use_memory_map;

#line 22 "jollyjack/jollyjack_cython.pyx"
  struct __pyx_obj_7pyarrow_3lib_CacheOptions *cache_options;

#line 22 "jollyjack/jollyjack_cython.pyx"
};

/* "jollyjack/jollyjack_cython.pyx":41
 *     return
 * 
 * cpdef void read_into_numpy (object source, FileMetaData metadata, cnp.ndarray np_array, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[int] crow_group_indices = row_group_indices
*/

#line 41 "jollyjack/jollyjack_cython.pyx"
struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_numpy {

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_n;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *row_ranges;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *column_indices;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *column_names;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *pre_buffer;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *use_threads;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *use_memory_map;

#line 41 "jollyjack/jollyjack_cython.pyx"
  struct __pyx_obj_7pyarrow_3lib_CacheOptions *cache_options;

#line 41 "jollyjack/jollyjack_cython.pyx"
};

/* "pyarrow/lib.pxd":37
 * 
 * 
 * cdef class _Weakrefable:             # <<<<<<<<<<<<<<
 *     cdef object __weakref__
 * 
*/

#line 37 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib__Weakrefable {

#line 37 "pyarrow/lib.pxd"
  PyObject_HEAD

#line 37 "pyarrow/lib.pxd"
  PyObject *__weakref__;

#line 37 "pyarrow/lib.pxd"
};

#line 37 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":41
 * 
 * 
 * cdef class IpcWriteOptions(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         CIpcWriteOptions c_options
*/

#line 41 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_IpcWriteOptions {

#line 41 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 41 "pyarrow/lib.pxd"
   arrow::ipc::IpcWriteOptions c_options;

#line 41 "pyarrow/lib.pxd"
};

#line 41 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":49
 * 
 * 
 * cdef class IpcReadOptions(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         CIpcReadOptions c_options
*/

#line 49 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_IpcReadOptions {

#line 49 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 49 "pyarrow/lib.pxd"
   arrow::ipc::IpcReadOptions c_options;

#line 49 "pyarrow/lib.pxd"
};

#line 49 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":60
 * 
 * 
 * cdef class Message(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         unique_ptr[CMessage] message
*/

#line 60 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Message {

#line 60 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 60 "pyarrow/lib.pxd"
  std::unique_ptr< arrow::ipc::Message>  message;

#line 60 "pyarrow/lib.pxd"
};

#line 60 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":65
 * 
 * 
 * cdef class MemoryPool(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         CMemoryPool* pool
*/

#line 65 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_MemoryPool {

#line 65 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 65 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_MemoryPool *__pyx_vtab;

#line 65 "pyarrow/lib.pxd"
   arrow::MemoryPool *pool;

#line 65 "pyarrow/lib.pxd"
};

#line 65 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":78
 * 
 * 
 * cdef class DataType(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CDataType] sp_type
*/

#line 78 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_DataType {

#line 78 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 78 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType *__pyx_vtab;

#line 78 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::DataType>  sp_type;

#line 78 "pyarrow/lib.pxd"
   arrow::DataType *type;

#line 78 "pyarrow/lib.pxd"
  PyObject *pep3118_format;

#line 78 "pyarrow/lib.pxd"
};

#line 78 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":88
 * 
 * 
 * cdef class ListType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CListType* list_type
*/

#line 88 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_ListType {

#line 88 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 88 "pyarrow/lib.pxd"
   arrow::ListType const *list_type;

#line 88 "pyarrow/lib.pxd"
};

#line 88 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":93
 * 
 * 
 * cdef class LargeListType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CLargeListType* list_type
*/

#line 93 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_LargeListType {

#line 93 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 93 "pyarrow/lib.pxd"
   arrow::LargeListType const *list_type;

#line 93 "pyarrow/lib.pxd"
};

#line 93 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":98
 * 
 * 
 * cdef class ListViewType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CListViewType* list_view_type
*/

#line 98 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_ListViewType {

#line 98 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 98 "pyarrow/lib.pxd"
   arrow::ListViewType const *list_view_type;

#line 98 "pyarrow/lib.pxd"
};

#line 98 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":103
 * 
 * 
 * cdef class LargeListViewType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CLargeListViewType* list_view_type
*/

#line 103 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_LargeListViewType {

#line 103 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 103 "pyarrow/lib.pxd"
   arrow::LargeListViewType const *list_view_type;

#line 103 "pyarrow/lib.pxd"
};

#line 103 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":108
 * 
 * 
 * cdef class MapType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CMapType* map_type
*/

#line 108 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_MapType {

#line 108 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 108 "pyarrow/lib.pxd"
   arrow::MapType const *map_type;

#line 108 "pyarrow/lib.pxd"
};

#line 108 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":113
 * 
 * 
 * cdef class FixedSizeListType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CFixedSizeListType* list_type
*/

#line 113 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_FixedSizeListType {

#line 113 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 113 "pyarrow/lib.pxd"
   arrow::FixedSizeListType const *list_type;

#line 113 "pyarrow/lib.pxd"
};

#line 113 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":118
 * 
 * 
 * cdef class StructType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CStructType* struct_type
*/

#line 118 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_StructType {

#line 118 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 118 "pyarrow/lib.pxd"
   arrow::StructType const *struct_type;

#line 118 "pyarrow/lib.pxd"
};

#line 118 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":125
 * 
 * 
 * cdef class DictionaryMemo(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         # Even though the CDictionaryMemo instance is private, we allocate
*/

#line 125 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_DictionaryMemo {

#line 125 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 125 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::ipc::DictionaryMemo>  sp_memo;

#line 125 "pyarrow/lib.pxd"
   arrow::ipc::DictionaryMemo *memo;

#line 125 "pyarrow/lib.pxd"
};

#line 125 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":133
 * 
 * 
 * cdef class DictionaryType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDictionaryType* dict_type
*/

#line 133 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_DictionaryType {

#line 133 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 133 "pyarrow/lib.pxd"
   arrow::DictionaryType const *dict_type;

#line 133 "pyarrow/lib.pxd"
};

#line 133 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":138
 * 
 * 
 * cdef class TimestampType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CTimestampType* ts_type
*/

#line 138 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_TimestampType {

#line 138 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 138 "pyarrow/lib.pxd"
   arrow::TimestampType const *ts_type;

#line 138 "pyarrow/lib.pxd"
};

#line 138 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":143
 * 
 * 
 * cdef class Time32Type(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CTime32Type* time_type
*/

#line 143 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Time32Type {

#line 143 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 143 "pyarrow/lib.pxd"
   arrow::Time32Type const *time_type;

#line 143 "pyarrow/lib.pxd"
};

#line 143 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":148
 * 
 * 
 * cdef class Time64Type(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CTime64Type* time_type
*/

#line 148 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Time64Type {

#line 148 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 148 "pyarrow/lib.pxd"
   arrow::Time64Type const *time_type;

#line 148 "pyarrow/lib.pxd"
};

#line 148 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":153
 * 
 * 
 * cdef class DurationType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDurationType* duration_type
*/

#line 153 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_DurationType {

#line 153 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 153 "pyarrow/lib.pxd"
   arrow::DurationType const *duration_type;

#line 153 "pyarrow/lib.pxd"
};

#line 153 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":158
 * 
 * 
 * cdef class FixedSizeBinaryType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CFixedSizeBinaryType* fixed_size_binary_type
*/

#line 158 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType {

#line 158 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 158 "pyarrow/lib.pxd"
   arrow::FixedSizeBinaryType const *fixed_size_binary_type;

#line 158 "pyarrow/lib.pxd"
};

#line 158 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":163
 * 
 * 
 * cdef class Decimal32Type(FixedSizeBinaryType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDecimal32Type* decimal32_type
*/

#line 163 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Decimal32Type {

#line 163 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType __pyx_base;

#line 163 "pyarrow/lib.pxd"
   arrow::Decimal32Type const *decimal32_type;

#line 163 "pyarrow/lib.pxd"
};

#line 163 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":168
 * 
 * 
 * cdef class Decimal64Type(FixedSizeBinaryType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDecimal64Type* decimal64_type
*/

#line 168 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Decimal64Type {

#line 168 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType __pyx_base;

#line 168 "pyarrow/lib.pxd"
   arrow::Decimal64Type const *decimal64_type;

#line 168 "pyarrow/lib.pxd"
};

#line 168 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":173
 * 
 * 
 * cdef class Decimal128Type(FixedSizeBinaryType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDecimal128Type* decimal128_type
*/

#line 173 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Decimal128Type {

#line 173 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType __pyx_base;

#line 173 "pyarrow/lib.pxd"
   arrow::Decimal128Type const *decimal128_type;

#line 173 "pyarrow/lib.pxd"
};

#line 173 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":178
 * 
 * 
 * cdef class Decimal256Type(FixedSizeBinaryType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDecimal256Type* decimal256_type
*/

#line 178 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Decimal256Type {

#line 178 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType __pyx_base;

#line 178 "pyarrow/lib.pxd"
   arrow::Decimal256Type const *decimal256_type;

#line 178 "pyarrow/lib.pxd"
};

#line 178 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":183
 * 
 * 
 * cdef class RunEndEncodedType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CRunEndEncodedType* run_end_encoded_type
*/

#line 183 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_RunEndEncodedType {

#line 183 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 183 "pyarrow/lib.pxd"
   arrow::RunEndEncodedType const *run_end_encoded_type;

#line 183 "pyarrow/lib.pxd"
};

#line 183 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":188
 * 
 * 
 * cdef class BaseExtensionType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CExtensionType* ext_type
*/

#line 188 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_BaseExtensionType {

#line 188 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType __pyx_base;

#line 188 "pyarrow/lib.pxd"
   arrow::ExtensionType const *ext_type;

#line 188 "pyarrow/lib.pxd"
};

#line 188 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":193
 * 
 * 
 * cdef class ExtensionType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CPyExtensionType* cpy_ext_type
*/

#line 193 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_ExtensionType {

#line 193 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 193 "pyarrow/lib.pxd"
   arrow::py::PyExtensionType const *cpy_ext_type;

#line 193 "pyarrow/lib.pxd"
};

#line 193 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":198
 * 
 * 
 * cdef class FixedShapeTensorType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CFixedShapeTensorType* tensor_ext_type
*/

#line 198 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_FixedShapeTensorType {

#line 198 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 198 "pyarrow/lib.pxd"
   arrow::extension::FixedShapeTensorType const *tensor_ext_type;

#line 198 "pyarrow/lib.pxd"
};

#line 198 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":202
 *         const CFixedShapeTensorType* tensor_ext_type
 * 
 * cdef class Bool8Type(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CBool8Type* bool8_ext_type
*/

#line 202 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Bool8Type {

#line 202 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 202 "pyarrow/lib.pxd"
   arrow::extension::Bool8Type const *bool8_ext_type;

#line 202 "pyarrow/lib.pxd"
};

#line 202 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":206
 *         const CBool8Type* bool8_ext_type
 * 
 * cdef class OpaqueType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const COpaqueType* opaque_ext_type
*/

#line 206 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_OpaqueType {

#line 206 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 206 "pyarrow/lib.pxd"
   arrow::extension::OpaqueType const *opaque_ext_type;

#line 206 "pyarrow/lib.pxd"
};

#line 206 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":210
 *         const COpaqueType* opaque_ext_type
 * 
 * cdef class UuidType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CUuidType* uuid_ext_type
*/

#line 210 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_UuidType {

#line 210 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 210 "pyarrow/lib.pxd"
   arrow::extension::UuidType const *uuid_ext_type;

#line 210 "pyarrow/lib.pxd"
};

#line 210 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":214
 *         const CUuidType* uuid_ext_type
 * 
 * cdef class JsonType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CJsonType* json_ext_type
*/

#line 214 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_JsonType {

#line 214 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 214 "pyarrow/lib.pxd"
   arrow::extension::JsonExtensionType const *json_ext_type;

#line 214 "pyarrow/lib.pxd"
};

#line 214 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":219
 * 
 * 
 * cdef class _Metadata(_Weakrefable):             # <<<<<<<<<<<<<<
 *     # required because KeyValueMetadata also extends collections.abc.Mapping
 *     # and the first parent class must be an extension type
*/

#line 219 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib__Metadata {

#line 219 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 219 "pyarrow/lib.pxd"
};

#line 219 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":225
 * 
 * 
 * cdef class KeyValueMetadata(_Metadata):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[const CKeyValueMetadata] wrapped
*/

#line 225 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata {

#line 225 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Metadata __pyx_base;

#line 225 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_KeyValueMetadata *__pyx_vtab;

#line 225 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::KeyValueMetadata const >  wrapped;

#line 225 "pyarrow/lib.pxd"
   arrow::KeyValueMetadata const *metadata;

#line 225 "pyarrow/lib.pxd"
};

#line 225 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":237
 * 
 * 
 * cdef class Field(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CField] sp_field
*/

#line 237 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Field {

#line 237 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 237 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Field *__pyx_vtab;

#line 237 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Field>  sp_field;

#line 237 "pyarrow/lib.pxd"
   arrow::Field *field;

#line 237 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType *type;

#line 237 "pyarrow/lib.pxd"
};

#line 237 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":248
 * 
 * 
 * cdef class Schema(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSchema] sp_schema
*/

#line 248 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Schema {

#line 248 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 248 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Schema *__pyx_vtab;

#line 248 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Schema>  sp_schema;

#line 248 "pyarrow/lib.pxd"
   arrow::Schema *schema;

#line 248 "pyarrow/lib.pxd"
};

#line 248 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":257
 * 
 * 
 * cdef class Scalar(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CScalar] wrapped
*/

#line 257 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Scalar {

#line 257 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 257 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Scalar *__pyx_vtab;

#line 257 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Scalar>  wrapped;

#line 257 "pyarrow/lib.pxd"
};

#line 257 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":269
 * 
 * 
 * cdef class ArrayStatistics(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CArrayStatistics] sp_statistics
*/

#line 269 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_ArrayStatistics {

#line 269 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 269 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_ArrayStatistics *__pyx_vtab;

#line 269 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::ArrayStatistics>  sp_statistics;

#line 269 "pyarrow/lib.pxd"
};

#line 269 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":277
 * 
 * 
 * cdef class _PandasConvertible(_Weakrefable):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 277 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib__PandasConvertible {

#line 277 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 277 "pyarrow/lib.pxd"
};

#line 277 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":281
 * 
 * 
 * cdef class Array(_PandasConvertible):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CArray] sp_array
*/

#line 281 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Array {

#line 281 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__PandasConvertible __pyx_base;

#line 281 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array *__pyx_vtab;

#line 281 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Array>  sp_array;

#line 281 "pyarrow/lib.pxd"
   arrow::Array *ap;

#line 281 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType *type;

#line 281 "pyarrow/lib.pxd"
  PyObject *_name;

#line 281 "pyarrow/lib.pxd"
};

#line 281 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":297
 * 
 * 
 * cdef class Tensor(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CTensor] sp_tensor
*/

#line 297 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Tensor {

#line 297 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 297 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Tensor *__pyx_vtab;

#line 297 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Tensor>  sp_tensor;

#line 297 "pyarrow/lib.pxd"
   arrow::Tensor *tp;

#line 297 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType *type;

#line 297 "pyarrow/lib.pxd"
  PyObject *_ssize_t_shape;

#line 297 "pyarrow/lib.pxd"
  PyObject *_ssize_t_strides;

#line 297 "pyarrow/lib.pxd"
};

#line 297 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":310
 * 
 * 
 * cdef class SparseCSRMatrix(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSparseCSRMatrix] sp_sparse_tensor
*/

#line 310 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_SparseCSRMatrix {

#line 310 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 310 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSRMatrix *__pyx_vtab;

#line 310 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::SparseCSRMatrix>  sp_sparse_tensor;

#line 310 "pyarrow/lib.pxd"
   arrow::SparseCSRMatrix *stp;

#line 310 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType *type;

#line 310 "pyarrow/lib.pxd"
};

#line 310 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":321
 * 
 * 
 * cdef class SparseCSCMatrix(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSparseCSCMatrix] sp_sparse_tensor
*/

#line 321 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_SparseCSCMatrix {

#line 321 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 321 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSCMatrix *__pyx_vtab;

#line 321 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::SparseCSCMatrix>  sp_sparse_tensor;

#line 321 "pyarrow/lib.pxd"
   arrow::SparseCSCMatrix *stp;

#line 321 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType *type;

#line 321 "pyarrow/lib.pxd"
};

#line 321 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":332
 * 
 * 
 * cdef class SparseCOOTensor(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSparseCOOTensor] sp_sparse_tensor
*/

#line 332 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_SparseCOOTensor {

#line 332 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 332 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_SparseCOOTensor *__pyx_vtab;

#line 332 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::SparseCOOTensor>  sp_sparse_tensor;

#line 332 "pyarrow/lib.pxd"
   arrow::SparseCOOTensor *stp;

#line 332 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType *type;

#line 332 "pyarrow/lib.pxd"
};

#line 332 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":343
 * 
 * 
 * cdef class SparseCSFTensor(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSparseCSFTensor] sp_sparse_tensor
*/

#line 343 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_SparseCSFTensor {

#line 343 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 343 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSFTensor *__pyx_vtab;

#line 343 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::SparseCSFTensor>  sp_sparse_tensor;

#line 343 "pyarrow/lib.pxd"
   arrow::SparseCSFTensor *stp;

#line 343 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_DataType *type;

#line 343 "pyarrow/lib.pxd"
};

#line 343 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":354
 * 
 * 
 * cdef class NullArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 354 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_NullArray {

#line 354 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 354 "pyarrow/lib.pxd"
};

#line 354 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":358
 * 
 * 
 * cdef class BooleanArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 358 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_BooleanArray {

#line 358 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 358 "pyarrow/lib.pxd"
};

#line 358 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":362
 * 
 * 
 * cdef class NumericArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 362 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_NumericArray {

#line 362 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 362 "pyarrow/lib.pxd"
};

#line 362 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":366
 * 
 * 
 * cdef class IntegerArray(NumericArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 366 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_IntegerArray {

#line 366 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_NumericArray __pyx_base;

#line 366 "pyarrow/lib.pxd"
};

#line 366 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":370
 * 
 * 
 * cdef class FloatingPointArray(NumericArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 370 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_FloatingPointArray {

#line 370 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_NumericArray __pyx_base;

#line 370 "pyarrow/lib.pxd"
};

#line 370 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":374
 * 
 * 
 * cdef class Int8Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 374 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Int8Array {

#line 374 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_IntegerArray __pyx_base;

#line 374 "pyarrow/lib.pxd"
};

#line 374 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":378
 * 
 * 
 * cdef class UInt8Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 378 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_UInt8Array {

#line 378 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_IntegerArray __pyx_base;

#line 378 "pyarrow/lib.pxd"
};

#line 378 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":382
 * 
 * 
 * cdef class Int16Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 382 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Int16Array {

#line 382 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_IntegerArray __pyx_base;

#line 382 "pyarrow/lib.pxd"
};

#line 382 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":386
 * 
 * 
 * cdef class UInt16Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 386 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_UInt16Array {

#line 386 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_IntegerArray __pyx_base;

#line 386 "pyarrow/lib.pxd"
};

#line 386 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":390
 * 
 * 
 * cdef class Int32Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 390 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Int32Array {

#line 390 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_IntegerArray __pyx_base;

#line 390 "pyarrow/lib.pxd"
};

#line 390 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":394
 * 
 * 
 * cdef class UInt32Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 394 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_UInt32Array {

#line 394 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_IntegerArray __pyx_base;

#line 394 "pyarrow/lib.pxd"
};

#line 394 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":398
 * 
 * 
 * cdef class Int64Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 398 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Int64Array {

#line 398 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_IntegerArray __pyx_base;

#line 398 "pyarrow/lib.pxd"
};

#line 398 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":402
 * 
 * 
 * cdef class UInt64Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 402 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_UInt64Array {

#line 402 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_IntegerArray __pyx_base;

#line 402 "pyarrow/lib.pxd"
};

#line 402 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":406
 * 
 * 
 * cdef class HalfFloatArray(FloatingPointArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 406 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_HalfFloatArray {

#line 406 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FloatingPointArray __pyx_base;

#line 406 "pyarrow/lib.pxd"
};

#line 406 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":410
 * 
 * 
 * cdef class FloatArray(FloatingPointArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 410 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_FloatArray {

#line 410 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FloatingPointArray __pyx_base;

#line 410 "pyarrow/lib.pxd"
};

#line 410 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":414
 * 
 * 
 * cdef class DoubleArray(FloatingPointArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 414 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_DoubleArray {

#line 414 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FloatingPointArray __pyx_base;

#line 414 "pyarrow/lib.pxd"
};

#line 414 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":418
 * 
 * 
 * cdef class FixedSizeBinaryArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 418 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray {

#line 418 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 418 "pyarrow/lib.pxd"
};

#line 418 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":422
 * 
 * 
 * cdef class Decimal32Array(FixedSizeBinaryArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 422 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Decimal32Array {

#line 422 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray __pyx_base;

#line 422 "pyarrow/lib.pxd"
};

#line 422 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":426
 * 
 * 
 * cdef class Decimal64Array(FixedSizeBinaryArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 426 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Decimal64Array {

#line 426 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray __pyx_base;

#line 426 "pyarrow/lib.pxd"
};

#line 426 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":430
 * 
 * 
 * cdef class Decimal128Array(FixedSizeBinaryArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 430 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Decimal128Array {

#line 430 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray __pyx_base;

#line 430 "pyarrow/lib.pxd"
};

#line 430 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":434
 * 
 * 
 * cdef class Decimal256Array(FixedSizeBinaryArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 434 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Decimal256Array {

#line 434 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray __pyx_base;

#line 434 "pyarrow/lib.pxd"
};

#line 434 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":438
 * 
 * 
 * cdef class StructArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 438 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_StructArray {

#line 438 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 438 "pyarrow/lib.pxd"
};

#line 438 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":442
 * 
 * 
 * cdef class BaseListArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 442 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_BaseListArray {

#line 442 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 442 "pyarrow/lib.pxd"
};

#line 442 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":446
 * 
 * 
 * cdef class ListArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 446 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_ListArray {

#line 446 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseListArray __pyx_base;

#line 446 "pyarrow/lib.pxd"
};

#line 446 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":450
 * 
 * 
 * cdef class LargeListArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 450 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_LargeListArray {

#line 450 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseListArray __pyx_base;

#line 450 "pyarrow/lib.pxd"
};

#line 450 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":454
 * 
 * 
 * cdef class ListViewArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 454 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_ListViewArray {

#line 454 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseListArray __pyx_base;

#line 454 "pyarrow/lib.pxd"
};

#line 454 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":458
 * 
 * 
 * cdef class LargeListViewArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 458 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_LargeListViewArray {

#line 458 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseListArray __pyx_base;

#line 458 "pyarrow/lib.pxd"
};

#line 458 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":462
 * 
 * 
 * cdef class MapArray(ListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 462 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_MapArray {

#line 462 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_ListArray __pyx_base;

#line 462 "pyarrow/lib.pxd"
};

#line 462 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":466
 * 
 * 
 * cdef class FixedSizeListArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 466 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_FixedSizeListArray {

#line 466 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_BaseListArray __pyx_base;

#line 466 "pyarrow/lib.pxd"
};

#line 466 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":470
 * 
 * 
 * cdef class UnionArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 470 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_UnionArray {

#line 470 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 470 "pyarrow/lib.pxd"
};

#line 470 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":474
 * 
 * 
 * cdef class StringArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 474 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_StringArray {

#line 474 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 474 "pyarrow/lib.pxd"
};

#line 474 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":478
 * 
 * 
 * cdef class BinaryArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 478 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_BinaryArray {

#line 478 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 478 "pyarrow/lib.pxd"
};

#line 478 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":482
 * 
 * 
 * cdef class StringViewArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 482 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_StringViewArray {

#line 482 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 482 "pyarrow/lib.pxd"
};

#line 482 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":486
 * 
 * 
 * cdef class BinaryViewArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 486 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_BinaryViewArray {

#line 486 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 486 "pyarrow/lib.pxd"
};

#line 486 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":490
 * 
 * 
 * cdef class DictionaryArray(Array):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _indices, _dictionary
*/

#line 490 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_DictionaryArray {

#line 490 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 490 "pyarrow/lib.pxd"
  PyObject *_indices;

#line 490 "pyarrow/lib.pxd"
  PyObject *_dictionary;

#line 490 "pyarrow/lib.pxd"
};

#line 490 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":495
 * 
 * 
 * cdef class ExtensionArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 495 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_ExtensionArray {

#line 495 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 495 "pyarrow/lib.pxd"
};

#line 495 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":499
 * 
 * 
 * cdef class MonthDayNanoIntervalArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 499 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_MonthDayNanoIntervalArray {

#line 499 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Array __pyx_base;

#line 499 "pyarrow/lib.pxd"
};

#line 499 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":507
 * 
 * 
 * cdef class ChunkedArray(_PandasConvertible):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CChunkedArray] sp_chunked_array
*/

#line 507 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_ChunkedArray {

#line 507 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__PandasConvertible __pyx_base;

#line 507 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_ChunkedArray *__pyx_vtab;

#line 507 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::ChunkedArray>  sp_chunked_array;

#line 507 "pyarrow/lib.pxd"
   arrow::ChunkedArray *chunked_array;

#line 507 "pyarrow/lib.pxd"
  bool _is_cpu;

#line 507 "pyarrow/lib.pxd"
  bool _init_is_cpu;

#line 507 "pyarrow/lib.pxd"
  PyObject *_name;

#line 507 "pyarrow/lib.pxd"
};

#line 507 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":522
 * 
 * 
 * cdef class _Tabular(_PandasConvertible):             # <<<<<<<<<<<<<<
 *     cdef void _assert_cpu(self) except *
 * 
*/

#line 522 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib__Tabular {

#line 522 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__PandasConvertible __pyx_base;

#line 522 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib__Tabular *__pyx_vtab;

#line 522 "pyarrow/lib.pxd"
};

#line 522 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":526
 * 
 * 
 * cdef class Table(_Tabular):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CTable] sp_table
*/

#line 526 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Table {

#line 526 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Tabular __pyx_base;

#line 526 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Table>  sp_table;

#line 526 "pyarrow/lib.pxd"
   arrow::Table *table;

#line 526 "pyarrow/lib.pxd"
  bool _is_cpu;

#line 526 "pyarrow/lib.pxd"
  bool _init_is_cpu;

#line 526 "pyarrow/lib.pxd"
};

#line 526 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":536
 * 
 * 
 * cdef class RecordBatch(_Tabular):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CRecordBatch] sp_batch
*/

#line 536 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_RecordBatch {

#line 536 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Tabular __pyx_base;

#line 536 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::RecordBatch>  sp_batch;

#line 536 "pyarrow/lib.pxd"
   arrow::RecordBatch *batch;

#line 536 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Schema *_schema;

#line 536 "pyarrow/lib.pxd"
};

#line 536 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":545
 * 
 * 
 * cdef class Device(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CDevice] device
*/

#line 545 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Device {

#line 545 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 545 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Device *__pyx_vtab;

#line 545 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Device>  device;

#line 545 "pyarrow/lib.pxd"
};

#line 545 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":557
 * 
 * 
 * cdef class MemoryManager(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CMemoryManager] memory_manager
*/

#line 557 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_MemoryManager {

#line 557 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 557 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_MemoryManager *__pyx_vtab;

#line 557 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::MemoryManager>  memory_manager;

#line 557 "pyarrow/lib.pxd"
};

#line 557 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":569
 * 
 * 
 * cdef class Buffer(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CBuffer] buffer
*/

#line 569 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Buffer {

#line 569 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 569 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Buffer *__pyx_vtab;

#line 569 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Buffer>  buffer;

#line 569 "pyarrow/lib.pxd"
  Py_ssize_t shape[1];

#line 569 "pyarrow/lib.pxd"
  Py_ssize_t strides[1];

#line 569 "pyarrow/lib.pxd"
};

#line 569 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":579
 * 
 * 
 * cdef class ResizableBuffer(Buffer):             # <<<<<<<<<<<<<<
 * 
 *     cdef void init_rz(self, const shared_ptr[CResizableBuffer]& buffer)
*/

#line 579 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_ResizableBuffer {

#line 579 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Buffer __pyx_base;

#line 579 "pyarrow/lib.pxd"
};

#line 579 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":584
 * 
 * 
 * cdef class NativeFile(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CInputStream] input_stream
*/

#line 584 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_NativeFile {

#line 584 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 584 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_NativeFile *__pyx_vtab;

#line 584 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::io::InputStream>  input_stream;

#line 584 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::io::RandomAccessFile>  random_access;

#line 584 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::io::OutputStream>  output_stream;

#line 584 "pyarrow/lib.pxd"
  int is_readable;

#line 584 "pyarrow/lib.pxd"
  int is_writable;

#line 584 "pyarrow/lib.pxd"
  int is_seekable;

#line 584 "pyarrow/lib.pxd"
  int _is_appending;

#line 584 "pyarrow/lib.pxd"
  int own_file;

#line 584 "pyarrow/lib.pxd"
};

#line 584 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":608
 * 
 * 
 * cdef class BufferedInputStream(NativeFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 608 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_BufferedInputStream {

#line 608 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_NativeFile __pyx_base;

#line 608 "pyarrow/lib.pxd"
};

#line 608 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":612
 * 
 * 
 * cdef class BufferedOutputStream(NativeFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 612 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_BufferedOutputStream {

#line 612 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_NativeFile __pyx_base;

#line 612 "pyarrow/lib.pxd"
};

#line 612 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":616
 * 
 * 
 * cdef class CompressedInputStream(NativeFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 616 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_CompressedInputStream {

#line 616 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_NativeFile __pyx_base;

#line 616 "pyarrow/lib.pxd"
};

#line 616 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":620
 * 
 * 
 * cdef class CompressedOutputStream(NativeFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 620 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_CompressedOutputStream {

#line 620 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_NativeFile __pyx_base;

#line 620 "pyarrow/lib.pxd"
};

#line 620 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":624
 * 
 * 
 * cdef class _CRecordBatchWriter(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         SharedPtrNoGIL[CRecordBatchWriter] writer
*/

#line 624 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib__CRecordBatchWriter {

#line 624 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 624 "pyarrow/lib.pxd"
  arrow::py::SharedPtrNoGIL< arrow::ipc::RecordBatchWriter>  writer;

#line 624 "pyarrow/lib.pxd"
};

#line 624 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":629
 * 
 * 
 * cdef class RecordBatchReader(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         SharedPtrNoGIL[CRecordBatchReader] reader
*/

#line 629 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_RecordBatchReader {

#line 629 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 629 "pyarrow/lib.pxd"
  arrow::py::SharedPtrNoGIL< arrow::RecordBatchReader>  reader;

#line 629 "pyarrow/lib.pxd"
};

#line 629 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":634
 * 
 * 
 * cdef class CacheOptions(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         CCacheOptions wrapped
*/

#line 634 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_CacheOptions {

#line 634 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 634 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_CacheOptions *__pyx_vtab;

#line 634 "pyarrow/lib.pxd"
  arrow::io::CacheOptions wrapped;

#line 634 "pyarrow/lib.pxd"
};

#line 634 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":646
 * 
 * 
 * cdef class Codec(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CCodec] wrapped
*/

#line 646 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_Codec {

#line 646 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 646 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Codec *__pyx_vtab;

#line 646 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::util::Codec>  wrapped;

#line 646 "pyarrow/lib.pxd"
};

#line 646 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":654
 * 
 * # This class is only used internally for now
 * cdef class StopToken:             # <<<<<<<<<<<<<<
 *     cdef:
 *         CStopToken stop_token
*/

#line 654 "pyarrow/lib.pxd"
struct __pyx_obj_7pyarrow_3lib_StopToken {

#line 654 "pyarrow/lib.pxd"
  PyObject_HEAD

#line 654 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_StopToken *__pyx_vtab;

#line 654 "pyarrow/lib.pxd"
  arrow::StopToken stop_token;

#line 654 "pyarrow/lib.pxd"
};

#line 654 "pyarrow/lib.pxd"


/* "pyarrow/_parquet.pxd":25
 * 
 * 
 * cdef class FileEncryptionProperties:             # <<<<<<<<<<<<<<
 *     """File-level encryption properties for the low-level API"""
 *     cdef:
*/

#line 25 "pyarrow/_parquet.pxd"
struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties {

#line 25 "pyarrow/_parquet.pxd"
  PyObject_HEAD

#line 25 "pyarrow/_parquet.pxd"
  struct __pyx_vtabstruct_7pyarrow_8_parquet_FileEncryptionProperties *__pyx_vtab;

#line 25 "pyarrow/_parquet.pxd"
  std::shared_ptr< parquet::FileEncryptionProperties>  properties;

#line 25 "pyarrow/_parquet.pxd"
};

#line 25 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":78
 * 
 * 
 * cdef class ParquetSchema(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         FileMetaData parent  # the FileMetaData owning the SchemaDescriptor
*/

#line 78 "pyarrow/_parquet.pxd"
struct __pyx_obj_7pyarrow_8_parquet_ParquetSchema {

#line 78 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 78 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *parent;

#line 78 "pyarrow/_parquet.pxd"
  parquet::SchemaDescriptor const *schema;

#line 78 "pyarrow/_parquet.pxd"
};

#line 78 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":83
 *         const SchemaDescriptor* schema
 * 
 * cdef class FileMetaData(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CFileMetaData] sp_metadata
*/

#line 83 "pyarrow/_parquet.pxd"
struct __pyx_obj_7pyarrow_8_parquet_FileMetaData {

#line 83 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 83 "pyarrow/_parquet.pxd"
  struct __pyx_vtabstruct_7pyarrow_8_parquet_FileMetaData *__pyx_vtab;

#line 83 "pyarrow/_parquet.pxd"
  std::shared_ptr< parquet::FileMetaData>  sp_metadata;

#line 83 "pyarrow/_parquet.pxd"
   parquet::FileMetaData *_metadata;

#line 83 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_ParquetSchema *_schema;

#line 83 "pyarrow/_parquet.pxd"
};

#line 83 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":93
 *         self._metadata = metadata.get()
 * 
 * cdef class RowGroupMetaData(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int index  # for pickling support
*/

#line 93 "pyarrow/_parquet.pxd"
struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData {

#line 93 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 93 "pyarrow/_parquet.pxd"
  struct __pyx_vtabstruct_7pyarrow_8_parquet_RowGroupMetaData *__pyx_vtab;

#line 93 "pyarrow/_parquet.pxd"
  int index;

#line 93 "pyarrow/_parquet.pxd"
  std::unique_ptr< parquet::RowGroupMetaData>  up_metadata;

#line 93 "pyarrow/_parquet.pxd"
   parquet::RowGroupMetaData *metadata;

#line 93 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *parent;

#line 93 "pyarrow/_parquet.pxd"
};

#line 93 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":109
 * 
 * 
 * cdef class ColumnChunkMetaData(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         unique_ptr[CColumnChunkMetaData] up_metadata
*/

#line 109 "pyarrow/_parquet.pxd"
struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData {

#line 109 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 109 "pyarrow/_parquet.pxd"
  struct __pyx_vtabstruct_7pyarrow_8_parquet_ColumnChunkMetaData *__pyx_vtab;

#line 109 "pyarrow/_parquet.pxd"
  std::unique_ptr< parquet::ColumnChunkMetaData>  up_metadata;

#line 109 "pyarrow/_parquet.pxd"
   parquet::ColumnChunkMetaData *metadata;

#line 109 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData *parent;

#line 109 "pyarrow/_parquet.pxd"
};

#line 109 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":120
 *         self.parent = parent
 * 
 * cdef class Statistics(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CStatistics] statistics
*/

#line 120 "pyarrow/_parquet.pxd"
struct __pyx_obj_7pyarrow_8_parquet_Statistics {

#line 120 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 120 "pyarrow/_parquet.pxd"
  struct __pyx_vtabstruct_7pyarrow_8_parquet_Statistics *__pyx_vtab;

#line 120 "pyarrow/_parquet.pxd"
  std::shared_ptr< parquet::Statistics>  statistics;

#line 120 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *parent;

#line 120 "pyarrow/_parquet.pxd"
};

#line 120 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":130
 *         self.parent = parent
 * 
 * cdef class GeoStatistics(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CParquetGeoStatistics] statistics
*/

#line 130 "pyarrow/_parquet.pxd"
struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics {

#line 130 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_3lib__Weakrefable __pyx_base;

#line 130 "pyarrow/_parquet.pxd"
  struct __pyx_vtabstruct_7pyarrow_8_parquet_GeoStatistics *__pyx_vtab;

#line 130 "pyarrow/_parquet.pxd"
  std::shared_ptr< parquet::geospatial::GeoStatistics>  statistics;

#line 130 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *parent;

#line 130 "pyarrow/_parquet.pxd"
};

#line 130 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":140
 *         self.parent = parent
 * 
 * cdef class FileDecryptionProperties:             # <<<<<<<<<<<<<<
 *     """File-level decryption properties for the low-level API"""
 *     cdef:
*/

#line 140 "pyarrow/_parquet.pxd"
struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties {

#line 140 "pyarrow/_parquet.pxd"
  PyObject_HEAD

#line 140 "pyarrow/_parquet.pxd"
  struct __pyx_vtabstruct_7pyarrow_8_parquet_FileDecryptionProperties *__pyx_vtab;

#line 140 "pyarrow/_parquet.pxd"
  std::shared_ptr< parquet::FileDecryptionProperties>  properties;

#line 140 "pyarrow/_parquet.pxd"
};

#line 140 "pyarrow/_parquet.pxd"


#line 140 "pyarrow/_parquet.pxd"


/* "pyarrow/lib.pxd":65
 * 
 * 
 * cdef class MemoryPool(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         CMemoryPool* pool
*/

#line 65 "pyarrow/lib.pxd"


#line 65 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_MemoryPool {

#line 65 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_MemoryPool *,  arrow::MemoryPool *);

#line 65 "pyarrow/lib.pxd"
};

#line 65 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_MemoryPool *__pyx_vtabptr_7pyarrow_3lib_MemoryPool;

#line 65 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":78
 * 
 * 
 * cdef class DataType(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CDataType] sp_type
*/

#line 78 "pyarrow/lib.pxd"


#line 78 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_DataType {

#line 78 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_DataType *, std::shared_ptr< arrow::DataType>  const &);

#line 78 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Field *(*field)(struct __pyx_obj_7pyarrow_3lib_DataType *, PyObject *, int __pyx_skip_dispatch);

#line 78 "pyarrow/lib.pxd"
};

#line 78 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_DataType *__pyx_vtabptr_7pyarrow_3lib_DataType;

#line 78 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":88
 * 
 * 
 * cdef class ListType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CListType* list_type
*/

#line 88 "pyarrow/lib.pxd"


#line 88 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_ListType {

#line 88 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 88 "pyarrow/lib.pxd"
};

#line 88 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_ListType *__pyx_vtabptr_7pyarrow_3lib_ListType;

#line 88 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":93
 * 
 * 
 * cdef class LargeListType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CLargeListType* list_type
*/

#line 93 "pyarrow/lib.pxd"


#line 93 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_LargeListType {

#line 93 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 93 "pyarrow/lib.pxd"
};

#line 93 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_LargeListType *__pyx_vtabptr_7pyarrow_3lib_LargeListType;

#line 93 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":98
 * 
 * 
 * cdef class ListViewType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CListViewType* list_view_type
*/

#line 98 "pyarrow/lib.pxd"


#line 98 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_ListViewType {

#line 98 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 98 "pyarrow/lib.pxd"
};

#line 98 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_ListViewType *__pyx_vtabptr_7pyarrow_3lib_ListViewType;

#line 98 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":103
 * 
 * 
 * cdef class LargeListViewType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CLargeListViewType* list_view_type
*/

#line 103 "pyarrow/lib.pxd"


#line 103 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_LargeListViewType {

#line 103 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 103 "pyarrow/lib.pxd"
};

#line 103 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_LargeListViewType *__pyx_vtabptr_7pyarrow_3lib_LargeListViewType;

#line 103 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":108
 * 
 * 
 * cdef class MapType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CMapType* map_type
*/

#line 108 "pyarrow/lib.pxd"


#line 108 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_MapType {

#line 108 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 108 "pyarrow/lib.pxd"
};

#line 108 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_MapType *__pyx_vtabptr_7pyarrow_3lib_MapType;

#line 108 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":113
 * 
 * 
 * cdef class FixedSizeListType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CFixedSizeListType* list_type
*/

#line 113 "pyarrow/lib.pxd"


#line 113 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeListType {

#line 113 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 113 "pyarrow/lib.pxd"
};

#line 113 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeListType *__pyx_vtabptr_7pyarrow_3lib_FixedSizeListType;

#line 113 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":118
 * 
 * 
 * cdef class StructType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CStructType* struct_type
*/

#line 118 "pyarrow/lib.pxd"


#line 118 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_StructType {

#line 118 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 118 "pyarrow/lib.pxd"
  struct __pyx_obj_7pyarrow_3lib_Field *(*field_by_name)(struct __pyx_obj_7pyarrow_3lib_StructType *, PyObject *);

#line 118 "pyarrow/lib.pxd"
};

#line 118 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_StructType *__pyx_vtabptr_7pyarrow_3lib_StructType;

#line 118 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":133
 * 
 * 
 * cdef class DictionaryType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDictionaryType* dict_type
*/

#line 133 "pyarrow/lib.pxd"


#line 133 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_DictionaryType {

#line 133 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 133 "pyarrow/lib.pxd"
};

#line 133 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_DictionaryType *__pyx_vtabptr_7pyarrow_3lib_DictionaryType;

#line 133 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":138
 * 
 * 
 * cdef class TimestampType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CTimestampType* ts_type
*/

#line 138 "pyarrow/lib.pxd"


#line 138 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_TimestampType {

#line 138 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 138 "pyarrow/lib.pxd"
};

#line 138 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_TimestampType *__pyx_vtabptr_7pyarrow_3lib_TimestampType;

#line 138 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":143
 * 
 * 
 * cdef class Time32Type(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CTime32Type* time_type
*/

#line 143 "pyarrow/lib.pxd"


#line 143 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Time32Type {

#line 143 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 143 "pyarrow/lib.pxd"
};

#line 143 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Time32Type *__pyx_vtabptr_7pyarrow_3lib_Time32Type;

#line 143 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":148
 * 
 * 
 * cdef class Time64Type(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CTime64Type* time_type
*/

#line 148 "pyarrow/lib.pxd"


#line 148 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Time64Type {

#line 148 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 148 "pyarrow/lib.pxd"
};

#line 148 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Time64Type *__pyx_vtabptr_7pyarrow_3lib_Time64Type;

#line 148 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":153
 * 
 * 
 * cdef class DurationType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDurationType* duration_type
*/

#line 153 "pyarrow/lib.pxd"


#line 153 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_DurationType {

#line 153 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 153 "pyarrow/lib.pxd"
};

#line 153 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_DurationType *__pyx_vtabptr_7pyarrow_3lib_DurationType;

#line 153 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":158
 * 
 * 
 * cdef class FixedSizeBinaryType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CFixedSizeBinaryType* fixed_size_binary_type
*/

#line 158 "pyarrow/lib.pxd"


#line 158 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryType {

#line 158 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 158 "pyarrow/lib.pxd"
};

#line 158 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryType *__pyx_vtabptr_7pyarrow_3lib_FixedSizeBinaryType;

#line 158 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":163
 * 
 * 
 * cdef class Decimal32Type(FixedSizeBinaryType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDecimal32Type* decimal32_type
*/

#line 163 "pyarrow/lib.pxd"


#line 163 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Decimal32Type {

#line 163 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryType __pyx_base;

#line 163 "pyarrow/lib.pxd"
};

#line 163 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Decimal32Type *__pyx_vtabptr_7pyarrow_3lib_Decimal32Type;

#line 163 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":168
 * 
 * 
 * cdef class Decimal64Type(FixedSizeBinaryType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDecimal64Type* decimal64_type
*/

#line 168 "pyarrow/lib.pxd"


#line 168 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Decimal64Type {

#line 168 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryType __pyx_base;

#line 168 "pyarrow/lib.pxd"
};

#line 168 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Decimal64Type *__pyx_vtabptr_7pyarrow_3lib_Decimal64Type;

#line 168 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":173
 * 
 * 
 * cdef class Decimal128Type(FixedSizeBinaryType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDecimal128Type* decimal128_type
*/

#line 173 "pyarrow/lib.pxd"


#line 173 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Decimal128Type {

#line 173 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryType __pyx_base;

#line 173 "pyarrow/lib.pxd"
};

#line 173 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Decimal128Type *__pyx_vtabptr_7pyarrow_3lib_Decimal128Type;

#line 173 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":178
 * 
 * 
 * cdef class Decimal256Type(FixedSizeBinaryType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CDecimal256Type* decimal256_type
*/

#line 178 "pyarrow/lib.pxd"


#line 178 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Decimal256Type {

#line 178 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryType __pyx_base;

#line 178 "pyarrow/lib.pxd"
};

#line 178 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Decimal256Type *__pyx_vtabptr_7pyarrow_3lib_Decimal256Type;

#line 178 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":183
 * 
 * 
 * cdef class RunEndEncodedType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CRunEndEncodedType* run_end_encoded_type
*/

#line 183 "pyarrow/lib.pxd"


#line 183 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_RunEndEncodedType {

#line 183 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 183 "pyarrow/lib.pxd"
};

#line 183 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_RunEndEncodedType *__pyx_vtabptr_7pyarrow_3lib_RunEndEncodedType;

#line 183 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":188
 * 
 * 
 * cdef class BaseExtensionType(DataType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CExtensionType* ext_type
*/

#line 188 "pyarrow/lib.pxd"


#line 188 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_BaseExtensionType {

#line 188 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_DataType __pyx_base;

#line 188 "pyarrow/lib.pxd"
};

#line 188 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_BaseExtensionType *__pyx_vtabptr_7pyarrow_3lib_BaseExtensionType;

#line 188 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":193
 * 
 * 
 * cdef class ExtensionType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CPyExtensionType* cpy_ext_type
*/

#line 193 "pyarrow/lib.pxd"


#line 193 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_ExtensionType {

#line 193 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 193 "pyarrow/lib.pxd"
};

#line 193 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_ExtensionType *__pyx_vtabptr_7pyarrow_3lib_ExtensionType;

#line 193 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":198
 * 
 * 
 * cdef class FixedShapeTensorType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CFixedShapeTensorType* tensor_ext_type
*/

#line 198 "pyarrow/lib.pxd"


#line 198 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_FixedShapeTensorType {

#line 198 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 198 "pyarrow/lib.pxd"
};

#line 198 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_FixedShapeTensorType *__pyx_vtabptr_7pyarrow_3lib_FixedShapeTensorType;

#line 198 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":202
 *         const CFixedShapeTensorType* tensor_ext_type
 * 
 * cdef class Bool8Type(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CBool8Type* bool8_ext_type
*/

#line 202 "pyarrow/lib.pxd"


#line 202 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Bool8Type {

#line 202 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 202 "pyarrow/lib.pxd"
};

#line 202 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Bool8Type *__pyx_vtabptr_7pyarrow_3lib_Bool8Type;

#line 202 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":206
 *         const CBool8Type* bool8_ext_type
 * 
 * cdef class OpaqueType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const COpaqueType* opaque_ext_type
*/

#line 206 "pyarrow/lib.pxd"


#line 206 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_OpaqueType {

#line 206 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 206 "pyarrow/lib.pxd"
};

#line 206 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_OpaqueType *__pyx_vtabptr_7pyarrow_3lib_OpaqueType;

#line 206 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":210
 *         const COpaqueType* opaque_ext_type
 * 
 * cdef class UuidType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CUuidType* uuid_ext_type
*/

#line 210 "pyarrow/lib.pxd"


#line 210 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_UuidType {

#line 210 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 210 "pyarrow/lib.pxd"
};

#line 210 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_UuidType *__pyx_vtabptr_7pyarrow_3lib_UuidType;

#line 210 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":214
 *         const CUuidType* uuid_ext_type
 * 
 * cdef class JsonType(BaseExtensionType):             # <<<<<<<<<<<<<<
 *     cdef:
 *         const CJsonType* json_ext_type
*/

#line 214 "pyarrow/lib.pxd"


#line 214 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_JsonType {

#line 214 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseExtensionType __pyx_base;

#line 214 "pyarrow/lib.pxd"
};

#line 214 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_JsonType *__pyx_vtabptr_7pyarrow_3lib_JsonType;

#line 214 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":225
 * 
 * 
 * cdef class KeyValueMetadata(_Metadata):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[const CKeyValueMetadata] wrapped
*/

#line 225 "pyarrow/lib.pxd"


#line 225 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_KeyValueMetadata {

#line 225 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata *, std::shared_ptr< arrow::KeyValueMetadata const >  const &);

#line 225 "pyarrow/lib.pxd"
  PyObject *(*wrap)(std::shared_ptr< arrow::KeyValueMetadata const >  const &);

#line 225 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::KeyValueMetadata const >  (*unwrap)(struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata *);

#line 225 "pyarrow/lib.pxd"
};

#line 225 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_KeyValueMetadata *__pyx_vtabptr_7pyarrow_3lib_KeyValueMetadata;

#line 225 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":237
 * 
 * 
 * cdef class Field(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CField] sp_field
*/

#line 237 "pyarrow/lib.pxd"


#line 237 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Field {

#line 237 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_Field *, std::shared_ptr< arrow::Field>  const &);

#line 237 "pyarrow/lib.pxd"
};

#line 237 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Field *__pyx_vtabptr_7pyarrow_3lib_Field;

#line 237 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":248
 * 
 * 
 * cdef class Schema(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSchema] sp_schema
*/

#line 248 "pyarrow/lib.pxd"


#line 248 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Schema {

#line 248 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_Schema *, std::vector<std::shared_ptr< arrow::Field> >  const &);

#line 248 "pyarrow/lib.pxd"
  void (*init_schema)(struct __pyx_obj_7pyarrow_3lib_Schema *, std::shared_ptr< arrow::Schema>  const &);

#line 248 "pyarrow/lib.pxd"
};

#line 248 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Schema *__pyx_vtabptr_7pyarrow_3lib_Schema;

#line 248 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":257
 * 
 * 
 * cdef class Scalar(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CScalar] wrapped
*/

#line 257 "pyarrow/lib.pxd"


#line 257 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Scalar {

#line 257 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_Scalar *, std::shared_ptr< arrow::Scalar>  const &);

#line 257 "pyarrow/lib.pxd"
  PyObject *(*wrap)(std::shared_ptr< arrow::Scalar>  const &);

#line 257 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Scalar>  (*unwrap)(struct __pyx_obj_7pyarrow_3lib_Scalar *);

#line 257 "pyarrow/lib.pxd"
};

#line 257 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Scalar *__pyx_vtabptr_7pyarrow_3lib_Scalar;

#line 257 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":269
 * 
 * 
 * cdef class ArrayStatistics(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CArrayStatistics] sp_statistics
*/

#line 269 "pyarrow/lib.pxd"


#line 269 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_ArrayStatistics {

#line 269 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_ArrayStatistics *, std::shared_ptr< arrow::ArrayStatistics>  const &);

#line 269 "pyarrow/lib.pxd"
  PyObject *(*_get_value)(struct __pyx_obj_7pyarrow_3lib_ArrayStatistics *, std::optional< std::variant<bool, int64_t, uint64_t, double, std::string>>  const &);

#line 269 "pyarrow/lib.pxd"
};

#line 269 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_ArrayStatistics *__pyx_vtabptr_7pyarrow_3lib_ArrayStatistics;

#line 269 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":281
 * 
 * 
 * cdef class Array(_PandasConvertible):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CArray] sp_array
*/

#line 281 "pyarrow/lib.pxd"


#line 281 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Array {

#line 281 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_Array *, std::shared_ptr< arrow::Array>  const &);

#line 281 "pyarrow/lib.pxd"
  PyObject *(*getitem)(struct __pyx_obj_7pyarrow_3lib_Array *, int64_t);

#line 281 "pyarrow/lib.pxd"
  int64_t (*length)(struct __pyx_obj_7pyarrow_3lib_Array *);

#line 281 "pyarrow/lib.pxd"
  void (*_assert_cpu)(struct __pyx_obj_7pyarrow_3lib_Array *);

#line 281 "pyarrow/lib.pxd"
};

#line 281 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Array *__pyx_vtabptr_7pyarrow_3lib_Array;

#line 281 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":297
 * 
 * 
 * cdef class Tensor(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CTensor] sp_tensor
*/

#line 297 "pyarrow/lib.pxd"


#line 297 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Tensor {

#line 297 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_Tensor *, std::shared_ptr< arrow::Tensor>  const &);

#line 297 "pyarrow/lib.pxd"
};

#line 297 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Tensor *__pyx_vtabptr_7pyarrow_3lib_Tensor;

#line 297 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":310
 * 
 * 
 * cdef class SparseCSRMatrix(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSparseCSRMatrix] sp_sparse_tensor
*/

#line 310 "pyarrow/lib.pxd"


#line 310 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSRMatrix {

#line 310 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_SparseCSRMatrix *, std::shared_ptr< arrow::SparseCSRMatrix>  const &);

#line 310 "pyarrow/lib.pxd"
};

#line 310 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSRMatrix *__pyx_vtabptr_7pyarrow_3lib_SparseCSRMatrix;

#line 310 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":321
 * 
 * 
 * cdef class SparseCSCMatrix(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSparseCSCMatrix] sp_sparse_tensor
*/

#line 321 "pyarrow/lib.pxd"


#line 321 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSCMatrix {

#line 321 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_SparseCSCMatrix *, std::shared_ptr< arrow::SparseCSCMatrix>  const &);

#line 321 "pyarrow/lib.pxd"
};

#line 321 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSCMatrix *__pyx_vtabptr_7pyarrow_3lib_SparseCSCMatrix;

#line 321 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":332
 * 
 * 
 * cdef class SparseCOOTensor(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSparseCOOTensor] sp_sparse_tensor
*/

#line 332 "pyarrow/lib.pxd"


#line 332 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_SparseCOOTensor {

#line 332 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_SparseCOOTensor *, std::shared_ptr< arrow::SparseCOOTensor>  const &);

#line 332 "pyarrow/lib.pxd"
};

#line 332 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_SparseCOOTensor *__pyx_vtabptr_7pyarrow_3lib_SparseCOOTensor;

#line 332 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":343
 * 
 * 
 * cdef class SparseCSFTensor(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CSparseCSFTensor] sp_sparse_tensor
*/

#line 343 "pyarrow/lib.pxd"


#line 343 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSFTensor {

#line 343 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_SparseCSFTensor *, std::shared_ptr< arrow::SparseCSFTensor>  const &);

#line 343 "pyarrow/lib.pxd"
};

#line 343 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSFTensor *__pyx_vtabptr_7pyarrow_3lib_SparseCSFTensor;

#line 343 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":354
 * 
 * 
 * cdef class NullArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 354 "pyarrow/lib.pxd"


#line 354 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_NullArray {

#line 354 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 354 "pyarrow/lib.pxd"
};

#line 354 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_NullArray *__pyx_vtabptr_7pyarrow_3lib_NullArray;

#line 354 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":358
 * 
 * 
 * cdef class BooleanArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 358 "pyarrow/lib.pxd"


#line 358 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_BooleanArray {

#line 358 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 358 "pyarrow/lib.pxd"
};

#line 358 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_BooleanArray *__pyx_vtabptr_7pyarrow_3lib_BooleanArray;

#line 358 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":362
 * 
 * 
 * cdef class NumericArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 362 "pyarrow/lib.pxd"


#line 362 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_NumericArray {

#line 362 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 362 "pyarrow/lib.pxd"
};

#line 362 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_NumericArray *__pyx_vtabptr_7pyarrow_3lib_NumericArray;

#line 362 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":366
 * 
 * 
 * cdef class IntegerArray(NumericArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 366 "pyarrow/lib.pxd"


#line 366 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray {

#line 366 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_NumericArray __pyx_base;

#line 366 "pyarrow/lib.pxd"
};

#line 366 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray *__pyx_vtabptr_7pyarrow_3lib_IntegerArray;

#line 366 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":370
 * 
 * 
 * cdef class FloatingPointArray(NumericArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 370 "pyarrow/lib.pxd"


#line 370 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_FloatingPointArray {

#line 370 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_NumericArray __pyx_base;

#line 370 "pyarrow/lib.pxd"
};

#line 370 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_FloatingPointArray *__pyx_vtabptr_7pyarrow_3lib_FloatingPointArray;

#line 370 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":374
 * 
 * 
 * cdef class Int8Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 374 "pyarrow/lib.pxd"


#line 374 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Int8Array {

#line 374 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray __pyx_base;

#line 374 "pyarrow/lib.pxd"
};

#line 374 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Int8Array *__pyx_vtabptr_7pyarrow_3lib_Int8Array;

#line 374 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":378
 * 
 * 
 * cdef class UInt8Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 378 "pyarrow/lib.pxd"


#line 378 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_UInt8Array {

#line 378 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray __pyx_base;

#line 378 "pyarrow/lib.pxd"
};

#line 378 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_UInt8Array *__pyx_vtabptr_7pyarrow_3lib_UInt8Array;

#line 378 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":382
 * 
 * 
 * cdef class Int16Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 382 "pyarrow/lib.pxd"


#line 382 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Int16Array {

#line 382 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray __pyx_base;

#line 382 "pyarrow/lib.pxd"
};

#line 382 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Int16Array *__pyx_vtabptr_7pyarrow_3lib_Int16Array;

#line 382 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":386
 * 
 * 
 * cdef class UInt16Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 386 "pyarrow/lib.pxd"


#line 386 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_UInt16Array {

#line 386 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray __pyx_base;

#line 386 "pyarrow/lib.pxd"
};

#line 386 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_UInt16Array *__pyx_vtabptr_7pyarrow_3lib_UInt16Array;

#line 386 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":390
 * 
 * 
 * cdef class Int32Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 390 "pyarrow/lib.pxd"


#line 390 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Int32Array {

#line 390 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray __pyx_base;

#line 390 "pyarrow/lib.pxd"
};

#line 390 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Int32Array *__pyx_vtabptr_7pyarrow_3lib_Int32Array;

#line 390 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":394
 * 
 * 
 * cdef class UInt32Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 394 "pyarrow/lib.pxd"


#line 394 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_UInt32Array {

#line 394 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray __pyx_base;

#line 394 "pyarrow/lib.pxd"
};

#line 394 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_UInt32Array *__pyx_vtabptr_7pyarrow_3lib_UInt32Array;

#line 394 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":398
 * 
 * 
 * cdef class Int64Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 398 "pyarrow/lib.pxd"


#line 398 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Int64Array {

#line 398 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray __pyx_base;

#line 398 "pyarrow/lib.pxd"
};

#line 398 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Int64Array *__pyx_vtabptr_7pyarrow_3lib_Int64Array;

#line 398 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":402
 * 
 * 
 * cdef class UInt64Array(IntegerArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 402 "pyarrow/lib.pxd"


#line 402 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_UInt64Array {

#line 402 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray __pyx_base;

#line 402 "pyarrow/lib.pxd"
};

#line 402 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_UInt64Array *__pyx_vtabptr_7pyarrow_3lib_UInt64Array;

#line 402 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":406
 * 
 * 
 * cdef class HalfFloatArray(FloatingPointArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 406 "pyarrow/lib.pxd"


#line 406 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_HalfFloatArray {

#line 406 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FloatingPointArray __pyx_base;

#line 406 "pyarrow/lib.pxd"
};

#line 406 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_HalfFloatArray *__pyx_vtabptr_7pyarrow_3lib_HalfFloatArray;

#line 406 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":410
 * 
 * 
 * cdef class FloatArray(FloatingPointArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 410 "pyarrow/lib.pxd"


#line 410 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_FloatArray {

#line 410 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FloatingPointArray __pyx_base;

#line 410 "pyarrow/lib.pxd"
};

#line 410 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_FloatArray *__pyx_vtabptr_7pyarrow_3lib_FloatArray;

#line 410 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":414
 * 
 * 
 * cdef class DoubleArray(FloatingPointArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 414 "pyarrow/lib.pxd"


#line 414 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_DoubleArray {

#line 414 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FloatingPointArray __pyx_base;

#line 414 "pyarrow/lib.pxd"
};

#line 414 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_DoubleArray *__pyx_vtabptr_7pyarrow_3lib_DoubleArray;

#line 414 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":418
 * 
 * 
 * cdef class FixedSizeBinaryArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 418 "pyarrow/lib.pxd"


#line 418 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryArray {

#line 418 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 418 "pyarrow/lib.pxd"
};

#line 418 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryArray *__pyx_vtabptr_7pyarrow_3lib_FixedSizeBinaryArray;

#line 418 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":422
 * 
 * 
 * cdef class Decimal32Array(FixedSizeBinaryArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 422 "pyarrow/lib.pxd"


#line 422 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Decimal32Array {

#line 422 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryArray __pyx_base;

#line 422 "pyarrow/lib.pxd"
};

#line 422 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Decimal32Array *__pyx_vtabptr_7pyarrow_3lib_Decimal32Array;

#line 422 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":426
 * 
 * 
 * cdef class Decimal64Array(FixedSizeBinaryArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 426 "pyarrow/lib.pxd"


#line 426 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Decimal64Array {

#line 426 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryArray __pyx_base;

#line 426 "pyarrow/lib.pxd"
};

#line 426 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Decimal64Array *__pyx_vtabptr_7pyarrow_3lib_Decimal64Array;

#line 426 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":430
 * 
 * 
 * cdef class Decimal128Array(FixedSizeBinaryArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 430 "pyarrow/lib.pxd"


#line 430 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Decimal128Array {

#line 430 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryArray __pyx_base;

#line 430 "pyarrow/lib.pxd"
};

#line 430 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Decimal128Array *__pyx_vtabptr_7pyarrow_3lib_Decimal128Array;

#line 430 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":434
 * 
 * 
 * cdef class Decimal256Array(FixedSizeBinaryArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 434 "pyarrow/lib.pxd"


#line 434 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Decimal256Array {

#line 434 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryArray __pyx_base;

#line 434 "pyarrow/lib.pxd"
};

#line 434 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Decimal256Array *__pyx_vtabptr_7pyarrow_3lib_Decimal256Array;

#line 434 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":438
 * 
 * 
 * cdef class StructArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 438 "pyarrow/lib.pxd"


#line 438 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_StructArray {

#line 438 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 438 "pyarrow/lib.pxd"
};

#line 438 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_StructArray *__pyx_vtabptr_7pyarrow_3lib_StructArray;

#line 438 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":442
 * 
 * 
 * cdef class BaseListArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 442 "pyarrow/lib.pxd"


#line 442 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_BaseListArray {

#line 442 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 442 "pyarrow/lib.pxd"
};

#line 442 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_BaseListArray *__pyx_vtabptr_7pyarrow_3lib_BaseListArray;

#line 442 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":446
 * 
 * 
 * cdef class ListArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 446 "pyarrow/lib.pxd"


#line 446 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_ListArray {

#line 446 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseListArray __pyx_base;

#line 446 "pyarrow/lib.pxd"
};

#line 446 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_ListArray *__pyx_vtabptr_7pyarrow_3lib_ListArray;

#line 446 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":450
 * 
 * 
 * cdef class LargeListArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 450 "pyarrow/lib.pxd"


#line 450 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_LargeListArray {

#line 450 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseListArray __pyx_base;

#line 450 "pyarrow/lib.pxd"
};

#line 450 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_LargeListArray *__pyx_vtabptr_7pyarrow_3lib_LargeListArray;

#line 450 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":454
 * 
 * 
 * cdef class ListViewArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 454 "pyarrow/lib.pxd"


#line 454 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_ListViewArray {

#line 454 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseListArray __pyx_base;

#line 454 "pyarrow/lib.pxd"
};

#line 454 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_ListViewArray *__pyx_vtabptr_7pyarrow_3lib_ListViewArray;

#line 454 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":458
 * 
 * 
 * cdef class LargeListViewArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 458 "pyarrow/lib.pxd"


#line 458 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_LargeListViewArray {

#line 458 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseListArray __pyx_base;

#line 458 "pyarrow/lib.pxd"
};

#line 458 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_LargeListViewArray *__pyx_vtabptr_7pyarrow_3lib_LargeListViewArray;

#line 458 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":462
 * 
 * 
 * cdef class MapArray(ListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 462 "pyarrow/lib.pxd"


#line 462 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_MapArray {

#line 462 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_ListArray __pyx_base;

#line 462 "pyarrow/lib.pxd"
};

#line 462 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_MapArray *__pyx_vtabptr_7pyarrow_3lib_MapArray;

#line 462 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":466
 * 
 * 
 * cdef class FixedSizeListArray(BaseListArray):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 466 "pyarrow/lib.pxd"


#line 466 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeListArray {

#line 466 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_BaseListArray __pyx_base;

#line 466 "pyarrow/lib.pxd"
};

#line 466 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeListArray *__pyx_vtabptr_7pyarrow_3lib_FixedSizeListArray;

#line 466 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":470
 * 
 * 
 * cdef class UnionArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 470 "pyarrow/lib.pxd"


#line 470 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_UnionArray {

#line 470 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 470 "pyarrow/lib.pxd"
};

#line 470 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_UnionArray *__pyx_vtabptr_7pyarrow_3lib_UnionArray;

#line 470 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":474
 * 
 * 
 * cdef class StringArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 474 "pyarrow/lib.pxd"


#line 474 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_StringArray {

#line 474 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 474 "pyarrow/lib.pxd"
};

#line 474 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_StringArray *__pyx_vtabptr_7pyarrow_3lib_StringArray;

#line 474 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":478
 * 
 * 
 * cdef class BinaryArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 478 "pyarrow/lib.pxd"


#line 478 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_BinaryArray {

#line 478 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 478 "pyarrow/lib.pxd"
};

#line 478 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_BinaryArray *__pyx_vtabptr_7pyarrow_3lib_BinaryArray;

#line 478 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":482
 * 
 * 
 * cdef class StringViewArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 482 "pyarrow/lib.pxd"


#line 482 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_StringViewArray {

#line 482 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 482 "pyarrow/lib.pxd"
};

#line 482 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_StringViewArray *__pyx_vtabptr_7pyarrow_3lib_StringViewArray;

#line 482 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":486
 * 
 * 
 * cdef class BinaryViewArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 486 "pyarrow/lib.pxd"


#line 486 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_BinaryViewArray {

#line 486 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 486 "pyarrow/lib.pxd"
};

#line 486 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_BinaryViewArray *__pyx_vtabptr_7pyarrow_3lib_BinaryViewArray;

#line 486 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":490
 * 
 * 
 * cdef class DictionaryArray(Array):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _indices, _dictionary
*/

#line 490 "pyarrow/lib.pxd"


#line 490 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_DictionaryArray {

#line 490 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 490 "pyarrow/lib.pxd"
};

#line 490 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_DictionaryArray *__pyx_vtabptr_7pyarrow_3lib_DictionaryArray;

#line 490 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":495
 * 
 * 
 * cdef class ExtensionArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 495 "pyarrow/lib.pxd"


#line 495 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_ExtensionArray {

#line 495 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 495 "pyarrow/lib.pxd"
};

#line 495 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_ExtensionArray *__pyx_vtabptr_7pyarrow_3lib_ExtensionArray;

#line 495 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":499
 * 
 * 
 * cdef class MonthDayNanoIntervalArray(Array):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 499 "pyarrow/lib.pxd"


#line 499 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_MonthDayNanoIntervalArray {

#line 499 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Array __pyx_base;

#line 499 "pyarrow/lib.pxd"
};

#line 499 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_MonthDayNanoIntervalArray *__pyx_vtabptr_7pyarrow_3lib_MonthDayNanoIntervalArray;

#line 499 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":507
 * 
 * 
 * cdef class ChunkedArray(_PandasConvertible):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CChunkedArray] sp_chunked_array
*/

#line 507 "pyarrow/lib.pxd"


#line 507 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_ChunkedArray {

#line 507 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_ChunkedArray *, std::shared_ptr< arrow::ChunkedArray>  const &);

#line 507 "pyarrow/lib.pxd"
  PyObject *(*getitem)(struct __pyx_obj_7pyarrow_3lib_ChunkedArray *, int64_t);

#line 507 "pyarrow/lib.pxd"
};

#line 507 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_ChunkedArray *__pyx_vtabptr_7pyarrow_3lib_ChunkedArray;

#line 507 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":522
 * 
 * 
 * cdef class _Tabular(_PandasConvertible):             # <<<<<<<<<<<<<<
 *     cdef void _assert_cpu(self) except *
 * 
*/

#line 522 "pyarrow/lib.pxd"


#line 522 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib__Tabular {

#line 522 "pyarrow/lib.pxd"
  void (*_assert_cpu)(struct __pyx_obj_7pyarrow_3lib__Tabular *);

#line 522 "pyarrow/lib.pxd"
};

#line 522 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib__Tabular *__pyx_vtabptr_7pyarrow_3lib__Tabular;

#line 522 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":526
 * 
 * 
 * cdef class Table(_Tabular):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CTable] sp_table
*/

#line 526 "pyarrow/lib.pxd"


#line 526 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Table {

#line 526 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib__Tabular __pyx_base;

#line 526 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_Table *, std::shared_ptr< arrow::Table>  const &);

#line 526 "pyarrow/lib.pxd"
};

#line 526 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Table *__pyx_vtabptr_7pyarrow_3lib_Table;

#line 526 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":536
 * 
 * 
 * cdef class RecordBatch(_Tabular):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CRecordBatch] sp_batch
*/

#line 536 "pyarrow/lib.pxd"


#line 536 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_RecordBatch {

#line 536 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib__Tabular __pyx_base;

#line 536 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_RecordBatch *, std::shared_ptr< arrow::RecordBatch>  const &);

#line 536 "pyarrow/lib.pxd"
};

#line 536 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_RecordBatch *__pyx_vtabptr_7pyarrow_3lib_RecordBatch;

#line 536 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":545
 * 
 * 
 * cdef class Device(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CDevice] device
*/

#line 545 "pyarrow/lib.pxd"


#line 545 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Device {

#line 545 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_Device *, std::shared_ptr< arrow::Device>  const &);

#line 545 "pyarrow/lib.pxd"
  PyObject *(*wrap)(std::shared_ptr< arrow::Device>  const &);

#line 545 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::Device>  (*unwrap)(struct __pyx_obj_7pyarrow_3lib_Device *);

#line 545 "pyarrow/lib.pxd"
};

#line 545 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Device *__pyx_vtabptr_7pyarrow_3lib_Device;

#line 545 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":557
 * 
 * 
 * cdef class MemoryManager(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CMemoryManager] memory_manager
*/

#line 557 "pyarrow/lib.pxd"


#line 557 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_MemoryManager {

#line 557 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_MemoryManager *, std::shared_ptr< arrow::MemoryManager>  const &);

#line 557 "pyarrow/lib.pxd"
  PyObject *(*wrap)(std::shared_ptr< arrow::MemoryManager>  const &);

#line 557 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::MemoryManager>  (*unwrap)(struct __pyx_obj_7pyarrow_3lib_MemoryManager *);

#line 557 "pyarrow/lib.pxd"
};

#line 557 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_MemoryManager *__pyx_vtabptr_7pyarrow_3lib_MemoryManager;

#line 557 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":569
 * 
 * 
 * cdef class Buffer(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CBuffer] buffer
*/

#line 569 "pyarrow/lib.pxd"


#line 569 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Buffer {

#line 569 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_Buffer *, std::shared_ptr< arrow::Buffer>  const &);

#line 569 "pyarrow/lib.pxd"
  PyObject *(*getitem)(struct __pyx_obj_7pyarrow_3lib_Buffer *, int64_t);

#line 569 "pyarrow/lib.pxd"
};

#line 569 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Buffer *__pyx_vtabptr_7pyarrow_3lib_Buffer;

#line 569 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":579
 * 
 * 
 * cdef class ResizableBuffer(Buffer):             # <<<<<<<<<<<<<<
 * 
 *     cdef void init_rz(self, const shared_ptr[CResizableBuffer]& buffer)
*/

#line 579 "pyarrow/lib.pxd"


#line 579 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_ResizableBuffer {

#line 579 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_Buffer __pyx_base;

#line 579 "pyarrow/lib.pxd"
  void (*init_rz)(struct __pyx_obj_7pyarrow_3lib_ResizableBuffer *, std::shared_ptr< arrow::ResizableBuffer>  const &);

#line 579 "pyarrow/lib.pxd"
};

#line 579 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_ResizableBuffer *__pyx_vtabptr_7pyarrow_3lib_ResizableBuffer;

#line 579 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":584
 * 
 * 
 * cdef class NativeFile(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CInputStream] input_stream
*/

#line 584 "pyarrow/lib.pxd"


#line 584 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_NativeFile {

#line 584 "pyarrow/lib.pxd"
  PyObject *(*set_random_access_file)(struct __pyx_obj_7pyarrow_3lib_NativeFile *, std::shared_ptr< arrow::io::RandomAccessFile> );

#line 584 "pyarrow/lib.pxd"
  PyObject *(*set_input_stream)(struct __pyx_obj_7pyarrow_3lib_NativeFile *, std::shared_ptr< arrow::io::InputStream> );

#line 584 "pyarrow/lib.pxd"
  PyObject *(*set_output_stream)(struct __pyx_obj_7pyarrow_3lib_NativeFile *, std::shared_ptr< arrow::io::OutputStream> );

#line 584 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::io::RandomAccessFile>  (*get_random_access_file)(struct __pyx_obj_7pyarrow_3lib_NativeFile *);

#line 584 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::io::InputStream>  (*get_input_stream)(struct __pyx_obj_7pyarrow_3lib_NativeFile *);

#line 584 "pyarrow/lib.pxd"
  std::shared_ptr< arrow::io::OutputStream>  (*get_output_stream)(struct __pyx_obj_7pyarrow_3lib_NativeFile *);

#line 584 "pyarrow/lib.pxd"
};

#line 584 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_NativeFile *__pyx_vtabptr_7pyarrow_3lib_NativeFile;

#line 584 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":608
 * 
 * 
 * cdef class BufferedInputStream(NativeFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 608 "pyarrow/lib.pxd"


#line 608 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_BufferedInputStream {

#line 608 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_NativeFile __pyx_base;

#line 608 "pyarrow/lib.pxd"
};

#line 608 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_BufferedInputStream *__pyx_vtabptr_7pyarrow_3lib_BufferedInputStream;

#line 608 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":612
 * 
 * 
 * cdef class BufferedOutputStream(NativeFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 612 "pyarrow/lib.pxd"


#line 612 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_BufferedOutputStream {

#line 612 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_NativeFile __pyx_base;

#line 612 "pyarrow/lib.pxd"
};

#line 612 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_BufferedOutputStream *__pyx_vtabptr_7pyarrow_3lib_BufferedOutputStream;

#line 612 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":616
 * 
 * 
 * cdef class CompressedInputStream(NativeFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 616 "pyarrow/lib.pxd"


#line 616 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_CompressedInputStream {

#line 616 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_NativeFile __pyx_base;

#line 616 "pyarrow/lib.pxd"
};

#line 616 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_CompressedInputStream *__pyx_vtabptr_7pyarrow_3lib_CompressedInputStream;

#line 616 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":620
 * 
 * 
 * cdef class CompressedOutputStream(NativeFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

#line 620 "pyarrow/lib.pxd"


#line 620 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_CompressedOutputStream {

#line 620 "pyarrow/lib.pxd"
  struct __pyx_vtabstruct_7pyarrow_3lib_NativeFile __pyx_base;

#line 620 "pyarrow/lib.pxd"
};

#line 620 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_CompressedOutputStream *__pyx_vtabptr_7pyarrow_3lib_CompressedOutputStream;

#line 620 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":634
 * 
 * 
 * cdef class CacheOptions(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         CCacheOptions wrapped
*/

#line 634 "pyarrow/lib.pxd"


#line 634 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_CacheOptions {

#line 634 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_CacheOptions *, arrow::io::CacheOptions);

#line 634 "pyarrow/lib.pxd"
  arrow::io::CacheOptions (*unwrap)(struct __pyx_obj_7pyarrow_3lib_CacheOptions *);

#line 634 "pyarrow/lib.pxd"
  PyObject *(*wrap)(arrow::io::CacheOptions const );

#line 634 "pyarrow/lib.pxd"
};

#line 634 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_CacheOptions *__pyx_vtabptr_7pyarrow_3lib_CacheOptions;

#line 634 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":646
 * 
 * 
 * cdef class Codec(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CCodec] wrapped
*/

#line 646 "pyarrow/lib.pxd"


#line 646 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_Codec {

#line 646 "pyarrow/lib.pxd"
   arrow::util::Codec *(*unwrap)(struct __pyx_obj_7pyarrow_3lib_Codec *);

#line 646 "pyarrow/lib.pxd"
};

#line 646 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_Codec *__pyx_vtabptr_7pyarrow_3lib_Codec;

#line 646 "pyarrow/lib.pxd"


/* "pyarrow/lib.pxd":654
 * 
 * # This class is only used internally for now
 * cdef class StopToken:             # <<<<<<<<<<<<<<
 *     cdef:
 *         CStopToken stop_token
*/

#line 654 "pyarrow/lib.pxd"


#line 654 "pyarrow/lib.pxd"
struct __pyx_vtabstruct_7pyarrow_3lib_StopToken {

#line 654 "pyarrow/lib.pxd"
  void (*init)(struct __pyx_obj_7pyarrow_3lib_StopToken *, arrow::StopToken);

#line 654 "pyarrow/lib.pxd"
};

#line 654 "pyarrow/lib.pxd"
static struct __pyx_vtabstruct_7pyarrow_3lib_StopToken *__pyx_vtabptr_7pyarrow_3lib_StopToken;

#line 654 "pyarrow/lib.pxd"


/* "pyarrow/_parquet.pxd":25
 * 
 * 
 * cdef class FileEncryptionProperties:             # <<<<<<<<<<<<<<
 *     """File-level encryption properties for the low-level API"""
 *     cdef:
*/

#line 25 "pyarrow/_parquet.pxd"


#line 25 "pyarrow/_parquet.pxd"
struct __pyx_vtabstruct_7pyarrow_8_parquet_FileEncryptionProperties {

#line 25 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *(*wrap)(std::shared_ptr< parquet::FileEncryptionProperties> );

#line 25 "pyarrow/_parquet.pxd"
  std::shared_ptr< parquet::FileEncryptionProperties>  (*unwrap)(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *);

#line 25 "pyarrow/_parquet.pxd"
};

#line 25 "pyarrow/_parquet.pxd"
static struct __pyx_vtabstruct_7pyarrow_8_parquet_FileEncryptionProperties *__pyx_vtabptr_7pyarrow_8_parquet_FileEncryptionProperties;

#line 25 "pyarrow/_parquet.pxd"
static CYTHON_INLINE struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *__pyx_f_7pyarrow_8_parquet_24FileEncryptionProperties_wrap(std::shared_ptr< parquet::FileEncryptionProperties> );

#line 25 "pyarrow/_parquet.pxd"
static CYTHON_INLINE std::shared_ptr< parquet::FileEncryptionProperties>  __pyx_f_7pyarrow_8_parquet_24FileEncryptionProperties_unwrap(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *);

#line 25 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":83
 *         const SchemaDescriptor* schema
 * 
 * cdef class FileMetaData(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CFileMetaData] sp_metadata
*/

#line 83 "pyarrow/_parquet.pxd"


#line 83 "pyarrow/_parquet.pxd"
struct __pyx_vtabstruct_7pyarrow_8_parquet_FileMetaData {

#line 83 "pyarrow/_parquet.pxd"
  PyObject *(*init)(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *, std::shared_ptr< parquet::FileMetaData>  const &);

#line 83 "pyarrow/_parquet.pxd"
};

#line 83 "pyarrow/_parquet.pxd"
static struct __pyx_vtabstruct_7pyarrow_8_parquet_FileMetaData *__pyx_vtabptr_7pyarrow_8_parquet_FileMetaData;

#line 83 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_12FileMetaData_init(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *, std::shared_ptr< parquet::FileMetaData>  const &);

#line 83 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":93
 *         self._metadata = metadata.get()
 * 
 * cdef class RowGroupMetaData(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int index  # for pickling support
*/

#line 93 "pyarrow/_parquet.pxd"


#line 93 "pyarrow/_parquet.pxd"
struct __pyx_vtabstruct_7pyarrow_8_parquet_RowGroupMetaData {

#line 93 "pyarrow/_parquet.pxd"
  PyObject *(*init)(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData *, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *, int);

#line 93 "pyarrow/_parquet.pxd"
};

#line 93 "pyarrow/_parquet.pxd"
static struct __pyx_vtabstruct_7pyarrow_8_parquet_RowGroupMetaData *__pyx_vtabptr_7pyarrow_8_parquet_RowGroupMetaData;

#line 93 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_16RowGroupMetaData_init(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData *, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *, int);

#line 93 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":109
 * 
 * 
 * cdef class ColumnChunkMetaData(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         unique_ptr[CColumnChunkMetaData] up_metadata
*/

#line 109 "pyarrow/_parquet.pxd"


#line 109 "pyarrow/_parquet.pxd"
struct __pyx_vtabstruct_7pyarrow_8_parquet_ColumnChunkMetaData {

#line 109 "pyarrow/_parquet.pxd"
  PyObject *(*init)(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *, struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData *, int);

#line 109 "pyarrow/_parquet.pxd"
};

#line 109 "pyarrow/_parquet.pxd"
static struct __pyx_vtabstruct_7pyarrow_8_parquet_ColumnChunkMetaData *__pyx_vtabptr_7pyarrow_8_parquet_ColumnChunkMetaData;

#line 109 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_19ColumnChunkMetaData_init(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *, struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData *, int);

#line 109 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":120
 *         self.parent = parent
 * 
 * cdef class Statistics(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CStatistics] statistics
*/

#line 120 "pyarrow/_parquet.pxd"


#line 120 "pyarrow/_parquet.pxd"
struct __pyx_vtabstruct_7pyarrow_8_parquet_Statistics {

#line 120 "pyarrow/_parquet.pxd"
  PyObject *(*init)(struct __pyx_obj_7pyarrow_8_parquet_Statistics *, std::shared_ptr< parquet::Statistics>  const &, struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *);

#line 120 "pyarrow/_parquet.pxd"
};

#line 120 "pyarrow/_parquet.pxd"
static struct __pyx_vtabstruct_7pyarrow_8_parquet_Statistics *__pyx_vtabptr_7pyarrow_8_parquet_Statistics;

#line 120 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_10Statistics_init(struct __pyx_obj_7pyarrow_8_parquet_Statistics *, std::shared_ptr< parquet::Statistics>  const &, struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *);

#line 120 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":130
 *         self.parent = parent
 * 
 * cdef class GeoStatistics(_Weakrefable):             # <<<<<<<<<<<<<<
 *     cdef:
 *         shared_ptr[CParquetGeoStatistics] statistics
*/

#line 130 "pyarrow/_parquet.pxd"


#line 130 "pyarrow/_parquet.pxd"
struct __pyx_vtabstruct_7pyarrow_8_parquet_GeoStatistics {

#line 130 "pyarrow/_parquet.pxd"
  PyObject *(*init)(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics *, std::shared_ptr< parquet::geospatial::GeoStatistics>  const &, struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *);

#line 130 "pyarrow/_parquet.pxd"
};

#line 130 "pyarrow/_parquet.pxd"
static struct __pyx_vtabstruct_7pyarrow_8_parquet_GeoStatistics *__pyx_vtabptr_7pyarrow_8_parquet_GeoStatistics;

#line 130 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_13GeoStatistics_init(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics *, std::shared_ptr< parquet::geospatial::GeoStatistics>  const &, struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *);

#line 130 "pyarrow/_parquet.pxd"


/* "pyarrow/_parquet.pxd":140
 *         self.parent = parent
 * 
 * cdef class FileDecryptionProperties:             # <<<<<<<<<<<<<<
 *     """File-level decryption properties for the low-level API"""
 *     cdef:
*/

#line 140 "pyarrow/_parquet.pxd"


#line 140 "pyarrow/_parquet.pxd"
struct __pyx_vtabstruct_7pyarrow_8_parquet_FileDecryptionProperties {

#line 140 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *(*wrap)(std::shared_ptr< parquet::FileDecryptionProperties> );

#line 140 "pyarrow/_parquet.pxd"
  std::shared_ptr< parquet::FileDecryptionProperties>  (*unwrap)(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *);

#line 140 "pyarrow/_parquet.pxd"
};

#line 140 "pyarrow/_parquet.pxd"
static struct __pyx_vtabstruct_7pyarrow_8_parquet_FileDecryptionProperties *__pyx_vtabptr_7pyarrow_8_parquet_FileDecryptionProperties;

#line 140 "pyarrow/_parquet.pxd"
static CYTHON_INLINE struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *__pyx_f_7pyarrow_8_parquet_24FileDecryptionProperties_wrap(std::shared_ptr< parquet::FileDecryptionProperties> );

#line 140 "pyarrow/_parquet.pxd"
static CYTHON_INLINE std::shared_ptr< parquet::FileDecryptionProperties>  __pyx_f_7pyarrow_8_parquet_24FileDecryptionProperties_unwrap(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *);
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyImportError_Check.proto */
#define __Pyx_PyExc_ImportError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ImportError)

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyIndexError_Check.proto */
#define __Pyx_PyExc_IndexError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_IndexError)

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* MoveIfSupported.proto */
#if CYTHON_USE_CPP_STD_MOVE
  #include <utility>
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) std::move(x)
#else
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) x
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_LIMITED_API  ||  PY_VERSION_HEX >= 0x030C0000
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  #if __clang__ || __GNUC__
  __attribute__((no_sanitize("thread")))
  #endif
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* PyAssertionError_Check.proto */
#define __Pyx_PyExc_AssertionError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_AssertionError)

/* BuildPyUnicode.proto (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* COrdinalToPyUnicode.proto (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value);
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t width, char padding_char);

/* GCCDiagnostics.proto (used by CIntToPyUnicode) */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* IncludeStdlibH.proto (used by CIntToPyUnicode) */
#include <stdlib.h>

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_Py_ssize_t(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_Py_ssize_t(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_Py_ssize_t(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_npy_intp(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_npy_intp(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_npy_intp(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_npy_intp(npy_intp value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_npy_intp(npy_intp value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* IterFinish.proto (used by dict_iter) */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by dict_iter) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto (used by UnpackTuple2) */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto (used by UnpackItemEndCheck) */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto (used by UnpackTuple2) */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto (used by UnpackTupleError) */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto (used by UnpackTuple2) */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto (used by dict_iter) */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
    PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple);
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyValueError_Check.proto */
#define __Pyx_PyExc_ValueError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ValueError)

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_int(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_int(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_int(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto (used by FetchCommonType) */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* CallTypeTraverse.proto (used by CythonFunctionShared) */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* LengthHint.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_PyObject_LengthHint(o, defaultval)  (defaultval)
#else
#define __Pyx_PyObject_LengthHint(o, defaultval)  PyObject_LengthHint(o, defaultval)
#endif

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_long__double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_long__double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_long__double(a, b) ((a)-(b))
    #define __Pyx_c_prod_long__double(a, b) ((a)*(b))
    #define __Pyx_c_quot_long__double(a, b) ((a)/(b))
    #define __Pyx_c_neg_long__double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_long__double(z) ((z)==(long double)0)
    #define __Pyx_c_conj_long__double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (::std::abs(z))
        #define __Pyx_c_pow_long__double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_long__double(z) ((z)==0)
    #define __Pyx_c_conj_long__double(z)    (conjl(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (cabsl(z))
        #define __Pyx_c_pow_long__double(a, b)  (cpowl(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex);
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex);
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    #endif
#endif

/* PyObjectVectorCallKwBuilder.proto (used by CIntToPy) */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_npy_intp(npy_intp value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self); /* proto*/
#if !CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
#endif
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_4year_year(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_5month_month(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_3day_day(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4hour_hour(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6minute_minute(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6second_second(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_11microsecond_microsecond(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_4time_6tzinfo_tzinfo(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4fold_fold(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4year_year(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_5month_month(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_3day_day(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4hour_hour(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6minute_minute(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6second_second(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_11microsecond_microsecond(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_8datetime_6tzinfo_tzinfo(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4fold_fold(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_3day_day(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_6second_second(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_11microsecond_microsecond(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *__pyx_f_7pyarrow_8_parquet_24FileEncryptionProperties_wrap(std::shared_ptr< parquet::FileEncryptionProperties>  __pyx_v_properties); /* proto*/
static CYTHON_INLINE std::shared_ptr< parquet::FileEncryptionProperties>  __pyx_f_7pyarrow_8_parquet_24FileEncryptionProperties_unwrap(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_12FileMetaData_init(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_self, std::shared_ptr< parquet::FileMetaData>  const &__pyx_v_metadata); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_16RowGroupMetaData_init(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData *__pyx_v_self, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_parent, int __pyx_v_index); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_19ColumnChunkMetaData_init(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *__pyx_v_self, struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData *__pyx_v_parent, int __pyx_v_i); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_10Statistics_init(struct __pyx_obj_7pyarrow_8_parquet_Statistics *__pyx_v_self, std::shared_ptr< parquet::Statistics>  const &__pyx_v_statistics, struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *__pyx_v_parent); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_13GeoStatistics_init(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics *__pyx_v_self, std::shared_ptr< parquet::geospatial::GeoStatistics>  const &__pyx_v_statistics, struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *__pyx_v_parent); /* proto*/
static CYTHON_INLINE struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *__pyx_f_7pyarrow_8_parquet_24FileDecryptionProperties_wrap(std::shared_ptr< parquet::FileDecryptionProperties>  __pyx_v_properties); /* proto*/
static CYTHON_INLINE std::shared_ptr< parquet::FileDecryptionProperties>  __pyx_f_7pyarrow_8_parquet_24FileDecryptionProperties_unwrap(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *__pyx_v_self); /* proto*/

/* Module declarations from "cython" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython.version" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "numpy" */

/* Module declarations from "numpy" */

/* Module declarations from "jollyjack" */

/* Module declarations from "libcpp.vector" */

/* Module declarations from "libcpp.string_view" */

/* Module declarations from "libcpp.string" */

/* Module declarations from "libcpp" */

/* Module declarations from "libcpp.memory" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "libcpp.functional" */

/* Module declarations from "libcpp.optional" */

/* Module declarations from "libcpp.utility" */

/* Module declarations from "libcpp.unordered_map" */

/* Module declarations from "libcpp.unordered_set" */

/* Module declarations from "datetime" */

/* Module declarations from "cpython.datetime" */

/* Module declarations from "pyarrow.includes.common" */

/* Module declarations from "pyarrow.includes.libarrow" */

/* Module declarations from "pyarrow.includes.libparquet" */

/* Module declarations from "cpython.slice" */

/* Module declarations from "libcpp.cast" */

/* Module declarations from "pyarrow.includes.libarrow_python" */

/* Module declarations from "pyarrow.lib" */
static PyObject *(*__pyx_f_7pyarrow_3lib_get_reader)(PyObject *, bool, std::shared_ptr< arrow::io::RandomAccessFile>  *); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_buffer)(std::shared_ptr< arrow::Buffer>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_resizable_buffer)(std::shared_ptr< arrow::ResizableBuffer>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_data_type)(std::shared_ptr< arrow::DataType>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_field)(std::shared_ptr< arrow::Field>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_schema)(std::shared_ptr< arrow::Schema>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_scalar)(std::shared_ptr< arrow::Scalar>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_array)(std::shared_ptr< arrow::Array>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_chunked_array)(std::shared_ptr< arrow::ChunkedArray>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_sparse_coo_tensor)(std::shared_ptr< arrow::SparseCOOTensor>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_sparse_csc_matrix)(std::shared_ptr< arrow::SparseCSCMatrix>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_sparse_csf_tensor)(std::shared_ptr< arrow::SparseCSFTensor>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_sparse_csr_matrix)(std::shared_ptr< arrow::SparseCSRMatrix>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_tensor)(std::shared_ptr< arrow::Tensor>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_batch)(std::shared_ptr< arrow::RecordBatch>  const &); /*proto*/
static PyObject *(*__pyx_f_7pyarrow_3lib_pyarrow_wrap_table)(std::shared_ptr< arrow::Table>  const &); /*proto*/
static std::shared_ptr< arrow::Buffer>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_buffer)(PyObject *); /*proto*/
static std::shared_ptr< arrow::DataType>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_data_type)(PyObject *); /*proto*/
static std::shared_ptr< arrow::Field>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_field)(PyObject *); /*proto*/
static std::shared_ptr< arrow::Schema>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_schema)(PyObject *); /*proto*/
static std::shared_ptr< arrow::Scalar>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_scalar)(PyObject *); /*proto*/
static std::shared_ptr< arrow::Array>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_array)(PyObject *); /*proto*/
static std::shared_ptr< arrow::ChunkedArray>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_chunked_array)(PyObject *); /*proto*/
static std::shared_ptr< arrow::SparseCOOTensor>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_sparse_coo_tensor)(PyObject *); /*proto*/
static std::shared_ptr< arrow::SparseCSCMatrix>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_sparse_csc_matrix)(PyObject *); /*proto*/
static std::shared_ptr< arrow::SparseCSFTensor>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_sparse_csf_tensor)(PyObject *); /*proto*/
static std::shared_ptr< arrow::SparseCSRMatrix>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_sparse_csr_matrix)(PyObject *); /*proto*/
static std::shared_ptr< arrow::Tensor>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_tensor)(PyObject *); /*proto*/
static std::shared_ptr< arrow::RecordBatch>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_batch)(PyObject *); /*proto*/
static std::shared_ptr< arrow::Table>  (*__pyx_f_7pyarrow_3lib_pyarrow_unwrap_table)(PyObject *); /*proto*/

/* Module declarations from "pyarrow._parquet" */

/* Module declarations from "jollyjack.cjollyjack" */

/* Module declarations from "jollyjack.jollyjack_cython" */
static void __pyx_f_9jollyjack_16jollyjack_cython_read_into_torch(PyObject *, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_torch *__pyx_optional_args); /*proto*/
static void __pyx_f_9jollyjack_16jollyjack_cython_read_into_numpy(PyObject *, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *, PyArrayObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_numpy *__pyx_optional_args); /*proto*/
static void __pyx_f_9jollyjack_16jollyjack_cython_copy_to_torch_row_major(PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_9jollyjack_16jollyjack_cython_copy_to_numpy_row_major(PyArrayObject *, PyArrayObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static std::string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(PyObject *); /*proto*/
static std::vector<int>  __pyx_convert_vector_from_py_int(PyObject *); /*proto*/
static std::vector<std::string>  __pyx_convert_vector_from_py_std_3a__3a_string(PyObject *); /*proto*/
static std::vector<int64_t>  __pyx_convert_vector_from_py_int64_t(PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "jollyjack.jollyjack_cython"
extern int __pyx_module_is_main_jollyjack__jollyjack_cython;
int __pyx_module_is_main_jollyjack__jollyjack_cython = 0;

/* Implementation of "jollyjack.jollyjack_cython" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_max;
/* #### Code section: string_decls ### */
/* #### Code section: decls ### */
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_is_iterable_of_iterables(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_2read_into_torch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_source, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_metadata, PyObject *__pyx_v_tensor, PyObject *__pyx_v_row_group_indices, PyObject *__pyx_v_row_ranges, PyObject *__pyx_v_column_indices, PyObject *__pyx_v_column_names, PyObject *__pyx_v_pre_buffer, PyObject *__pyx_v_use_threads, PyObject *__pyx_v_use_memory_map, struct __pyx_obj_7pyarrow_3lib_CacheOptions *__pyx_v_cache_options); /* proto */
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_4read_into_numpy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_source, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_metadata, PyArrayObject *__pyx_v_np_array, PyObject *__pyx_v_row_group_indices, PyObject *__pyx_v_row_ranges, PyObject *__pyx_v_column_indices, PyObject *__pyx_v_column_names, PyObject *__pyx_v_pre_buffer, PyObject *__pyx_v_use_threads, PyObject *__pyx_v_use_memory_map, struct __pyx_obj_7pyarrow_3lib_CacheOptions *__pyx_v_cache_options); /* proto */
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_6copy_to_torch_row_major(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src_tensor, PyObject *__pyx_v_dst_tensor, PyObject *__pyx_v_row_indices); /* proto */
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_8copy_to_numpy_row_major(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_src_array, PyArrayObject *__pyx_v_dst_array, PyObject *__pyx_v_row_indices); /* proto */
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  PyTypeObject *__pyx_ptype_5numpy_dtype;
  PyTypeObject *__pyx_ptype_5numpy_flatiter;
  PyTypeObject *__pyx_ptype_5numpy_broadcast;
  PyTypeObject *__pyx_ptype_5numpy_ndarray;
  PyTypeObject *__pyx_ptype_5numpy_generic;
  PyTypeObject *__pyx_ptype_5numpy_number;
  PyTypeObject *__pyx_ptype_5numpy_integer;
  PyTypeObject *__pyx_ptype_5numpy_signedinteger;
  PyTypeObject *__pyx_ptype_5numpy_unsignedinteger;
  PyTypeObject *__pyx_ptype_5numpy_inexact;
  PyTypeObject *__pyx_ptype_5numpy_floating;
  PyTypeObject *__pyx_ptype_5numpy_complexfloating;
  PyTypeObject *__pyx_ptype_5numpy_flexible;
  PyTypeObject *__pyx_ptype_5numpy_character;
  PyTypeObject *__pyx_ptype_5numpy_ufunc;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_date;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_time;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_datetime;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_timedelta;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_tzinfo;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib__Weakrefable;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_IpcWriteOptions;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_IpcReadOptions;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Message;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_MemoryPool;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_DataType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_ListType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_LargeListType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_ListViewType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_LargeListViewType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_MapType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_FixedSizeListType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_StructType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_DictionaryMemo;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_DictionaryType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_TimestampType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Time32Type;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Time64Type;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_DurationType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Decimal32Type;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Decimal64Type;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Decimal128Type;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Decimal256Type;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_RunEndEncodedType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_BaseExtensionType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_ExtensionType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_FixedShapeTensorType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Bool8Type;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_OpaqueType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_UuidType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_JsonType;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib__Metadata;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_KeyValueMetadata;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Field;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Schema;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Scalar;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_ArrayStatistics;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib__PandasConvertible;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Tensor;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_SparseCSRMatrix;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_SparseCSCMatrix;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_SparseCOOTensor;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_SparseCSFTensor;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_NullArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_BooleanArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_NumericArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_IntegerArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_FloatingPointArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Int8Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_UInt8Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Int16Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_UInt16Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Int32Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_UInt32Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Int64Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_UInt64Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_HalfFloatArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_FloatArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_DoubleArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Decimal32Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Decimal64Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Decimal128Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Decimal256Array;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_StructArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_BaseListArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_ListArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_LargeListArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_ListViewArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_LargeListViewArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_MapArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_FixedSizeListArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_UnionArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_StringArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_BinaryArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_StringViewArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_BinaryViewArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_DictionaryArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_ExtensionArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_MonthDayNanoIntervalArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_ChunkedArray;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib__Tabular;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Table;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_RecordBatch;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Device;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_MemoryManager;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Buffer;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_ResizableBuffer;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_NativeFile;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_BufferedInputStream;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_BufferedOutputStream;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_CompressedInputStream;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_CompressedOutputStream;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib__CRecordBatchWriter;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_RecordBatchReader;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_CacheOptions;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_Codec;
  PyTypeObject *__pyx_ptype_7pyarrow_3lib_StopToken;
  PyTypeObject *__pyx_ptype_7pyarrow_8_parquet_FileEncryptionProperties;
  PyTypeObject *__pyx_ptype_7pyarrow_8_parquet_ParquetSchema;
  PyTypeObject *__pyx_ptype_7pyarrow_8_parquet_FileMetaData;
  PyTypeObject *__pyx_ptype_7pyarrow_8_parquet_RowGroupMetaData;
  PyTypeObject *__pyx_ptype_7pyarrow_8_parquet_ColumnChunkMetaData;
  PyTypeObject *__pyx_ptype_7pyarrow_8_parquet_Statistics;
  PyTypeObject *__pyx_ptype_7pyarrow_8_parquet_GeoStatistics;
  PyTypeObject *__pyx_ptype_7pyarrow_8_parquet_FileDecryptionProperties;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  PyObject *__pyx_k_;
  PyObject *__pyx_k__2;
  PyObject *__pyx_k__3;
  PyObject *__pyx_k__4;
  PyObject *__pyx_k__5;
  PyObject *__pyx_k__6;
  PyObject *__pyx_codeobj_tab[5];
  PyObject *__pyx_string_tab[102];
  PyObject *__pyx_number_tab[2];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_0_out_of_bounds __pyx_string_tab[0]
#define __pyx_kp_u_2 __pyx_string_tab[1]
#define __pyx_kp_u_Either_column_indices_or_column __pyx_string_tab[2]
#define __pyx_kp_u_Expected_array_in_a_Fortran_colu __pyx_string_tab[3]
#define __pyx_kp_u_Expected_destination_array_in_a __pyx_string_tab[4]
#define __pyx_kp_u_Expected_source_array_in_a_Fortr __pyx_string_tab[5]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[6]
#define __pyx_kp_u_Requested_to_read __pyx_string_tab[7]
#define __pyx_kp_u_Row_index __pyx_string_tab[8]
#define __pyx_kp_u_Row_range __pyx_string_tab[9]
#define __pyx_kp_u_Source_and_destination_arrays_ha __pyx_string_tab[10]
#define __pyx_kp_u_Unexpected_dst_array_ndim __pyx_string_tab[11]
#define __pyx_kp_u_Unexpected_len_of_row_indices __pyx_string_tab[12]
#define __pyx_kp_u_Unexpected_np_array_ndim __pyx_string_tab[13]
#define __pyx_kp_u_Unexpected_src_array_ndim __pyx_string_tab[14]
#define __pyx_kp_u_Unsupported_JJ_READER_BACKEND __pyx_string_tab[15]
#define __pyx_kp_u__10 __pyx_string_tab[16]
#define __pyx_kp_u__11 __pyx_string_tab[17]
#define __pyx_kp_u__7 __pyx_string_tab[18]
#define __pyx_kp_u__8 __pyx_string_tab[19]
#define __pyx_kp_u__9 __pyx_string_tab[20]
#define __pyx_kp_u_add_note __pyx_string_tab[21]
#define __pyx_kp_u_columns_but_the_number_of_colum __pyx_string_tab[22]
#define __pyx_kp_u_is_not_a_valid_range __pyx_string_tab[23]
#define __pyx_kp_u_is_not_contiguous __pyx_string_tab[24]
#define __pyx_kp_u_is_not_in_the_expected_range_0 __pyx_string_tab[25]
#define __pyx_kp_u_jollyjack_jollyjack_cython_pyx __pyx_string_tab[26]
#define __pyx_kp_u_numpy__core_multiarray_failed_to __pyx_string_tab[27]
#define __pyx_kp_u_numpy__core_umath_failed_to_impo __pyx_string_tab[28]
#define __pyx_kp_u_src_array_shape_0_dst_array_shap __pyx_string_tab[29]
#define __pyx_kp_u_src_array_shape_1_dst_array_shap __pyx_string_tab[30]
#define __pyx_kp_u_utf_8 __pyx_string_tab[31]
#define __pyx_n_u_Iterable __pyx_string_tab[32]
#define __pyx_n_u_JJ_READER_BACKEND __pyx_string_tab[33]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[34]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[35]
#define __pyx_n_u_cache_options __pyx_string_tab[36]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[37]
#define __pyx_n_u_collections_abc __pyx_string_tab[38]
#define __pyx_n_u_column_indices __pyx_string_tab[39]
#define __pyx_n_u_column_names __pyx_string_tab[40]
#define __pyx_n_u_copy_to_numpy_row_major __pyx_string_tab[41]
#define __pyx_n_u_copy_to_torch_row_major __pyx_string_tab[42]
#define __pyx_n_u_dst_array __pyx_string_tab[43]
#define __pyx_n_u_dst_tensor __pyx_string_tab[44]
#define __pyx_n_u_dtype __pyx_string_tab[45]
#define __pyx_n_u_encode __pyx_string_tab[46]
#define __pyx_n_u_environ __pyx_string_tab[47]
#define __pyx_n_u_format __pyx_string_tab[48]
#define __pyx_n_u_func __pyx_string_tab[49]
#define __pyx_n_u_get __pyx_string_tab[50]
#define __pyx_n_u_io_uring __pyx_string_tab[51]
#define __pyx_n_u_io_uring_odirect __pyx_string_tab[52]
#define __pyx_n_u_is_coroutine __pyx_string_tab[53]
#define __pyx_n_u_is_iterable_of_iterables __pyx_string_tab[54]
#define __pyx_n_u_items __pyx_string_tab[55]
#define __pyx_n_u_jollyjack_jollyjack_cython __pyx_string_tab[56]
#define __pyx_n_u_keys __pyx_string_tab[57]
#define __pyx_n_u_main __pyx_string_tab[58]
#define __pyx_n_u_max __pyx_string_tab[59]
#define __pyx_n_u_metadata __pyx_string_tab[60]
#define __pyx_n_u_min __pyx_string_tab[61]
#define __pyx_n_u_module __pyx_string_tab[62]
#define __pyx_n_u_name __pyx_string_tab[63]
#define __pyx_n_u_np_array __pyx_string_tab[64]
#define __pyx_n_u_num_row_groups __pyx_string_tab[65]
#define __pyx_n_u_numpy __pyx_string_tab[66]
#define __pyx_n_u_obj __pyx_string_tab[67]
#define __pyx_n_u_os __pyx_string_tab[68]
#define __pyx_n_u_parquet __pyx_string_tab[69]
#define __pyx_n_u_pop __pyx_string_tab[70]
#define __pyx_n_u_pq __pyx_string_tab[71]
#define __pyx_n_u_pre_buffer __pyx_string_tab[72]
#define __pyx_n_u_pyarrow_parquet __pyx_string_tab[73]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[74]
#define __pyx_n_u_qualname __pyx_string_tab[75]
#define __pyx_n_u_read_into_numpy __pyx_string_tab[76]
#define __pyx_n_u_read_into_torch __pyx_string_tab[77]
#define __pyx_n_u_row_group_indices __pyx_string_tab[78]
#define __pyx_n_u_row_indices __pyx_string_tab[79]
#define __pyx_n_u_row_ranges __pyx_string_tab[80]
#define __pyx_n_u_set_name __pyx_string_tab[81]
#define __pyx_n_u_setdefault __pyx_string_tab[82]
#define __pyx_n_u_source __pyx_string_tab[83]
#define __pyx_n_u_src_array __pyx_string_tab[84]
#define __pyx_n_u_src_tensor __pyx_string_tab[85]
#define __pyx_n_u_start __pyx_string_tab[86]
#define __pyx_n_u_step __pyx_string_tab[87]
#define __pyx_n_u_stop __pyx_string_tab[88]
#define __pyx_n_u_tensor __pyx_string_tab[89]
#define __pyx_n_u_test __pyx_string_tab[90]
#define __pyx_n_u_torch __pyx_string_tab[91]
#define __pyx_n_u_use_memory_map __pyx_string_tab[92]
#define __pyx_n_u_use_threads __pyx_string_tab[93]
#define __pyx_n_u_utf8 __pyx_string_tab[94]
#define __pyx_n_u_values __pyx_string_tab[95]
#define __pyx_kp_b_PyObject_PyObject_bool_std_share __pyx_string_tab[96]
#define __pyx_kp_b_iso88591_M_6_a __pyx_string_tab[97]
#define __pyx_kp_b_iso88591_Qe_T_1Cq_Jd_jXYY_aab __pyx_string_tab[98]
#define __pyx_kp_b_iso88591_jj_S_S_d_d_y_y_P_P_t_t_u_Ja_6 __pyx_string_tab[99]
#define __pyx_kp_b_iso88591_x_y_N_N_a_a_r_r_G_G_B_B_C_1_1_x __pyx_string_tab[100]
#define __pyx_kp_b_iso88591_y_as_Q_9F_S_Qiq_9F_S_Qiq_9F_3c __pyx_string_tab[101]
#define __pyx_int_0 __pyx_number_tab[0]
#define __pyx_int_1 __pyx_number_tab[1]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_dtype);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flatiter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_broadcast);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ndarray);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_generic);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_number);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_integer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_inexact);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_floating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flexible);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_character);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ufunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_date);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_time);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_datetime);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_timedelta);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_tzinfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib__Weakrefable);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_IpcWriteOptions);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_IpcReadOptions);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Message);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_MemoryPool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_DataType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_ListType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_LargeListType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_ListViewType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_LargeListViewType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_MapType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_FixedSizeListType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_StructType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_DictionaryMemo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_DictionaryType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_TimestampType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Time32Type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Time64Type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_DurationType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Decimal32Type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Decimal64Type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Decimal128Type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Decimal256Type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_RunEndEncodedType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_BaseExtensionType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_ExtensionType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_FixedShapeTensorType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Bool8Type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_OpaqueType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_UuidType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_JsonType);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib__Metadata);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_KeyValueMetadata);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Field);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Schema);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Scalar);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_ArrayStatistics);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib__PandasConvertible);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Tensor);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_SparseCSRMatrix);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_SparseCSCMatrix);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_SparseCOOTensor);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_SparseCSFTensor);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_NullArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_BooleanArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_NumericArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_IntegerArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_FloatingPointArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Int8Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_UInt8Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Int16Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_UInt16Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Int32Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_UInt32Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Int64Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_UInt64Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_HalfFloatArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_FloatArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_DoubleArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Decimal32Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Decimal64Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Decimal128Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Decimal256Array);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_StructArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_BaseListArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_ListArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_LargeListArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_ListViewArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_LargeListViewArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_MapArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_FixedSizeListArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_UnionArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_StringArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_BinaryArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_StringViewArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_BinaryViewArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_DictionaryArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_ExtensionArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_MonthDayNanoIntervalArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_ChunkedArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib__Tabular);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Table);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_RecordBatch);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Device);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_MemoryManager);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Buffer);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_ResizableBuffer);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_NativeFile);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_BufferedInputStream);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_BufferedOutputStream);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_CompressedInputStream);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_CompressedOutputStream);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib__CRecordBatchWriter);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_RecordBatchReader);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_CacheOptions);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_Codec);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_3lib_StopToken);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_8_parquet_FileEncryptionProperties);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_8_parquet_ParquetSchema);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_8_parquet_FileMetaData);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_8_parquet_RowGroupMetaData);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_8_parquet_ColumnChunkMetaData);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_8_parquet_Statistics);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_8_parquet_GeoStatistics);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyarrow_8_parquet_FileDecryptionProperties);
  Py_CLEAR(clear_module_state->__pyx_k_);
  Py_CLEAR(clear_module_state->__pyx_k__2);
  Py_CLEAR(clear_module_state->__pyx_k__3);
  Py_CLEAR(clear_module_state->__pyx_k__4);
  Py_CLEAR(clear_module_state->__pyx_k__5);
  Py_CLEAR(clear_module_state->__pyx_k__6);
  for (int i=0; i<5; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<102; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<2; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_dtype);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flatiter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_broadcast);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ndarray);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_generic);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_number);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_integer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_inexact);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_floating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flexible);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_character);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ufunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_date);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_time);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_datetime);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_timedelta);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_tzinfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib__Weakrefable);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_IpcWriteOptions);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_IpcReadOptions);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Message);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_MemoryPool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_DataType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_ListType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_LargeListType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_ListViewType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_LargeListViewType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_MapType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_FixedSizeListType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_StructType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_DictionaryMemo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_DictionaryType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_TimestampType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Time32Type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Time64Type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_DurationType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Decimal32Type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Decimal64Type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Decimal128Type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Decimal256Type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_RunEndEncodedType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_BaseExtensionType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_ExtensionType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_FixedShapeTensorType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Bool8Type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_OpaqueType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_UuidType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_JsonType);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib__Metadata);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_KeyValueMetadata);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Field);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Schema);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Scalar);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_ArrayStatistics);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib__PandasConvertible);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Tensor);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_SparseCSRMatrix);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_SparseCSCMatrix);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_SparseCOOTensor);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_SparseCSFTensor);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_NullArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_BooleanArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_NumericArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_IntegerArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_FloatingPointArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Int8Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_UInt8Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Int16Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_UInt16Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Int32Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_UInt32Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Int64Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_UInt64Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_HalfFloatArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_FloatArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_DoubleArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Decimal32Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Decimal64Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Decimal128Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Decimal256Array);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_StructArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_BaseListArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_ListArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_LargeListArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_ListViewArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_LargeListViewArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_MapArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_FixedSizeListArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_UnionArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_StringArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_BinaryArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_StringViewArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_BinaryViewArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_DictionaryArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_ExtensionArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_MonthDayNanoIntervalArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_ChunkedArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib__Tabular);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Table);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_RecordBatch);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Device);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_MemoryManager);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Buffer);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_ResizableBuffer);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_NativeFile);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_BufferedInputStream);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_BufferedOutputStream);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_CompressedInputStream);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_CompressedOutputStream);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib__CRecordBatchWriter);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_RecordBatchReader);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_CacheOptions);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_Codec);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_3lib_StopToken);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_8_parquet_FileEncryptionProperties);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_8_parquet_ParquetSchema);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_8_parquet_FileMetaData);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_8_parquet_RowGroupMetaData);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_8_parquet_ColumnChunkMetaData);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_8_parquet_Statistics);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_8_parquet_GeoStatistics);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyarrow_8_parquet_FileDecryptionProperties);
  Py_VISIT(traverse_module_state->__pyx_k_);
  Py_VISIT(traverse_module_state->__pyx_k__2);
  Py_VISIT(traverse_module_state->__pyx_k__3);
  Py_VISIT(traverse_module_state->__pyx_k__4);
  Py_VISIT(traverse_module_state->__pyx_k__5);
  Py_VISIT(traverse_module_state->__pyx_k__6);
  for (int i=0; i<5; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<102; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<2; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "string.from_py":12
 *     cdef const char* __Pyx_PyObject_AsStringAndSize(object, Py_ssize_t*) except NULL
 * 
 * @cname("__pyx_convert_string_from_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0
*/

#line 12 "string.from_py"


#line 12 "string.from_py"
static std::string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(PyObject *__pyx_v_o) {
  Py_ssize_t __pyx_v_length
#line 12 "string.from_py"
;
  char const *__pyx_v_data
#line 12 "string.from_py"
;

#line 12 "string.from_py"
  std::string __pyx_r;

#line 12 "string.from_py"
  char const *__pyx_t_1;

#line 12 "string.from_py"
  std::string __pyx_t_2;

#line 12 "string.from_py"
  int __pyx_lineno = 0;

#line 12 "string.from_py"
  const char *__pyx_filename = NULL;

#line 12 "string.from_py"
  int __pyx_clineno = 0;

  /* "string.from_py":14
 * @cname("__pyx_convert_string_from_py_6libcpp_6string_std__in_string")
 * cdef string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0             # <<<<<<<<<<<<<<
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, <size_t> length)
*/

#line 14 "string.from_py"
  __pyx_v_length = 0;

  /* "string.from_py":15
 * cdef string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)             # <<<<<<<<<<<<<<
 *     return string(data, <size_t> length)
 * 
*/

#line 15 "string.from_py"
  __pyx_t_1 = __Pyx_PyObject_AsStringAndSize(__pyx_v_o, (&__pyx_v_length)); if (unlikely(__pyx_t_1 == ((void *)NULL))) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "string.from_py"
  __pyx_v_data = __pyx_t_1;

  /* "string.from_py":16
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, <size_t> length)             # <<<<<<<<<<<<<<
 * 
*/

#line 16 "string.from_py"
  try {

#line 16 "string.from_py"
    __pyx_t_2 = std::string(__pyx_v_data, ((size_t)__pyx_v_length));

#line 16 "string.from_py"
  } catch(...) {

#line 16 "string.from_py"
    __Pyx_CppExn2PyErr();

#line 16 "string.from_py"
    __PYX_ERR(1, 16, __pyx_L1_error)

#line 16 "string.from_py"
  }

#line 16 "string.from_py"
  __pyx_r = __pyx_t_2;

#line 16 "string.from_py"
  goto __pyx_L0;

  /* "string.from_py":12
 *     cdef const char* __Pyx_PyObject_AsStringAndSize(object, Py_ssize_t*) except NULL
 * 
 * @cname("__pyx_convert_string_from_py_6libcpp_6string_std__in_string")             # <<<<<<<<<<<<<<
 * cdef string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0
*/

#line 12 "string.from_py"


#line 12 "string.from_py"
  /* function exit code */

#line 12 "string.from_py"
  __pyx_L1_error:;

#line 12 "string.from_py"
  __Pyx_AddTraceback("string.from_py.__pyx_convert_string_from_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 12 "string.from_py"
  __Pyx_pretend_to_initialize(&__pyx_r);

#line 12 "string.from_py"
  __pyx_L0:;

#line 12 "string.from_py"
  return __pyx_r;

#line 12 "string.from_py"
}

/* "vector.from_py":51
 *     cdef Py_ssize_t __Pyx_PyObject_LengthHint(object o, Py_ssize_t defaultval) except -1
 * 
 * @cname("__pyx_convert_vector_from_py_int")             # <<<<<<<<<<<<<<
 * cdef vector[X] __pyx_convert_vector_from_py_int(object o) except *:
 * 
*/

#line 51 "vector.from_py"


#line 51 "vector.from_py"
static std::vector<int>  __pyx_convert_vector_from_py_int(PyObject *__pyx_v_o) {
  std::vector<int>  __pyx_v_v
#line 51 "vector.from_py"
;
  Py_ssize_t __pyx_v_s
#line 51 "vector.from_py"
;
  PyObject *__pyx_v_item = NULL
#line 51 "vector.from_py"
;

#line 51 "vector.from_py"
  std::vector<int>  __pyx_r;

#line 51 "vector.from_py"
  __Pyx_RefNannyDeclarations

#line 51 "vector.from_py"
  Py_ssize_t __pyx_t_1;

#line 51 "vector.from_py"
  int __pyx_t_2;

#line 51 "vector.from_py"
  PyObject *__pyx_t_3 = NULL;

#line 51 "vector.from_py"
  PyObject *(*__pyx_t_4)(PyObject *);

#line 51 "vector.from_py"
  PyObject *__pyx_t_5 = NULL;

#line 51 "vector.from_py"
  int __pyx_t_6;

#line 51 "vector.from_py"
  int __pyx_lineno = 0;

#line 51 "vector.from_py"
  const char *__pyx_filename = NULL;

#line 51 "vector.from_py"
  int __pyx_clineno = 0;

#line 51 "vector.from_py"
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_int", 0);

  /* "vector.from_py":55
 * 
 *     cdef vector[X] v
 *     cdef Py_ssize_t s = __Pyx_PyObject_LengthHint(o, 0)             # <<<<<<<<<<<<<<
 * 
 *     if s > 0:
*/

#line 55 "vector.from_py"
  __pyx_t_1 = __Pyx_PyObject_LengthHint(__pyx_v_o, 0); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(1, 55, __pyx_L1_error)

#line 55 "vector.from_py"
  __pyx_v_s = __pyx_t_1;

  /* "vector.from_py":57
 *     cdef Py_ssize_t s = __Pyx_PyObject_LengthHint(o, 0)
 * 
 *     if s > 0:             # <<<<<<<<<<<<<<
 *         v.reserve(<size_t> s)
 * 
*/

#line 57 "vector.from_py"
  __pyx_t_2 = (__pyx_v_s > 0);

#line 57 "vector.from_py"
  if (__pyx_t_2) {

    /* "vector.from_py":58
 * 
 *     if s > 0:
 *         v.reserve(<size_t> s)             # <<<<<<<<<<<<<<
 * 
 *     for item in o:
*/

#line 58 "vector.from_py"
    try {

#line 58 "vector.from_py"
      __pyx_v_v.reserve(((size_t)__pyx_v_s));

#line 58 "vector.from_py"
    } catch(...) {

#line 58 "vector.from_py"
      __Pyx_CppExn2PyErr();

#line 58 "vector.from_py"
      __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "vector.from_py"
    }

    /* "vector.from_py":57
 *     cdef Py_ssize_t s = __Pyx_PyObject_LengthHint(o, 0)
 * 
 *     if s > 0:             # <<<<<<<<<<<<<<
 *         v.reserve(<size_t> s)
 * 
*/

#line 57 "vector.from_py"
  }

  /* "vector.from_py":60
 *         v.reserve(<size_t> s)
 * 
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 * 
*/

#line 60 "vector.from_py"
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {

#line 60 "vector.from_py"
    __pyx_t_3 = __pyx_v_o; __Pyx_INCREF(__pyx_t_3);

#line 60 "vector.from_py"
    __pyx_t_1 = 0;

#line 60 "vector.from_py"
    __pyx_t_4 = NULL;

#line 60 "vector.from_py"
  } else {
    __pyx_t_1 = -1; 
#line 60 "vector.from_py"
__pyx_t_3 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); 
#line 60 "vector.from_py"
if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
  }
  for (;;
#line 60 "vector.from_py"
) {

#line 60 "vector.from_py"
    if (likely(!__pyx_t_4)) {

#line 60 "vector.from_py"
      if (likely(PyList_CheckExact(__pyx_t_3))) {

#line 60 "vector.from_py"
        {

#line 60 "vector.from_py"
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);

#line 60 "vector.from_py"
          #if !CYTHON_ASSUME_SAFE_SIZE

#line 60 "vector.from_py"
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
          #endif

#line 60 "vector.from_py"
          if (__pyx_t_1 >= __pyx_temp) break;

#line 60 "vector.from_py"
        }

#line 60 "vector.from_py"
        __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_1, __Pyx_ReferenceSharing_OwnStrongReference);

#line 60 "vector.from_py"
        ++__pyx_t_1;

#line 60 "vector.from_py"
      } else {

#line 60 "vector.from_py"
        {

#line 60 "vector.from_py"
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);

#line 60 "vector.from_py"
          #if !CYTHON_ASSUME_SAFE_SIZE

#line 60 "vector.from_py"
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
          #endif

#line 60 "vector.from_py"
          if (__pyx_t_1 >= __pyx_temp) break;

#line 60 "vector.from_py"
        }

#line 60 "vector.from_py"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 60 "vector.from_py"
        __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1));

#line 60 "vector.from_py"
        #else

#line 60 "vector.from_py"
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_1);

#line 60 "vector.from_py"
        #endif

#line 60 "vector.from_py"
        ++__pyx_t_1;

#line 60 "vector.from_py"
      }

#line 60 "vector.from_py"
      if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 60, __pyx_L1_error)
    } else 
#line 60 "vector.from_py"
{

#line 60 "vector.from_py"
      __pyx_t_5 = __pyx_t_4(__pyx_t_3);

#line 60 "vector.from_py"
      if (unlikely(!__pyx_t_5)) {

#line 60 "vector.from_py"
        PyObject* exc_type = PyErr_Occurred();

#line 60 "vector.from_py"
        if (exc_type) {

#line 60 "vector.from_py"
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
          PyErr_Clear();

#line 60 "vector.from_py"
        }

#line 60 "vector.from_py"
        break;

#line 60 "vector.from_py"
      }

#line 60 "vector.from_py"
    }

#line 60 "vector.from_py"
    __Pyx_GOTREF(__pyx_t_5);

#line 60 "vector.from_py"
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_5);

#line 60 "vector.from_py"
    __pyx_t_5 = 0;

    /* "vector.from_py":61
 * 
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 * 
 *     return v
*/

#line 61 "vector.from_py"
    __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_item); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 61, __pyx_L1_error)

#line 61 "vector.from_py"
    try {

#line 61 "vector.from_py"
      __pyx_v_v.push_back(((int)__pyx_t_6));

#line 61 "vector.from_py"
    } catch(...) {

#line 61 "vector.from_py"
      __Pyx_CppExn2PyErr();

#line 61 "vector.from_py"
      __PYX_ERR(1, 61, __pyx_L1_error)

#line 61 "vector.from_py"
    }

    /* "vector.from_py":60
 *         v.reserve(<size_t> s)
 * 
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 * 
*/

#line 60 "vector.from_py"
  }

#line 60 "vector.from_py"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "vector.from_py":63
 *         v.push_back(<X>item)
 * 
 *     return v             # <<<<<<<<<<<<<<
 * 
*/

#line 63 "vector.from_py"
  __pyx_r = __pyx_v_v;

#line 63 "vector.from_py"
  goto __pyx_L0;

  /* "vector.from_py":51
 *     cdef Py_ssize_t __Pyx_PyObject_LengthHint(object o, Py_ssize_t defaultval) except -1
 * 
 * @cname("__pyx_convert_vector_from_py_int")             # <<<<<<<<<<<<<<
 * cdef vector[X] __pyx_convert_vector_from_py_int(object o) except *:
 * 
*/

#line 51 "vector.from_py"


#line 51 "vector.from_py"
  /* function exit code */

#line 51 "vector.from_py"
  __pyx_L1_error:;

#line 51 "vector.from_py"
  __Pyx_XDECREF(__pyx_t_3);

#line 51 "vector.from_py"
  __Pyx_XDECREF(__pyx_t_5);

#line 51 "vector.from_py"
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_int", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 51 "vector.from_py"
  __Pyx_pretend_to_initialize(&__pyx_r);

#line 51 "vector.from_py"
  __pyx_L0:;

#line 51 "vector.from_py"
  __Pyx_XDECREF(__pyx_v_item);

#line 51 "vector.from_py"
  __Pyx_RefNannyFinishContext();

#line 51 "vector.from_py"
  return __pyx_r;

#line 51 "vector.from_py"
}

#line 51 "vector.from_py"


#line 51 "vector.from_py"
static std::vector<std::string>  __pyx_convert_vector_from_py_std_3a__3a_string(PyObject *__pyx_v_o) {
  std::vector<std::string>  __pyx_v_v
#line 51 "vector.from_py"
;
  Py_ssize_t __pyx_v_s
#line 51 "vector.from_py"
;
  PyObject *__pyx_v_item = NULL
#line 51 "vector.from_py"
;

#line 51 "vector.from_py"
  std::vector<std::string>  __pyx_r;

#line 51 "vector.from_py"
  __Pyx_RefNannyDeclarations

#line 51 "vector.from_py"
  Py_ssize_t __pyx_t_1;

#line 51 "vector.from_py"
  int __pyx_t_2;

#line 51 "vector.from_py"
  PyObject *__pyx_t_3 = NULL;

#line 51 "vector.from_py"
  PyObject *(*__pyx_t_4)(PyObject *);

#line 51 "vector.from_py"
  PyObject *__pyx_t_5 = NULL;

#line 51 "vector.from_py"
  std::string __pyx_t_6;

#line 51 "vector.from_py"
  int __pyx_lineno = 0;

#line 51 "vector.from_py"
  const char *__pyx_filename = NULL;

#line 51 "vector.from_py"
  int __pyx_clineno = 0;

#line 51 "vector.from_py"
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_std_3a__3a_string", 0);

  /* "vector.from_py":55
 * 
 *     cdef vector[X] v
 *     cdef Py_ssize_t s = __Pyx_PyObject_LengthHint(o, 0)             # <<<<<<<<<<<<<<
 * 
 *     if s > 0:
*/

#line 55 "vector.from_py"
  __pyx_t_1 = __Pyx_PyObject_LengthHint(__pyx_v_o, 0); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(1, 55, __pyx_L1_error)

#line 55 "vector.from_py"
  __pyx_v_s = __pyx_t_1;

  /* "vector.from_py":57
 *     cdef Py_ssize_t s = __Pyx_PyObject_LengthHint(o, 0)
 * 
 *     if s > 0:             # <<<<<<<<<<<<<<
 *         v.reserve(<size_t> s)
 * 
*/

#line 57 "vector.from_py"
  __pyx_t_2 = (__pyx_v_s > 0);

#line 57 "vector.from_py"
  if (__pyx_t_2) {

    /* "vector.from_py":58
 * 
 *     if s > 0:
 *         v.reserve(<size_t> s)             # <<<<<<<<<<<<<<
 * 
 *     for item in o:
*/

#line 58 "vector.from_py"
    try {

#line 58 "vector.from_py"
      __pyx_v_v.reserve(((size_t)__pyx_v_s));

#line 58 "vector.from_py"
    } catch(...) {

#line 58 "vector.from_py"
      __Pyx_CppExn2PyErr();

#line 58 "vector.from_py"
      __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "vector.from_py"
    }

    /* "vector.from_py":57
 *     cdef Py_ssize_t s = __Pyx_PyObject_LengthHint(o, 0)
 * 
 *     if s > 0:             # <<<<<<<<<<<<<<
 *         v.reserve(<size_t> s)
 * 
*/

#line 57 "vector.from_py"
  }

  /* "vector.from_py":60
 *         v.reserve(<size_t> s)
 * 
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 * 
*/

#line 60 "vector.from_py"
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {

#line 60 "vector.from_py"
    __pyx_t_3 = __pyx_v_o; __Pyx_INCREF(__pyx_t_3);

#line 60 "vector.from_py"
    __pyx_t_1 = 0;

#line 60 "vector.from_py"
    __pyx_t_4 = NULL;

#line 60 "vector.from_py"
  } else {
    __pyx_t_1 = -1; 
#line 60 "vector.from_py"
__pyx_t_3 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); 
#line 60 "vector.from_py"
if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
  }
  for (;;
#line 60 "vector.from_py"
) {

#line 60 "vector.from_py"
    if (likely(!__pyx_t_4)) {

#line 60 "vector.from_py"
      if (likely(PyList_CheckExact(__pyx_t_3))) {

#line 60 "vector.from_py"
        {

#line 60 "vector.from_py"
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);

#line 60 "vector.from_py"
          #if !CYTHON_ASSUME_SAFE_SIZE

#line 60 "vector.from_py"
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
          #endif

#line 60 "vector.from_py"
          if (__pyx_t_1 >= __pyx_temp) break;

#line 60 "vector.from_py"
        }

#line 60 "vector.from_py"
        __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_1, __Pyx_ReferenceSharing_OwnStrongReference);

#line 60 "vector.from_py"
        ++__pyx_t_1;

#line 60 "vector.from_py"
      } else {

#line 60 "vector.from_py"
        {

#line 60 "vector.from_py"
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);

#line 60 "vector.from_py"
          #if !CYTHON_ASSUME_SAFE_SIZE

#line 60 "vector.from_py"
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
          #endif

#line 60 "vector.from_py"
          if (__pyx_t_1 >= __pyx_temp) break;

#line 60 "vector.from_py"
        }

#line 60 "vector.from_py"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 60 "vector.from_py"
        __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1));

#line 60 "vector.from_py"
        #else

#line 60 "vector.from_py"
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_1);

#line 60 "vector.from_py"
        #endif

#line 60 "vector.from_py"
        ++__pyx_t_1;

#line 60 "vector.from_py"
      }

#line 60 "vector.from_py"
      if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 60, __pyx_L1_error)
    } else 
#line 60 "vector.from_py"
{

#line 60 "vector.from_py"
      __pyx_t_5 = __pyx_t_4(__pyx_t_3);

#line 60 "vector.from_py"
      if (unlikely(!__pyx_t_5)) {

#line 60 "vector.from_py"
        PyObject* exc_type = PyErr_Occurred();

#line 60 "vector.from_py"
        if (exc_type) {

#line 60 "vector.from_py"
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
          PyErr_Clear();

#line 60 "vector.from_py"
        }

#line 60 "vector.from_py"
        break;

#line 60 "vector.from_py"
      }

#line 60 "vector.from_py"
    }

#line 60 "vector.from_py"
    __Pyx_GOTREF(__pyx_t_5);

#line 60 "vector.from_py"
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_5);

#line 60 "vector.from_py"
    __pyx_t_5 = 0;

    /* "vector.from_py":61
 * 
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 * 
 *     return v
*/

#line 61 "vector.from_py"
    __pyx_t_6 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 61, __pyx_L1_error)

#line 61 "vector.from_py"
    try {

#line 61 "vector.from_py"
      __pyx_v_v.push_back(((std::string)__pyx_t_6));

#line 61 "vector.from_py"
    } catch(...) {

#line 61 "vector.from_py"
      __Pyx_CppExn2PyErr();

#line 61 "vector.from_py"
      __PYX_ERR(1, 61, __pyx_L1_error)

#line 61 "vector.from_py"
    }

    /* "vector.from_py":60
 *         v.reserve(<size_t> s)
 * 
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 * 
*/

#line 60 "vector.from_py"
  }

#line 60 "vector.from_py"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "vector.from_py":63
 *         v.push_back(<X>item)
 * 
 *     return v             # <<<<<<<<<<<<<<
 * 
*/

#line 63 "vector.from_py"
  __pyx_r = __pyx_v_v;

#line 63 "vector.from_py"
  goto __pyx_L0;

  /* "vector.from_py":51
 *     cdef Py_ssize_t __Pyx_PyObject_LengthHint(object o, Py_ssize_t defaultval) except -1
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_string")             # <<<<<<<<<<<<<<
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_string(object o) except *:
 * 
*/

#line 51 "vector.from_py"


#line 51 "vector.from_py"
  /* function exit code */

#line 51 "vector.from_py"
  __pyx_L1_error:;

#line 51 "vector.from_py"
  __Pyx_XDECREF(__pyx_t_3);

#line 51 "vector.from_py"
  __Pyx_XDECREF(__pyx_t_5);

#line 51 "vector.from_py"
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_std_3a__3a_string", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 51 "vector.from_py"
  __Pyx_pretend_to_initialize(&__pyx_r);

#line 51 "vector.from_py"
  __pyx_L0:;

#line 51 "vector.from_py"
  __Pyx_XDECREF(__pyx_v_item);

#line 51 "vector.from_py"
  __Pyx_RefNannyFinishContext();

#line 51 "vector.from_py"
  return __pyx_r;

#line 51 "vector.from_py"
}

#line 51 "vector.from_py"


#line 51 "vector.from_py"
static std::vector<int64_t>  __pyx_convert_vector_from_py_int64_t(PyObject *__pyx_v_o) {
  std::vector<int64_t>  __pyx_v_v
#line 51 "vector.from_py"
;
  Py_ssize_t __pyx_v_s
#line 51 "vector.from_py"
;
  PyObject *__pyx_v_item = NULL
#line 51 "vector.from_py"
;

#line 51 "vector.from_py"
  std::vector<int64_t>  __pyx_r;

#line 51 "vector.from_py"
  __Pyx_RefNannyDeclarations

#line 51 "vector.from_py"
  Py_ssize_t __pyx_t_1;

#line 51 "vector.from_py"
  int __pyx_t_2;

#line 51 "vector.from_py"
  PyObject *__pyx_t_3 = NULL;

#line 51 "vector.from_py"
  PyObject *(*__pyx_t_4)(PyObject *);

#line 51 "vector.from_py"
  PyObject *__pyx_t_5 = NULL;

#line 51 "vector.from_py"
  int64_t __pyx_t_6;

#line 51 "vector.from_py"
  int __pyx_lineno = 0;

#line 51 "vector.from_py"
  const char *__pyx_filename = NULL;

#line 51 "vector.from_py"
  int __pyx_clineno = 0;

#line 51 "vector.from_py"
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_int64_t", 0);

  /* "vector.from_py":55
 * 
 *     cdef vector[X] v
 *     cdef Py_ssize_t s = __Pyx_PyObject_LengthHint(o, 0)             # <<<<<<<<<<<<<<
 * 
 *     if s > 0:
*/

#line 55 "vector.from_py"
  __pyx_t_1 = __Pyx_PyObject_LengthHint(__pyx_v_o, 0); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(1, 55, __pyx_L1_error)

#line 55 "vector.from_py"
  __pyx_v_s = __pyx_t_1;

  /* "vector.from_py":57
 *     cdef Py_ssize_t s = __Pyx_PyObject_LengthHint(o, 0)
 * 
 *     if s > 0:             # <<<<<<<<<<<<<<
 *         v.reserve(<size_t> s)
 * 
*/

#line 57 "vector.from_py"
  __pyx_t_2 = (__pyx_v_s > 0);

#line 57 "vector.from_py"
  if (__pyx_t_2) {

    /* "vector.from_py":58
 * 
 *     if s > 0:
 *         v.reserve(<size_t> s)             # <<<<<<<<<<<<<<
 * 
 *     for item in o:
*/

#line 58 "vector.from_py"
    try {

#line 58 "vector.from_py"
      __pyx_v_v.reserve(((size_t)__pyx_v_s));

#line 58 "vector.from_py"
    } catch(...) {

#line 58 "vector.from_py"
      __Pyx_CppExn2PyErr();

#line 58 "vector.from_py"
      __PYX_ERR(1, 58, __pyx_L1_error)

#line 58 "vector.from_py"
    }

    /* "vector.from_py":57
 *     cdef Py_ssize_t s = __Pyx_PyObject_LengthHint(o, 0)
 * 
 *     if s > 0:             # <<<<<<<<<<<<<<
 *         v.reserve(<size_t> s)
 * 
*/

#line 57 "vector.from_py"
  }

  /* "vector.from_py":60
 *         v.reserve(<size_t> s)
 * 
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 * 
*/

#line 60 "vector.from_py"
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {

#line 60 "vector.from_py"
    __pyx_t_3 = __pyx_v_o; __Pyx_INCREF(__pyx_t_3);

#line 60 "vector.from_py"
    __pyx_t_1 = 0;

#line 60 "vector.from_py"
    __pyx_t_4 = NULL;

#line 60 "vector.from_py"
  } else {
    __pyx_t_1 = -1; 
#line 60 "vector.from_py"
__pyx_t_3 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); 
#line 60 "vector.from_py"
if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
  }
  for (;;
#line 60 "vector.from_py"
) {

#line 60 "vector.from_py"
    if (likely(!__pyx_t_4)) {

#line 60 "vector.from_py"
      if (likely(PyList_CheckExact(__pyx_t_3))) {

#line 60 "vector.from_py"
        {

#line 60 "vector.from_py"
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);

#line 60 "vector.from_py"
          #if !CYTHON_ASSUME_SAFE_SIZE

#line 60 "vector.from_py"
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
          #endif

#line 60 "vector.from_py"
          if (__pyx_t_1 >= __pyx_temp) break;

#line 60 "vector.from_py"
        }

#line 60 "vector.from_py"
        __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_1, __Pyx_ReferenceSharing_OwnStrongReference);

#line 60 "vector.from_py"
        ++__pyx_t_1;

#line 60 "vector.from_py"
      } else {

#line 60 "vector.from_py"
        {

#line 60 "vector.from_py"
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);

#line 60 "vector.from_py"
          #if !CYTHON_ASSUME_SAFE_SIZE

#line 60 "vector.from_py"
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
          #endif

#line 60 "vector.from_py"
          if (__pyx_t_1 >= __pyx_temp) break;

#line 60 "vector.from_py"
        }

#line 60 "vector.from_py"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 60 "vector.from_py"
        __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1));

#line 60 "vector.from_py"
        #else

#line 60 "vector.from_py"
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_1);

#line 60 "vector.from_py"
        #endif

#line 60 "vector.from_py"
        ++__pyx_t_1;

#line 60 "vector.from_py"
      }

#line 60 "vector.from_py"
      if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 60, __pyx_L1_error)
    } else 
#line 60 "vector.from_py"
{

#line 60 "vector.from_py"
      __pyx_t_5 = __pyx_t_4(__pyx_t_3);

#line 60 "vector.from_py"
      if (unlikely(!__pyx_t_5)) {

#line 60 "vector.from_py"
        PyObject* exc_type = PyErr_Occurred();

#line 60 "vector.from_py"
        if (exc_type) {

#line 60 "vector.from_py"
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(1, 60, __pyx_L1_error)

#line 60 "vector.from_py"
          PyErr_Clear();

#line 60 "vector.from_py"
        }

#line 60 "vector.from_py"
        break;

#line 60 "vector.from_py"
      }

#line 60 "vector.from_py"
    }

#line 60 "vector.from_py"
    __Pyx_GOTREF(__pyx_t_5);

#line 60 "vector.from_py"
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_5);

#line 60 "vector.from_py"
    __pyx_t_5 = 0;

    /* "vector.from_py":61
 * 
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 * 
 *     return v
*/

#line 61 "vector.from_py"
    __pyx_t_6 = __Pyx_PyLong_As_int64_t(__pyx_v_item); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 61, __pyx_L1_error)

#line 61 "vector.from_py"
    try {

#line 61 "vector.from_py"
      __pyx_v_v.push_back(((int64_t)__pyx_t_6));

#line 61 "vector.from_py"
    } catch(...) {

#line 61 "vector.from_py"
      __Pyx_CppExn2PyErr();

#line 61 "vector.from_py"
      __PYX_ERR(1, 61, __pyx_L1_error)

#line 61 "vector.from_py"
    }

    /* "vector.from_py":60
 *         v.reserve(<size_t> s)
 * 
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 * 
*/

#line 60 "vector.from_py"
  }

#line 60 "vector.from_py"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "vector.from_py":63
 *         v.push_back(<X>item)
 * 
 *     return v             # <<<<<<<<<<<<<<
 * 
*/

#line 63 "vector.from_py"
  __pyx_r = __pyx_v_v;

#line 63 "vector.from_py"
  goto __pyx_L0;

  /* "vector.from_py":51
 *     cdef Py_ssize_t __Pyx_PyObject_LengthHint(object o, Py_ssize_t defaultval) except -1
 * 
 * @cname("__pyx_convert_vector_from_py_int64_t")             # <<<<<<<<<<<<<<
 * cdef vector[X] __pyx_convert_vector_from_py_int64_t(object o) except *:
 * 
*/

#line 51 "vector.from_py"


#line 51 "vector.from_py"
  /* function exit code */

#line 51 "vector.from_py"
  __pyx_L1_error:;

#line 51 "vector.from_py"
  __Pyx_XDECREF(__pyx_t_3);

#line 51 "vector.from_py"
  __Pyx_XDECREF(__pyx_t_5);

#line 51 "vector.from_py"
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_int64_t", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 51 "vector.from_py"
  __Pyx_pretend_to_initialize(&__pyx_r);

#line 51 "vector.from_py"
  __pyx_L0:;

#line 51 "vector.from_py"
  __Pyx_XDECREF(__pyx_v_item);

#line 51 "vector.from_py"
  __Pyx_RefNannyFinishContext();

#line 51 "vector.from_py"
  return __pyx_r;

#line 51 "vector.from_py"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":243
 *         cdef int type_num
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)
*/

#line 243 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 243 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self) {

#line 243 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":245
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 245 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyDataType_ELSIZE(__pyx_v_self);

#line 245 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":243
 *         cdef int type_num
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)
*/

#line 243 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 243 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 243 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 243 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 243 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":247
 *             return PyDataType_ELSIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)
*/

#line 247 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 247 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self) {

#line 247 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":249
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use fields/names with care as they may be NULL.  You must check
*/

#line 249 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyDataType_ALIGNMENT(__pyx_v_self);

#line 249 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":247
 *             return PyDataType_ELSIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)
*/

#line 247 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 247 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 247 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 247 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 247 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":253
 *         # Use fields/names with care as they may be NULL.  You must check
 *         # for this using PyDataType_HASFIELDS.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)
*/

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self) {

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r = NULL;

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1;

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("fields", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":255
 *         @property
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 255 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 255 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyDataType_FIELDS(__pyx_v_self);

#line 255 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_t_1));

#line 255 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = ((PyObject *)__pyx_t_1);

#line 255 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":253
 *         # Use fields/names with care as they may be NULL.  You must check
 *         # for this using PyDataType_HASFIELDS.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)
*/

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 253 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 *             return <object>PyDataType_FIELDS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)
*/

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self) {

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r = NULL;

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1;

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("names", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":259
 *         @property
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use PyDataType_HASSUBARRAY to test whether this field is
*/

#line 259 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 259 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyDataType_NAMES(__pyx_v_self);

#line 259 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_INCREF(((PyObject*)__pyx_t_1));

#line 259 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = ((PyObject*)__pyx_t_1);

#line 259 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 *             return <object>PyDataType_FIELDS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)
*/

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 257 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":264
 *         # valid (the pointer can be NULL). Most users should access
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)
*/

#line 264 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 264 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self) {

#line 264 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyArray_ArrayDescr *__pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":266
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 266 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyDataType_SUBARRAY(__pyx_v_self);

#line 266 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":264
 *         # valid (the pointer can be NULL). Most users should access
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)
*/

#line 264 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 264 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 264 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 264 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 264 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 *             return PyDataType_SUBARRAY(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
*/

#line 268 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 268 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self) {

#line 268 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_uint64 __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":271
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 271 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyDataType_FLAGS(__pyx_v_self);

#line 271 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 *             return PyDataType_SUBARRAY(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
*/

#line 268 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 268 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 268 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 268 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 268 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":280
 *     ctypedef class numpy.broadcast [object PyArrayMultiIterObject, check_size ignore]:
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
*/

#line 280 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 280 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self) {

#line 280 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":283
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 283 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_MultiIter_NUMITER(__pyx_v_self);

#line 283 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":280
 *     ctypedef class numpy.broadcast [object PyArrayMultiIterObject, check_size ignore]:
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
*/

#line 280 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 280 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 280 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 280 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 280 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":285
 *             return PyArray_MultiIter_NUMITER(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
*/

#line 285 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 285 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self) {

#line 285 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":288
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 288 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_MultiIter_SIZE(__pyx_v_self);

#line 288 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":285
 *             return PyArray_MultiIter_NUMITER(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
*/

#line 285 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 285 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 285 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 285 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 285 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 *             return PyArray_MultiIter_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
*/

#line 290 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 290 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self) {

#line 290 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":293
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 293 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_MultiIter_INDEX(__pyx_v_self);

#line 293 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 *             return PyArray_MultiIter_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
*/

#line 290 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 290 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 290 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 290 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 290 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":295
 *             return PyArray_MultiIter_INDEX(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
*/

#line 295 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 295 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self) {

#line 295 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":298
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 298 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_MultiIter_NDIM(__pyx_v_self);

#line 298 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":295
 *             return PyArray_MultiIter_INDEX(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
*/

#line 295 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 295 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 295 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 295 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 295 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":300
 *             return PyArray_MultiIter_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
*/

#line 300 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 300 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self) {

#line 300 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_intp *__pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":303
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 303 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_MultiIter_DIMS(__pyx_v_self);

#line 303 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":300
 *             return PyArray_MultiIter_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
*/

#line 300 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 300 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 300 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 300 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 300 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":305
 *             return PyArray_MultiIter_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline void** iters(self) noexcept nogil:
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
*/

#line 305 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 305 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self) {

#line 305 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  void **__pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":309
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 *             return PyArray_MultiIter_ITERS(self)             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 309 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_MultiIter_ITERS(__pyx_v_self);

#line 309 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":305
 *             return PyArray_MultiIter_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline void** iters(self) noexcept nogil:
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
*/

#line 305 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 305 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 305 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 305 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 305 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":323
 *         # Instead, we use properties that map to the corresponding C-API functions.
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyObject* base(self) noexcept nogil:
 *             """Returns a borrowed reference to the object owning the data/memory.
*/

#line 323 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 323 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self) {

#line 323 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":327
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 *             return PyArray_BASE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 327 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_BASE(__pyx_v_self);

#line 327 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":323
 *         # Instead, we use properties that map to the corresponding C-API functions.
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyObject* base(self) noexcept nogil:
 *             """Returns a borrowed reference to the object owning the data/memory.
*/

#line 323 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 323 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 323 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 323 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 323 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 *             return PyArray_BASE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline dtype descr(self):
 *             """Returns an owned reference to the dtype of the array.
*/

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self) {

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyArray_Descr *__pyx_r = NULL;

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyArray_Descr *__pyx_t_1;

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("descr", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":333
 *             """Returns an owned reference to the dtype of the array.
 *             """
 *             return <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 333 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_r);

#line 333 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyArray_DESCR(__pyx_v_self);

#line 333 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_INCREF((PyObject *)((PyArray_Descr *)__pyx_t_1));

#line 333 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = ((PyArray_Descr *)__pyx_t_1);

#line 333 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 *             return PyArray_BASE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline dtype descr(self):
 *             """Returns an owned reference to the dtype of the array.
*/

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 329 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":335
 *             return <dtype>PyArray_DESCR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int ndim(self) noexcept nogil:
 *             """Returns the number of dimensions in the array.
*/

#line 335 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 335 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self) {

#line 335 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":339
 *             """Returns the number of dimensions in the array.
 *             """
 *             return PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 339 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_NDIM(__pyx_v_self);

#line 339 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":335
 *             return <dtype>PyArray_DESCR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int ndim(self) noexcept nogil:
 *             """Returns the number of dimensions in the array.
*/

#line 335 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 335 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 335 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 335 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 335 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":341
 *             return PyArray_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *shape(self) noexcept nogil:
 *             """Returns a pointer to the dimensions/shape of the array.
*/

#line 341 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 341 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self) {

#line 341 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_intp *__pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":347
 *             Can return NULL for 0-dimensional arrays.
 *             """
 *             return PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 347 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_DIMS(__pyx_v_self);

#line 347 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":341
 *             return PyArray_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *shape(self) noexcept nogil:
 *             """Returns a pointer to the dimensions/shape of the array.
*/

#line 341 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 341 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 341 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 341 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 341 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 *             return PyArray_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *strides(self) noexcept nogil:
 *             """Returns a pointer to the strides of the array.
*/

#line 349 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 349 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self) {

#line 349 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_intp *__pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":354
 *             The number of elements matches the number of dimensions of the array (ndim).
 *             """
 *             return PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 354 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_STRIDES(__pyx_v_self);

#line 354 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 *             return PyArray_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *strides(self) noexcept nogil:
 *             """Returns a pointer to the strides of the array.
*/

#line 349 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 349 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 349 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 349 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 349 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":356
 *             return PyArray_STRIDES(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """Returns the total size (in number of elements) of the array.
*/

#line 356 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 356 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self) {

#line 356 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":360
 *             """Returns the total size (in number of elements) of the array.
 *             """
 *             return PyArray_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 360 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_SIZE(__pyx_v_self);

#line 360 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":356
 *             return PyArray_STRIDES(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """Returns the total size (in number of elements) of the array.
*/

#line 356 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 356 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 356 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 356 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 356 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":362
 *             return PyArray_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline char* data(self) noexcept nogil:
 *             """The pointer to the data buffer as a char*.
*/

#line 362 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 362 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self) {

#line 362 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  char *__pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":369
 *             of `PyArray_DATA()` instead, which returns a 'void*'.
 *             """
 *             return PyArray_BYTES(self)             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 369 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyArray_BYTES(__pyx_v_self);

#line 369 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":362
 *             return PyArray_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline char* data(self) noexcept nogil:
 *             """The pointer to the data buffer as a char*.
*/

#line 362 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 362 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 362 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 362 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 362 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":777
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
*/

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r = NULL;

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_lineno = 0;

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  const char *__pyx_filename = NULL;

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_clineno = 0;

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":778
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
*/

#line 778 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 778 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 778, __pyx_L1_error)

#line 778 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 778 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = __pyx_t_1;

#line 778 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = 0;

#line 778 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":777
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
*/

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L1_error:;

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = 0;

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 777 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":780
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
*/

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r = NULL;

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_lineno = 0;

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  const char *__pyx_filename = NULL;

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_clineno = 0;

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":781
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
*/

#line 781 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 781 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 781, __pyx_L1_error)

#line 781 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 781 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = __pyx_t_1;

#line 781 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = 0;

#line 781 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":780
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
*/

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L1_error:;

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = 0;

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 780 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":783
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
*/

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r = NULL;

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_lineno = 0;

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  const char *__pyx_filename = NULL;

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_clineno = 0;

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":784
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
*/

#line 784 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 784 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 784, __pyx_L1_error)

#line 784 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 784 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = __pyx_t_1;

#line 784 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = 0;

#line 784 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":783
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
*/

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L1_error:;

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = 0;

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 783 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":786
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
*/

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r = NULL;

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_lineno = 0;

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  const char *__pyx_filename = NULL;

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_clineno = 0;

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":787
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
*/

#line 787 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 787 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 787, __pyx_L1_error)

#line 787 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 787 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = __pyx_t_1;

#line 787 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = 0;

#line 787 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":786
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
*/

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L1_error:;

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = 0;

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 786 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
*/

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r = NULL;

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_lineno = 0;

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  const char *__pyx_filename = NULL;

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_clineno = 0;

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":790
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
*/

#line 790 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 790 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 790, __pyx_L1_error)

#line 790 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 790 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = __pyx_t_1;

#line 790 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = 0;

#line 790 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
*/

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L1_error:;

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = 0;

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 789 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
*/

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r = NULL;

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_t_1;

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_2;

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":793
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
*/

#line 793 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyDataType_HASSUBARRAY(__pyx_v_d);

#line 793 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":794
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
*/

#line 794 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 794 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_t_2 = __pyx_f_5numpy_5dtype_8subarray_subarray(__pyx_v_d)->shape;

#line 794 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_INCREF(((PyObject*)__pyx_t_2));

#line 794 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_r = ((PyObject*)__pyx_t_2);

#line 794 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":793
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
*/

#line 793 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  }

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":796
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 796 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /*else*/ {

#line 796 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 796 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);

#line 796 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_r = __pyx_mstate_global->__pyx_empty_tuple;

#line 796 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L0;

#line 796 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  }

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
*/

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 792 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":995
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
*/

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_t_1;

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_lineno = 0;

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  const char *__pyx_filename = NULL;

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_clineno = 0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":996
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
*/

#line 996 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  Py_INCREF(__pyx_v_base);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":997
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
*/

#line 997 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 997, __pyx_L1_error)

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":995
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
*/

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L1_error:;

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_AddTraceback("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 995 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
*/

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base
#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
;

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_r = NULL;

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_t_1;

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1000
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
*/

#line 1000 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1001
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
*/

#line 1001 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_t_1 = (__pyx_v_base == NULL);

#line 1001 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1002
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
*/

#line 1002 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 1002 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 1002 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1001
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
*/

#line 1001 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  }

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1003
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
*/

#line 1003 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 1003 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_base));

#line 1003 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = ((PyObject *)__pyx_v_base);

#line 1003 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
*/

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 999 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1007
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
*/

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_r;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_t_4;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_9 = NULL;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  size_t __pyx_t_10;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_lineno = 0;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  const char *__pyx_filename = NULL;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_clineno = 0;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
*/

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  {

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_PyThreadState_declare

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_PyThreadState_assign

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    /*try:*/ {

      /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1009
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")
*/

#line 1009 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1009, __pyx_L3_error)

      /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
*/

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    }

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L8_try_end;

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1010
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
*/

#line 1010 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));

#line 1010 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    if (__pyx_t_4) {

#line 1010 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1010 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1010, __pyx_L5_except_error)

#line 1010 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_XGOTREF(__pyx_t_5);

#line 1010 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_XGOTREF(__pyx_t_6);

#line 1010 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1011
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
*/

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __pyx_t_9 = NULL;

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __pyx_t_10 = 1;

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      {

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_kp_u_numpy__core_multiarray_failed_to};

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ImportError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1011, __pyx_L5_except_error)

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        __Pyx_GOTREF(__pyx_t_8);

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      }

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __PYX_ERR(2, 1011, __pyx_L5_except_error)

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    }

#line 1011 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L5_except_error;

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
*/

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_L5_except_error:;

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L1_error;

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_L8_try_end:;

#line 1008 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  }

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1007
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
*/

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = 0;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L1_error:;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_9);

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = -1;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 1007 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1013
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_r;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_t_4;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_9 = NULL;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  size_t __pyx_t_10;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_lineno = 0;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  const char *__pyx_filename = NULL;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_clineno = 0;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  {

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_PyThreadState_declare

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_PyThreadState_assign

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    /*try:*/ {

      /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1015
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
*/

#line 1015 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1015, __pyx_L3_error)

      /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    }

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L8_try_end;

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1016
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
*/

#line 1016 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));

#line 1016 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    if (__pyx_t_4) {

#line 1016 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1016 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1016, __pyx_L5_except_error)

#line 1016 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_XGOTREF(__pyx_t_5);

#line 1016 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_XGOTREF(__pyx_t_6);

#line 1016 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1017
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
*/

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __pyx_t_9 = NULL;

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __pyx_t_10 = 1;

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      {

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_kp_u_numpy__core_umath_failed_to_impo};

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ImportError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1017, __pyx_L5_except_error)

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        __Pyx_GOTREF(__pyx_t_8);

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      }

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __PYX_ERR(2, 1017, __pyx_L5_except_error)

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    }

#line 1017 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L5_except_error;

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_L5_except_error:;

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L1_error;

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_L8_try_end:;

#line 1014 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  }

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1013
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = 0;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L1_error:;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_9);

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = -1;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 1013 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1019
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_r;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyDeclarations

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_t_4;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  PyObject *__pyx_t_9 = NULL;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  size_t __pyx_t_10;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_lineno = 0;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  const char *__pyx_filename = NULL;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_clineno = 0;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  {

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_PyThreadState_declare

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_PyThreadState_assign

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    /*try:*/ {

      /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1021
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
*/

#line 1021 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1021, __pyx_L3_error)

      /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    }

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L8_try_end;

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1022
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
*/

#line 1022 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));

#line 1022 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    if (__pyx_t_4) {

#line 1022 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1022 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1022, __pyx_L5_except_error)

#line 1022 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_XGOTREF(__pyx_t_5);

#line 1022 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_XGOTREF(__pyx_t_6);

#line 1022 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1023
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __pyx_t_9 = NULL;

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __pyx_t_10 = 1;

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      {

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_kp_u_numpy__core_umath_failed_to_impo};

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ImportError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1023, __pyx_L5_except_error)

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
        __Pyx_GOTREF(__pyx_t_8);

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      }

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
      __PYX_ERR(2, 1023, __pyx_L5_except_error)

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    }

#line 1023 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L5_except_error;

    /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_L5_except_error:;

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    goto __pyx_L1_error;

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
    __pyx_L8_try_end:;

#line 1020 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  }

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1019
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = 0;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L1_error:;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_XDECREF(__pyx_t_9);

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = -1;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __Pyx_RefNannyFinishContext();

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 1019 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1026
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
*/

#line 1026 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1026 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {

#line 1026 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1038
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 1038 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));

#line 1038 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1026
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
*/

#line 1026 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1026 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 1026 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 1026 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 1026 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1041
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
*/

#line 1041 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1041 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {

#line 1041 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  int __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1053
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 1053 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));

#line 1053 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1041
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
*/

#line 1041 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1041 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 1041 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 1041 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 1041 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1056
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
*/

#line 1056 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1056 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {

#line 1056 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_datetime __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1063
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 1063 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;

#line 1063 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1056
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
*/

#line 1056 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1056 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 1056 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 1056 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 1056 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1066
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
*/

#line 1066 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1066 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {

#line 1066 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  npy_timedelta __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1070
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 1070 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;

#line 1070 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1066
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
*/

#line 1066 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1066 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 1066 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 1066 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 1066 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1073
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
*/

#line 1073 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1073 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {

#line 1073 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1077
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 1077 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);

#line 1077 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1073
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
*/

#line 1073 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"


#line 1073 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  /* function exit code */

#line 1073 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  __pyx_L0:;

#line 1073 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
  return __pyx_r;

#line 1073 "../../../../../tmp/build-env-7lykdjnz/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd"
}

/* "cpython/complex.pxd":20
 * 
 *         # unavailable in limited API
 *         @property             # <<<<<<<<<<<<<<
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double real(self) noexcept:
*/

#line 20 "cpython/complex.pxd"


#line 20 "cpython/complex.pxd"
#if !CYTHON_COMPILING_IN_LIMITED_API

#line 20 "cpython/complex.pxd"
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {

#line 20 "cpython/complex.pxd"
  double __pyx_r;

  /* "cpython/complex.pxd":23
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double real(self) noexcept:
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         # unavailable in limited API
*/

#line 23 "cpython/complex.pxd"
  __pyx_r = __pyx_v_self->cval.real;

#line 23 "cpython/complex.pxd"
  goto __pyx_L0;

  /* "cpython/complex.pxd":20
 * 
 *         # unavailable in limited API
 *         @property             # <<<<<<<<<<<<<<
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double real(self) noexcept:
*/

#line 20 "cpython/complex.pxd"


#line 20 "cpython/complex.pxd"
  /* function exit code */

#line 20 "cpython/complex.pxd"
  __pyx_L0:;

#line 20 "cpython/complex.pxd"
  return __pyx_r;

#line 20 "cpython/complex.pxd"
}

#line 20 "cpython/complex.pxd"
#endif /*!(#if !CYTHON_COMPILING_IN_LIMITED_API)*/

/* "cpython/complex.pxd":26
 * 
 *         # unavailable in limited API
 *         @property             # <<<<<<<<<<<<<<
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double imag(self) noexcept:
*/

#line 26 "cpython/complex.pxd"


#line 26 "cpython/complex.pxd"
#if !CYTHON_COMPILING_IN_LIMITED_API

#line 26 "cpython/complex.pxd"
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {

#line 26 "cpython/complex.pxd"
  double __pyx_r;

  /* "cpython/complex.pxd":29
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double imag(self) noexcept:
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
*/

#line 29 "cpython/complex.pxd"
  __pyx_r = __pyx_v_self->cval.imag;

#line 29 "cpython/complex.pxd"
  goto __pyx_L0;

  /* "cpython/complex.pxd":26
 * 
 *         # unavailable in limited API
 *         @property             # <<<<<<<<<<<<<<
 *         @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 *         cdef inline double imag(self) noexcept:
*/

#line 26 "cpython/complex.pxd"


#line 26 "cpython/complex.pxd"
  /* function exit code */

#line 26 "cpython/complex.pxd"
  __pyx_L0:;

#line 26 "cpython/complex.pxd"
  return __pyx_r;

#line 26 "cpython/complex.pxd"
}

#line 26 "cpython/complex.pxd"
#endif /*!(#if !CYTHON_COMPILING_IN_LIMITED_API)*/

/* "cpython/contextvars.pxd":115
 * 
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")             # <<<<<<<<<<<<<<
 * cdef inline object get_value(var, default_value=None):
 *     """Return a new reference to the value of the context variable,
*/

#line 115 "cpython/contextvars.pxd"


#line 115 "cpython/contextvars.pxd"
#if !CYTHON_COMPILING_IN_LIMITED_API

#line 115 "cpython/contextvars.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {

  /* "cpython/contextvars.pxd":116
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
*/

#line 116 "cpython/contextvars.pxd"
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value
#line 116 "cpython/contextvars.pxd"
;
  PyObject *__pyx_v_pyvalue = NULL
#line 116 "cpython/contextvars.pxd"
;

#line 116 "cpython/contextvars.pxd"
  PyObject *__pyx_r = NULL;

#line 116 "cpython/contextvars.pxd"
  __Pyx_RefNannyDeclarations

#line 116 "cpython/contextvars.pxd"
  int __pyx_t_1;

#line 116 "cpython/contextvars.pxd"
  int __pyx_t_2;

#line 116 "cpython/contextvars.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 116 "cpython/contextvars.pxd"
  int __pyx_lineno = 0;

#line 116 "cpython/contextvars.pxd"
  const char *__pyx_filename = NULL;

#line 116 "cpython/contextvars.pxd"
  int __pyx_clineno = 0;

#line 116 "cpython/contextvars.pxd"
  __Pyx_RefNannySetupContext("get_value", 0);

#line 116 "cpython/contextvars.pxd"
  if (__pyx_optional_args) {

#line 116 "cpython/contextvars.pxd"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 116 "cpython/contextvars.pxd"
      __pyx_v_default_value = __pyx_optional_args->default_value;

#line 116 "cpython/contextvars.pxd"
    }

#line 116 "cpython/contextvars.pxd"
  }

  /* "cpython/contextvars.pxd":121
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
*/

#line 121 "cpython/contextvars.pxd"
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":122
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
*/

#line 122 "cpython/contextvars.pxd"
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 122, __pyx_L1_error)

  /* "cpython/contextvars.pxd":123
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
*/

#line 123 "cpython/contextvars.pxd"
  __pyx_t_2 = (__pyx_v_value == NULL);

#line 123 "cpython/contextvars.pxd"
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":125
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
*/

#line 125 "cpython/contextvars.pxd"
    __Pyx_INCREF(__pyx_v_default_value);

#line 125 "cpython/contextvars.pxd"
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":123
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
*/

#line 123 "cpython/contextvars.pxd"
    goto __pyx_L3;

#line 123 "cpython/contextvars.pxd"
  }

  /* "cpython/contextvars.pxd":128
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
*/

#line 128 "cpython/contextvars.pxd"
  /*else*/ {

#line 128 "cpython/contextvars.pxd"
    __pyx_t_3 = ((PyObject *)__pyx_v_value);

#line 128 "cpython/contextvars.pxd"
    __Pyx_INCREF(__pyx_t_3);

#line 128 "cpython/contextvars.pxd"
    __pyx_v_pyvalue = __pyx_t_3;

#line 128 "cpython/contextvars.pxd"
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":129
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
*/

#line 129 "cpython/contextvars.pxd"
    Py_XDECREF(__pyx_v_value);

#line 129 "cpython/contextvars.pxd"
  }

#line 129 "cpython/contextvars.pxd"
  __pyx_L3:;

  /* "cpython/contextvars.pxd":130
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 130 "cpython/contextvars.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 130 "cpython/contextvars.pxd"
  __Pyx_INCREF(__pyx_v_pyvalue);

#line 130 "cpython/contextvars.pxd"
  __pyx_r = __pyx_v_pyvalue;

#line 130 "cpython/contextvars.pxd"
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":115
 * 
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")             # <<<<<<<<<<<<<<
 * cdef inline object get_value(var, default_value=None):
 *     """Return a new reference to the value of the context variable,
*/

#line 115 "cpython/contextvars.pxd"


#line 115 "cpython/contextvars.pxd"
  /* function exit code */

#line 115 "cpython/contextvars.pxd"
  __pyx_L1_error:;

#line 115 "cpython/contextvars.pxd"
  __Pyx_XDECREF(__pyx_t_3);

#line 115 "cpython/contextvars.pxd"
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 115 "cpython/contextvars.pxd"
  __pyx_r = 0;

#line 115 "cpython/contextvars.pxd"
  __pyx_L0:;

#line 115 "cpython/contextvars.pxd"
  __Pyx_XDECREF(__pyx_v_pyvalue);

#line 115 "cpython/contextvars.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 115 "cpython/contextvars.pxd"
  __Pyx_RefNannyFinishContext();

#line 115 "cpython/contextvars.pxd"
  return __pyx_r;

#line 115 "cpython/contextvars.pxd"
}

#line 115 "cpython/contextvars.pxd"
#endif /*!(#if !CYTHON_COMPILING_IN_LIMITED_API)*/

/* "cpython/contextvars.pxd":133
 * 
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")             # <<<<<<<<<<<<<<
 * cdef inline object get_value_no_default(var, default_value=None):
 *     """Return a new reference to the value of the context variable,
*/

#line 133 "cpython/contextvars.pxd"


#line 133 "cpython/contextvars.pxd"
#if !CYTHON_COMPILING_IN_LIMITED_API

#line 133 "cpython/contextvars.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {

  /* "cpython/contextvars.pxd":134
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
*/

#line 134 "cpython/contextvars.pxd"
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value
#line 134 "cpython/contextvars.pxd"
;
  PyObject *__pyx_v_pyvalue = NULL
#line 134 "cpython/contextvars.pxd"
;

#line 134 "cpython/contextvars.pxd"
  PyObject *__pyx_r = NULL;

#line 134 "cpython/contextvars.pxd"
  __Pyx_RefNannyDeclarations

#line 134 "cpython/contextvars.pxd"
  int __pyx_t_1;

#line 134 "cpython/contextvars.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 134 "cpython/contextvars.pxd"
  int __pyx_lineno = 0;

#line 134 "cpython/contextvars.pxd"
  const char *__pyx_filename = NULL;

#line 134 "cpython/contextvars.pxd"
  int __pyx_clineno = 0;

#line 134 "cpython/contextvars.pxd"
  __Pyx_RefNannySetupContext("get_value_no_default", 0);

#line 134 "cpython/contextvars.pxd"
  if (__pyx_optional_args) {

#line 134 "cpython/contextvars.pxd"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 134 "cpython/contextvars.pxd"
      __pyx_v_default_value = __pyx_optional_args->default_value;

#line 134 "cpython/contextvars.pxd"
    }

#line 134 "cpython/contextvars.pxd"
  }

  /* "cpython/contextvars.pxd":140
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
*/

#line 140 "cpython/contextvars.pxd"
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":141
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
*/

#line 141 "cpython/contextvars.pxd"
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 141, __pyx_L1_error)

  /* "cpython/contextvars.pxd":143
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
*/

#line 143 "cpython/contextvars.pxd"
  __pyx_t_2 = ((PyObject *)__pyx_v_value);

#line 143 "cpython/contextvars.pxd"
  __Pyx_INCREF(__pyx_t_2);

#line 143 "cpython/contextvars.pxd"
  __pyx_v_pyvalue = __pyx_t_2;

#line 143 "cpython/contextvars.pxd"
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":144
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
*/

#line 144 "cpython/contextvars.pxd"
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":145
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
*/

#line 145 "cpython/contextvars.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 145 "cpython/contextvars.pxd"
  __Pyx_INCREF(__pyx_v_pyvalue);

#line 145 "cpython/contextvars.pxd"
  __pyx_r = __pyx_v_pyvalue;

#line 145 "cpython/contextvars.pxd"
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":133
 * 
 * 
 * @_cython.c_compile_guard("!CYTHON_COMPILING_IN_LIMITED_API")             # <<<<<<<<<<<<<<
 * cdef inline object get_value_no_default(var, default_value=None):
 *     """Return a new reference to the value of the context variable,
*/

#line 133 "cpython/contextvars.pxd"


#line 133 "cpython/contextvars.pxd"
  /* function exit code */

#line 133 "cpython/contextvars.pxd"
  __pyx_L1_error:;

#line 133 "cpython/contextvars.pxd"
  __Pyx_XDECREF(__pyx_t_2);

#line 133 "cpython/contextvars.pxd"
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 133 "cpython/contextvars.pxd"
  __pyx_r = 0;

#line 133 "cpython/contextvars.pxd"
  __pyx_L0:;

#line 133 "cpython/contextvars.pxd"
  __Pyx_XDECREF(__pyx_v_pyvalue);

#line 133 "cpython/contextvars.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 133 "cpython/contextvars.pxd"
  __Pyx_RefNannyFinishContext();

#line 133 "cpython/contextvars.pxd"
  return __pyx_r;

#line 133 "cpython/contextvars.pxd"
}

#line 133 "cpython/contextvars.pxd"
#endif /*!(#if !CYTHON_COMPILING_IN_LIMITED_API)*/

/* "cpython/datetime.pxd":44
 * 
 *     ctypedef extern class datetime.date[object PyDateTime_Date]:
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int year(self) noexcept:
 *             return PyDateTime_GET_YEAR(self)
*/

#line 44 "cpython/datetime.pxd"


#line 44 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_4year_year(PyDateTime_Date *__pyx_v_self) {

#line 44 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":46
 *         @property
 *         cdef inline int year(self) noexcept:
 *             return PyDateTime_GET_YEAR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 46 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_YEAR(((PyObject *)__pyx_v_self));

#line 46 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":44
 * 
 *     ctypedef extern class datetime.date[object PyDateTime_Date]:
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int year(self) noexcept:
 *             return PyDateTime_GET_YEAR(self)
*/

#line 44 "cpython/datetime.pxd"


#line 44 "cpython/datetime.pxd"
  /* function exit code */

#line 44 "cpython/datetime.pxd"
  __pyx_L0:;

#line 44 "cpython/datetime.pxd"
  return __pyx_r;

#line 44 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":48
 *             return PyDateTime_GET_YEAR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int month(self) noexcept:
 *             return PyDateTime_GET_MONTH(self)
*/

#line 48 "cpython/datetime.pxd"


#line 48 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_5month_month(PyDateTime_Date *__pyx_v_self) {

#line 48 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":50
 *         @property
 *         cdef inline int month(self) noexcept:
 *             return PyDateTime_GET_MONTH(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 50 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_MONTH(((PyObject *)__pyx_v_self));

#line 50 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":48
 *             return PyDateTime_GET_YEAR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int month(self) noexcept:
 *             return PyDateTime_GET_MONTH(self)
*/

#line 48 "cpython/datetime.pxd"


#line 48 "cpython/datetime.pxd"
  /* function exit code */

#line 48 "cpython/datetime.pxd"
  __pyx_L0:;

#line 48 "cpython/datetime.pxd"
  return __pyx_r;

#line 48 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":52
 *             return PyDateTime_GET_MONTH(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_GET_DAY(self)
*/

#line 52 "cpython/datetime.pxd"


#line 52 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_3day_day(PyDateTime_Date *__pyx_v_self) {

#line 52 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":54
 *         @property
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_GET_DAY(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
*/

#line 54 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_DAY(((PyObject *)__pyx_v_self));

#line 54 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":52
 *             return PyDateTime_GET_MONTH(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_GET_DAY(self)
*/

#line 52 "cpython/datetime.pxd"


#line 52 "cpython/datetime.pxd"
  /* function exit code */

#line 52 "cpython/datetime.pxd"
  __pyx_L0:;

#line 52 "cpython/datetime.pxd"
  return __pyx_r;

#line 52 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":57
 * 
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int hour(self) noexcept:
 *             return PyDateTime_TIME_GET_HOUR(self)
*/

#line 57 "cpython/datetime.pxd"


#line 57 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4hour_hour(PyDateTime_Time *__pyx_v_self) {

#line 57 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":59
 *         @property
 *         cdef inline int hour(self) noexcept:
 *             return PyDateTime_TIME_GET_HOUR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 59 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_HOUR(((PyObject *)__pyx_v_self));

#line 59 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":57
 * 
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int hour(self) noexcept:
 *             return PyDateTime_TIME_GET_HOUR(self)
*/

#line 57 "cpython/datetime.pxd"


#line 57 "cpython/datetime.pxd"
  /* function exit code */

#line 57 "cpython/datetime.pxd"
  __pyx_L0:;

#line 57 "cpython/datetime.pxd"
  return __pyx_r;

#line 57 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":61
 *             return PyDateTime_TIME_GET_HOUR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int minute(self) noexcept:
 *             return PyDateTime_TIME_GET_MINUTE(self)
*/

#line 61 "cpython/datetime.pxd"


#line 61 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6minute_minute(PyDateTime_Time *__pyx_v_self) {

#line 61 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":63
 *         @property
 *         cdef inline int minute(self) noexcept:
 *             return PyDateTime_TIME_GET_MINUTE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 63 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_MINUTE(((PyObject *)__pyx_v_self));

#line 63 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":61
 *             return PyDateTime_TIME_GET_HOUR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int minute(self) noexcept:
 *             return PyDateTime_TIME_GET_MINUTE(self)
*/

#line 61 "cpython/datetime.pxd"


#line 61 "cpython/datetime.pxd"
  /* function exit code */

#line 61 "cpython/datetime.pxd"
  __pyx_L0:;

#line 61 "cpython/datetime.pxd"
  return __pyx_r;

#line 61 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":65
 *             return PyDateTime_TIME_GET_MINUTE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_TIME_GET_SECOND(self)
*/

#line 65 "cpython/datetime.pxd"


#line 65 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6second_second(PyDateTime_Time *__pyx_v_self) {

#line 65 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":67
 *         @property
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_TIME_GET_SECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 67 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_SECOND(((PyObject *)__pyx_v_self));

#line 67 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":65
 *             return PyDateTime_TIME_GET_MINUTE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_TIME_GET_SECOND(self)
*/

#line 65 "cpython/datetime.pxd"


#line 65 "cpython/datetime.pxd"
  /* function exit code */

#line 65 "cpython/datetime.pxd"
  __pyx_L0:;

#line 65 "cpython/datetime.pxd"
  return __pyx_r;

#line 65 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":69
 *             return PyDateTime_TIME_GET_SECOND(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
*/

#line 69 "cpython/datetime.pxd"


#line 69 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_11microsecond_microsecond(PyDateTime_Time *__pyx_v_self) {

#line 69 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":71
 *         @property
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_TIME_GET_MICROSECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 71 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_MICROSECOND(((PyObject *)__pyx_v_self));

#line 71 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":69
 *             return PyDateTime_TIME_GET_SECOND(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
*/

#line 69 "cpython/datetime.pxd"


#line 69 "cpython/datetime.pxd"
  /* function exit code */

#line 69 "cpython/datetime.pxd"
  __pyx_L0:;

#line 69 "cpython/datetime.pxd"
  return __pyx_r;

#line 69 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":73
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
*/

#line 73 "cpython/datetime.pxd"


#line 73 "cpython/datetime.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_4time_6tzinfo_tzinfo(PyDateTime_Time *__pyx_v_self) {

#line 73 "cpython/datetime.pxd"
  PyObject *__pyx_r = NULL;

#line 73 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 73 "cpython/datetime.pxd"
  PyObject *__pyx_t_1;

#line 73 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("tzinfo", 0);

  /* "cpython/datetime.pxd":75
 *         @property
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 75 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 75 "cpython/datetime.pxd"
  __pyx_t_1 = PyDateTime_TIME_GET_TZINFO(((PyObject *)__pyx_v_self));

#line 75 "cpython/datetime.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_t_1));

#line 75 "cpython/datetime.pxd"
  __pyx_r = ((PyObject *)__pyx_t_1);

#line 75 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":73
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
*/

#line 73 "cpython/datetime.pxd"


#line 73 "cpython/datetime.pxd"
  /* function exit code */

#line 73 "cpython/datetime.pxd"
  __pyx_L0:;

#line 73 "cpython/datetime.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 73 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 73 "cpython/datetime.pxd"
  return __pyx_r;

#line 73 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":77
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int fold(self) noexcept:
 *             # For Python < 3.6 this returns 0 no matter what
*/

#line 77 "cpython/datetime.pxd"


#line 77 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4fold_fold(PyDateTime_Time *__pyx_v_self) {

#line 77 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":80
 *         cdef inline int fold(self) noexcept:
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
*/

#line 80 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_FOLD(((PyObject *)__pyx_v_self));

#line 80 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":77
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int fold(self) noexcept:
 *             # For Python < 3.6 this returns 0 no matter what
*/

#line 77 "cpython/datetime.pxd"


#line 77 "cpython/datetime.pxd"
  /* function exit code */

#line 77 "cpython/datetime.pxd"
  __pyx_L0:;

#line 77 "cpython/datetime.pxd"
  return __pyx_r;

#line 77 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":83
 * 
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int year(self) noexcept:
 *             return PyDateTime_GET_YEAR(self)
*/

#line 83 "cpython/datetime.pxd"


#line 83 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4year_year(PyDateTime_DateTime *__pyx_v_self) {

#line 83 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":85
 *         @property
 *         cdef inline int year(self) noexcept:
 *             return PyDateTime_GET_YEAR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 85 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_YEAR(((PyObject *)__pyx_v_self));

#line 85 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":83
 * 
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int year(self) noexcept:
 *             return PyDateTime_GET_YEAR(self)
*/

#line 83 "cpython/datetime.pxd"


#line 83 "cpython/datetime.pxd"
  /* function exit code */

#line 83 "cpython/datetime.pxd"
  __pyx_L0:;

#line 83 "cpython/datetime.pxd"
  return __pyx_r;

#line 83 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":87
 *             return PyDateTime_GET_YEAR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int month(self) noexcept:
 *             return PyDateTime_GET_MONTH(self)
*/

#line 87 "cpython/datetime.pxd"


#line 87 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_5month_month(PyDateTime_DateTime *__pyx_v_self) {

#line 87 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":89
 *         @property
 *         cdef inline int month(self) noexcept:
 *             return PyDateTime_GET_MONTH(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 89 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_MONTH(((PyObject *)__pyx_v_self));

#line 89 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":87
 *             return PyDateTime_GET_YEAR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int month(self) noexcept:
 *             return PyDateTime_GET_MONTH(self)
*/

#line 87 "cpython/datetime.pxd"


#line 87 "cpython/datetime.pxd"
  /* function exit code */

#line 87 "cpython/datetime.pxd"
  __pyx_L0:;

#line 87 "cpython/datetime.pxd"
  return __pyx_r;

#line 87 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":91
 *             return PyDateTime_GET_MONTH(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_GET_DAY(self)
*/

#line 91 "cpython/datetime.pxd"


#line 91 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_3day_day(PyDateTime_DateTime *__pyx_v_self) {

#line 91 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":93
 *         @property
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_GET_DAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 93 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_DAY(((PyObject *)__pyx_v_self));

#line 93 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":91
 *             return PyDateTime_GET_MONTH(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_GET_DAY(self)
*/

#line 91 "cpython/datetime.pxd"


#line 91 "cpython/datetime.pxd"
  /* function exit code */

#line 91 "cpython/datetime.pxd"
  __pyx_L0:;

#line 91 "cpython/datetime.pxd"
  return __pyx_r;

#line 91 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":95
 *             return PyDateTime_GET_DAY(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int hour(self) noexcept:
 *             return PyDateTime_DATE_GET_HOUR(self)
*/

#line 95 "cpython/datetime.pxd"


#line 95 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4hour_hour(PyDateTime_DateTime *__pyx_v_self) {

#line 95 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":97
 *         @property
 *         cdef inline int hour(self) noexcept:
 *             return PyDateTime_DATE_GET_HOUR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 97 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_HOUR(((PyObject *)__pyx_v_self));

#line 97 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":95
 *             return PyDateTime_GET_DAY(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int hour(self) noexcept:
 *             return PyDateTime_DATE_GET_HOUR(self)
*/

#line 95 "cpython/datetime.pxd"


#line 95 "cpython/datetime.pxd"
  /* function exit code */

#line 95 "cpython/datetime.pxd"
  __pyx_L0:;

#line 95 "cpython/datetime.pxd"
  return __pyx_r;

#line 95 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":99
 *             return PyDateTime_DATE_GET_HOUR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int minute(self) noexcept:
 *             return PyDateTime_DATE_GET_MINUTE(self)
*/

#line 99 "cpython/datetime.pxd"


#line 99 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6minute_minute(PyDateTime_DateTime *__pyx_v_self) {

#line 99 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":101
 *         @property
 *         cdef inline int minute(self) noexcept:
 *             return PyDateTime_DATE_GET_MINUTE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 101 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_MINUTE(((PyObject *)__pyx_v_self));

#line 101 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":99
 *             return PyDateTime_DATE_GET_HOUR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int minute(self) noexcept:
 *             return PyDateTime_DATE_GET_MINUTE(self)
*/

#line 99 "cpython/datetime.pxd"


#line 99 "cpython/datetime.pxd"
  /* function exit code */

#line 99 "cpython/datetime.pxd"
  __pyx_L0:;

#line 99 "cpython/datetime.pxd"
  return __pyx_r;

#line 99 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":103
 *             return PyDateTime_DATE_GET_MINUTE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_DATE_GET_SECOND(self)
*/

#line 103 "cpython/datetime.pxd"


#line 103 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6second_second(PyDateTime_DateTime *__pyx_v_self) {

#line 103 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":105
 *         @property
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_DATE_GET_SECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 105 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_SECOND(((PyObject *)__pyx_v_self));

#line 105 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":103
 *             return PyDateTime_DATE_GET_MINUTE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_DATE_GET_SECOND(self)
*/

#line 103 "cpython/datetime.pxd"


#line 103 "cpython/datetime.pxd"
  /* function exit code */

#line 103 "cpython/datetime.pxd"
  __pyx_L0:;

#line 103 "cpython/datetime.pxd"
  return __pyx_r;

#line 103 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":107
 *             return PyDateTime_DATE_GET_SECOND(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
*/

#line 107 "cpython/datetime.pxd"


#line 107 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_11microsecond_microsecond(PyDateTime_DateTime *__pyx_v_self) {

#line 107 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":109
 *         @property
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_DATE_GET_MICROSECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 109 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_MICROSECOND(((PyObject *)__pyx_v_self));

#line 109 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":107
 *             return PyDateTime_DATE_GET_SECOND(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
*/

#line 107 "cpython/datetime.pxd"


#line 107 "cpython/datetime.pxd"
  /* function exit code */

#line 107 "cpython/datetime.pxd"
  __pyx_L0:;

#line 107 "cpython/datetime.pxd"
  return __pyx_r;

#line 107 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":111
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
*/

#line 111 "cpython/datetime.pxd"


#line 111 "cpython/datetime.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_8datetime_6tzinfo_tzinfo(PyDateTime_DateTime *__pyx_v_self) {

#line 111 "cpython/datetime.pxd"
  PyObject *__pyx_r = NULL;

#line 111 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 111 "cpython/datetime.pxd"
  PyObject *__pyx_t_1;

#line 111 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("tzinfo", 0);

  /* "cpython/datetime.pxd":113
 *         @property
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 113 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 113 "cpython/datetime.pxd"
  __pyx_t_1 = PyDateTime_DATE_GET_TZINFO(((PyObject *)__pyx_v_self));

#line 113 "cpython/datetime.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_t_1));

#line 113 "cpython/datetime.pxd"
  __pyx_r = ((PyObject *)__pyx_t_1);

#line 113 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":111
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
*/

#line 111 "cpython/datetime.pxd"


#line 111 "cpython/datetime.pxd"
  /* function exit code */

#line 111 "cpython/datetime.pxd"
  __pyx_L0:;

#line 111 "cpython/datetime.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 111 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 111 "cpython/datetime.pxd"
  return __pyx_r;

#line 111 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":115
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int fold(self) noexcept:
 *             # For Python < 3.6 this returns 0 no matter what
*/

#line 115 "cpython/datetime.pxd"


#line 115 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4fold_fold(PyDateTime_DateTime *__pyx_v_self) {

#line 115 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":118
 *         cdef inline int fold(self) noexcept:
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
*/

#line 118 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_FOLD(((PyObject *)__pyx_v_self));

#line 118 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":115
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int fold(self) noexcept:
 *             # For Python < 3.6 this returns 0 no matter what
*/

#line 115 "cpython/datetime.pxd"


#line 115 "cpython/datetime.pxd"
  /* function exit code */

#line 115 "cpython/datetime.pxd"
  __pyx_L0:;

#line 115 "cpython/datetime.pxd"
  return __pyx_r;

#line 115 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":121
 * 
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_DELTA_GET_DAYS(self)
*/

#line 121 "cpython/datetime.pxd"


#line 121 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_3day_day(PyDateTime_Delta *__pyx_v_self) {

#line 121 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":123
 *         @property
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_DELTA_GET_DAYS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 123 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DELTA_GET_DAYS(((PyObject *)__pyx_v_self));

#line 123 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":121
 * 
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_DELTA_GET_DAYS(self)
*/

#line 121 "cpython/datetime.pxd"


#line 121 "cpython/datetime.pxd"
  /* function exit code */

#line 121 "cpython/datetime.pxd"
  __pyx_L0:;

#line 121 "cpython/datetime.pxd"
  return __pyx_r;

#line 121 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":125
 *             return PyDateTime_DELTA_GET_DAYS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_DELTA_GET_SECONDS(self)
*/

#line 125 "cpython/datetime.pxd"


#line 125 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_6second_second(PyDateTime_Delta *__pyx_v_self) {

#line 125 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":127
 *         @property
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_DELTA_GET_SECONDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/

#line 127 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DELTA_GET_SECONDS(((PyObject *)__pyx_v_self));

#line 127 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":125
 *             return PyDateTime_DELTA_GET_DAYS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_DELTA_GET_SECONDS(self)
*/

#line 125 "cpython/datetime.pxd"


#line 125 "cpython/datetime.pxd"
  /* function exit code */

#line 125 "cpython/datetime.pxd"
  __pyx_L0:;

#line 125 "cpython/datetime.pxd"
  return __pyx_r;

#line 125 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":129
 *             return PyDateTime_DELTA_GET_SECONDS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)
*/

#line 129 "cpython/datetime.pxd"


#line 129 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_11microsecond_microsecond(PyDateTime_Delta *__pyx_v_self) {

#line 129 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":131
 *         @property
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.tzinfo[object PyDateTime_TZInfo]:
*/

#line 131 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DELTA_GET_MICROSECONDS(((PyObject *)__pyx_v_self));

#line 131 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":129
 *             return PyDateTime_DELTA_GET_SECONDS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)
*/

#line 129 "cpython/datetime.pxd"


#line 129 "cpython/datetime.pxd"
  /* function exit code */

#line 129 "cpython/datetime.pxd"
  __pyx_L0:;

#line 129 "cpython/datetime.pxd"
  return __pyx_r;

#line 129 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":255
 * # Datetime C API initialization function.
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime() noexcept:             # <<<<<<<<<<<<<<
 *     PyDateTime_IMPORT
 * 
*/

#line 255 "cpython/datetime.pxd"


#line 255 "cpython/datetime.pxd"
static CYTHON_INLINE void __pyx_f_7cpython_8datetime_import_datetime(void) {

  /* "cpython/datetime.pxd":256
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime() noexcept:
 *     PyDateTime_IMPORT             # <<<<<<<<<<<<<<
 * 
 * # Create date object using DateTime CAPI factory function.
*/

#line 256 "cpython/datetime.pxd"
  (void)(PyDateTime_IMPORT);

  /* "cpython/datetime.pxd":255
 * # Datetime C API initialization function.
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime() noexcept:             # <<<<<<<<<<<<<<
 *     PyDateTime_IMPORT
 * 
*/

#line 255 "cpython/datetime.pxd"


#line 255 "cpython/datetime.pxd"
  /* function exit code */

#line 255 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":260
 * # Create date object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)
 * 
*/

#line 260 "cpython/datetime.pxd"


#line 260 "cpython/datetime.pxd"
static CYTHON_INLINE PyDateTime_Date *__pyx_f_7cpython_8datetime_date_new(int __pyx_v_year, int __pyx_v_month, int __pyx_v_day) {

#line 260 "cpython/datetime.pxd"
  PyDateTime_Date *__pyx_r = NULL;

#line 260 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 260 "cpython/datetime.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 260 "cpython/datetime.pxd"
  int __pyx_lineno = 0;

#line 260 "cpython/datetime.pxd"
  const char *__pyx_filename = NULL;

#line 260 "cpython/datetime.pxd"
  int __pyx_clineno = 0;

#line 260 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("date_new", 0);

  /* "cpython/datetime.pxd":261
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)             # <<<<<<<<<<<<<<
 * 
 * # Create time object using DateTime CAPI factory function
*/

#line 261 "cpython/datetime.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_r);

#line 261 "cpython/datetime.pxd"
  __pyx_t_1 = ((PyObject *)PyDateTimeAPI->Date_FromDate(__pyx_v_year, __pyx_v_month, __pyx_v_day, PyDateTimeAPI->DateType)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 261, __pyx_L1_error)

#line 261 "cpython/datetime.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 261 "cpython/datetime.pxd"
  __pyx_r = ((PyDateTime_Date *)__pyx_t_1);

#line 261 "cpython/datetime.pxd"
  __pyx_t_1 = 0;

#line 261 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":260
 * # Create date object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)
 * 
*/

#line 260 "cpython/datetime.pxd"


#line 260 "cpython/datetime.pxd"
  /* function exit code */

#line 260 "cpython/datetime.pxd"
  __pyx_L1_error:;

#line 260 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 260 "cpython/datetime.pxd"
  __Pyx_AddTraceback("cpython.datetime.date_new", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 260 "cpython/datetime.pxd"
  __pyx_r = 0;

#line 260 "cpython/datetime.pxd"
  __pyx_L0:;

#line 260 "cpython/datetime.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 260 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 260 "cpython/datetime.pxd"
  return __pyx_r;

#line 260 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":265
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
*/

#line 265 "cpython/datetime.pxd"


#line 265 "cpython/datetime.pxd"
static CYTHON_INLINE PyDateTime_Time *__pyx_f_7cpython_8datetime_time_new(int __pyx_v_hour, int __pyx_v_minute, int __pyx_v_second, int __pyx_v_microsecond, PyObject *__pyx_v_tz, struct __pyx_opt_args_7cpython_8datetime_time_new *__pyx_optional_args) {

#line 265 "cpython/datetime.pxd"
  int __pyx_v_fold = ((int)0);

#line 265 "cpython/datetime.pxd"
  PyDateTime_Time *__pyx_r = NULL;

#line 265 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 265 "cpython/datetime.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 265 "cpython/datetime.pxd"
  int __pyx_lineno = 0;

#line 265 "cpython/datetime.pxd"
  const char *__pyx_filename = NULL;

#line 265 "cpython/datetime.pxd"
  int __pyx_clineno = 0;

#line 265 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("time_new", 0);

#line 265 "cpython/datetime.pxd"
  if (__pyx_optional_args) {

#line 265 "cpython/datetime.pxd"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 265 "cpython/datetime.pxd"
      __pyx_v_fold = __pyx_optional_args->fold;

#line 265 "cpython/datetime.pxd"
    }

#line 265 "cpython/datetime.pxd"
  }

  /* "cpython/datetime.pxd":266
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             # <<<<<<<<<<<<<<
 * 
 * # Create datetime object using DateTime CAPI factory function.
*/

#line 266 "cpython/datetime.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_r);

#line 266 "cpython/datetime.pxd"
  __pyx_t_1 = ((PyObject *)__Pyx_DateTime_TimeWithFold(__pyx_v_hour, __pyx_v_minute, __pyx_v_second, __pyx_v_microsecond, __pyx_v_tz, __pyx_v_fold)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 266, __pyx_L1_error)

#line 266 "cpython/datetime.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 266 "cpython/datetime.pxd"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_time))))) __PYX_ERR(4, 266, __pyx_L1_error)

#line 266 "cpython/datetime.pxd"
  __pyx_r = ((PyDateTime_Time *)__pyx_t_1);

#line 266 "cpython/datetime.pxd"
  __pyx_t_1 = 0;

#line 266 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":265
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
*/

#line 265 "cpython/datetime.pxd"


#line 265 "cpython/datetime.pxd"
  /* function exit code */

#line 265 "cpython/datetime.pxd"
  __pyx_L1_error:;

#line 265 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 265 "cpython/datetime.pxd"
  __Pyx_AddTraceback("cpython.datetime.time_new", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 265 "cpython/datetime.pxd"
  __pyx_r = 0;

#line 265 "cpython/datetime.pxd"
  __pyx_L0:;

#line 265 "cpython/datetime.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 265 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 265 "cpython/datetime.pxd"
  return __pyx_r;

#line 265 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":270
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
*/

#line 270 "cpython/datetime.pxd"


#line 270 "cpython/datetime.pxd"
static CYTHON_INLINE PyDateTime_DateTime *__pyx_f_7cpython_8datetime_datetime_new(int __pyx_v_year, int __pyx_v_month, int __pyx_v_day, int __pyx_v_hour, int __pyx_v_minute, int __pyx_v_second, int __pyx_v_microsecond, PyObject *__pyx_v_tz, struct __pyx_opt_args_7cpython_8datetime_datetime_new *__pyx_optional_args) {

#line 270 "cpython/datetime.pxd"
  int __pyx_v_fold = ((int)0);

#line 270 "cpython/datetime.pxd"
  PyDateTime_DateTime *__pyx_r = NULL;

#line 270 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 270 "cpython/datetime.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 270 "cpython/datetime.pxd"
  int __pyx_lineno = 0;

#line 270 "cpython/datetime.pxd"
  const char *__pyx_filename = NULL;

#line 270 "cpython/datetime.pxd"
  int __pyx_clineno = 0;

#line 270 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("datetime_new", 0);

#line 270 "cpython/datetime.pxd"
  if (__pyx_optional_args) {

#line 270 "cpython/datetime.pxd"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 270 "cpython/datetime.pxd"
      __pyx_v_fold = __pyx_optional_args->fold;

#line 270 "cpython/datetime.pxd"
    }

#line 270 "cpython/datetime.pxd"
  }

  /* "cpython/datetime.pxd":271
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             # <<<<<<<<<<<<<<
 * 
 * # Create timedelta object using DateTime CAPI factory function.
*/

#line 271 "cpython/datetime.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_r);

#line 271 "cpython/datetime.pxd"
  __pyx_t_1 = ((PyObject *)__Pyx_DateTime_DateTimeWithFold(__pyx_v_year, __pyx_v_month, __pyx_v_day, __pyx_v_hour, __pyx_v_minute, __pyx_v_second, __pyx_v_microsecond, __pyx_v_tz, __pyx_v_fold)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 271, __pyx_L1_error)

#line 271 "cpython/datetime.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 271 "cpython/datetime.pxd"
  __pyx_r = ((PyDateTime_DateTime *)__pyx_t_1);

#line 271 "cpython/datetime.pxd"
  __pyx_t_1 = 0;

#line 271 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":270
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
*/

#line 270 "cpython/datetime.pxd"


#line 270 "cpython/datetime.pxd"
  /* function exit code */

#line 270 "cpython/datetime.pxd"
  __pyx_L1_error:;

#line 270 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 270 "cpython/datetime.pxd"
  __Pyx_AddTraceback("cpython.datetime.datetime_new", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 270 "cpython/datetime.pxd"
  __pyx_r = 0;

#line 270 "cpython/datetime.pxd"
  __pyx_L0:;

#line 270 "cpython/datetime.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 270 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 270 "cpython/datetime.pxd"
  return __pyx_r;

#line 270 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":275
 * # Create timedelta object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)
 * 
*/

#line 275 "cpython/datetime.pxd"


#line 275 "cpython/datetime.pxd"
static CYTHON_INLINE PyDateTime_Delta *__pyx_f_7cpython_8datetime_timedelta_new(int __pyx_v_days, int __pyx_v_seconds, int __pyx_v_useconds) {

#line 275 "cpython/datetime.pxd"
  PyDateTime_Delta *__pyx_r = NULL;

#line 275 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 275 "cpython/datetime.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 275 "cpython/datetime.pxd"
  int __pyx_lineno = 0;

#line 275 "cpython/datetime.pxd"
  const char *__pyx_filename = NULL;

#line 275 "cpython/datetime.pxd"
  int __pyx_clineno = 0;

#line 275 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("timedelta_new", 0);

  /* "cpython/datetime.pxd":276
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)             # <<<<<<<<<<<<<<
 * 
 * # Create timedelta object using DateTime CAPI factory function.
*/

#line 276 "cpython/datetime.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_r);

#line 276 "cpython/datetime.pxd"
  __pyx_t_1 = ((PyObject *)PyDateTimeAPI->Delta_FromDelta(__pyx_v_days, __pyx_v_seconds, __pyx_v_useconds, 1, PyDateTimeAPI->DeltaType)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 276, __pyx_L1_error)

#line 276 "cpython/datetime.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 276 "cpython/datetime.pxd"
  __pyx_r = ((PyDateTime_Delta *)__pyx_t_1);

#line 276 "cpython/datetime.pxd"
  __pyx_t_1 = 0;

#line 276 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":275
 * # Create timedelta object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)
 * 
*/

#line 275 "cpython/datetime.pxd"


#line 275 "cpython/datetime.pxd"
  /* function exit code */

#line 275 "cpython/datetime.pxd"
  __pyx_L1_error:;

#line 275 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 275 "cpython/datetime.pxd"
  __Pyx_AddTraceback("cpython.datetime.timedelta_new", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 275 "cpython/datetime.pxd"
  __pyx_r = 0;

#line 275 "cpython/datetime.pxd"
  __pyx_L0:;

#line 275 "cpython/datetime.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 275 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 275 "cpython/datetime.pxd"
  return __pyx_r;

#line 275 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":279
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 * 
*/

#line 279 "cpython/datetime.pxd"


#line 279 "cpython/datetime.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_timezone_new(PyObject *__pyx_v_offset, struct __pyx_opt_args_7cpython_8datetime_timezone_new *__pyx_optional_args) {

#line 279 "cpython/datetime.pxd"
  PyObject *__pyx_v_name = ((PyObject *)Py_None);

#line 279 "cpython/datetime.pxd"
  PyObject *__pyx_r = NULL;

#line 279 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 279 "cpython/datetime.pxd"
  PyObject *__pyx_t_1;

#line 279 "cpython/datetime.pxd"
  int __pyx_t_2;

#line 279 "cpython/datetime.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 279 "cpython/datetime.pxd"
  int __pyx_lineno = 0;

#line 279 "cpython/datetime.pxd"
  const char *__pyx_filename = NULL;

#line 279 "cpython/datetime.pxd"
  int __pyx_clineno = 0;

#line 279 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("timezone_new", 0);

#line 279 "cpython/datetime.pxd"
  if (__pyx_optional_args) {

#line 279 "cpython/datetime.pxd"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 279 "cpython/datetime.pxd"
      __pyx_v_name = __pyx_optional_args->name;

#line 279 "cpython/datetime.pxd"
    }

#line 279 "cpython/datetime.pxd"
  }

  /* "cpython/datetime.pxd":280
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)             # <<<<<<<<<<<<<<
 * 
 * # Create datetime object using DB API constructor.
*/

#line 280 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 280 "cpython/datetime.pxd"
  __pyx_t_2 = (__pyx_v_name != Py_None);

#line 280 "cpython/datetime.pxd"
  if (__pyx_t_2) {

#line 280 "cpython/datetime.pxd"
    __pyx_t_1 = ((PyObject *)__pyx_v_name);

#line 280 "cpython/datetime.pxd"
  } else {

#line 280 "cpython/datetime.pxd"
    __pyx_t_1 = NULL;

#line 280 "cpython/datetime.pxd"
  }

#line 280 "cpython/datetime.pxd"
  __pyx_t_3 = __Pyx_TimeZone_FromOffsetAndName(__pyx_v_offset, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 280, __pyx_L1_error)

#line 280 "cpython/datetime.pxd"
  __Pyx_GOTREF(__pyx_t_3);

#line 280 "cpython/datetime.pxd"
  __pyx_r = __pyx_t_3;

#line 280 "cpython/datetime.pxd"
  __pyx_t_3 = 0;

#line 280 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":279
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 * 
*/

#line 279 "cpython/datetime.pxd"


#line 279 "cpython/datetime.pxd"
  /* function exit code */

#line 279 "cpython/datetime.pxd"
  __pyx_L1_error:;

#line 279 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_t_3);

#line 279 "cpython/datetime.pxd"
  __Pyx_AddTraceback("cpython.datetime.timezone_new", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 279 "cpython/datetime.pxd"
  __pyx_r = 0;

#line 279 "cpython/datetime.pxd"
  __pyx_L0:;

#line 279 "cpython/datetime.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 279 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 279 "cpython/datetime.pxd"
  return __pyx_r;

#line 279 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":283
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
*/

#line 283 "cpython/datetime.pxd"


#line 283 "cpython/datetime.pxd"
static CYTHON_INLINE PyDateTime_DateTime *__pyx_f_7cpython_8datetime_datetime_from_timestamp(PyObject *__pyx_v_timestamp, struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp *__pyx_optional_args) {

#line 283 "cpython/datetime.pxd"
  PyObject *__pyx_v_tz = ((PyObject *)Py_None);

#line 283 "cpython/datetime.pxd"
  PyDateTime_DateTime *__pyx_r = NULL;

#line 283 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 283 "cpython/datetime.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 283 "cpython/datetime.pxd"
  int __pyx_t_2;

#line 283 "cpython/datetime.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 283 "cpython/datetime.pxd"
  int __pyx_lineno = 0;

#line 283 "cpython/datetime.pxd"
  const char *__pyx_filename = NULL;

#line 283 "cpython/datetime.pxd"
  int __pyx_clineno = 0;

#line 283 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("datetime_from_timestamp", 0);

#line 283 "cpython/datetime.pxd"
  if (__pyx_optional_args) {

#line 283 "cpython/datetime.pxd"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 283 "cpython/datetime.pxd"
      __pyx_v_tz = __pyx_optional_args->tz;

#line 283 "cpython/datetime.pxd"
    }

#line 283 "cpython/datetime.pxd"
  }

  /* "cpython/datetime.pxd":284
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 * 
*/

#line 284 "cpython/datetime.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "cpython/datetime.pxd":285
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)             # <<<<<<<<<<<<<<
 * 
 * # Create date object using DB API constructor.
*/

#line 285 "cpython/datetime.pxd"
  __pyx_t_2 = (__pyx_v_tz != Py_None);

#line 285 "cpython/datetime.pxd"
  if (__pyx_t_2) {

#line 285 "cpython/datetime.pxd"
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 285, __pyx_L1_error)

#line 285 "cpython/datetime.pxd"
    __Pyx_GOTREF(__pyx_t_3);

#line 285 "cpython/datetime.pxd"
    __Pyx_INCREF(__pyx_v_timestamp);

#line 285 "cpython/datetime.pxd"
    __Pyx_GIVEREF(__pyx_v_timestamp);

#line 285 "cpython/datetime.pxd"
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_timestamp) != (0)) __PYX_ERR(4, 285, __pyx_L1_error);

#line 285 "cpython/datetime.pxd"
    __Pyx_INCREF(__pyx_v_tz);

#line 285 "cpython/datetime.pxd"
    __Pyx_GIVEREF(__pyx_v_tz);

#line 285 "cpython/datetime.pxd"
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_tz) != (0)) __PYX_ERR(4, 285, __pyx_L1_error);

#line 285 "cpython/datetime.pxd"
    __pyx_t_1 = __pyx_t_3;

#line 285 "cpython/datetime.pxd"
    __pyx_t_3 = 0;

#line 285 "cpython/datetime.pxd"
  } else {

#line 285 "cpython/datetime.pxd"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 285, __pyx_L1_error)

#line 285 "cpython/datetime.pxd"
    __Pyx_GOTREF(__pyx_t_3);

#line 285 "cpython/datetime.pxd"
    __Pyx_INCREF(__pyx_v_timestamp);

#line 285 "cpython/datetime.pxd"
    __Pyx_GIVEREF(__pyx_v_timestamp);

#line 285 "cpython/datetime.pxd"
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_timestamp) != (0)) __PYX_ERR(4, 285, __pyx_L1_error);

#line 285 "cpython/datetime.pxd"
    __pyx_t_1 = __pyx_t_3;

#line 285 "cpython/datetime.pxd"
    __pyx_t_3 = 0;

#line 285 "cpython/datetime.pxd"
  }

  /* "cpython/datetime.pxd":284
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 * 
*/

#line 284 "cpython/datetime.pxd"
  __pyx_t_3 = ((PyObject *)PyDateTimeAPI->DateTime_FromTimestamp(((PyObject *)PyDateTimeAPI->DateTimeType), __pyx_t_1, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 284, __pyx_L1_error)

#line 284 "cpython/datetime.pxd"
  __Pyx_GOTREF(__pyx_t_3);

#line 284 "cpython/datetime.pxd"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 284 "cpython/datetime.pxd"
  __pyx_r = ((PyDateTime_DateTime *)__pyx_t_3);

#line 284 "cpython/datetime.pxd"
  __pyx_t_3 = 0;

#line 284 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":283
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
*/

#line 283 "cpython/datetime.pxd"


#line 283 "cpython/datetime.pxd"
  /* function exit code */

#line 283 "cpython/datetime.pxd"
  __pyx_L1_error:;

#line 283 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 283 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_t_3);

#line 283 "cpython/datetime.pxd"
  __Pyx_AddTraceback("cpython.datetime.datetime_from_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 283 "cpython/datetime.pxd"
  __pyx_r = 0;

#line 283 "cpython/datetime.pxd"
  __pyx_L0:;

#line 283 "cpython/datetime.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 283 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 283 "cpython/datetime.pxd"
  return __pyx_r;

#line 283 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":288
 * 
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))
 * 
*/

#line 288 "cpython/datetime.pxd"


#line 288 "cpython/datetime.pxd"
static CYTHON_INLINE PyDateTime_Date *__pyx_f_7cpython_8datetime_date_from_timestamp(PyObject *__pyx_v_timestamp) {

#line 288 "cpython/datetime.pxd"
  PyDateTime_Date *__pyx_r = NULL;

#line 288 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 288 "cpython/datetime.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 288 "cpython/datetime.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 288 "cpython/datetime.pxd"
  int __pyx_lineno = 0;

#line 288 "cpython/datetime.pxd"
  const char *__pyx_filename = NULL;

#line 288 "cpython/datetime.pxd"
  int __pyx_clineno = 0;

#line 288 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("date_from_timestamp", 0);

  /* "cpython/datetime.pxd":289
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))             # <<<<<<<<<<<<<<
 * 
 * # More recognizable getters for date/time/datetime/timedelta.
*/

#line 289 "cpython/datetime.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_r);

#line 289 "cpython/datetime.pxd"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 289, __pyx_L1_error)

#line 289 "cpython/datetime.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 289 "cpython/datetime.pxd"
  __Pyx_INCREF(__pyx_v_timestamp);

#line 289 "cpython/datetime.pxd"
  __Pyx_GIVEREF(__pyx_v_timestamp);

#line 289 "cpython/datetime.pxd"
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_timestamp) != (0)) __PYX_ERR(4, 289, __pyx_L1_error);

#line 289 "cpython/datetime.pxd"
  __pyx_t_2 = ((PyObject *)PyDateTimeAPI->Date_FromTimestamp(((PyObject *)PyDateTimeAPI->DateType), __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 289, __pyx_L1_error)

#line 289 "cpython/datetime.pxd"
  __Pyx_GOTREF(__pyx_t_2);

#line 289 "cpython/datetime.pxd"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 289 "cpython/datetime.pxd"
  __pyx_r = ((PyDateTime_Date *)__pyx_t_2);

#line 289 "cpython/datetime.pxd"
  __pyx_t_2 = 0;

#line 289 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":288
 * 
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))
 * 
*/

#line 288 "cpython/datetime.pxd"


#line 288 "cpython/datetime.pxd"
  /* function exit code */

#line 288 "cpython/datetime.pxd"
  __pyx_L1_error:;

#line 288 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 288 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_t_2);

#line 288 "cpython/datetime.pxd"
  __Pyx_AddTraceback("cpython.datetime.date_from_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 288 "cpython/datetime.pxd"
  __pyx_r = 0;

#line 288 "cpython/datetime.pxd"
  __pyx_L0:;

#line 288 "cpython/datetime.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 288 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 288 "cpython/datetime.pxd"
  return __pyx_r;

#line 288 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":297
 * 
 * # Get UTC singleton
 * cdef inline object get_utc():             # <<<<<<<<<<<<<<
 *     return <object>__Pyx_TimeZone_UTC
 * 
*/

#line 297 "cpython/datetime.pxd"


#line 297 "cpython/datetime.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_get_utc(void) {

#line 297 "cpython/datetime.pxd"
  PyObject *__pyx_r = NULL;

#line 297 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 297 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("get_utc", 0);

  /* "cpython/datetime.pxd":298
 * # Get UTC singleton
 * cdef inline object get_utc():
 *     return <object>__Pyx_TimeZone_UTC             # <<<<<<<<<<<<<<
 * 
 * # Get tzinfo of time
*/

#line 298 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 298 "cpython/datetime.pxd"
  __Pyx_INCREF(((PyObject *)__Pyx_TimeZone_UTC));

#line 298 "cpython/datetime.pxd"
  __pyx_r = ((PyObject *)__Pyx_TimeZone_UTC);

#line 298 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":297
 * 
 * # Get UTC singleton
 * cdef inline object get_utc():             # <<<<<<<<<<<<<<
 *     return <object>__Pyx_TimeZone_UTC
 * 
*/

#line 297 "cpython/datetime.pxd"


#line 297 "cpython/datetime.pxd"
  /* function exit code */

#line 297 "cpython/datetime.pxd"
  __pyx_L0:;

#line 297 "cpython/datetime.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 297 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 297 "cpython/datetime.pxd"
  return __pyx_r;

#line 297 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":301
 * 
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)
 * 
*/

#line 301 "cpython/datetime.pxd"


#line 301 "cpython/datetime.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_time_tzinfo(PyObject *__pyx_v_o) {

#line 301 "cpython/datetime.pxd"
  PyObject *__pyx_r = NULL;

#line 301 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 301 "cpython/datetime.pxd"
  PyObject *__pyx_t_1;

#line 301 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("time_tzinfo", 0);

  /* "cpython/datetime.pxd":302
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)             # <<<<<<<<<<<<<<
 * 
 * # Get tzinfo of datetime
*/

#line 302 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 302 "cpython/datetime.pxd"
  __pyx_t_1 = PyDateTime_TIME_GET_TZINFO(__pyx_v_o);

#line 302 "cpython/datetime.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_t_1));

#line 302 "cpython/datetime.pxd"
  __pyx_r = ((PyObject *)__pyx_t_1);

#line 302 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":301
 * 
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)
 * 
*/

#line 301 "cpython/datetime.pxd"


#line 301 "cpython/datetime.pxd"
  /* function exit code */

#line 301 "cpython/datetime.pxd"
  __pyx_L0:;

#line 301 "cpython/datetime.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 301 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 301 "cpython/datetime.pxd"
  return __pyx_r;

#line 301 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":305
 * 
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)
 * 
*/

#line 305 "cpython/datetime.pxd"


#line 305 "cpython/datetime.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_datetime_tzinfo(PyObject *__pyx_v_o) {

#line 305 "cpython/datetime.pxd"
  PyObject *__pyx_r = NULL;

#line 305 "cpython/datetime.pxd"
  __Pyx_RefNannyDeclarations

#line 305 "cpython/datetime.pxd"
  PyObject *__pyx_t_1;

#line 305 "cpython/datetime.pxd"
  __Pyx_RefNannySetupContext("datetime_tzinfo", 0);

  /* "cpython/datetime.pxd":306
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)             # <<<<<<<<<<<<<<
 * 
 * # Get year of date
*/

#line 306 "cpython/datetime.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 306 "cpython/datetime.pxd"
  __pyx_t_1 = PyDateTime_DATE_GET_TZINFO(__pyx_v_o);

#line 306 "cpython/datetime.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_t_1));

#line 306 "cpython/datetime.pxd"
  __pyx_r = ((PyObject *)__pyx_t_1);

#line 306 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":305
 * 
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)
 * 
*/

#line 305 "cpython/datetime.pxd"


#line 305 "cpython/datetime.pxd"
  /* function exit code */

#line 305 "cpython/datetime.pxd"
  __pyx_L0:;

#line 305 "cpython/datetime.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 305 "cpython/datetime.pxd"
  __Pyx_RefNannyFinishContext();

#line 305 "cpython/datetime.pxd"
  return __pyx_r;

#line 305 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":309
 * 
 * # Get year of date
 * cdef inline int date_year(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
*/

#line 309 "cpython/datetime.pxd"


#line 309 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_year(PyObject *__pyx_v_o) {

#line 309 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":310
 * # Get year of date
 * cdef inline int date_year(object o) noexcept:
 *     return PyDateTime_GET_YEAR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get month of date
*/

#line 310 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_YEAR(__pyx_v_o);

#line 310 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":309
 * 
 * # Get year of date
 * cdef inline int date_year(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
*/

#line 309 "cpython/datetime.pxd"


#line 309 "cpython/datetime.pxd"
  /* function exit code */

#line 309 "cpython/datetime.pxd"
  __pyx_L0:;

#line 309 "cpython/datetime.pxd"
  return __pyx_r;

#line 309 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":313
 * 
 * # Get month of date
 * cdef inline int date_month(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
*/

#line 313 "cpython/datetime.pxd"


#line 313 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_month(PyObject *__pyx_v_o) {

#line 313 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":314
 * # Get month of date
 * cdef inline int date_month(object o) noexcept:
 *     return PyDateTime_GET_MONTH(o)             # <<<<<<<<<<<<<<
 * 
 * # Get day of date
*/

#line 314 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_MONTH(__pyx_v_o);

#line 314 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":313
 * 
 * # Get month of date
 * cdef inline int date_month(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
*/

#line 313 "cpython/datetime.pxd"


#line 313 "cpython/datetime.pxd"
  /* function exit code */

#line 313 "cpython/datetime.pxd"
  __pyx_L0:;

#line 313 "cpython/datetime.pxd"
  return __pyx_r;

#line 313 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":317
 * 
 * # Get day of date
 * cdef inline int date_day(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
*/

#line 317 "cpython/datetime.pxd"


#line 317 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_day(PyObject *__pyx_v_o) {

#line 317 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":318
 * # Get day of date
 * cdef inline int date_day(object o) noexcept:
 *     return PyDateTime_GET_DAY(o)             # <<<<<<<<<<<<<<
 * 
 * # Get year of datetime
*/

#line 318 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_DAY(__pyx_v_o);

#line 318 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":317
 * 
 * # Get day of date
 * cdef inline int date_day(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
*/

#line 317 "cpython/datetime.pxd"


#line 317 "cpython/datetime.pxd"
  /* function exit code */

#line 317 "cpython/datetime.pxd"
  __pyx_L0:;

#line 317 "cpython/datetime.pxd"
  return __pyx_r;

#line 317 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":321
 * 
 * # Get year of datetime
 * cdef inline int datetime_year(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
*/

#line 321 "cpython/datetime.pxd"


#line 321 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_year(PyObject *__pyx_v_o) {

#line 321 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":322
 * # Get year of datetime
 * cdef inline int datetime_year(object o) noexcept:
 *     return PyDateTime_GET_YEAR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get month of datetime
*/

#line 322 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_YEAR(__pyx_v_o);

#line 322 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":321
 * 
 * # Get year of datetime
 * cdef inline int datetime_year(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
*/

#line 321 "cpython/datetime.pxd"


#line 321 "cpython/datetime.pxd"
  /* function exit code */

#line 321 "cpython/datetime.pxd"
  __pyx_L0:;

#line 321 "cpython/datetime.pxd"
  return __pyx_r;

#line 321 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":325
 * 
 * # Get month of datetime
 * cdef inline int datetime_month(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
*/

#line 325 "cpython/datetime.pxd"


#line 325 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_month(PyObject *__pyx_v_o) {

#line 325 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":326
 * # Get month of datetime
 * cdef inline int datetime_month(object o) noexcept:
 *     return PyDateTime_GET_MONTH(o)             # <<<<<<<<<<<<<<
 * 
 * # Get day of datetime
*/

#line 326 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_MONTH(__pyx_v_o);

#line 326 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":325
 * 
 * # Get month of datetime
 * cdef inline int datetime_month(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
*/

#line 325 "cpython/datetime.pxd"


#line 325 "cpython/datetime.pxd"
  /* function exit code */

#line 325 "cpython/datetime.pxd"
  __pyx_L0:;

#line 325 "cpython/datetime.pxd"
  return __pyx_r;

#line 325 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":329
 * 
 * # Get day of datetime
 * cdef inline int datetime_day(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
*/

#line 329 "cpython/datetime.pxd"


#line 329 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_day(PyObject *__pyx_v_o) {

#line 329 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":330
 * # Get day of datetime
 * cdef inline int datetime_day(object o) noexcept:
 *     return PyDateTime_GET_DAY(o)             # <<<<<<<<<<<<<<
 * 
 * # Get hour of time
*/

#line 330 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_GET_DAY(__pyx_v_o);

#line 330 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":329
 * 
 * # Get day of datetime
 * cdef inline int datetime_day(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
*/

#line 329 "cpython/datetime.pxd"


#line 329 "cpython/datetime.pxd"
  /* function exit code */

#line 329 "cpython/datetime.pxd"
  __pyx_L0:;

#line 329 "cpython/datetime.pxd"
  return __pyx_r;

#line 329 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":333
 * 
 * # Get hour of time
 * cdef inline int time_hour(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_HOUR(o)
 * 
*/

#line 333 "cpython/datetime.pxd"


#line 333 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_hour(PyObject *__pyx_v_o) {

#line 333 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":334
 * # Get hour of time
 * cdef inline int time_hour(object o) noexcept:
 *     return PyDateTime_TIME_GET_HOUR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get minute of time
*/

#line 334 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_HOUR(__pyx_v_o);

#line 334 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":333
 * 
 * # Get hour of time
 * cdef inline int time_hour(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_HOUR(o)
 * 
*/

#line 333 "cpython/datetime.pxd"


#line 333 "cpython/datetime.pxd"
  /* function exit code */

#line 333 "cpython/datetime.pxd"
  __pyx_L0:;

#line 333 "cpython/datetime.pxd"
  return __pyx_r;

#line 333 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":337
 * 
 * # Get minute of time
 * cdef inline int time_minute(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MINUTE(o)
 * 
*/

#line 337 "cpython/datetime.pxd"


#line 337 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_minute(PyObject *__pyx_v_o) {

#line 337 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":338
 * # Get minute of time
 * cdef inline int time_minute(object o) noexcept:
 *     return PyDateTime_TIME_GET_MINUTE(o)             # <<<<<<<<<<<<<<
 * 
 * # Get second of time
*/

#line 338 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_MINUTE(__pyx_v_o);

#line 338 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":337
 * 
 * # Get minute of time
 * cdef inline int time_minute(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MINUTE(o)
 * 
*/

#line 337 "cpython/datetime.pxd"


#line 337 "cpython/datetime.pxd"
  /* function exit code */

#line 337 "cpython/datetime.pxd"
  __pyx_L0:;

#line 337 "cpython/datetime.pxd"
  return __pyx_r;

#line 337 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":341
 * 
 * # Get second of time
 * cdef inline int time_second(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_SECOND(o)
 * 
*/

#line 341 "cpython/datetime.pxd"


#line 341 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_second(PyObject *__pyx_v_o) {

#line 341 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":342
 * # Get second of time
 * cdef inline int time_second(object o) noexcept:
 *     return PyDateTime_TIME_GET_SECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get microsecond of time
*/

#line 342 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_SECOND(__pyx_v_o);

#line 342 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":341
 * 
 * # Get second of time
 * cdef inline int time_second(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_SECOND(o)
 * 
*/

#line 341 "cpython/datetime.pxd"


#line 341 "cpython/datetime.pxd"
  /* function exit code */

#line 341 "cpython/datetime.pxd"
  __pyx_L0:;

#line 341 "cpython/datetime.pxd"
  return __pyx_r;

#line 341 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":345
 * 
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MICROSECOND(o)
 * 
*/

#line 345 "cpython/datetime.pxd"


#line 345 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_microsecond(PyObject *__pyx_v_o) {

#line 345 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":346
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o) noexcept:
 *     return PyDateTime_TIME_GET_MICROSECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get fold of time
*/

#line 346 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_MICROSECOND(__pyx_v_o);

#line 346 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":345
 * 
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MICROSECOND(o)
 * 
*/

#line 345 "cpython/datetime.pxd"


#line 345 "cpython/datetime.pxd"
  /* function exit code */

#line 345 "cpython/datetime.pxd"
  __pyx_L0:;

#line 345 "cpython/datetime.pxd"
  return __pyx_r;

#line 345 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":349
 * 
 * # Get fold of time
 * cdef inline int time_fold(object o) noexcept:             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)
*/

#line 349 "cpython/datetime.pxd"


#line 349 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_fold(PyObject *__pyx_v_o) {

#line 349 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":351
 * cdef inline int time_fold(object o) noexcept:
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)             # <<<<<<<<<<<<<<
 * 
 * # Get hour of datetime
*/

#line 351 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_TIME_GET_FOLD(__pyx_v_o);

#line 351 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":349
 * 
 * # Get fold of time
 * cdef inline int time_fold(object o) noexcept:             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)
*/

#line 349 "cpython/datetime.pxd"


#line 349 "cpython/datetime.pxd"
  /* function exit code */

#line 349 "cpython/datetime.pxd"
  __pyx_L0:;

#line 349 "cpython/datetime.pxd"
  return __pyx_r;

#line 349 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":354
 * 
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_HOUR(o)
 * 
*/

#line 354 "cpython/datetime.pxd"


#line 354 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_hour(PyObject *__pyx_v_o) {

#line 354 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":355
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o) noexcept:
 *     return PyDateTime_DATE_GET_HOUR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get minute of datetime
*/

#line 355 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_HOUR(__pyx_v_o);

#line 355 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":354
 * 
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_HOUR(o)
 * 
*/

#line 354 "cpython/datetime.pxd"


#line 354 "cpython/datetime.pxd"
  /* function exit code */

#line 354 "cpython/datetime.pxd"
  __pyx_L0:;

#line 354 "cpython/datetime.pxd"
  return __pyx_r;

#line 354 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":358
 * 
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MINUTE(o)
 * 
*/

#line 358 "cpython/datetime.pxd"


#line 358 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_minute(PyObject *__pyx_v_o) {

#line 358 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":359
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o) noexcept:
 *     return PyDateTime_DATE_GET_MINUTE(o)             # <<<<<<<<<<<<<<
 * 
 * # Get second of datetime
*/

#line 359 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_MINUTE(__pyx_v_o);

#line 359 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":358
 * 
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MINUTE(o)
 * 
*/

#line 358 "cpython/datetime.pxd"


#line 358 "cpython/datetime.pxd"
  /* function exit code */

#line 358 "cpython/datetime.pxd"
  __pyx_L0:;

#line 358 "cpython/datetime.pxd"
  return __pyx_r;

#line 358 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":362
 * 
 * # Get second of datetime
 * cdef inline int datetime_second(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_SECOND(o)
 * 
*/

#line 362 "cpython/datetime.pxd"


#line 362 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_second(PyObject *__pyx_v_o) {

#line 362 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":363
 * # Get second of datetime
 * cdef inline int datetime_second(object o) noexcept:
 *     return PyDateTime_DATE_GET_SECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get microsecond of datetime
*/

#line 363 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_SECOND(__pyx_v_o);

#line 363 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":362
 * 
 * # Get second of datetime
 * cdef inline int datetime_second(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_SECOND(o)
 * 
*/

#line 362 "cpython/datetime.pxd"


#line 362 "cpython/datetime.pxd"
  /* function exit code */

#line 362 "cpython/datetime.pxd"
  __pyx_L0:;

#line 362 "cpython/datetime.pxd"
  return __pyx_r;

#line 362 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":366
 * 
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MICROSECOND(o)
 * 
*/

#line 366 "cpython/datetime.pxd"


#line 366 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_microsecond(PyObject *__pyx_v_o) {

#line 366 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":367
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o) noexcept:
 *     return PyDateTime_DATE_GET_MICROSECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get fold of datetime
*/

#line 367 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_MICROSECOND(__pyx_v_o);

#line 367 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":366
 * 
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MICROSECOND(o)
 * 
*/

#line 366 "cpython/datetime.pxd"


#line 366 "cpython/datetime.pxd"
  /* function exit code */

#line 366 "cpython/datetime.pxd"
  __pyx_L0:;

#line 366 "cpython/datetime.pxd"
  return __pyx_r;

#line 366 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":370
 * 
 * # Get fold of datetime
 * cdef inline int datetime_fold(object o) noexcept:             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)
*/

#line 370 "cpython/datetime.pxd"


#line 370 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_fold(PyObject *__pyx_v_o) {

#line 370 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":372
 * cdef inline int datetime_fold(object o) noexcept:
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)             # <<<<<<<<<<<<<<
 * 
 * # Get days of timedelta
*/

#line 372 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DATE_GET_FOLD(__pyx_v_o);

#line 372 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":370
 * 
 * # Get fold of datetime
 * cdef inline int datetime_fold(object o) noexcept:             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)
*/

#line 370 "cpython/datetime.pxd"


#line 370 "cpython/datetime.pxd"
  /* function exit code */

#line 370 "cpython/datetime.pxd"
  __pyx_L0:;

#line 370 "cpython/datetime.pxd"
  return __pyx_r;

#line 370 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":375
 * 
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DELTA_GET_DAYS(o)
 * 
*/

#line 375 "cpython/datetime.pxd"


#line 375 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_days(PyObject *__pyx_v_o) {

#line 375 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":376
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o) noexcept:
 *     return PyDateTime_DELTA_GET_DAYS(o)             # <<<<<<<<<<<<<<
 * 
 * # Get seconds of timedelta
*/

#line 376 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DELTA_GET_DAYS(__pyx_v_o);

#line 376 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":375
 * 
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DELTA_GET_DAYS(o)
 * 
*/

#line 375 "cpython/datetime.pxd"


#line 375 "cpython/datetime.pxd"
  /* function exit code */

#line 375 "cpython/datetime.pxd"
  __pyx_L0:;

#line 375 "cpython/datetime.pxd"
  return __pyx_r;

#line 375 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":379
 * 
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DELTA_GET_SECONDS(o)
 * 
*/

#line 379 "cpython/datetime.pxd"


#line 379 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_seconds(PyObject *__pyx_v_o) {

#line 379 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":380
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o) noexcept:
 *     return PyDateTime_DELTA_GET_SECONDS(o)             # <<<<<<<<<<<<<<
 * 
 * # Get microseconds of timedelta
*/

#line 380 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DELTA_GET_SECONDS(__pyx_v_o);

#line 380 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":379
 * 
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DELTA_GET_SECONDS(o)
 * 
*/

#line 379 "cpython/datetime.pxd"


#line 379 "cpython/datetime.pxd"
  /* function exit code */

#line 379 "cpython/datetime.pxd"
  __pyx_L0:;

#line 379 "cpython/datetime.pxd"
  return __pyx_r;

#line 379 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":383
 * 
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DELTA_GET_MICROSECONDS(o)
 * 
*/

#line 383 "cpython/datetime.pxd"


#line 383 "cpython/datetime.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_microseconds(PyObject *__pyx_v_o) {

#line 383 "cpython/datetime.pxd"
  int __pyx_r;

  /* "cpython/datetime.pxd":384
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o) noexcept:
 *     return PyDateTime_DELTA_GET_MICROSECONDS(o)             # <<<<<<<<<<<<<<
 * 
 * cdef inline double total_seconds(timedelta obj) noexcept:
*/

#line 384 "cpython/datetime.pxd"
  __pyx_r = PyDateTime_DELTA_GET_MICROSECONDS(__pyx_v_o);

#line 384 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":383
 * 
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DELTA_GET_MICROSECONDS(o)
 * 
*/

#line 383 "cpython/datetime.pxd"


#line 383 "cpython/datetime.pxd"
  /* function exit code */

#line 383 "cpython/datetime.pxd"
  __pyx_L0:;

#line 383 "cpython/datetime.pxd"
  return __pyx_r;

#line 383 "cpython/datetime.pxd"
}

/* "cpython/datetime.pxd":386
 *     return PyDateTime_DELTA_GET_MICROSECONDS(o)
 * 
 * cdef inline double total_seconds(timedelta obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Mirrors the "timedelta.total_seconds()" method.
 *     # Note that this implementation is not guaranteed to give *exactly* the same
*/

#line 386 "cpython/datetime.pxd"


#line 386 "cpython/datetime.pxd"
static CYTHON_INLINE double __pyx_f_7cpython_8datetime_total_seconds(PyDateTime_Delta *__pyx_v_obj) {
  double __pyx_v_days
#line 386 "cpython/datetime.pxd"
;
  double __pyx_v_seconds
#line 386 "cpython/datetime.pxd"
;
  double __pyx_v_micros
#line 386 "cpython/datetime.pxd"
;

#line 386 "cpython/datetime.pxd"
  double __pyx_r;

  /* "cpython/datetime.pxd":392
 *     cdef:
 *         double days, seconds, micros
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)             # <<<<<<<<<<<<<<
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
*/

#line 392 "cpython/datetime.pxd"
  __pyx_v_days = ((double)PyDateTime_DELTA_GET_DAYS(((PyObject *)__pyx_v_obj)));

  /* "cpython/datetime.pxd":393
 *         double days, seconds, micros
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)             # <<<<<<<<<<<<<<
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 *     return days * 24 * 3600 + seconds + micros / 1_000_000
*/

#line 393 "cpython/datetime.pxd"
  __pyx_v_seconds = ((double)PyDateTime_DELTA_GET_SECONDS(((PyObject *)__pyx_v_obj)));

  /* "cpython/datetime.pxd":394
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)             # <<<<<<<<<<<<<<
 *     return days * 24 * 3600 + seconds + micros / 1_000_000
*/

#line 394 "cpython/datetime.pxd"
  __pyx_v_micros = ((double)PyDateTime_DELTA_GET_MICROSECONDS(((PyObject *)__pyx_v_obj)));

  /* "cpython/datetime.pxd":395
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 *     return days * 24 * 3600 + seconds + micros / 1_000_000             # <<<<<<<<<<<<<<
*/

#line 395 "cpython/datetime.pxd"
  __pyx_r = ((((__pyx_v_days * 24.0) * 3600.0) + __pyx_v_seconds) + (__pyx_v_micros / 1000000.0));

#line 395 "cpython/datetime.pxd"
  goto __pyx_L0;

  /* "cpython/datetime.pxd":386
 *     return PyDateTime_DELTA_GET_MICROSECONDS(o)
 * 
 * cdef inline double total_seconds(timedelta obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Mirrors the "timedelta.total_seconds()" method.
 *     # Note that this implementation is not guaranteed to give *exactly* the same
*/

#line 386 "cpython/datetime.pxd"


#line 386 "cpython/datetime.pxd"
  /* function exit code */

#line 386 "cpython/datetime.pxd"
  __pyx_L0:;

#line 386 "cpython/datetime.pxd"
  return __pyx_r;

#line 386 "cpython/datetime.pxd"
}

/* "pyarrow/includes/common.pxd":151
 * 
 * 
 * cdef inline object PyObject_to_object(PyObject* o):             # <<<<<<<<<<<<<<
 *     # Cast to "object" increments reference count
 *     cdef object result = <object> o
*/

#line 151 "pyarrow/includes/common.pxd"


#line 151 "pyarrow/includes/common.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8includes_6common_PyObject_to_object(PyObject *__pyx_v_o) {
  PyObject *__pyx_v_result = 0
#line 151 "pyarrow/includes/common.pxd"
;

#line 151 "pyarrow/includes/common.pxd"
  PyObject *__pyx_r = NULL;

#line 151 "pyarrow/includes/common.pxd"
  __Pyx_RefNannyDeclarations

#line 151 "pyarrow/includes/common.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 151 "pyarrow/includes/common.pxd"
  __Pyx_RefNannySetupContext("PyObject_to_object", 0);

  /* "pyarrow/includes/common.pxd":153
 * cdef inline object PyObject_to_object(PyObject* o):
 *     # Cast to "object" increments reference count
 *     cdef object result = <object> o             # <<<<<<<<<<<<<<
 *     cpython.Py_DECREF(result)
 *     return result
*/

#line 153 "pyarrow/includes/common.pxd"
  __pyx_t_1 = ((PyObject *)__pyx_v_o);

#line 153 "pyarrow/includes/common.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 153 "pyarrow/includes/common.pxd"
  __pyx_v_result = __pyx_t_1;

#line 153 "pyarrow/includes/common.pxd"
  __pyx_t_1 = 0;

  /* "pyarrow/includes/common.pxd":154
 *     # Cast to "object" increments reference count
 *     cdef object result = <object> o
 *     cpython.Py_DECREF(result)             # <<<<<<<<<<<<<<
 *     return result
*/

#line 154 "pyarrow/includes/common.pxd"
  Py_DECREF(__pyx_v_result);

  /* "pyarrow/includes/common.pxd":155
 *     cdef object result = <object> o
 *     cpython.Py_DECREF(result)
 *     return result             # <<<<<<<<<<<<<<
*/

#line 155 "pyarrow/includes/common.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 155 "pyarrow/includes/common.pxd"
  __Pyx_INCREF(__pyx_v_result);

#line 155 "pyarrow/includes/common.pxd"
  __pyx_r = __pyx_v_result;

#line 155 "pyarrow/includes/common.pxd"
  goto __pyx_L0;

  /* "pyarrow/includes/common.pxd":151
 * 
 * 
 * cdef inline object PyObject_to_object(PyObject* o):             # <<<<<<<<<<<<<<
 *     # Cast to "object" increments reference count
 *     cdef object result = <object> o
*/

#line 151 "pyarrow/includes/common.pxd"


#line 151 "pyarrow/includes/common.pxd"
  /* function exit code */

#line 151 "pyarrow/includes/common.pxd"
  __pyx_L0:;

#line 151 "pyarrow/includes/common.pxd"
  __Pyx_XDECREF(__pyx_v_result);

#line 151 "pyarrow/includes/common.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 151 "pyarrow/includes/common.pxd"
  __Pyx_RefNannyFinishContext();

#line 151 "pyarrow/includes/common.pxd"
  return __pyx_r;

#line 151 "pyarrow/includes/common.pxd"
}

/* "pyarrow/_parquet.pxd":30
 *         shared_ptr[CFileEncryptionProperties] properties
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline FileEncryptionProperties wrap(
 *             shared_ptr[CFileEncryptionProperties] properties):
*/

#line 30 "pyarrow/_parquet.pxd"


#line 30 "pyarrow/_parquet.pxd"
static CYTHON_INLINE struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *__pyx_f_7pyarrow_8_parquet_24FileEncryptionProperties_wrap(std::shared_ptr< parquet::FileEncryptionProperties>  __pyx_v_properties) {
  struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *__pyx_v_result = NULL
#line 30 "pyarrow/_parquet.pxd"
;

#line 30 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *__pyx_r = NULL;

#line 30 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyDeclarations

#line 30 "pyarrow/_parquet.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 30 "pyarrow/_parquet.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 30 "pyarrow/_parquet.pxd"
  size_t __pyx_t_3;

#line 30 "pyarrow/_parquet.pxd"
  int __pyx_lineno = 0;

#line 30 "pyarrow/_parquet.pxd"
  const char *__pyx_filename = NULL;

#line 30 "pyarrow/_parquet.pxd"
  int __pyx_clineno = 0;

#line 30 "pyarrow/_parquet.pxd"
  __Pyx_RefNannySetupContext("wrap", 0);

  /* "pyarrow/_parquet.pxd":34
 *             shared_ptr[CFileEncryptionProperties] properties):
 * 
 *         result = FileEncryptionProperties()             # <<<<<<<<<<<<<<
 *         result.properties = properties
 *         return result
*/

#line 34 "pyarrow/_parquet.pxd"
  __pyx_t_2 = NULL;

#line 34 "pyarrow/_parquet.pxd"
  __pyx_t_3 = 1;

#line 34 "pyarrow/_parquet.pxd"
  {

#line 34 "pyarrow/_parquet.pxd"
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};

#line 34 "pyarrow/_parquet.pxd"
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyarrow_8_parquet_FileEncryptionProperties, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 34 "pyarrow/_parquet.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 34 "pyarrow/_parquet.pxd"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 34, __pyx_L1_error)

#line 34 "pyarrow/_parquet.pxd"
    __Pyx_GOTREF((PyObject *)__pyx_t_1);

#line 34 "pyarrow/_parquet.pxd"
  }

#line 34 "pyarrow/_parquet.pxd"
  __pyx_v_result = ((struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *)__pyx_t_1);

#line 34 "pyarrow/_parquet.pxd"
  __pyx_t_1 = 0;

  /* "pyarrow/_parquet.pxd":35
 * 
 *         result = FileEncryptionProperties()
 *         result.properties = properties             # <<<<<<<<<<<<<<
 *         return result
 * 
*/

#line 35 "pyarrow/_parquet.pxd"
  __pyx_v_result->properties = __pyx_v_properties;

  /* "pyarrow/_parquet.pxd":36
 *         result = FileEncryptionProperties()
 *         result.properties = properties
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef inline shared_ptr[CFileEncryptionProperties] unwrap(self):
*/

#line 36 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_r);

#line 36 "pyarrow/_parquet.pxd"
  __Pyx_INCREF((PyObject *)__pyx_v_result);

#line 36 "pyarrow/_parquet.pxd"
  __pyx_r = __pyx_v_result;

#line 36 "pyarrow/_parquet.pxd"
  goto __pyx_L0;

  /* "pyarrow/_parquet.pxd":30
 *         shared_ptr[CFileEncryptionProperties] properties
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline FileEncryptionProperties wrap(
 *             shared_ptr[CFileEncryptionProperties] properties):
*/

#line 30 "pyarrow/_parquet.pxd"


#line 30 "pyarrow/_parquet.pxd"
  /* function exit code */

#line 30 "pyarrow/_parquet.pxd"
  __pyx_L1_error:;

#line 30 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 30 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF(__pyx_t_2);

#line 30 "pyarrow/_parquet.pxd"
  __Pyx_AddTraceback("pyarrow._parquet.FileEncryptionProperties.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 30 "pyarrow/_parquet.pxd"
  __pyx_r = 0;

#line 30 "pyarrow/_parquet.pxd"
  __pyx_L0:;

#line 30 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 30 "pyarrow/_parquet.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 30 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyFinishContext();

#line 30 "pyarrow/_parquet.pxd"
  return __pyx_r;

#line 30 "pyarrow/_parquet.pxd"
}

/* "pyarrow/_parquet.pxd":38
 *         return result
 * 
 *     cdef inline shared_ptr[CFileEncryptionProperties] unwrap(self):             # <<<<<<<<<<<<<<
 *         return self.properties
 * 
*/

#line 38 "pyarrow/_parquet.pxd"


#line 38 "pyarrow/_parquet.pxd"
static CYTHON_INLINE std::shared_ptr< parquet::FileEncryptionProperties>  __pyx_f_7pyarrow_8_parquet_24FileEncryptionProperties_unwrap(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties *__pyx_v_self) {

#line 38 "pyarrow/_parquet.pxd"
  std::shared_ptr< parquet::FileEncryptionProperties>  __pyx_r;

  /* "pyarrow/_parquet.pxd":39
 * 
 *     cdef inline shared_ptr[CFileEncryptionProperties] unwrap(self):
 *         return self.properties             # <<<<<<<<<<<<<<
 * 
 * cdef shared_ptr[WriterProperties] _create_writer_properties(
*/

#line 39 "pyarrow/_parquet.pxd"
  __pyx_r = __pyx_v_self->properties;

#line 39 "pyarrow/_parquet.pxd"
  goto __pyx_L0;

  /* "pyarrow/_parquet.pxd":38
 *         return result
 * 
 *     cdef inline shared_ptr[CFileEncryptionProperties] unwrap(self):             # <<<<<<<<<<<<<<
 *         return self.properties
 * 
*/

#line 38 "pyarrow/_parquet.pxd"


#line 38 "pyarrow/_parquet.pxd"
  /* function exit code */

#line 38 "pyarrow/_parquet.pxd"
  __pyx_L0:;

#line 38 "pyarrow/_parquet.pxd"
  return __pyx_r;

#line 38 "pyarrow/_parquet.pxd"
}

/* "pyarrow/_parquet.pxd":89
 *         ParquetSchema _schema
 * 
 *     cdef inline init(self, const shared_ptr[CFileMetaData]& metadata):             # <<<<<<<<<<<<<<
 *         self.sp_metadata = metadata
 *         self._metadata = metadata.get()
*/

#line 89 "pyarrow/_parquet.pxd"


#line 89 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_12FileMetaData_init(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_self, std::shared_ptr< parquet::FileMetaData>  const &__pyx_v_metadata) {

#line 89 "pyarrow/_parquet.pxd"
  PyObject *__pyx_r = NULL;

#line 89 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyDeclarations

#line 89 "pyarrow/_parquet.pxd"
  __Pyx_RefNannySetupContext("init", 0);

  /* "pyarrow/_parquet.pxd":90
 * 
 *     cdef inline init(self, const shared_ptr[CFileMetaData]& metadata):
 *         self.sp_metadata = metadata             # <<<<<<<<<<<<<<
 *         self._metadata = metadata.get()
 * 
*/

#line 90 "pyarrow/_parquet.pxd"
  __pyx_v_self->sp_metadata = __pyx_v_metadata;

  /* "pyarrow/_parquet.pxd":91
 *     cdef inline init(self, const shared_ptr[CFileMetaData]& metadata):
 *         self.sp_metadata = metadata
 *         self._metadata = metadata.get()             # <<<<<<<<<<<<<<
 * 
 * cdef class RowGroupMetaData(_Weakrefable):
*/

#line 91 "pyarrow/_parquet.pxd"
  __pyx_v_self->_metadata = __pyx_v_metadata.get();

  /* "pyarrow/_parquet.pxd":89
 *         ParquetSchema _schema
 * 
 *     cdef inline init(self, const shared_ptr[CFileMetaData]& metadata):             # <<<<<<<<<<<<<<
 *         self.sp_metadata = metadata
 *         self._metadata = metadata.get()
*/

#line 89 "pyarrow/_parquet.pxd"


#line 89 "pyarrow/_parquet.pxd"
  /* function exit code */

#line 89 "pyarrow/_parquet.pxd"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 89 "pyarrow/_parquet.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 89 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyFinishContext();

#line 89 "pyarrow/_parquet.pxd"
  return __pyx_r;

#line 89 "pyarrow/_parquet.pxd"
}

/* "pyarrow/_parquet.pxd":100
 *         FileMetaData parent
 * 
 *     cdef inline init(self, FileMetaData parent, int index):             # <<<<<<<<<<<<<<
 *         if index < 0 or index >= parent.num_row_groups:
 *             raise IndexError('{0} out of bounds'.format(index))
*/

#line 100 "pyarrow/_parquet.pxd"


#line 100 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_16RowGroupMetaData_init(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData *__pyx_v_self, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_parent, int __pyx_v_index) {

#line 100 "pyarrow/_parquet.pxd"
  PyObject *__pyx_r = NULL;

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyDeclarations

#line 100 "pyarrow/_parquet.pxd"
  int __pyx_t_1;

#line 100 "pyarrow/_parquet.pxd"
  int __pyx_t_2;

#line 100 "pyarrow/_parquet.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 100 "pyarrow/_parquet.pxd"
  PyObject *__pyx_t_4 = NULL;

#line 100 "pyarrow/_parquet.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 100 "pyarrow/_parquet.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 100 "pyarrow/_parquet.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 100 "pyarrow/_parquet.pxd"
  size_t __pyx_t_8;

#line 100 "pyarrow/_parquet.pxd"
  int __pyx_lineno = 0;

#line 100 "pyarrow/_parquet.pxd"
  const char *__pyx_filename = NULL;

#line 100 "pyarrow/_parquet.pxd"
  int __pyx_clineno = 0;

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_RefNannySetupContext("init", 0);

  /* "pyarrow/_parquet.pxd":101
 * 
 *     cdef inline init(self, FileMetaData parent, int index):
 *         if index < 0 or index >= parent.num_row_groups:             # <<<<<<<<<<<<<<
 *             raise IndexError('{0} out of bounds'.format(index))
 *         self.up_metadata = parent._metadata.RowGroup(index)
*/

#line 101 "pyarrow/_parquet.pxd"
  __pyx_t_2 = (__pyx_v_index < 0);

#line 101 "pyarrow/_parquet.pxd"
  if (!__pyx_t_2) {

#line 101 "pyarrow/_parquet.pxd"
  } else {

#line 101 "pyarrow/_parquet.pxd"
    __pyx_t_1 = __pyx_t_2;

#line 101 "pyarrow/_parquet.pxd"
    goto __pyx_L4_bool_binop_done;

#line 101 "pyarrow/_parquet.pxd"
  }

#line 101 "pyarrow/_parquet.pxd"
  __pyx_t_3 = __Pyx_PyLong_From_int(__pyx_v_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 101, __pyx_L1_error)

#line 101 "pyarrow/_parquet.pxd"
  __Pyx_GOTREF(__pyx_t_3);

#line 101 "pyarrow/_parquet.pxd"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_parent), __pyx_mstate_global->__pyx_n_u_num_row_groups); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 101, __pyx_L1_error)

#line 101 "pyarrow/_parquet.pxd"
  __Pyx_GOTREF(__pyx_t_4);

#line 101 "pyarrow/_parquet.pxd"
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_GE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 101, __pyx_L1_error)

#line 101 "pyarrow/_parquet.pxd"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 101 "pyarrow/_parquet.pxd"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 101 "pyarrow/_parquet.pxd"
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(5, 101, __pyx_L1_error)

#line 101 "pyarrow/_parquet.pxd"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 101 "pyarrow/_parquet.pxd"
  __pyx_t_1 = __pyx_t_2;

#line 101 "pyarrow/_parquet.pxd"
  __pyx_L4_bool_binop_done:;

#line 101 "pyarrow/_parquet.pxd"
  if (unlikely(__pyx_t_1)) {

    /* "pyarrow/_parquet.pxd":102
 *     cdef inline init(self, FileMetaData parent, int index):
 *         if index < 0 or index >= parent.num_row_groups:
 *             raise IndexError('{0} out of bounds'.format(index))             # <<<<<<<<<<<<<<
 *         self.up_metadata = parent._metadata.RowGroup(index)
 *         self.metadata = self.up_metadata.get()
*/

#line 102 "pyarrow/_parquet.pxd"
    __pyx_t_4 = NULL;

#line 102 "pyarrow/_parquet.pxd"
    __pyx_t_6 = __pyx_mstate_global->__pyx_kp_u_0_out_of_bounds;

#line 102 "pyarrow/_parquet.pxd"
    __Pyx_INCREF(__pyx_t_6);

#line 102 "pyarrow/_parquet.pxd"
    __pyx_t_7 = __Pyx_PyLong_From_int(__pyx_v_index); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 102, __pyx_L1_error)

#line 102 "pyarrow/_parquet.pxd"
    __Pyx_GOTREF(__pyx_t_7);

#line 102 "pyarrow/_parquet.pxd"
    __pyx_t_8 = 0;

#line 102 "pyarrow/_parquet.pxd"
    {

#line 102 "pyarrow/_parquet.pxd"
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_7};

#line 102 "pyarrow/_parquet.pxd"
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 102 "pyarrow/_parquet.pxd"
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 102 "pyarrow/_parquet.pxd"
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 102 "pyarrow/_parquet.pxd"
      if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 102, __pyx_L1_error)

#line 102 "pyarrow/_parquet.pxd"
      __Pyx_GOTREF(__pyx_t_3);

#line 102 "pyarrow/_parquet.pxd"
    }

#line 102 "pyarrow/_parquet.pxd"
    __pyx_t_8 = 1;

#line 102 "pyarrow/_parquet.pxd"
    {

#line 102 "pyarrow/_parquet.pxd"
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};

#line 102 "pyarrow/_parquet.pxd"
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_IndexError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 102 "pyarrow/_parquet.pxd"
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 102 "pyarrow/_parquet.pxd"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 102 "pyarrow/_parquet.pxd"
      if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 102, __pyx_L1_error)

#line 102 "pyarrow/_parquet.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 102 "pyarrow/_parquet.pxd"
    }

#line 102 "pyarrow/_parquet.pxd"
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);

#line 102 "pyarrow/_parquet.pxd"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 102 "pyarrow/_parquet.pxd"
    __PYX_ERR(5, 102, __pyx_L1_error)

    /* "pyarrow/_parquet.pxd":101
 * 
 *     cdef inline init(self, FileMetaData parent, int index):
 *         if index < 0 or index >= parent.num_row_groups:             # <<<<<<<<<<<<<<
 *             raise IndexError('{0} out of bounds'.format(index))
 *         self.up_metadata = parent._metadata.RowGroup(index)
*/

#line 101 "pyarrow/_parquet.pxd"
  }

  /* "pyarrow/_parquet.pxd":103
 *         if index < 0 or index >= parent.num_row_groups:
 *             raise IndexError('{0} out of bounds'.format(index))
 *         self.up_metadata = parent._metadata.RowGroup(index)             # <<<<<<<<<<<<<<
 *         self.metadata = self.up_metadata.get()
 *         self.parent = parent
*/

#line 103 "pyarrow/_parquet.pxd"
  __pyx_v_self->up_metadata = __pyx_v_parent->_metadata->RowGroup(__pyx_v_index);

  /* "pyarrow/_parquet.pxd":104
 *             raise IndexError('{0} out of bounds'.format(index))
 *         self.up_metadata = parent._metadata.RowGroup(index)
 *         self.metadata = self.up_metadata.get()             # <<<<<<<<<<<<<<
 *         self.parent = parent
 *         self.index = index
*/

#line 104 "pyarrow/_parquet.pxd"
  __pyx_v_self->metadata = __pyx_v_self->up_metadata.get();

  /* "pyarrow/_parquet.pxd":105
 *         self.up_metadata = parent._metadata.RowGroup(index)
 *         self.metadata = self.up_metadata.get()
 *         self.parent = parent             # <<<<<<<<<<<<<<
 *         self.index = index
 * 
*/

#line 105 "pyarrow/_parquet.pxd"
  __Pyx_INCREF((PyObject *)__pyx_v_parent);

#line 105 "pyarrow/_parquet.pxd"
  __Pyx_GIVEREF((PyObject *)__pyx_v_parent);

#line 105 "pyarrow/_parquet.pxd"
  __Pyx_GOTREF((PyObject *)__pyx_v_self->parent);

#line 105 "pyarrow/_parquet.pxd"
  __Pyx_DECREF((PyObject *)__pyx_v_self->parent);

#line 105 "pyarrow/_parquet.pxd"
  __pyx_v_self->parent = __pyx_v_parent;

  /* "pyarrow/_parquet.pxd":106
 *         self.metadata = self.up_metadata.get()
 *         self.parent = parent
 *         self.index = index             # <<<<<<<<<<<<<<
 * 
 * 
*/

#line 106 "pyarrow/_parquet.pxd"
  __pyx_v_self->index = __pyx_v_index;

  /* "pyarrow/_parquet.pxd":100
 *         FileMetaData parent
 * 
 *     cdef inline init(self, FileMetaData parent, int index):             # <<<<<<<<<<<<<<
 *         if index < 0 or index >= parent.num_row_groups:
 *             raise IndexError('{0} out of bounds'.format(index))
*/

#line 100 "pyarrow/_parquet.pxd"


#line 100 "pyarrow/_parquet.pxd"
  /* function exit code */

#line 100 "pyarrow/_parquet.pxd"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 100 "pyarrow/_parquet.pxd"
  goto __pyx_L0;

#line 100 "pyarrow/_parquet.pxd"
  __pyx_L1_error:;

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF(__pyx_t_3);

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF(__pyx_t_4);

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_AddTraceback("pyarrow._parquet.RowGroupMetaData.init", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 100 "pyarrow/_parquet.pxd"
  __pyx_r = 0;

#line 100 "pyarrow/_parquet.pxd"
  __pyx_L0:;

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 100 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyFinishContext();

#line 100 "pyarrow/_parquet.pxd"
  return __pyx_r;

#line 100 "pyarrow/_parquet.pxd"
}

/* "pyarrow/_parquet.pxd":115
 *         RowGroupMetaData parent
 * 
 *     cdef inline init(self, RowGroupMetaData parent, int i):             # <<<<<<<<<<<<<<
 *         self.up_metadata = parent.metadata.ColumnChunk(i)
 *         self.metadata = self.up_metadata.get()
*/

#line 115 "pyarrow/_parquet.pxd"


#line 115 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_19ColumnChunkMetaData_init(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *__pyx_v_self, struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData *__pyx_v_parent, int __pyx_v_i) {

#line 115 "pyarrow/_parquet.pxd"
  PyObject *__pyx_r = NULL;

#line 115 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyDeclarations

#line 115 "pyarrow/_parquet.pxd"
  __Pyx_RefNannySetupContext("init", 0);

  /* "pyarrow/_parquet.pxd":116
 * 
 *     cdef inline init(self, RowGroupMetaData parent, int i):
 *         self.up_metadata = parent.metadata.ColumnChunk(i)             # <<<<<<<<<<<<<<
 *         self.metadata = self.up_metadata.get()
 *         self.parent = parent
*/

#line 116 "pyarrow/_parquet.pxd"
  __pyx_v_self->up_metadata = __pyx_v_parent->metadata->ColumnChunk(__pyx_v_i);

  /* "pyarrow/_parquet.pxd":117
 *     cdef inline init(self, RowGroupMetaData parent, int i):
 *         self.up_metadata = parent.metadata.ColumnChunk(i)
 *         self.metadata = self.up_metadata.get()             # <<<<<<<<<<<<<<
 *         self.parent = parent
 * 
*/

#line 117 "pyarrow/_parquet.pxd"
  __pyx_v_self->metadata = __pyx_v_self->up_metadata.get();

  /* "pyarrow/_parquet.pxd":118
 *         self.up_metadata = parent.metadata.ColumnChunk(i)
 *         self.metadata = self.up_metadata.get()
 *         self.parent = parent             # <<<<<<<<<<<<<<
 * 
 * cdef class Statistics(_Weakrefable):
*/

#line 118 "pyarrow/_parquet.pxd"
  __Pyx_INCREF((PyObject *)__pyx_v_parent);

#line 118 "pyarrow/_parquet.pxd"
  __Pyx_GIVEREF((PyObject *)__pyx_v_parent);

#line 118 "pyarrow/_parquet.pxd"
  __Pyx_GOTREF((PyObject *)__pyx_v_self->parent);

#line 118 "pyarrow/_parquet.pxd"
  __Pyx_DECREF((PyObject *)__pyx_v_self->parent);

#line 118 "pyarrow/_parquet.pxd"
  __pyx_v_self->parent = __pyx_v_parent;

  /* "pyarrow/_parquet.pxd":115
 *         RowGroupMetaData parent
 * 
 *     cdef inline init(self, RowGroupMetaData parent, int i):             # <<<<<<<<<<<<<<
 *         self.up_metadata = parent.metadata.ColumnChunk(i)
 *         self.metadata = self.up_metadata.get()
*/

#line 115 "pyarrow/_parquet.pxd"


#line 115 "pyarrow/_parquet.pxd"
  /* function exit code */

#line 115 "pyarrow/_parquet.pxd"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 115 "pyarrow/_parquet.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 115 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyFinishContext();

#line 115 "pyarrow/_parquet.pxd"
  return __pyx_r;

#line 115 "pyarrow/_parquet.pxd"
}

/* "pyarrow/_parquet.pxd":125
 *         ColumnChunkMetaData parent
 * 
 *     cdef inline init(self, const shared_ptr[CStatistics]& statistics,             # <<<<<<<<<<<<<<
 *                      ColumnChunkMetaData parent):
 *         self.statistics = statistics
*/

#line 125 "pyarrow/_parquet.pxd"


#line 125 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_10Statistics_init(struct __pyx_obj_7pyarrow_8_parquet_Statistics *__pyx_v_self, std::shared_ptr< parquet::Statistics>  const &__pyx_v_statistics, struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *__pyx_v_parent) {

#line 125 "pyarrow/_parquet.pxd"
  PyObject *__pyx_r = NULL;

#line 125 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyDeclarations

#line 125 "pyarrow/_parquet.pxd"
  __Pyx_RefNannySetupContext("init", 0);

  /* "pyarrow/_parquet.pxd":127
 *     cdef inline init(self, const shared_ptr[CStatistics]& statistics,
 *                      ColumnChunkMetaData parent):
 *         self.statistics = statistics             # <<<<<<<<<<<<<<
 *         self.parent = parent
 * 
*/

#line 127 "pyarrow/_parquet.pxd"
  __pyx_v_self->statistics = __pyx_v_statistics;

  /* "pyarrow/_parquet.pxd":128
 *                      ColumnChunkMetaData parent):
 *         self.statistics = statistics
 *         self.parent = parent             # <<<<<<<<<<<<<<
 * 
 * cdef class GeoStatistics(_Weakrefable):
*/

#line 128 "pyarrow/_parquet.pxd"
  __Pyx_INCREF((PyObject *)__pyx_v_parent);

#line 128 "pyarrow/_parquet.pxd"
  __Pyx_GIVEREF((PyObject *)__pyx_v_parent);

#line 128 "pyarrow/_parquet.pxd"
  __Pyx_GOTREF((PyObject *)__pyx_v_self->parent);

#line 128 "pyarrow/_parquet.pxd"
  __Pyx_DECREF((PyObject *)__pyx_v_self->parent);

#line 128 "pyarrow/_parquet.pxd"
  __pyx_v_self->parent = __pyx_v_parent;

  /* "pyarrow/_parquet.pxd":125
 *         ColumnChunkMetaData parent
 * 
 *     cdef inline init(self, const shared_ptr[CStatistics]& statistics,             # <<<<<<<<<<<<<<
 *                      ColumnChunkMetaData parent):
 *         self.statistics = statistics
*/

#line 125 "pyarrow/_parquet.pxd"


#line 125 "pyarrow/_parquet.pxd"
  /* function exit code */

#line 125 "pyarrow/_parquet.pxd"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 125 "pyarrow/_parquet.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 125 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyFinishContext();

#line 125 "pyarrow/_parquet.pxd"
  return __pyx_r;

#line 125 "pyarrow/_parquet.pxd"
}

/* "pyarrow/_parquet.pxd":135
 *         ColumnChunkMetaData parent
 * 
 *     cdef inline init(self, const shared_ptr[CParquetGeoStatistics]& statistics,             # <<<<<<<<<<<<<<
 *                      ColumnChunkMetaData parent):
 *         self.statistics = statistics
*/

#line 135 "pyarrow/_parquet.pxd"


#line 135 "pyarrow/_parquet.pxd"
static CYTHON_INLINE PyObject *__pyx_f_7pyarrow_8_parquet_13GeoStatistics_init(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics *__pyx_v_self, std::shared_ptr< parquet::geospatial::GeoStatistics>  const &__pyx_v_statistics, struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData *__pyx_v_parent) {

#line 135 "pyarrow/_parquet.pxd"
  PyObject *__pyx_r = NULL;

#line 135 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyDeclarations

#line 135 "pyarrow/_parquet.pxd"
  __Pyx_RefNannySetupContext("init", 0);

  /* "pyarrow/_parquet.pxd":137
 *     cdef inline init(self, const shared_ptr[CParquetGeoStatistics]& statistics,
 *                      ColumnChunkMetaData parent):
 *         self.statistics = statistics             # <<<<<<<<<<<<<<
 *         self.parent = parent
 * 
*/

#line 137 "pyarrow/_parquet.pxd"
  __pyx_v_self->statistics = __pyx_v_statistics;

  /* "pyarrow/_parquet.pxd":138
 *                      ColumnChunkMetaData parent):
 *         self.statistics = statistics
 *         self.parent = parent             # <<<<<<<<<<<<<<
 * 
 * cdef class FileDecryptionProperties:
*/

#line 138 "pyarrow/_parquet.pxd"
  __Pyx_INCREF((PyObject *)__pyx_v_parent);

#line 138 "pyarrow/_parquet.pxd"
  __Pyx_GIVEREF((PyObject *)__pyx_v_parent);

#line 138 "pyarrow/_parquet.pxd"
  __Pyx_GOTREF((PyObject *)__pyx_v_self->parent);

#line 138 "pyarrow/_parquet.pxd"
  __Pyx_DECREF((PyObject *)__pyx_v_self->parent);

#line 138 "pyarrow/_parquet.pxd"
  __pyx_v_self->parent = __pyx_v_parent;

  /* "pyarrow/_parquet.pxd":135
 *         ColumnChunkMetaData parent
 * 
 *     cdef inline init(self, const shared_ptr[CParquetGeoStatistics]& statistics,             # <<<<<<<<<<<<<<
 *                      ColumnChunkMetaData parent):
 *         self.statistics = statistics
*/

#line 135 "pyarrow/_parquet.pxd"


#line 135 "pyarrow/_parquet.pxd"
  /* function exit code */

#line 135 "pyarrow/_parquet.pxd"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 135 "pyarrow/_parquet.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 135 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyFinishContext();

#line 135 "pyarrow/_parquet.pxd"
  return __pyx_r;

#line 135 "pyarrow/_parquet.pxd"
}

/* "pyarrow/_parquet.pxd":145
 *         shared_ptr[CFileDecryptionProperties] properties
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline FileDecryptionProperties wrap(
 *             shared_ptr[CFileDecryptionProperties] properties):
*/

#line 145 "pyarrow/_parquet.pxd"


#line 145 "pyarrow/_parquet.pxd"
static CYTHON_INLINE struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *__pyx_f_7pyarrow_8_parquet_24FileDecryptionProperties_wrap(std::shared_ptr< parquet::FileDecryptionProperties>  __pyx_v_properties) {
  struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *__pyx_v_result = NULL
#line 145 "pyarrow/_parquet.pxd"
;

#line 145 "pyarrow/_parquet.pxd"
  struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *__pyx_r = NULL;

#line 145 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyDeclarations

#line 145 "pyarrow/_parquet.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 145 "pyarrow/_parquet.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 145 "pyarrow/_parquet.pxd"
  size_t __pyx_t_3;

#line 145 "pyarrow/_parquet.pxd"
  int __pyx_lineno = 0;

#line 145 "pyarrow/_parquet.pxd"
  const char *__pyx_filename = NULL;

#line 145 "pyarrow/_parquet.pxd"
  int __pyx_clineno = 0;

#line 145 "pyarrow/_parquet.pxd"
  __Pyx_RefNannySetupContext("wrap", 0);

  /* "pyarrow/_parquet.pxd":149
 *             shared_ptr[CFileDecryptionProperties] properties):
 * 
 *         result = FileDecryptionProperties()             # <<<<<<<<<<<<<<
 *         result.properties = properties
 *         return result
*/

#line 149 "pyarrow/_parquet.pxd"
  __pyx_t_2 = NULL;

#line 149 "pyarrow/_parquet.pxd"
  __pyx_t_3 = 1;

#line 149 "pyarrow/_parquet.pxd"
  {

#line 149 "pyarrow/_parquet.pxd"
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};

#line 149 "pyarrow/_parquet.pxd"
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyarrow_8_parquet_FileDecryptionProperties, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 149 "pyarrow/_parquet.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 149 "pyarrow/_parquet.pxd"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 149, __pyx_L1_error)

#line 149 "pyarrow/_parquet.pxd"
    __Pyx_GOTREF((PyObject *)__pyx_t_1);

#line 149 "pyarrow/_parquet.pxd"
  }

#line 149 "pyarrow/_parquet.pxd"
  __pyx_v_result = ((struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *)__pyx_t_1);

#line 149 "pyarrow/_parquet.pxd"
  __pyx_t_1 = 0;

  /* "pyarrow/_parquet.pxd":150
 * 
 *         result = FileDecryptionProperties()
 *         result.properties = properties             # <<<<<<<<<<<<<<
 *         return result
 * 
*/

#line 150 "pyarrow/_parquet.pxd"
  __pyx_v_result->properties = __pyx_v_properties;

  /* "pyarrow/_parquet.pxd":151
 *         result = FileDecryptionProperties()
 *         result.properties = properties
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef inline shared_ptr[CFileDecryptionProperties] unwrap(self):
*/

#line 151 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_r);

#line 151 "pyarrow/_parquet.pxd"
  __Pyx_INCREF((PyObject *)__pyx_v_result);

#line 151 "pyarrow/_parquet.pxd"
  __pyx_r = __pyx_v_result;

#line 151 "pyarrow/_parquet.pxd"
  goto __pyx_L0;

  /* "pyarrow/_parquet.pxd":145
 *         shared_ptr[CFileDecryptionProperties] properties
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     cdef inline FileDecryptionProperties wrap(
 *             shared_ptr[CFileDecryptionProperties] properties):
*/

#line 145 "pyarrow/_parquet.pxd"


#line 145 "pyarrow/_parquet.pxd"
  /* function exit code */

#line 145 "pyarrow/_parquet.pxd"
  __pyx_L1_error:;

#line 145 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 145 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF(__pyx_t_2);

#line 145 "pyarrow/_parquet.pxd"
  __Pyx_AddTraceback("pyarrow._parquet.FileDecryptionProperties.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 145 "pyarrow/_parquet.pxd"
  __pyx_r = 0;

#line 145 "pyarrow/_parquet.pxd"
  __pyx_L0:;

#line 145 "pyarrow/_parquet.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 145 "pyarrow/_parquet.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 145 "pyarrow/_parquet.pxd"
  __Pyx_RefNannyFinishContext();

#line 145 "pyarrow/_parquet.pxd"
  return __pyx_r;

#line 145 "pyarrow/_parquet.pxd"
}

/* "pyarrow/_parquet.pxd":153
 *         return result
 * 
 *     cdef inline shared_ptr[CFileDecryptionProperties] unwrap(self):             # <<<<<<<<<<<<<<
 *         return self.properties
*/

#line 153 "pyarrow/_parquet.pxd"


#line 153 "pyarrow/_parquet.pxd"
static CYTHON_INLINE std::shared_ptr< parquet::FileDecryptionProperties>  __pyx_f_7pyarrow_8_parquet_24FileDecryptionProperties_unwrap(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties *__pyx_v_self) {

#line 153 "pyarrow/_parquet.pxd"
  std::shared_ptr< parquet::FileDecryptionProperties>  __pyx_r;

  /* "pyarrow/_parquet.pxd":154
 * 
 *     cdef inline shared_ptr[CFileDecryptionProperties] unwrap(self):
 *         return self.properties             # <<<<<<<<<<<<<<
*/

#line 154 "pyarrow/_parquet.pxd"
  __pyx_r = __pyx_v_self->properties;

#line 154 "pyarrow/_parquet.pxd"
  goto __pyx_L0;

  /* "pyarrow/_parquet.pxd":153
 *         return result
 * 
 *     cdef inline shared_ptr[CFileDecryptionProperties] unwrap(self):             # <<<<<<<<<<<<<<
 *         return self.properties
*/

#line 153 "pyarrow/_parquet.pxd"


#line 153 "pyarrow/_parquet.pxd"
  /* function exit code */

#line 153 "pyarrow/_parquet.pxd"
  __pyx_L0:;

#line 153 "pyarrow/_parquet.pxd"
  return __pyx_r;

#line 153 "pyarrow/_parquet.pxd"
}

/* "jollyjack/jollyjack_cython.pyx":19
 * from collections.abc import Iterable
 * 
 * def is_iterable_of_iterables(obj):             # <<<<<<<<<<<<<<
 *     return isinstance(obj, Iterable) and isinstance(obj[0], Iterable) and not isinstance(obj[0], str)
 * 
*/

#line 19 "jollyjack/jollyjack_cython.pyx"


#line 19 "jollyjack/jollyjack_cython.pyx"
/* Python wrapper */

#line 19 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_1is_iterable_of_iterables(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/

#line 19 "jollyjack/jollyjack_cython.pyx"
PyDoc_STRVAR(__pyx_doc_9jollyjack_16jollyjack_cython_is_iterable_of_iterables, "is_iterable_of_iterables(obj)");
static PyMethodDef __pyx_mdef_9jollyjack_16jollyjack_cython_1is_iterable_of_iterables = 
#line 19 "jollyjack/jollyjack_cython.pyx"
{"is_iterable_of_iterables", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9jollyjack_16jollyjack_cython_1is_iterable_of_iterables, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9jollyjack_16jollyjack_cython_is_iterable_of_iterables};

#line 19 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_1is_iterable_of_iterables(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_obj = 0
#line 19 "jollyjack/jollyjack_cython.pyx"
;

#line 19 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 19 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;

#line 19 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 19 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;

#line 19 "jollyjack/jollyjack_cython.pyx"
  PyObject* values[1] = {0};

#line 19 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 19 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 19 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 19 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = 0;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("is_iterable_of_iterables (wrapper)", 0);

#line 19 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 19 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_ASSUME_SAFE_SIZE

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);

#line 19 "jollyjack/jollyjack_cython.pyx"
  #else

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;

#line 19 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 19 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);

#line 19 "jollyjack/jollyjack_cython.pyx"
  {

#line 19 "jollyjack/jollyjack_cython.pyx"
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_obj,0};

#line 19 "jollyjack/jollyjack_cython.pyx"
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;

#line 19 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 19, __pyx_L3_error)

#line 19 "jollyjack/jollyjack_cython.pyx"
    if (__pyx_kwds_len > 0) {

#line 19 "jollyjack/jollyjack_cython.pyx"
      switch (__pyx_nargs) {

#line 19 "jollyjack/jollyjack_cython.pyx"
        case  1:

#line 19 "jollyjack/jollyjack_cython.pyx"
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 19 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 19, __pyx_L3_error)

#line 19 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 19 "jollyjack/jollyjack_cython.pyx"
break;
        default: 
#line 19 "jollyjack/jollyjack_cython.pyx"
goto __pyx_L5_argtuple_error;

#line 19 "jollyjack/jollyjack_cython.pyx"
      }

#line 19 "jollyjack/jollyjack_cython.pyx"
      const Py_ssize_t kwd_pos_args = __pyx_nargs;

#line 19 "jollyjack/jollyjack_cython.pyx"
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "is_iterable_of_iterables", 0) < (0)) __PYX_ERR(0, 19, __pyx_L3_error)

#line 19 "jollyjack/jollyjack_cython.pyx"
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {

#line 19 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("is_iterable_of_iterables", 1, 1, 1, i); __PYX_ERR(0, 19, __pyx_L3_error) }

#line 19 "jollyjack/jollyjack_cython.pyx"
      }

#line 19 "jollyjack/jollyjack_cython.pyx"
    } else if (unlikely(__pyx_nargs != 1)) {

#line 19 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L5_argtuple_error;

#line 19 "jollyjack/jollyjack_cython.pyx"
    } else {

#line 19 "jollyjack/jollyjack_cython.pyx"
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 19 "jollyjack/jollyjack_cython.pyx"
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 19, __pyx_L3_error)

#line 19 "jollyjack/jollyjack_cython.pyx"
    }

#line 19 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_obj = values[0];

#line 19 "jollyjack/jollyjack_cython.pyx"
  }

#line 19 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L6_skip;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_L5_argtuple_error:;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RaiseArgtupleInvalid("is_iterable_of_iterables", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 19, __pyx_L3_error)

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_L6_skip:;

#line 19 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_L3_error:;

#line 19 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 19 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 19 "jollyjack/jollyjack_cython.pyx"
  }

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.is_iterable_of_iterables", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 19 "jollyjack/jollyjack_cython.pyx"
  return NULL;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 19 "jollyjack/jollyjack_cython.pyx"
__pyx_pf_9jollyjack_16jollyjack_cython_is_iterable_of_iterables(__pyx_self, __pyx_v_obj);

#line 19 "jollyjack/jollyjack_cython.pyx"


#line 19 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 19 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 19 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 19 "jollyjack/jollyjack_cython.pyx"
  }

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 19 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 19 "jollyjack/jollyjack_cython.pyx"
}

#line 19 "jollyjack/jollyjack_cython.pyx"


#line 19 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_is_iterable_of_iterables(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {

#line 19 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = NULL;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 19 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 19 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 19 "jollyjack/jollyjack_cython.pyx"
  int __pyx_t_3;

#line 19 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 19 "jollyjack/jollyjack_cython.pyx"
  int __pyx_t_5;

#line 19 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 19 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 19 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("is_iterable_of_iterables", 0);

  /* "jollyjack/jollyjack_cython.pyx":20
 * 
 * def is_iterable_of_iterables(obj):
 *     return isinstance(obj, Iterable) and isinstance(obj[0], Iterable) and not isinstance(obj[0], str)             # <<<<<<<<<<<<<<
 * 
 * cpdef void read_into_torch (object source, FileMetaData metadata, tensor, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):
*/

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Iterable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 20 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

#line 20 "jollyjack/jollyjack_cython.pyx"
  } else {

#line 20 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 20 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __pyx_t_2;

#line 20 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_2 = 0;

#line 20 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L3_bool_binop_done;

#line 20 "jollyjack/jollyjack_cython.pyx"
  }

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_obj, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Iterable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 20 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

#line 20 "jollyjack/jollyjack_cython.pyx"
  } else {

#line 20 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 20 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __pyx_t_4;

#line 20 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_4 = 0;

#line 20 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L3_bool_binop_done;

#line 20 "jollyjack/jollyjack_cython.pyx"
  }

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_obj, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = PyUnicode_Check(__pyx_t_4); 

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = (!__pyx_t_3);

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = __pyx_t_4;

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_4 = 0;

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_L3_bool_binop_done:;

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = __pyx_t_1;

#line 20 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = 0;

#line 20 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

  /* "jollyjack/jollyjack_cython.pyx":19
 * from collections.abc import Iterable
 * 
 * def is_iterable_of_iterables(obj):             # <<<<<<<<<<<<<<
 *     return isinstance(obj, Iterable) and isinstance(obj[0], Iterable) and not isinstance(obj[0], str)
 * 
*/

#line 19 "jollyjack/jollyjack_cython.pyx"


#line 19 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.is_iterable_of_iterables", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = NULL;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 19 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 19 "jollyjack/jollyjack_cython.pyx"
}

/* "jollyjack/jollyjack_cython.pyx":22
 *     return isinstance(obj, Iterable) and isinstance(obj[0], Iterable) and not isinstance(obj[0], str)
 * 
 * cpdef void read_into_torch (object source, FileMetaData metadata, tensor, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):             # <<<<<<<<<<<<<<
 * 
 *     import torch
*/

#line 22 "jollyjack/jollyjack_cython.pyx"


#line 22 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_3read_into_torch(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/

#line 22 "jollyjack/jollyjack_cython.pyx"
static void __pyx_f_9jollyjack_16jollyjack_cython_read_into_torch(PyObject *__pyx_v_source, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_metadata, PyObject *__pyx_v_tensor, PyObject *__pyx_v_row_group_indices, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_torch *__pyx_optional_args) {

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_row_ranges = __pyx_mstate_global->__pyx_k_;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_column_indices = __pyx_mstate_global->__pyx_k__2;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_column_names = __pyx_mstate_global->__pyx_k__3;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_pre_buffer = ((PyObject *)Py_False);

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_use_threads = ((PyObject *)Py_True);

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_use_memory_map = ((PyObject *)Py_False);

#line 22 "jollyjack/jollyjack_cython.pyx"
  struct __pyx_obj_7pyarrow_3lib_CacheOptions *__pyx_v_cache_options = ((struct __pyx_obj_7pyarrow_3lib_CacheOptions *)Py_None);
  CYTHON_UNUSED PyObject *__pyx_v_torch = NULL
#line 22 "jollyjack/jollyjack_cython.pyx"
;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  size_t __pyx_t_4;

#line 22 "jollyjack/jollyjack_cython.pyx"
  struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_numpy __pyx_t_5;

#line 22 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("read_into_torch", 0);

#line 22 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_optional_args) {

#line 22 "jollyjack/jollyjack_cython.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 22 "jollyjack/jollyjack_cython.pyx"
      __pyx_v_row_ranges = __pyx_optional_args->row_ranges;

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 22 "jollyjack/jollyjack_cython.pyx"
        __pyx_v_column_indices = __pyx_optional_args->column_indices;

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 22 "jollyjack/jollyjack_cython.pyx"
          __pyx_v_column_names = __pyx_optional_args->column_names;

#line 22 "jollyjack/jollyjack_cython.pyx"
          if (__pyx_optional_args->__pyx_n > 3) {

#line 22 "jollyjack/jollyjack_cython.pyx"
            __pyx_v_pre_buffer = __pyx_optional_args->pre_buffer;

#line 22 "jollyjack/jollyjack_cython.pyx"
            if (__pyx_optional_args->__pyx_n > 4) {

#line 22 "jollyjack/jollyjack_cython.pyx"
              __pyx_v_use_threads = __pyx_optional_args->use_threads;

#line 22 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_optional_args->__pyx_n > 5) {

#line 22 "jollyjack/jollyjack_cython.pyx"
                __pyx_v_use_memory_map = __pyx_optional_args->use_memory_map;

#line 22 "jollyjack/jollyjack_cython.pyx"
                if (__pyx_optional_args->__pyx_n > 6) {

#line 22 "jollyjack/jollyjack_cython.pyx"
                  __pyx_v_cache_options = __pyx_optional_args->cache_options;

#line 22 "jollyjack/jollyjack_cython.pyx"
                }

#line 22 "jollyjack/jollyjack_cython.pyx"
              }

#line 22 "jollyjack/jollyjack_cython.pyx"
            }

#line 22 "jollyjack/jollyjack_cython.pyx"
          }

#line 22 "jollyjack/jollyjack_cython.pyx"
        }

#line 22 "jollyjack/jollyjack_cython.pyx"
      }

#line 22 "jollyjack/jollyjack_cython.pyx"
    }

#line 22 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":24
 * cpdef void read_into_torch (object source, FileMetaData metadata, tensor, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):
 * 
 *     import torch             # <<<<<<<<<<<<<<
 * 
 *     read_into_numpy (source = source
*/

#line 24 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_torch, 0, 0, NULL, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)

#line 24 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = __pyx_t_2;

#line 24 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 24 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_torch = __pyx_t_1;

#line 24 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = 0;

  /* "jollyjack/jollyjack_cython.pyx":28
 *     read_into_numpy (source = source
 *         , metadata = metadata
 *         , np_array = tensor.numpy()             # <<<<<<<<<<<<<<
 *         , row_group_indices = row_group_indices
 *         , row_ranges = row_ranges
*/

#line 28 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __pyx_v_tensor;

#line 28 "jollyjack/jollyjack_cython.pyx"
  __Pyx_INCREF(__pyx_t_3);

#line 28 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_4 = 0;

#line 28 "jollyjack/jollyjack_cython.pyx"
  {

#line 28 "jollyjack/jollyjack_cython.pyx"
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};

#line 28 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_numpy, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 28 "jollyjack/jollyjack_cython.pyx"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 28 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)

#line 28 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 28 "jollyjack/jollyjack_cython.pyx"
  }

#line 28 "jollyjack/jollyjack_cython.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 28, __pyx_L1_error)

  /* "jollyjack/jollyjack_cython.pyx":26
 *     import torch
 * 
 *     read_into_numpy (source = source             # <<<<<<<<<<<<<<
 *         , metadata = metadata
 *         , np_array = tensor.numpy()
*/

#line 26 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5.__pyx_n = 7;

#line 26 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5.row_ranges = __pyx_v_row_ranges;

#line 26 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5.column_indices = __pyx_v_column_indices;

#line 26 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5.column_names = __pyx_v_column_names;

#line 26 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5.pre_buffer = __pyx_v_pre_buffer;

#line 26 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5.use_threads = __pyx_v_use_threads;

#line 26 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5.use_memory_map = __pyx_v_use_memory_map;

#line 26 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5.cache_options = __pyx_v_cache_options;

#line 26 "jollyjack/jollyjack_cython.pyx"
  __pyx_f_9jollyjack_16jollyjack_cython_read_into_numpy(__pyx_v_source, __pyx_v_metadata, ((PyArrayObject *)__pyx_t_1), __pyx_v_row_group_indices, 0, &__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 26, __pyx_L1_error)

#line 26 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "jollyjack/jollyjack_cython.pyx":39
 *     )
 * 
 *     return             # <<<<<<<<<<<<<<
 * 
 * cpdef void read_into_numpy (object source, FileMetaData metadata, cnp.ndarray np_array, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):
*/

#line 39 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

  /* "jollyjack/jollyjack_cython.pyx":22
 *     return isinstance(obj, Iterable) and isinstance(obj[0], Iterable) and not isinstance(obj[0], str)
 * 
 * cpdef void read_into_torch (object source, FileMetaData metadata, tensor, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):             # <<<<<<<<<<<<<<
 * 
 *     import torch
*/

#line 22 "jollyjack/jollyjack_cython.pyx"


#line 22 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.read_into_torch", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_v_torch);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 22 "jollyjack/jollyjack_cython.pyx"
}

#line 22 "jollyjack/jollyjack_cython.pyx"


#line 22 "jollyjack/jollyjack_cython.pyx"
/* Python wrapper */

#line 22 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_3read_into_torch(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/

#line 22 "jollyjack/jollyjack_cython.pyx"
PyDoc_STRVAR(__pyx_doc_9jollyjack_16jollyjack_cython_2read_into_torch, "read_into_torch(source, FileMetaData metadata, tensor, row_group_indices, row_ranges=[], column_indices=[], column_names=[], pre_buffer=False, use_threads=True, use_memory_map=False, CacheOptions cache_options=None) -> void");
static PyMethodDef __pyx_mdef_9jollyjack_16jollyjack_cython_3read_into_torch = 
#line 22 "jollyjack/jollyjack_cython.pyx"
{"read_into_torch", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9jollyjack_16jollyjack_cython_3read_into_torch, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9jollyjack_16jollyjack_cython_2read_into_torch};

#line 22 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_3read_into_torch(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_source = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_metadata = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_tensor = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_row_group_indices = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_row_ranges = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_column_indices = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_column_names = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_pre_buffer = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_use_threads = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_use_memory_map = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;
  struct __pyx_obj_7pyarrow_3lib_CacheOptions *__pyx_v_cache_options = 0
#line 22 "jollyjack/jollyjack_cython.pyx"
;

#line 22 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 22 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;

#line 22 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 22 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};

#line 22 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("read_into_torch (wrapper)", 0);

#line 22 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 22 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_ASSUME_SAFE_SIZE

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);

#line 22 "jollyjack/jollyjack_cython.pyx"
  #else

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 22 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);

#line 22 "jollyjack/jollyjack_cython.pyx"
  {

#line 22 "jollyjack/jollyjack_cython.pyx"
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_source,&__pyx_mstate_global->__pyx_n_u_metadata,&__pyx_mstate_global->__pyx_n_u_tensor,&__pyx_mstate_global->__pyx_n_u_row_group_indices,&__pyx_mstate_global->__pyx_n_u_row_ranges,&__pyx_mstate_global->__pyx_n_u_column_indices,&__pyx_mstate_global->__pyx_n_u_column_names,&__pyx_mstate_global->__pyx_n_u_pre_buffer,&__pyx_mstate_global->__pyx_n_u_use_threads,&__pyx_mstate_global->__pyx_n_u_use_memory_map,&__pyx_mstate_global->__pyx_n_u_cache_options,0};

#line 22 "jollyjack/jollyjack_cython.pyx"
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
    if (__pyx_kwds_len > 0) {

#line 22 "jollyjack/jollyjack_cython.pyx"
      switch (__pyx_nargs) {

#line 22 "jollyjack/jollyjack_cython.pyx"
        case 11:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case 10:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  9:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  8:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  7:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  6:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  5:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  4:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  3:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  2:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  1:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 22 "jollyjack/jollyjack_cython.pyx"
break;
        default: 
#line 22 "jollyjack/jollyjack_cython.pyx"
goto __pyx_L5_argtuple_error;

#line 22 "jollyjack/jollyjack_cython.pyx"
      }

#line 22 "jollyjack/jollyjack_cython.pyx"
      const Py_ssize_t kwd_pos_args = __pyx_nargs;

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "read_into_torch", 0) < (0)) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[4]) values[4] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k_);

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[5]) values[5] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__2);

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[6]) values[6] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__3);

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)Py_False));

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)Py_True));

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[9]) values[9] = __Pyx_NewRef(((PyObject *)Py_False));

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[10]) values[10] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyarrow_3lib_CacheOptions *)Py_None));

#line 22 "jollyjack/jollyjack_cython.pyx"
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("read_into_torch", 0, 4, 11, i); __PYX_ERR(0, 22, __pyx_L3_error) }

#line 22 "jollyjack/jollyjack_cython.pyx"
      }

#line 22 "jollyjack/jollyjack_cython.pyx"
    } else {

#line 22 "jollyjack/jollyjack_cython.pyx"
      switch (__pyx_nargs) {

#line 22 "jollyjack/jollyjack_cython.pyx"
        case 11:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case 10:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  9:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  8:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  7:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  6:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  5:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 22 "jollyjack/jollyjack_cython.pyx"
        case  4:

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 22 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
        break;
        default: 
#line 22 "jollyjack/jollyjack_cython.pyx"
goto __pyx_L5_argtuple_error;

#line 22 "jollyjack/jollyjack_cython.pyx"
      }

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[4]) values[4] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k_);

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[5]) values[5] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__2);

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[6]) values[6] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__3);

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)Py_False));

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)Py_True));

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[9]) values[9] = __Pyx_NewRef(((PyObject *)Py_False));

#line 22 "jollyjack/jollyjack_cython.pyx"
      if (!values[10]) values[10] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyarrow_3lib_CacheOptions *)Py_None));

#line 22 "jollyjack/jollyjack_cython.pyx"
    }

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_source = values[0];

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_metadata = ((struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *)values[1]);

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_tensor = values[2];

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_row_group_indices = values[3];

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_row_ranges = values[4];

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_column_indices = values[5];

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_column_names = values[6];

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_pre_buffer = values[7];

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_use_threads = values[8];

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_use_memory_map = values[9];

#line 22 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_cache_options = ((struct __pyx_obj_7pyarrow_3lib_CacheOptions *)values[10]);

#line 22 "jollyjack/jollyjack_cython.pyx"
  }

#line 22 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L6_skip;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L5_argtuple_error:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RaiseArgtupleInvalid("read_into_torch", 0, 4, 11, __pyx_nargs); __PYX_ERR(0, 22, __pyx_L3_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L6_skip:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L3_error:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 22 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 22 "jollyjack/jollyjack_cython.pyx"
  }

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.read_into_torch", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 22 "jollyjack/jollyjack_cython.pyx"
  return NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_metadata), __pyx_mstate_global->__pyx_ptype_7pyarrow_8_parquet_FileMetaData, 1, "metadata", 0))) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cache_options), __pyx_mstate_global->__pyx_ptype_7pyarrow_3lib_CacheOptions, 1, "cache_options", 0))) __PYX_ERR(0, 22, __pyx_L1_error)
  __pyx_r = 
#line 22 "jollyjack/jollyjack_cython.pyx"
__pyx_pf_9jollyjack_16jollyjack_cython_2read_into_torch(__pyx_self, __pyx_v_source, __pyx_v_metadata, __pyx_v_tensor, __pyx_v_row_group_indices, __pyx_v_row_ranges, __pyx_v_column_indices, __pyx_v_column_names, __pyx_v_pre_buffer, __pyx_v_use_threads, __pyx_v_use_memory_map, __pyx_v_cache_options);

#line 22 "jollyjack/jollyjack_cython.pyx"


#line 22 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 22 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 22 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 22 "jollyjack/jollyjack_cython.pyx"
  }

#line 22 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L7_cleaned_up;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 22 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 22 "jollyjack/jollyjack_cython.pyx"
  }

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L7_cleaned_up:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 22 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 22 "jollyjack/jollyjack_cython.pyx"
}

#line 22 "jollyjack/jollyjack_cython.pyx"


#line 22 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_2read_into_torch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_source, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_metadata, PyObject *__pyx_v_tensor, PyObject *__pyx_v_row_group_indices, PyObject *__pyx_v_row_ranges, PyObject *__pyx_v_column_indices, PyObject *__pyx_v_column_names, PyObject *__pyx_v_pre_buffer, PyObject *__pyx_v_use_threads, PyObject *__pyx_v_use_memory_map, struct __pyx_obj_7pyarrow_3lib_CacheOptions *__pyx_v_cache_options) {

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 22 "jollyjack/jollyjack_cython.pyx"
  struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_torch __pyx_t_1;

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("read_into_torch", 0);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.__pyx_n = 7;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.row_ranges = __pyx_v_row_ranges;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.column_indices = __pyx_v_column_indices;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.column_names = __pyx_v_column_names;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.pre_buffer = __pyx_v_pre_buffer;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.use_threads = __pyx_v_use_threads;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.use_memory_map = __pyx_v_use_memory_map;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.cache_options = __pyx_v_cache_options;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_f_9jollyjack_16jollyjack_cython_read_into_torch(__pyx_v_source, __pyx_v_metadata, __pyx_v_tensor, __pyx_v_row_group_indices, 1, &__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = __pyx_t_2;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

#line 22 "jollyjack/jollyjack_cython.pyx"


#line 22 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.read_into_torch", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = NULL;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 22 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 22 "jollyjack/jollyjack_cython.pyx"
}

/* "jollyjack/jollyjack_cython.pyx":41
 *     return
 * 
 * cpdef void read_into_numpy (object source, FileMetaData metadata, cnp.ndarray np_array, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[int] crow_group_indices = row_group_indices
*/

#line 41 "jollyjack/jollyjack_cython.pyx"


#line 41 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_5read_into_numpy(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/

#line 41 "jollyjack/jollyjack_cython.pyx"
static void __pyx_f_9jollyjack_16jollyjack_cython_read_into_numpy(PyObject *__pyx_v_source, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_metadata, PyArrayObject *__pyx_v_np_array, PyObject *__pyx_v_row_group_indices, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_numpy *__pyx_optional_args) {

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_row_ranges = __pyx_mstate_global->__pyx_k__4;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_column_indices = __pyx_mstate_global->__pyx_k__5;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_column_names = __pyx_mstate_global->__pyx_k__6;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_pre_buffer = ((PyObject *)Py_False);

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_use_threads = ((PyObject *)Py_True);

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_v_use_memory_map = ((PyObject *)Py_False);

#line 41 "jollyjack/jollyjack_cython.pyx"
  struct __pyx_obj_7pyarrow_3lib_CacheOptions *__pyx_v_cache_options = ((struct __pyx_obj_7pyarrow_3lib_CacheOptions *)Py_None);
  std::vector<int>  __pyx_v_crow_group_indices
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  std::vector<int>  __pyx_v_ccolumn_indices
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  uint64_t __pyx_v_cstride0_size
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  uint64_t __pyx_v_cstride1_size
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  CYTHON_UNUSED void *__pyx_v_cdata
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  bool __pyx_v_cpre_buffer
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  bool __pyx_v_cuse_threads
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  std::vector<std::string>  __pyx_v_ccolumn_names
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  uint64_t __pyx_v_cbuffer_size
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  std::shared_ptr< parquet::FileMetaData>  __pyx_v_c_metadata
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  std::vector<int>  __pyx_v_ctarget_column_indices
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  std::vector<int64_t>  __pyx_v_ctarget_row_ranges
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_sl = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  int64_t __pyx_v_cexpected_rows
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  arrow::io::CacheOptions __pyx_v_c_cache_options
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  std::shared_ptr< arrow::io::RandomAccessFile>  __pyx_v_rd_handle
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  std::string __pyx_v_pathstr
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  bool __pyx_v_cuse_o_direct
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_jj_reader_backend = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_7genexpr__pyx_v_item = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_8genexpr1__pyx_v_item = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_8genexpr2__pyx_v_c = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_8genexpr3__pyx_v_item = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_8genexpr4__pyx_v_item = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_8genexpr5__pyx_v_c = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_8genexpr6__pyx_v_sl = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_8genexpr6__pyx_v_x = NULL
#line 41 "jollyjack/jollyjack_cython.pyx"
;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 41 "jollyjack/jollyjack_cython.pyx"
  std::vector<int>  __pyx_t_1;

#line 41 "jollyjack/jollyjack_cython.pyx"
  bool __pyx_t_2;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_t_3;

#line 41 "jollyjack/jollyjack_cython.pyx"
  std::shared_ptr< parquet::FileMetaData>  __pyx_t_4;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_t_5;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  size_t __pyx_t_8;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  Py_ssize_t __pyx_t_10;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *(*__pyx_t_11)(PyObject *);

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_12[4];

#line 41 "jollyjack/jollyjack_cython.pyx"
  Py_ssize_t __pyx_t_13;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_t_14;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_t_15;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_16 = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  std::vector<std::string>  __pyx_t_17;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_18 = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_19[3];

#line 41 "jollyjack/jollyjack_cython.pyx"
  std::vector<int64_t>  __pyx_t_20;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_t_21;

#line 41 "jollyjack/jollyjack_cython.pyx"
  arrow::io::CacheOptions __pyx_t_22;

#line 41 "jollyjack/jollyjack_cython.pyx"
  std::string __pyx_t_23;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("read_into_numpy", 0);

#line 41 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_optional_args) {

#line 41 "jollyjack/jollyjack_cython.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 41 "jollyjack/jollyjack_cython.pyx"
      __pyx_v_row_ranges = __pyx_optional_args->row_ranges;

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 41 "jollyjack/jollyjack_cython.pyx"
        __pyx_v_column_indices = __pyx_optional_args->column_indices;

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (__pyx_optional_args->__pyx_n > 2) {

#line 41 "jollyjack/jollyjack_cython.pyx"
          __pyx_v_column_names = __pyx_optional_args->column_names;

#line 41 "jollyjack/jollyjack_cython.pyx"
          if (__pyx_optional_args->__pyx_n > 3) {

#line 41 "jollyjack/jollyjack_cython.pyx"
            __pyx_v_pre_buffer = __pyx_optional_args->pre_buffer;

#line 41 "jollyjack/jollyjack_cython.pyx"
            if (__pyx_optional_args->__pyx_n > 4) {

#line 41 "jollyjack/jollyjack_cython.pyx"
              __pyx_v_use_threads = __pyx_optional_args->use_threads;

#line 41 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_optional_args->__pyx_n > 5) {

#line 41 "jollyjack/jollyjack_cython.pyx"
                __pyx_v_use_memory_map = __pyx_optional_args->use_memory_map;

#line 41 "jollyjack/jollyjack_cython.pyx"
                if (__pyx_optional_args->__pyx_n > 6) {

#line 41 "jollyjack/jollyjack_cython.pyx"
                  __pyx_v_cache_options = __pyx_optional_args->cache_options;

#line 41 "jollyjack/jollyjack_cython.pyx"
                }

#line 41 "jollyjack/jollyjack_cython.pyx"
              }

#line 41 "jollyjack/jollyjack_cython.pyx"
            }

#line 41 "jollyjack/jollyjack_cython.pyx"
          }

#line 41 "jollyjack/jollyjack_cython.pyx"
        }

#line 41 "jollyjack/jollyjack_cython.pyx"
      }

#line 41 "jollyjack/jollyjack_cython.pyx"
    }

#line 41 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":43
 * cpdef void read_into_numpy (object source, FileMetaData metadata, cnp.ndarray np_array, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):
 * 
 *     cdef vector[int] crow_group_indices = row_group_indices             # <<<<<<<<<<<<<<
 *     cdef vector[int] ccolumn_indices
 *     cdef uint64_t cstride0_size = np_array.strides[0]
*/

#line 43 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = __pyx_convert_vector_from_py_int(__pyx_v_row_group_indices); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 43, __pyx_L1_error)

#line 43 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_crow_group_indices = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "jollyjack/jollyjack_cython.pyx":45
 *     cdef vector[int] crow_group_indices = row_group_indices
 *     cdef vector[int] ccolumn_indices
 *     cdef uint64_t cstride0_size = np_array.strides[0]             # <<<<<<<<<<<<<<
 *     cdef uint64_t cstride1_size = np_array.strides[1]
 *     cdef void* cdata = np_array.data
*/

#line 45 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_cstride0_size = (__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_np_array)[0]);

  /* "jollyjack/jollyjack_cython.pyx":46
 *     cdef vector[int] ccolumn_indices
 *     cdef uint64_t cstride0_size = np_array.strides[0]
 *     cdef uint64_t cstride1_size = np_array.strides[1]             # <<<<<<<<<<<<<<
 *     cdef void* cdata = np_array.data
 *     cdef bool cpre_buffer = pre_buffer
*/

#line 46 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_cstride1_size = (__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_np_array)[1]);

  /* "jollyjack/jollyjack_cython.pyx":47
 *     cdef uint64_t cstride0_size = np_array.strides[0]
 *     cdef uint64_t cstride1_size = np_array.strides[1]
 *     cdef void* cdata = np_array.data             # <<<<<<<<<<<<<<
 *     cdef bool cpre_buffer = pre_buffer
 *     cdef bool cuse_threads = use_threads
*/

#line 47 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_cdata = __pyx_f_5numpy_7ndarray_4data_data(__pyx_v_np_array);

  /* "jollyjack/jollyjack_cython.pyx":48
 *     cdef uint64_t cstride1_size = np_array.strides[1]
 *     cdef void* cdata = np_array.data
 *     cdef bool cpre_buffer = pre_buffer             # <<<<<<<<<<<<<<
 *     cdef bool cuse_threads = use_threads
 *     cdef vector[string] ccolumn_names
*/

#line 48 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_pre_buffer); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)

#line 48 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_cpre_buffer = __pyx_t_2;

  /* "jollyjack/jollyjack_cython.pyx":49
 *     cdef void* cdata = np_array.data
 *     cdef bool cpre_buffer = pre_buffer
 *     cdef bool cuse_threads = use_threads             # <<<<<<<<<<<<<<
 *     cdef vector[string] ccolumn_names
 *     cdef uint64_t cbuffer_size = (np_array.shape[0]) * cstride0_size + (np_array.shape[1] - 1) * cstride1_size
*/

#line 49 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_use_threads); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_cuse_threads = __pyx_t_2;

  /* "jollyjack/jollyjack_cython.pyx":51
 *     cdef bool cuse_threads = use_threads
 *     cdef vector[string] ccolumn_names
 *     cdef uint64_t cbuffer_size = (np_array.shape[0]) * cstride0_size + (np_array.shape[1] - 1) * cstride1_size             # <<<<<<<<<<<<<<
 *     cdef shared_ptr[CFileMetaData] c_metadata
 *     cdef vector[int] ctarget_column_indices
*/

#line 51 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_cbuffer_size = (((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_np_array)[0]) * __pyx_v_cstride0_size) + (((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_np_array)[1]) - 1) * __pyx_v_cstride1_size));

  /* "jollyjack/jollyjack_cython.pyx":56
 *     cdef vector[int64_t] ctarget_row_ranges
 * 
 *     if metadata is not None:             # <<<<<<<<<<<<<<
 *         c_metadata = metadata.sp_metadata
 * 
*/

#line 56 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = (((PyObject *)__pyx_v_metadata) != Py_None);

#line 56 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":57
 * 
 *     if metadata is not None:
 *         c_metadata = metadata.sp_metadata             # <<<<<<<<<<<<<<
 * 
 *     if column_indices and isinstance(column_indices, dict):
*/

#line 57 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_4 = __pyx_v_metadata->sp_metadata;

#line 57 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_c_metadata = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4);

    /* "jollyjack/jollyjack_cython.pyx":56
 *     cdef vector[int64_t] ctarget_row_ranges
 * 
 *     if metadata is not None:             # <<<<<<<<<<<<<<
 *         c_metadata = metadata.sp_metadata
 * 
*/

#line 56 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":59
 *         c_metadata = metadata.sp_metadata
 * 
 *     if column_indices and isinstance(column_indices, dict):             # <<<<<<<<<<<<<<
 *         ccolumn_indices = column_indices.keys()
 *         ctarget_column_indices = column_indices.values()
*/

#line 59 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_column_indices); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 59, __pyx_L1_error)

#line 59 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_5) {

#line 59 "jollyjack/jollyjack_cython.pyx"
  } else {

#line 59 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_3 = __pyx_t_5;

#line 59 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L5_bool_binop_done;

#line 59 "jollyjack/jollyjack_cython.pyx"
  }

#line 59 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = PyDict_Check(__pyx_v_column_indices); 

#line 59 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __pyx_t_5;

#line 59 "jollyjack/jollyjack_cython.pyx"
  __pyx_L5_bool_binop_done:;

#line 59 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":60
 * 
 *     if column_indices and isinstance(column_indices, dict):
 *         ccolumn_indices = column_indices.keys()             # <<<<<<<<<<<<<<
 *         ctarget_column_indices = column_indices.values()
 *     elif column_indices and is_iterable_of_iterables(column_indices):
*/

#line 60 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_7 = __pyx_v_column_indices;

#line 60 "jollyjack/jollyjack_cython.pyx"
    __Pyx_INCREF(__pyx_t_7);

#line 60 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_8 = 0;

#line 60 "jollyjack/jollyjack_cython.pyx"
    {

#line 60 "jollyjack/jollyjack_cython.pyx"
      PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};

#line 60 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_keys, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 60 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 60 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 60, __pyx_L1_error)

#line 60 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 60 "jollyjack/jollyjack_cython.pyx"
    }

#line 60 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __pyx_convert_vector_from_py_int(__pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 60, __pyx_L1_error)

#line 60 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 60 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ccolumn_indices = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

    /* "jollyjack/jollyjack_cython.pyx":61
 *     if column_indices and isinstance(column_indices, dict):
 *         ccolumn_indices = column_indices.keys()
 *         ctarget_column_indices = column_indices.values()             # <<<<<<<<<<<<<<
 *     elif column_indices and is_iterable_of_iterables(column_indices):
 *         ccolumn_indices = [item[0] for item in column_indices]
*/

#line 61 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_7 = __pyx_v_column_indices;

#line 61 "jollyjack/jollyjack_cython.pyx"
    __Pyx_INCREF(__pyx_t_7);

#line 61 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_8 = 0;

#line 61 "jollyjack/jollyjack_cython.pyx"
    {

#line 61 "jollyjack/jollyjack_cython.pyx"
      PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};

#line 61 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_values, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 61 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 61 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 61, __pyx_L1_error)

#line 61 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 61 "jollyjack/jollyjack_cython.pyx"
    }

#line 61 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __pyx_convert_vector_from_py_int(__pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L1_error)

#line 61 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 61 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ctarget_column_indices = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

    /* "jollyjack/jollyjack_cython.pyx":59
 *         c_metadata = metadata.sp_metadata
 * 
 *     if column_indices and isinstance(column_indices, dict):             # <<<<<<<<<<<<<<
 *         ccolumn_indices = column_indices.keys()
 *         ctarget_column_indices = column_indices.values()
*/

#line 59 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L4;

#line 59 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":62
 *         ccolumn_indices = column_indices.keys()
 *         ctarget_column_indices = column_indices.values()
 *     elif column_indices and is_iterable_of_iterables(column_indices):             # <<<<<<<<<<<<<<
 *         ccolumn_indices = [item[0] for item in column_indices]
 *         ctarget_column_indices = [item[1] for item in column_indices]
*/

#line 62 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_column_indices); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 62, __pyx_L1_error)

#line 62 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_5) {

#line 62 "jollyjack/jollyjack_cython.pyx"
  } else {

#line 62 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_3 = __pyx_t_5;

#line 62 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L7_bool_binop_done;

#line 62 "jollyjack/jollyjack_cython.pyx"
  }

#line 62 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_7 = NULL;

#line 62 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_is_iterable_of_iterables); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 62, __pyx_L1_error)

#line 62 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_9);

#line 62 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_8 = 1;

#line 62 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_UNPACK_METHODS

#line 62 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(PyMethod_Check(__pyx_t_9))) {

#line 62 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);

#line 62 "jollyjack/jollyjack_cython.pyx"
    assert(__pyx_t_7);

#line 62 "jollyjack/jollyjack_cython.pyx"
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);

#line 62 "jollyjack/jollyjack_cython.pyx"
    __Pyx_INCREF(__pyx_t_7);

#line 62 "jollyjack/jollyjack_cython.pyx"
    __Pyx_INCREF(__pyx__function);

#line 62 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);

#line 62 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_8 = 0;

#line 62 "jollyjack/jollyjack_cython.pyx"
  }

#line 62 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 62 "jollyjack/jollyjack_cython.pyx"
  {

#line 62 "jollyjack/jollyjack_cython.pyx"
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_column_indices};

#line 62 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_9, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 62 "jollyjack/jollyjack_cython.pyx"
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 62 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 62 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 62, __pyx_L1_error)

#line 62 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 62 "jollyjack/jollyjack_cython.pyx"
  }

#line 62 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 62, __pyx_L1_error)

#line 62 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 62 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __pyx_t_5;

#line 62 "jollyjack/jollyjack_cython.pyx"
  __pyx_L7_bool_binop_done:;

#line 62 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":63
 *         ctarget_column_indices = column_indices.values()
 *     elif column_indices and is_iterable_of_iterables(column_indices):
 *         ccolumn_indices = [item[0] for item in column_indices]             # <<<<<<<<<<<<<<
 *         ctarget_column_indices = [item[1] for item in column_indices]
 *     elif column_indices:
*/

#line 63 "jollyjack/jollyjack_cython.pyx"
    { /* enter inner scope */

#line 63 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 63, __pyx_L11_error)

#line 63 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 63 "jollyjack/jollyjack_cython.pyx"
      if (likely(PyList_CheckExact(__pyx_v_column_indices)) || PyTuple_CheckExact(__pyx_v_column_indices)) {

#line 63 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_9 = __pyx_v_column_indices; __Pyx_INCREF(__pyx_t_9);

#line 63 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_10 = 0;

#line 63 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_11 = NULL;

#line 63 "jollyjack/jollyjack_cython.pyx"
      } else {
        __pyx_t_10 = -1; 
#line 63 "jollyjack/jollyjack_cython.pyx"
__pyx_t_9 = PyObject_GetIter(__pyx_v_column_indices); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 63, __pyx_L11_error)

#line 63 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9); 
#line 63 "jollyjack/jollyjack_cython.pyx"
if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 63, __pyx_L11_error)

#line 63 "jollyjack/jollyjack_cython.pyx"
      }
      for (;;
#line 63 "jollyjack/jollyjack_cython.pyx"
) {

#line 63 "jollyjack/jollyjack_cython.pyx"
        if (likely(!__pyx_t_11)) {

#line 63 "jollyjack/jollyjack_cython.pyx"
          if (likely(PyList_CheckExact(__pyx_t_9))) {

#line 63 "jollyjack/jollyjack_cython.pyx"
            {

#line 63 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);

#line 63 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 63 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 63, __pyx_L11_error)

#line 63 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 63 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_10 >= __pyx_temp) break;

#line 63 "jollyjack/jollyjack_cython.pyx"
            }

#line 63 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_7 = __Pyx_PyList_GetItemRefFast(__pyx_t_9, __pyx_t_10, __Pyx_ReferenceSharing_OwnStrongReference);

#line 63 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_10;

#line 63 "jollyjack/jollyjack_cython.pyx"
          } else {

#line 63 "jollyjack/jollyjack_cython.pyx"
            {

#line 63 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_9);

#line 63 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 63 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 63, __pyx_L11_error)

#line 63 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 63 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_10 >= __pyx_temp) break;

#line 63 "jollyjack/jollyjack_cython.pyx"
            }

#line 63 "jollyjack/jollyjack_cython.pyx"
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 63 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_7 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_10));

#line 63 "jollyjack/jollyjack_cython.pyx"
            #else

#line 63 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_10);

#line 63 "jollyjack/jollyjack_cython.pyx"
            #endif

#line 63 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_10;

#line 63 "jollyjack/jollyjack_cython.pyx"
          }

#line 63 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 63, __pyx_L11_error)
        } else 
#line 63 "jollyjack/jollyjack_cython.pyx"
{

#line 63 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_7 = __pyx_t_11(__pyx_t_9);

#line 63 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_7)) {

#line 63 "jollyjack/jollyjack_cython.pyx"
            PyObject* exc_type = PyErr_Occurred();

#line 63 "jollyjack/jollyjack_cython.pyx"
            if (exc_type) {

#line 63 "jollyjack/jollyjack_cython.pyx"
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 63, __pyx_L11_error)

#line 63 "jollyjack/jollyjack_cython.pyx"
              PyErr_Clear();

#line 63 "jollyjack/jollyjack_cython.pyx"
            }

#line 63 "jollyjack/jollyjack_cython.pyx"
            break;

#line 63 "jollyjack/jollyjack_cython.pyx"
          }

#line 63 "jollyjack/jollyjack_cython.pyx"
        }

#line 63 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 63 "jollyjack/jollyjack_cython.pyx"
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_item, __pyx_t_7);

#line 63 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = 0;

#line 63 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_7genexpr__pyx_v_item, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 63, __pyx_L11_error)

#line 63 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 63 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 63, __pyx_L11_error)

#line 63 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 63 "jollyjack/jollyjack_cython.pyx"
      }

#line 63 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 63 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_item); __pyx_7genexpr__pyx_v_item = 0;

#line 63 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L15_exit_scope;

#line 63 "jollyjack/jollyjack_cython.pyx"
      __pyx_L11_error:;

#line 63 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_item); __pyx_7genexpr__pyx_v_item = 0;

#line 63 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L1_error;

#line 63 "jollyjack/jollyjack_cython.pyx"
      __pyx_L15_exit_scope:;

#line 63 "jollyjack/jollyjack_cython.pyx"
    } /* exit inner scope */

#line 63 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __pyx_convert_vector_from_py_int(__pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 63, __pyx_L1_error)

#line 63 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 63 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ccolumn_indices = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

    /* "jollyjack/jollyjack_cython.pyx":64
 *     elif column_indices and is_iterable_of_iterables(column_indices):
 *         ccolumn_indices = [item[0] for item in column_indices]
 *         ctarget_column_indices = [item[1] for item in column_indices]             # <<<<<<<<<<<<<<
 *     elif column_indices:
 *         assert len(column_indices) == np_array.shape[1], f"Requested to read {len(column_indices)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
*/

#line 64 "jollyjack/jollyjack_cython.pyx"
    { /* enter inner scope */

#line 64 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 64, __pyx_L18_error)

#line 64 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 64 "jollyjack/jollyjack_cython.pyx"
      if (likely(PyList_CheckExact(__pyx_v_column_indices)) || PyTuple_CheckExact(__pyx_v_column_indices)) {

#line 64 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_9 = __pyx_v_column_indices; __Pyx_INCREF(__pyx_t_9);

#line 64 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_10 = 0;

#line 64 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_11 = NULL;

#line 64 "jollyjack/jollyjack_cython.pyx"
      } else {
        __pyx_t_10 = -1; 
#line 64 "jollyjack/jollyjack_cython.pyx"
__pyx_t_9 = PyObject_GetIter(__pyx_v_column_indices); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 64, __pyx_L18_error)

#line 64 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9); 
#line 64 "jollyjack/jollyjack_cython.pyx"
if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 64, __pyx_L18_error)

#line 64 "jollyjack/jollyjack_cython.pyx"
      }
      for (;;
#line 64 "jollyjack/jollyjack_cython.pyx"
) {

#line 64 "jollyjack/jollyjack_cython.pyx"
        if (likely(!__pyx_t_11)) {

#line 64 "jollyjack/jollyjack_cython.pyx"
          if (likely(PyList_CheckExact(__pyx_t_9))) {

#line 64 "jollyjack/jollyjack_cython.pyx"
            {

#line 64 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);

#line 64 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 64 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 64, __pyx_L18_error)

#line 64 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 64 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_10 >= __pyx_temp) break;

#line 64 "jollyjack/jollyjack_cython.pyx"
            }

#line 64 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_7 = __Pyx_PyList_GetItemRefFast(__pyx_t_9, __pyx_t_10, __Pyx_ReferenceSharing_OwnStrongReference);

#line 64 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_10;

#line 64 "jollyjack/jollyjack_cython.pyx"
          } else {

#line 64 "jollyjack/jollyjack_cython.pyx"
            {

#line 64 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_9);

#line 64 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 64 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 64, __pyx_L18_error)

#line 64 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 64 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_10 >= __pyx_temp) break;

#line 64 "jollyjack/jollyjack_cython.pyx"
            }

#line 64 "jollyjack/jollyjack_cython.pyx"
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 64 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_7 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_10));

#line 64 "jollyjack/jollyjack_cython.pyx"
            #else

#line 64 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_10);

#line 64 "jollyjack/jollyjack_cython.pyx"
            #endif

#line 64 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_10;

#line 64 "jollyjack/jollyjack_cython.pyx"
          }

#line 64 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 64, __pyx_L18_error)
        } else 
#line 64 "jollyjack/jollyjack_cython.pyx"
{

#line 64 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_7 = __pyx_t_11(__pyx_t_9);

#line 64 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_7)) {

#line 64 "jollyjack/jollyjack_cython.pyx"
            PyObject* exc_type = PyErr_Occurred();

#line 64 "jollyjack/jollyjack_cython.pyx"
            if (exc_type) {

#line 64 "jollyjack/jollyjack_cython.pyx"
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 64, __pyx_L18_error)

#line 64 "jollyjack/jollyjack_cython.pyx"
              PyErr_Clear();

#line 64 "jollyjack/jollyjack_cython.pyx"
            }

#line 64 "jollyjack/jollyjack_cython.pyx"
            break;

#line 64 "jollyjack/jollyjack_cython.pyx"
          }

#line 64 "jollyjack/jollyjack_cython.pyx"
        }

#line 64 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 64 "jollyjack/jollyjack_cython.pyx"
        __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_item, __pyx_t_7);

#line 64 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = 0;

#line 64 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_8genexpr1__pyx_v_item, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 64, __pyx_L18_error)

#line 64 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 64 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 64, __pyx_L18_error)

#line 64 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 64 "jollyjack/jollyjack_cython.pyx"
      }

#line 64 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 64 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_item); __pyx_8genexpr1__pyx_v_item = 0;

#line 64 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L22_exit_scope;

#line 64 "jollyjack/jollyjack_cython.pyx"
      __pyx_L18_error:;

#line 64 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_item); __pyx_8genexpr1__pyx_v_item = 0;

#line 64 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L1_error;

#line 64 "jollyjack/jollyjack_cython.pyx"
      __pyx_L22_exit_scope:;

#line 64 "jollyjack/jollyjack_cython.pyx"
    } /* exit inner scope */

#line 64 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __pyx_convert_vector_from_py_int(__pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L1_error)

#line 64 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 64 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ctarget_column_indices = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

    /* "jollyjack/jollyjack_cython.pyx":62
 *         ccolumn_indices = column_indices.keys()
 *         ctarget_column_indices = column_indices.values()
 *     elif column_indices and is_iterable_of_iterables(column_indices):             # <<<<<<<<<<<<<<
 *         ccolumn_indices = [item[0] for item in column_indices]
 *         ctarget_column_indices = [item[1] for item in column_indices]
*/

#line 62 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L4;

#line 62 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":65
 *         ccolumn_indices = [item[0] for item in column_indices]
 *         ctarget_column_indices = [item[1] for item in column_indices]
 *     elif column_indices:             # <<<<<<<<<<<<<<
 *         assert len(column_indices) == np_array.shape[1], f"Requested to read {len(column_indices)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
 *         ccolumn_indices = column_indices
*/

#line 65 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_column_indices); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 65, __pyx_L1_error)

#line 65 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":66
 *         ctarget_column_indices = [item[1] for item in column_indices]
 *     elif column_indices:
 *         assert len(column_indices) == np_array.shape[1], f"Requested to read {len(column_indices)} columns, but the number of columns in numpy array is {np_array.shape[1]}"             # <<<<<<<<<<<<<<
 *         ccolumn_indices = column_indices
 * 
*/

#line 66 "jollyjack/jollyjack_cython.pyx"
    #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 66 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(__pyx_assertions_enabled())) {

#line 66 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_10 = PyObject_Length(__pyx_v_column_indices); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 66, __pyx_L1_error)

#line 66 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3 = (__pyx_t_10 == (__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_np_array)[1]));

#line 66 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_3)) {

#line 66 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_10 = PyObject_Length(__pyx_v_column_indices); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 66, __pyx_L1_error)

#line 66 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_10, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 66, __pyx_L1_error)

#line 66 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 66 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_9 = __Pyx_PyUnicode_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_np_array)[1]), 0, ' ', 'd'); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 66, __pyx_L1_error)

#line 66 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_9);

#line 66 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_12[0] = __pyx_mstate_global->__pyx_kp_u_Requested_to_read;

#line 66 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_12[1] = __pyx_t_6;

#line 66 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_12[2] = __pyx_mstate_global->__pyx_kp_u_columns_but_the_number_of_colum;

#line 66 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_12[3] = __pyx_t_9;

#line 66 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_12, 4, 18 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 54 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9), 127);

#line 66 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 66, __pyx_L1_error)

#line 66 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 66 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 66 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 66 "jollyjack/jollyjack_cython.pyx"
        __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_7, 0, 0);

#line 66 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 66 "jollyjack/jollyjack_cython.pyx"
        __PYX_ERR(0, 66, __pyx_L1_error)

#line 66 "jollyjack/jollyjack_cython.pyx"
      }

#line 66 "jollyjack/jollyjack_cython.pyx"
    }

#line 66 "jollyjack/jollyjack_cython.pyx"
    #else

#line 66 "jollyjack/jollyjack_cython.pyx"
    if ((1)); else __PYX_ERR(0, 66, __pyx_L1_error)

#line 66 "jollyjack/jollyjack_cython.pyx"
    #endif

    /* "jollyjack/jollyjack_cython.pyx":67
 *     elif column_indices:
 *         assert len(column_indices) == np_array.shape[1], f"Requested to read {len(column_indices)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
 *         ccolumn_indices = column_indices             # <<<<<<<<<<<<<<
 * 
 *     if column_names and isinstance(column_names, dict):
*/

#line 67 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __pyx_convert_vector_from_py_int(__pyx_v_column_indices); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L1_error)

#line 67 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ccolumn_indices = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

    /* "jollyjack/jollyjack_cython.pyx":65
 *         ccolumn_indices = [item[0] for item in column_indices]
 *         ctarget_column_indices = [item[1] for item in column_indices]
 *     elif column_indices:             # <<<<<<<<<<<<<<
 *         assert len(column_indices) == np_array.shape[1], f"Requested to read {len(column_indices)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
 *         ccolumn_indices = column_indices
*/

#line 65 "jollyjack/jollyjack_cython.pyx"
  }

#line 65 "jollyjack/jollyjack_cython.pyx"
  __pyx_L4:;

  /* "jollyjack/jollyjack_cython.pyx":69
 *         ccolumn_indices = column_indices
 * 
 *     if column_names and isinstance(column_names, dict):             # <<<<<<<<<<<<<<
 *         ccolumn_names = [c.encode('utf8') for c in column_names.keys()]
 *         ctarget_column_indices = column_names.values()
*/

#line 69 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_column_names); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 69, __pyx_L1_error)

#line 69 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_5) {

#line 69 "jollyjack/jollyjack_cython.pyx"
  } else {

#line 69 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_3 = __pyx_t_5;

#line 69 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L24_bool_binop_done;

#line 69 "jollyjack/jollyjack_cython.pyx"
  }

#line 69 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = PyDict_Check(__pyx_v_column_names); 

#line 69 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __pyx_t_5;

#line 69 "jollyjack/jollyjack_cython.pyx"
  __pyx_L24_bool_binop_done:;

#line 69 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":70
 * 
 *     if column_names and isinstance(column_names, dict):
 *         ccolumn_names = [c.encode('utf8') for c in column_names.keys()]             # <<<<<<<<<<<<<<
 *         ctarget_column_indices = column_names.values()
 *     elif column_names and is_iterable_of_iterables(column_names):
*/

#line 70 "jollyjack/jollyjack_cython.pyx"
    { /* enter inner scope */

#line 70 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 70, __pyx_L28_error)

#line 70 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 70 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_10 = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(__pyx_v_column_names == Py_None)) {

#line 70 "jollyjack/jollyjack_cython.pyx"
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");

#line 70 "jollyjack/jollyjack_cython.pyx"
        __PYX_ERR(0, 70, __pyx_L28_error)

#line 70 "jollyjack/jollyjack_cython.pyx"
      }

#line 70 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_column_names, 0, __pyx_mstate_global->__pyx_n_u_keys, (&__pyx_t_13), (&__pyx_t_14)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 70, __pyx_L28_error)

#line 70 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 70 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_t_9);

#line 70 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9 = __pyx_t_6;

#line 70 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
      while (1) {

#line 70 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_15 = __Pyx_dict_iter_next(__pyx_t_9, __pyx_t_13, &__pyx_t_10, &__pyx_t_6, NULL, NULL, __pyx_t_14);

#line 70 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(__pyx_t_15 == 0)) break;

#line 70 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(__pyx_t_15 == -1)) __PYX_ERR(0, 70, __pyx_L28_error)

#line 70 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 70 "jollyjack/jollyjack_cython.pyx"
        __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_c, __pyx_t_6);

#line 70 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_16 = __pyx_8genexpr2__pyx_v_c;

#line 70 "jollyjack/jollyjack_cython.pyx"
        __Pyx_INCREF(__pyx_t_16);

#line 70 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_8 = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
        {

#line 70 "jollyjack/jollyjack_cython.pyx"
          PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_mstate_global->__pyx_n_u_utf8};

#line 70 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 70 "jollyjack/jollyjack_cython.pyx"
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 70, __pyx_L28_error)

#line 70 "jollyjack/jollyjack_cython.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 70 "jollyjack/jollyjack_cython.pyx"
        }

#line 70 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 70, __pyx_L28_error)

#line 70 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
      }

#line 70 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_c); __pyx_8genexpr2__pyx_v_c = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L31_exit_scope;

#line 70 "jollyjack/jollyjack_cython.pyx"
      __pyx_L28_error:;

#line 70 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_c); __pyx_8genexpr2__pyx_v_c = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L1_error;

#line 70 "jollyjack/jollyjack_cython.pyx"
      __pyx_L31_exit_scope:;

#line 70 "jollyjack/jollyjack_cython.pyx"
    } /* exit inner scope */

#line 70 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_17 = __pyx_convert_vector_from_py_std_3a__3a_string(__pyx_t_7); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 70, __pyx_L1_error)

#line 70 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 70 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ccolumn_names = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_17);

    /* "jollyjack/jollyjack_cython.pyx":71
 *     if column_names and isinstance(column_names, dict):
 *         ccolumn_names = [c.encode('utf8') for c in column_names.keys()]
 *         ctarget_column_indices = column_names.values()             # <<<<<<<<<<<<<<
 *     elif column_names and is_iterable_of_iterables(column_names):
 *         ccolumn_names = [item[0].encode('utf8') for item in column_names]
*/

#line 71 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_9 = __pyx_v_column_names;

#line 71 "jollyjack/jollyjack_cython.pyx"
    __Pyx_INCREF(__pyx_t_9);

#line 71 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_8 = 0;

#line 71 "jollyjack/jollyjack_cython.pyx"
    {

#line 71 "jollyjack/jollyjack_cython.pyx"
      PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};

#line 71 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_values, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 71 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 71 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 71, __pyx_L1_error)

#line 71 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 71 "jollyjack/jollyjack_cython.pyx"
    }

#line 71 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __pyx_convert_vector_from_py_int(__pyx_t_7); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 71, __pyx_L1_error)

#line 71 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 71 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ctarget_column_indices = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

    /* "jollyjack/jollyjack_cython.pyx":69
 *         ccolumn_indices = column_indices
 * 
 *     if column_names and isinstance(column_names, dict):             # <<<<<<<<<<<<<<
 *         ccolumn_names = [c.encode('utf8') for c in column_names.keys()]
 *         ctarget_column_indices = column_names.values()
*/

#line 69 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L23;

#line 69 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":72
 *         ccolumn_names = [c.encode('utf8') for c in column_names.keys()]
 *         ctarget_column_indices = column_names.values()
 *     elif column_names and is_iterable_of_iterables(column_names):             # <<<<<<<<<<<<<<
 *         ccolumn_names = [item[0].encode('utf8') for item in column_names]
 *         ctarget_column_indices = [item[1] for item in column_names]
*/

#line 72 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_column_names); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 72, __pyx_L1_error)

#line 72 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_5) {

#line 72 "jollyjack/jollyjack_cython.pyx"
  } else {

#line 72 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_3 = __pyx_t_5;

#line 72 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L32_bool_binop_done;

#line 72 "jollyjack/jollyjack_cython.pyx"
  }

#line 72 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_9 = NULL;

#line 72 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_is_iterable_of_iterables); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 72, __pyx_L1_error)

#line 72 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 72 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_8 = 1;

#line 72 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_UNPACK_METHODS

#line 72 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(PyMethod_Check(__pyx_t_6))) {

#line 72 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);

#line 72 "jollyjack/jollyjack_cython.pyx"
    assert(__pyx_t_9);

#line 72 "jollyjack/jollyjack_cython.pyx"
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 72 "jollyjack/jollyjack_cython.pyx"
    __Pyx_INCREF(__pyx_t_9);

#line 72 "jollyjack/jollyjack_cython.pyx"
    __Pyx_INCREF(__pyx__function);

#line 72 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);

#line 72 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_8 = 0;

#line 72 "jollyjack/jollyjack_cython.pyx"
  }

#line 72 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 72 "jollyjack/jollyjack_cython.pyx"
  {

#line 72 "jollyjack/jollyjack_cython.pyx"
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_column_names};

#line 72 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_7 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 72 "jollyjack/jollyjack_cython.pyx"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 72 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 72 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 72, __pyx_L1_error)

#line 72 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_7);

#line 72 "jollyjack/jollyjack_cython.pyx"
  }

#line 72 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 72, __pyx_L1_error)

#line 72 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 72 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __pyx_t_5;

#line 72 "jollyjack/jollyjack_cython.pyx"
  __pyx_L32_bool_binop_done:;

#line 72 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":73
 *         ctarget_column_indices = column_names.values()
 *     elif column_names and is_iterable_of_iterables(column_names):
 *         ccolumn_names = [item[0].encode('utf8') for item in column_names]             # <<<<<<<<<<<<<<
 *         ctarget_column_indices = [item[1] for item in column_names]
 *     elif column_names:
*/

#line 73 "jollyjack/jollyjack_cython.pyx"
    { /* enter inner scope */

#line 73 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 73, __pyx_L36_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 73 "jollyjack/jollyjack_cython.pyx"
      if (likely(PyList_CheckExact(__pyx_v_column_names)) || PyTuple_CheckExact(__pyx_v_column_names)) {

#line 73 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = __pyx_v_column_names; __Pyx_INCREF(__pyx_t_6);

#line 73 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_13 = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_11 = NULL;

#line 73 "jollyjack/jollyjack_cython.pyx"
      } else {
        __pyx_t_13 = -1; 
#line 73 "jollyjack/jollyjack_cython.pyx"
__pyx_t_6 = PyObject_GetIter(__pyx_v_column_names); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 73, __pyx_L36_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); 
#line 73 "jollyjack/jollyjack_cython.pyx"
if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 73, __pyx_L36_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
      }
      for (;;
#line 73 "jollyjack/jollyjack_cython.pyx"
) {

#line 73 "jollyjack/jollyjack_cython.pyx"
        if (likely(!__pyx_t_11)) {

#line 73 "jollyjack/jollyjack_cython.pyx"
          if (likely(PyList_CheckExact(__pyx_t_6))) {

#line 73 "jollyjack/jollyjack_cython.pyx"
            {

#line 73 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);

#line 73 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 73 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 73, __pyx_L36_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 73 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_13 >= __pyx_temp) break;

#line 73 "jollyjack/jollyjack_cython.pyx"
            }

#line 73 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_9 = __Pyx_PyList_GetItemRefFast(__pyx_t_6, __pyx_t_13, __Pyx_ReferenceSharing_OwnStrongReference);

#line 73 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_13;

#line 73 "jollyjack/jollyjack_cython.pyx"
          } else {

#line 73 "jollyjack/jollyjack_cython.pyx"
            {

#line 73 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);

#line 73 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 73 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 73, __pyx_L36_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 73 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_13 >= __pyx_temp) break;

#line 73 "jollyjack/jollyjack_cython.pyx"
            }

#line 73 "jollyjack/jollyjack_cython.pyx"
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 73 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_9 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_13));

#line 73 "jollyjack/jollyjack_cython.pyx"
            #else

#line 73 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_9 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_13);

#line 73 "jollyjack/jollyjack_cython.pyx"
            #endif

#line 73 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_13;

#line 73 "jollyjack/jollyjack_cython.pyx"
          }

#line 73 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 73, __pyx_L36_error)
        } else 
#line 73 "jollyjack/jollyjack_cython.pyx"
{

#line 73 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_9 = __pyx_t_11(__pyx_t_6);

#line 73 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_9)) {

#line 73 "jollyjack/jollyjack_cython.pyx"
            PyObject* exc_type = PyErr_Occurred();

#line 73 "jollyjack/jollyjack_cython.pyx"
            if (exc_type) {

#line 73 "jollyjack/jollyjack_cython.pyx"
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 73, __pyx_L36_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
              PyErr_Clear();

#line 73 "jollyjack/jollyjack_cython.pyx"
            }

#line 73 "jollyjack/jollyjack_cython.pyx"
            break;

#line 73 "jollyjack/jollyjack_cython.pyx"
          }

#line 73 "jollyjack/jollyjack_cython.pyx"
        }

#line 73 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_9);

#line 73 "jollyjack/jollyjack_cython.pyx"
        __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_item, __pyx_t_9);

#line 73 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_9 = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_18 = __Pyx_GetItemInt(__pyx_8genexpr3__pyx_v_item, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 73, __pyx_L36_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_18);

#line 73 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_16 = __pyx_t_18;

#line 73 "jollyjack/jollyjack_cython.pyx"
        __Pyx_INCREF(__pyx_t_16);

#line 73 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_8 = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
        {

#line 73 "jollyjack/jollyjack_cython.pyx"
          PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_mstate_global->__pyx_n_u_utf8};

#line 73 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_9 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 73 "jollyjack/jollyjack_cython.pyx"
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 73, __pyx_L36_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
          __Pyx_GOTREF(__pyx_t_9);

#line 73 "jollyjack/jollyjack_cython.pyx"
        }

#line 73 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 73, __pyx_L36_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
      }

#line 73 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_item); __pyx_8genexpr3__pyx_v_item = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L40_exit_scope;

#line 73 "jollyjack/jollyjack_cython.pyx"
      __pyx_L36_error:;

#line 73 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_item); __pyx_8genexpr3__pyx_v_item = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L1_error;

#line 73 "jollyjack/jollyjack_cython.pyx"
      __pyx_L40_exit_scope:;

#line 73 "jollyjack/jollyjack_cython.pyx"
    } /* exit inner scope */

#line 73 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_17 = __pyx_convert_vector_from_py_std_3a__3a_string(__pyx_t_7); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L1_error)

#line 73 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 73 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ccolumn_names = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_17);

    /* "jollyjack/jollyjack_cython.pyx":74
 *     elif column_names and is_iterable_of_iterables(column_names):
 *         ccolumn_names = [item[0].encode('utf8') for item in column_names]
 *         ctarget_column_indices = [item[1] for item in column_names]             # <<<<<<<<<<<<<<
 *     elif column_names:
 *         assert len(column_names) == np_array.shape[1], f"Requested to read {len(column_names)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
*/

#line 74 "jollyjack/jollyjack_cython.pyx"
    { /* enter inner scope */

#line 74 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 74, __pyx_L43_error)

#line 74 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 74 "jollyjack/jollyjack_cython.pyx"
      if (likely(PyList_CheckExact(__pyx_v_column_names)) || PyTuple_CheckExact(__pyx_v_column_names)) {

#line 74 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = __pyx_v_column_names; __Pyx_INCREF(__pyx_t_6);

#line 74 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_13 = 0;

#line 74 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_11 = NULL;

#line 74 "jollyjack/jollyjack_cython.pyx"
      } else {
        __pyx_t_13 = -1; 
#line 74 "jollyjack/jollyjack_cython.pyx"
__pyx_t_6 = PyObject_GetIter(__pyx_v_column_names); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 74, __pyx_L43_error)

#line 74 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); 
#line 74 "jollyjack/jollyjack_cython.pyx"
if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 74, __pyx_L43_error)

#line 74 "jollyjack/jollyjack_cython.pyx"
      }
      for (;;
#line 74 "jollyjack/jollyjack_cython.pyx"
) {

#line 74 "jollyjack/jollyjack_cython.pyx"
        if (likely(!__pyx_t_11)) {

#line 74 "jollyjack/jollyjack_cython.pyx"
          if (likely(PyList_CheckExact(__pyx_t_6))) {

#line 74 "jollyjack/jollyjack_cython.pyx"
            {

#line 74 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);

#line 74 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 74 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 74, __pyx_L43_error)

#line 74 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 74 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_13 >= __pyx_temp) break;

#line 74 "jollyjack/jollyjack_cython.pyx"
            }

#line 74 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_9 = __Pyx_PyList_GetItemRefFast(__pyx_t_6, __pyx_t_13, __Pyx_ReferenceSharing_OwnStrongReference);

#line 74 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_13;

#line 74 "jollyjack/jollyjack_cython.pyx"
          } else {

#line 74 "jollyjack/jollyjack_cython.pyx"
            {

#line 74 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);

#line 74 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 74 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 74, __pyx_L43_error)

#line 74 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 74 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_13 >= __pyx_temp) break;

#line 74 "jollyjack/jollyjack_cython.pyx"
            }

#line 74 "jollyjack/jollyjack_cython.pyx"
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 74 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_9 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_13));

#line 74 "jollyjack/jollyjack_cython.pyx"
            #else

#line 74 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_9 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_13);

#line 74 "jollyjack/jollyjack_cython.pyx"
            #endif

#line 74 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_13;

#line 74 "jollyjack/jollyjack_cython.pyx"
          }

#line 74 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 74, __pyx_L43_error)
        } else 
#line 74 "jollyjack/jollyjack_cython.pyx"
{

#line 74 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_9 = __pyx_t_11(__pyx_t_6);

#line 74 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_9)) {

#line 74 "jollyjack/jollyjack_cython.pyx"
            PyObject* exc_type = PyErr_Occurred();

#line 74 "jollyjack/jollyjack_cython.pyx"
            if (exc_type) {

#line 74 "jollyjack/jollyjack_cython.pyx"
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 74, __pyx_L43_error)

#line 74 "jollyjack/jollyjack_cython.pyx"
              PyErr_Clear();

#line 74 "jollyjack/jollyjack_cython.pyx"
            }

#line 74 "jollyjack/jollyjack_cython.pyx"
            break;

#line 74 "jollyjack/jollyjack_cython.pyx"
          }

#line 74 "jollyjack/jollyjack_cython.pyx"
        }

#line 74 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_9);

#line 74 "jollyjack/jollyjack_cython.pyx"
        __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_item, __pyx_t_9);

#line 74 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_9 = 0;

#line 74 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_9 = __Pyx_GetItemInt(__pyx_8genexpr4__pyx_v_item, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 74, __pyx_L43_error)

#line 74 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_9);

#line 74 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 74, __pyx_L43_error)

#line 74 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 74 "jollyjack/jollyjack_cython.pyx"
      }

#line 74 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 74 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_item); __pyx_8genexpr4__pyx_v_item = 0;

#line 74 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L47_exit_scope;

#line 74 "jollyjack/jollyjack_cython.pyx"
      __pyx_L43_error:;

#line 74 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_item); __pyx_8genexpr4__pyx_v_item = 0;

#line 74 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L1_error;

#line 74 "jollyjack/jollyjack_cython.pyx"
      __pyx_L47_exit_scope:;

#line 74 "jollyjack/jollyjack_cython.pyx"
    } /* exit inner scope */

#line 74 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __pyx_convert_vector_from_py_int(__pyx_t_7); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 74, __pyx_L1_error)

#line 74 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 74 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ctarget_column_indices = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

    /* "jollyjack/jollyjack_cython.pyx":72
 *         ccolumn_names = [c.encode('utf8') for c in column_names.keys()]
 *         ctarget_column_indices = column_names.values()
 *     elif column_names and is_iterable_of_iterables(column_names):             # <<<<<<<<<<<<<<
 *         ccolumn_names = [item[0].encode('utf8') for item in column_names]
 *         ctarget_column_indices = [item[1] for item in column_names]
*/

#line 72 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L23;

#line 72 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":75
 *         ccolumn_names = [item[0].encode('utf8') for item in column_names]
 *         ctarget_column_indices = [item[1] for item in column_names]
 *     elif column_names:             # <<<<<<<<<<<<<<
 *         assert len(column_names) == np_array.shape[1], f"Requested to read {len(column_names)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
 *         ccolumn_names = [c.encode('utf8') for c in column_names]
*/

#line 75 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_column_names); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 75, __pyx_L1_error)

#line 75 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":76
 *         ctarget_column_indices = [item[1] for item in column_names]
 *     elif column_names:
 *         assert len(column_names) == np_array.shape[1], f"Requested to read {len(column_names)} columns, but the number of columns in numpy array is {np_array.shape[1]}"             # <<<<<<<<<<<<<<
 *         ccolumn_names = [c.encode('utf8') for c in column_names]
 * 
*/

#line 76 "jollyjack/jollyjack_cython.pyx"
    #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 76 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(__pyx_assertions_enabled())) {

#line 76 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_13 = PyObject_Length(__pyx_v_column_names); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 76, __pyx_L1_error)

#line 76 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3 = (__pyx_t_13 == (__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_np_array)[1]));

#line 76 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_3)) {

#line 76 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_13 = PyObject_Length(__pyx_v_column_names); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 76, __pyx_L1_error)

#line 76 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_13, 0, ' ', 'd'); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 76, __pyx_L1_error)

#line 76 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 76 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = __Pyx_PyUnicode_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_np_array)[1]), 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 76, __pyx_L1_error)

#line 76 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 76 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_12[0] = __pyx_mstate_global->__pyx_kp_u_Requested_to_read;

#line 76 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_12[1] = __pyx_t_7;

#line 76 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_12[2] = __pyx_mstate_global->__pyx_kp_u_columns_but_the_number_of_colum;

#line 76 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_12[3] = __pyx_t_6;

#line 76 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_12, 4, 18 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + 54 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6), 127);

#line 76 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 76, __pyx_L1_error)

#line 76 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_9);

#line 76 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 76 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 76 "jollyjack/jollyjack_cython.pyx"
        __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_9, 0, 0);

#line 76 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 76 "jollyjack/jollyjack_cython.pyx"
        __PYX_ERR(0, 76, __pyx_L1_error)

#line 76 "jollyjack/jollyjack_cython.pyx"
      }

#line 76 "jollyjack/jollyjack_cython.pyx"
    }

#line 76 "jollyjack/jollyjack_cython.pyx"
    #else

#line 76 "jollyjack/jollyjack_cython.pyx"
    if ((1)); else __PYX_ERR(0, 76, __pyx_L1_error)

#line 76 "jollyjack/jollyjack_cython.pyx"
    #endif

    /* "jollyjack/jollyjack_cython.pyx":77
 *     elif column_names:
 *         assert len(column_names) == np_array.shape[1], f"Requested to read {len(column_names)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
 *         ccolumn_names = [c.encode('utf8') for c in column_names]             # <<<<<<<<<<<<<<
 * 
 *     if row_ranges:
*/

#line 77 "jollyjack/jollyjack_cython.pyx"
    { /* enter inner scope */

#line 77 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 77, __pyx_L50_error)

#line 77 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 77 "jollyjack/jollyjack_cython.pyx"
      if (likely(PyList_CheckExact(__pyx_v_column_names)) || PyTuple_CheckExact(__pyx_v_column_names)) {

#line 77 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = __pyx_v_column_names; __Pyx_INCREF(__pyx_t_6);

#line 77 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_13 = 0;

#line 77 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_11 = NULL;

#line 77 "jollyjack/jollyjack_cython.pyx"
      } else {
        __pyx_t_13 = -1; 
#line 77 "jollyjack/jollyjack_cython.pyx"
__pyx_t_6 = PyObject_GetIter(__pyx_v_column_names); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 77, __pyx_L50_error)

#line 77 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); 
#line 77 "jollyjack/jollyjack_cython.pyx"
if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 77, __pyx_L50_error)

#line 77 "jollyjack/jollyjack_cython.pyx"
      }
      for (;;
#line 77 "jollyjack/jollyjack_cython.pyx"
) {

#line 77 "jollyjack/jollyjack_cython.pyx"
        if (likely(!__pyx_t_11)) {

#line 77 "jollyjack/jollyjack_cython.pyx"
          if (likely(PyList_CheckExact(__pyx_t_6))) {

#line 77 "jollyjack/jollyjack_cython.pyx"
            {

#line 77 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);

#line 77 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 77 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 77, __pyx_L50_error)

#line 77 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 77 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_13 >= __pyx_temp) break;

#line 77 "jollyjack/jollyjack_cython.pyx"
            }

#line 77 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_7 = __Pyx_PyList_GetItemRefFast(__pyx_t_6, __pyx_t_13, __Pyx_ReferenceSharing_OwnStrongReference);

#line 77 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_13;

#line 77 "jollyjack/jollyjack_cython.pyx"
          } else {

#line 77 "jollyjack/jollyjack_cython.pyx"
            {

#line 77 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);

#line 77 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 77 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 77, __pyx_L50_error)

#line 77 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 77 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_13 >= __pyx_temp) break;

#line 77 "jollyjack/jollyjack_cython.pyx"
            }

#line 77 "jollyjack/jollyjack_cython.pyx"
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 77 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_7 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_13));

#line 77 "jollyjack/jollyjack_cython.pyx"
            #else

#line 77 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_13);

#line 77 "jollyjack/jollyjack_cython.pyx"
            #endif

#line 77 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_13;

#line 77 "jollyjack/jollyjack_cython.pyx"
          }

#line 77 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 77, __pyx_L50_error)
        } else 
#line 77 "jollyjack/jollyjack_cython.pyx"
{

#line 77 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_7 = __pyx_t_11(__pyx_t_6);

#line 77 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_7)) {

#line 77 "jollyjack/jollyjack_cython.pyx"
            PyObject* exc_type = PyErr_Occurred();

#line 77 "jollyjack/jollyjack_cython.pyx"
            if (exc_type) {

#line 77 "jollyjack/jollyjack_cython.pyx"
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 77, __pyx_L50_error)

#line 77 "jollyjack/jollyjack_cython.pyx"
              PyErr_Clear();

#line 77 "jollyjack/jollyjack_cython.pyx"
            }

#line 77 "jollyjack/jollyjack_cython.pyx"
            break;

#line 77 "jollyjack/jollyjack_cython.pyx"
          }

#line 77 "jollyjack/jollyjack_cython.pyx"
        }

#line 77 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 77 "jollyjack/jollyjack_cython.pyx"
        __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_c, __pyx_t_7);

#line 77 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = 0;

#line 77 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_18 = __pyx_8genexpr5__pyx_v_c;

#line 77 "jollyjack/jollyjack_cython.pyx"
        __Pyx_INCREF(__pyx_t_18);

#line 77 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_8 = 0;

#line 77 "jollyjack/jollyjack_cython.pyx"
        {

#line 77 "jollyjack/jollyjack_cython.pyx"
          PyObject *__pyx_callargs[2] = {__pyx_t_18, __pyx_mstate_global->__pyx_n_u_utf8};

#line 77 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 77 "jollyjack/jollyjack_cython.pyx"
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 77 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 77, __pyx_L50_error)

#line 77 "jollyjack/jollyjack_cython.pyx"
          __Pyx_GOTREF(__pyx_t_7);

#line 77 "jollyjack/jollyjack_cython.pyx"
        }

#line 77 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_9, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 77, __pyx_L50_error)

#line 77 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 77 "jollyjack/jollyjack_cython.pyx"
      }

#line 77 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 77 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_c); __pyx_8genexpr5__pyx_v_c = 0;

#line 77 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L54_exit_scope;

#line 77 "jollyjack/jollyjack_cython.pyx"
      __pyx_L50_error:;

#line 77 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_c); __pyx_8genexpr5__pyx_v_c = 0;

#line 77 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L1_error;

#line 77 "jollyjack/jollyjack_cython.pyx"
      __pyx_L54_exit_scope:;

#line 77 "jollyjack/jollyjack_cython.pyx"
    } /* exit inner scope */

#line 77 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_17 = __pyx_convert_vector_from_py_std_3a__3a_string(__pyx_t_9); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 77, __pyx_L1_error)

#line 77 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 77 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ccolumn_names = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_17);

    /* "jollyjack/jollyjack_cython.pyx":75
 *         ccolumn_names = [item[0].encode('utf8') for item in column_names]
 *         ctarget_column_indices = [item[1] for item in column_names]
 *     elif column_names:             # <<<<<<<<<<<<<<
 *         assert len(column_names) == np_array.shape[1], f"Requested to read {len(column_names)} columns, but the number of columns in numpy array is {np_array.shape[1]}"
 *         ccolumn_names = [c.encode('utf8') for c in column_names]
*/

#line 75 "jollyjack/jollyjack_cython.pyx"
  }

#line 75 "jollyjack/jollyjack_cython.pyx"
  __pyx_L23:;

  /* "jollyjack/jollyjack_cython.pyx":79
 *         ccolumn_names = [c.encode('utf8') for c in column_names]
 * 
 *     if row_ranges:             # <<<<<<<<<<<<<<
 *         for sl in row_ranges:
 *             if sl.start is None or sl.stop is None or sl.start < 0 or sl.stop < 0 or sl.stop <= sl.start:
*/

#line 79 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_row_ranges); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 79, __pyx_L1_error)

#line 79 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":80
 * 
 *     if row_ranges:
 *         for sl in row_ranges:             # <<<<<<<<<<<<<<
 *             if sl.start is None or sl.stop is None or sl.start < 0 or sl.stop < 0 or sl.stop <= sl.start:
 *                 raise ValueError(f"Row range '{sl}' is not a valid range")
*/

#line 80 "jollyjack/jollyjack_cython.pyx"
    if (likely(PyList_CheckExact(__pyx_v_row_ranges)) || PyTuple_CheckExact(__pyx_v_row_ranges)) {

#line 80 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9 = __pyx_v_row_ranges; __Pyx_INCREF(__pyx_t_9);

#line 80 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_13 = 0;

#line 80 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_11 = NULL;

#line 80 "jollyjack/jollyjack_cython.pyx"
    } else {
      __pyx_t_13 = -1; 
#line 80 "jollyjack/jollyjack_cython.pyx"
__pyx_t_9 = PyObject_GetIter(__pyx_v_row_ranges); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9); 
#line 80 "jollyjack/jollyjack_cython.pyx"
if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "jollyjack/jollyjack_cython.pyx"
    }
    for (;;
#line 80 "jollyjack/jollyjack_cython.pyx"
) {

#line 80 "jollyjack/jollyjack_cython.pyx"
      if (likely(!__pyx_t_11)) {

#line 80 "jollyjack/jollyjack_cython.pyx"
        if (likely(PyList_CheckExact(__pyx_t_9))) {

#line 80 "jollyjack/jollyjack_cython.pyx"
          {

#line 80 "jollyjack/jollyjack_cython.pyx"
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);

#line 80 "jollyjack/jollyjack_cython.pyx"
            #if !CYTHON_ASSUME_SAFE_SIZE

#line 80 "jollyjack/jollyjack_cython.pyx"
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "jollyjack/jollyjack_cython.pyx"
            #endif

#line 80 "jollyjack/jollyjack_cython.pyx"
            if (__pyx_t_13 >= __pyx_temp) break;

#line 80 "jollyjack/jollyjack_cython.pyx"
          }

#line 80 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_6 = __Pyx_PyList_GetItemRefFast(__pyx_t_9, __pyx_t_13, __Pyx_ReferenceSharing_OwnStrongReference);

#line 80 "jollyjack/jollyjack_cython.pyx"
          ++__pyx_t_13;

#line 80 "jollyjack/jollyjack_cython.pyx"
        } else {

#line 80 "jollyjack/jollyjack_cython.pyx"
          {

#line 80 "jollyjack/jollyjack_cython.pyx"
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_9);

#line 80 "jollyjack/jollyjack_cython.pyx"
            #if !CYTHON_ASSUME_SAFE_SIZE

#line 80 "jollyjack/jollyjack_cython.pyx"
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "jollyjack/jollyjack_cython.pyx"
            #endif

#line 80 "jollyjack/jollyjack_cython.pyx"
            if (__pyx_t_13 >= __pyx_temp) break;

#line 80 "jollyjack/jollyjack_cython.pyx"
          }

#line 80 "jollyjack/jollyjack_cython.pyx"
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 80 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_6 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_13));

#line 80 "jollyjack/jollyjack_cython.pyx"
          #else

#line 80 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_13);

#line 80 "jollyjack/jollyjack_cython.pyx"
          #endif

#line 80 "jollyjack/jollyjack_cython.pyx"
          ++__pyx_t_13;

#line 80 "jollyjack/jollyjack_cython.pyx"
        }

#line 80 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 80, __pyx_L1_error)
      } else 
#line 80 "jollyjack/jollyjack_cython.pyx"
{

#line 80 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = __pyx_t_11(__pyx_t_9);

#line 80 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!__pyx_t_6)) {

#line 80 "jollyjack/jollyjack_cython.pyx"
          PyObject* exc_type = PyErr_Occurred();

#line 80 "jollyjack/jollyjack_cython.pyx"
          if (exc_type) {

#line 80 "jollyjack/jollyjack_cython.pyx"
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "jollyjack/jollyjack_cython.pyx"
            PyErr_Clear();

#line 80 "jollyjack/jollyjack_cython.pyx"
          }

#line 80 "jollyjack/jollyjack_cython.pyx"
          break;

#line 80 "jollyjack/jollyjack_cython.pyx"
        }

#line 80 "jollyjack/jollyjack_cython.pyx"
      }

#line 80 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 80 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF_SET(__pyx_v_sl, __pyx_t_6);

#line 80 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = 0;

      /* "jollyjack/jollyjack_cython.pyx":81
 *     if row_ranges:
 *         for sl in row_ranges:
 *             if sl.start is None or sl.stop is None or sl.start < 0 or sl.stop < 0 or sl.stop <= sl.start:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Row range '{sl}' is not a valid range")
 * 
*/

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5 = (__pyx_t_6 == Py_None);

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 81 "jollyjack/jollyjack_cython.pyx"
      if (!__pyx_t_5) {

#line 81 "jollyjack/jollyjack_cython.pyx"
      } else {

#line 81 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_3 = __pyx_t_5;

#line 81 "jollyjack/jollyjack_cython.pyx"
        goto __pyx_L59_bool_binop_done;

#line 81 "jollyjack/jollyjack_cython.pyx"
      }

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_stop); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5 = (__pyx_t_6 == Py_None);

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 81 "jollyjack/jollyjack_cython.pyx"
      if (!__pyx_t_5) {

#line 81 "jollyjack/jollyjack_cython.pyx"
      } else {

#line 81 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_3 = __pyx_t_5;

#line 81 "jollyjack/jollyjack_cython.pyx"
        goto __pyx_L59_bool_binop_done;

#line 81 "jollyjack/jollyjack_cython.pyx"
      }

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_6, __pyx_mstate_global->__pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 81 "jollyjack/jollyjack_cython.pyx"
      if (!__pyx_t_5) {

#line 81 "jollyjack/jollyjack_cython.pyx"
      } else {

#line 81 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_3 = __pyx_t_5;

#line 81 "jollyjack/jollyjack_cython.pyx"
        goto __pyx_L59_bool_binop_done;

#line 81 "jollyjack/jollyjack_cython.pyx"
      }

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_stop); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = PyObject_RichCompare(__pyx_t_7, __pyx_mstate_global->__pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 81 "jollyjack/jollyjack_cython.pyx"
      if (!__pyx_t_5) {

#line 81 "jollyjack/jollyjack_cython.pyx"
      } else {

#line 81 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_3 = __pyx_t_5;

#line 81 "jollyjack/jollyjack_cython.pyx"
        goto __pyx_L59_bool_binop_done;

#line 81 "jollyjack/jollyjack_cython.pyx"
      }

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_stop); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_start); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_18 = PyObject_RichCompare(__pyx_t_6, __pyx_t_7, Py_LE); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 81, __pyx_L1_error)

#line 81 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3 = __pyx_t_5;

#line 81 "jollyjack/jollyjack_cython.pyx"
      __pyx_L59_bool_binop_done:;

#line 81 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(__pyx_t_3)) {

        /* "jollyjack/jollyjack_cython.pyx":82
 *         for sl in row_ranges:
 *             if sl.start is None or sl.stop is None or sl.start < 0 or sl.stop < 0 or sl.stop <= sl.start:
 *                 raise ValueError(f"Row range '{sl}' is not a valid range")             # <<<<<<<<<<<<<<
 * 
 *             if sl.step is not None and sl.step != 1:
*/

#line 82 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = NULL;

#line 82 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_sl, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 82, __pyx_L1_error)

#line 82 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 82 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_19[0] = __pyx_mstate_global->__pyx_kp_u_Row_range;

#line 82 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_19[1] = __pyx_t_6;

#line 82 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_19[2] = __pyx_mstate_global->__pyx_kp_u_is_not_a_valid_range;

#line 82 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_16 = __Pyx_PyUnicode_Join(__pyx_t_19, 3, 11 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 22, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6));

#line 82 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 82, __pyx_L1_error)

#line 82 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_16);

#line 82 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 82 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_8 = 1;

#line 82 "jollyjack/jollyjack_cython.pyx"
        {

#line 82 "jollyjack/jollyjack_cython.pyx"
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_16};

#line 82 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_18 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 82 "jollyjack/jollyjack_cython.pyx"
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 82 "jollyjack/jollyjack_cython.pyx"
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

#line 82 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 82, __pyx_L1_error)

#line 82 "jollyjack/jollyjack_cython.pyx"
          __Pyx_GOTREF(__pyx_t_18);

#line 82 "jollyjack/jollyjack_cython.pyx"
        }

#line 82 "jollyjack/jollyjack_cython.pyx"
        __Pyx_Raise(__pyx_t_18, 0, 0, 0);

#line 82 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 82 "jollyjack/jollyjack_cython.pyx"
        __PYX_ERR(0, 82, __pyx_L1_error)

        /* "jollyjack/jollyjack_cython.pyx":81
 *     if row_ranges:
 *         for sl in row_ranges:
 *             if sl.start is None or sl.stop is None or sl.start < 0 or sl.stop < 0 or sl.stop <= sl.start:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Row range '{sl}' is not a valid range")
 * 
*/

#line 81 "jollyjack/jollyjack_cython.pyx"
      }

      /* "jollyjack/jollyjack_cython.pyx":84
 *                 raise ValueError(f"Row range '{sl}' is not a valid range")
 * 
 *             if sl.step is not None and sl.step != 1:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Row range '{sl}' is not contiguous")
 * 
*/

#line 84 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_step); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 84, __pyx_L1_error)

#line 84 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_18);

#line 84 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5 = (__pyx_t_18 != Py_None);

#line 84 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 84 "jollyjack/jollyjack_cython.pyx"
      if (__pyx_t_5) {

#line 84 "jollyjack/jollyjack_cython.pyx"
      } else {

#line 84 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_3 = __pyx_t_5;

#line 84 "jollyjack/jollyjack_cython.pyx"
        goto __pyx_L65_bool_binop_done;

#line 84 "jollyjack/jollyjack_cython.pyx"
      }

#line 84 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_step); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 84, __pyx_L1_error)

#line 84 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_18);

#line 84 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5 = (__Pyx_PyLong_BoolNeObjC(__pyx_t_18, __pyx_mstate_global->__pyx_int_1, 1, 0)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 84, __pyx_L1_error)

#line 84 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 84 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3 = __pyx_t_5;

#line 84 "jollyjack/jollyjack_cython.pyx"
      __pyx_L65_bool_binop_done:;

#line 84 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(__pyx_t_3)) {

        /* "jollyjack/jollyjack_cython.pyx":85
 * 
 *             if sl.step is not None and sl.step != 1:
 *                 raise ValueError(f"Row range '{sl}' is not contiguous")             # <<<<<<<<<<<<<<
 * 
 *         ctarget_row_ranges = [x for sl in row_ranges for x in (sl.start, sl.stop)]
*/

#line 85 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_16 = NULL;

#line 85 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_v_sl, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 85, __pyx_L1_error)

#line 85 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 85 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_19[0] = __pyx_mstate_global->__pyx_kp_u_Row_range;

#line 85 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_19[1] = __pyx_t_7;

#line 85 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_19[2] = __pyx_mstate_global->__pyx_kp_u_is_not_contiguous;

#line 85 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_19, 3, 11 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + 19, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7));

#line 85 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 85, __pyx_L1_error)

#line 85 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 85 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 85 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_8 = 1;

#line 85 "jollyjack/jollyjack_cython.pyx"
        {

#line 85 "jollyjack/jollyjack_cython.pyx"
          PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_t_6};

#line 85 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_18 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 85 "jollyjack/jollyjack_cython.pyx"
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;

#line 85 "jollyjack/jollyjack_cython.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 85 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 85, __pyx_L1_error)

#line 85 "jollyjack/jollyjack_cython.pyx"
          __Pyx_GOTREF(__pyx_t_18);

#line 85 "jollyjack/jollyjack_cython.pyx"
        }

#line 85 "jollyjack/jollyjack_cython.pyx"
        __Pyx_Raise(__pyx_t_18, 0, 0, 0);

#line 85 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 85 "jollyjack/jollyjack_cython.pyx"
        __PYX_ERR(0, 85, __pyx_L1_error)

        /* "jollyjack/jollyjack_cython.pyx":84
 *                 raise ValueError(f"Row range '{sl}' is not a valid range")
 * 
 *             if sl.step is not None and sl.step != 1:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Row range '{sl}' is not contiguous")
 * 
*/

#line 84 "jollyjack/jollyjack_cython.pyx"
      }

      /* "jollyjack/jollyjack_cython.pyx":80
 * 
 *     if row_ranges:
 *         for sl in row_ranges:             # <<<<<<<<<<<<<<
 *             if sl.start is None or sl.stop is None or sl.start < 0 or sl.stop < 0 or sl.stop <= sl.start:
 *                 raise ValueError(f"Row range '{sl}' is not a valid range")
*/

#line 80 "jollyjack/jollyjack_cython.pyx"
    }

#line 80 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "jollyjack/jollyjack_cython.pyx":87
 *                 raise ValueError(f"Row range '{sl}' is not contiguous")
 * 
 *         ctarget_row_ranges = [x for sl in row_ranges for x in (sl.start, sl.stop)]             # <<<<<<<<<<<<<<
 * 
 *     # Ensure that only one input is set
*/

#line 87 "jollyjack/jollyjack_cython.pyx"
    { /* enter inner scope */

#line 87 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 87 "jollyjack/jollyjack_cython.pyx"
      if (likely(PyList_CheckExact(__pyx_v_row_ranges)) || PyTuple_CheckExact(__pyx_v_row_ranges)) {

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_18 = __pyx_v_row_ranges; __Pyx_INCREF(__pyx_t_18);

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_13 = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_11 = NULL;

#line 87 "jollyjack/jollyjack_cython.pyx"
      } else {
        __pyx_t_13 = -1; 
#line 87 "jollyjack/jollyjack_cython.pyx"
__pyx_t_18 = PyObject_GetIter(__pyx_v_row_ranges); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_18); 
#line 87 "jollyjack/jollyjack_cython.pyx"
if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
      }
      for (;;
#line 87 "jollyjack/jollyjack_cython.pyx"
) {

#line 87 "jollyjack/jollyjack_cython.pyx"
        if (likely(!__pyx_t_11)) {

#line 87 "jollyjack/jollyjack_cython.pyx"
          if (likely(PyList_CheckExact(__pyx_t_18))) {

#line 87 "jollyjack/jollyjack_cython.pyx"
            {

#line 87 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_18);

#line 87 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 87 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 87 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_13 >= __pyx_temp) break;

#line 87 "jollyjack/jollyjack_cython.pyx"
            }

#line 87 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_6 = __Pyx_PyList_GetItemRefFast(__pyx_t_18, __pyx_t_13, __Pyx_ReferenceSharing_OwnStrongReference);

#line 87 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_13;

#line 87 "jollyjack/jollyjack_cython.pyx"
          } else {

#line 87 "jollyjack/jollyjack_cython.pyx"
            {

#line 87 "jollyjack/jollyjack_cython.pyx"
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_18);

#line 87 "jollyjack/jollyjack_cython.pyx"
              #if !CYTHON_ASSUME_SAFE_SIZE

#line 87 "jollyjack/jollyjack_cython.pyx"
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
              #endif

#line 87 "jollyjack/jollyjack_cython.pyx"
              if (__pyx_t_13 >= __pyx_temp) break;

#line 87 "jollyjack/jollyjack_cython.pyx"
            }

#line 87 "jollyjack/jollyjack_cython.pyx"
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 87 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_6 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_18, __pyx_t_13));

#line 87 "jollyjack/jollyjack_cython.pyx"
            #else

#line 87 "jollyjack/jollyjack_cython.pyx"
            __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_18, __pyx_t_13);

#line 87 "jollyjack/jollyjack_cython.pyx"
            #endif

#line 87 "jollyjack/jollyjack_cython.pyx"
            ++__pyx_t_13;

#line 87 "jollyjack/jollyjack_cython.pyx"
          }

#line 87 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 87, __pyx_L70_error)
        } else 
#line 87 "jollyjack/jollyjack_cython.pyx"
{

#line 87 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_6 = __pyx_t_11(__pyx_t_18);

#line 87 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_6)) {

#line 87 "jollyjack/jollyjack_cython.pyx"
            PyObject* exc_type = PyErr_Occurred();

#line 87 "jollyjack/jollyjack_cython.pyx"
            if (exc_type) {

#line 87 "jollyjack/jollyjack_cython.pyx"
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
              PyErr_Clear();

#line 87 "jollyjack/jollyjack_cython.pyx"
            }

#line 87 "jollyjack/jollyjack_cython.pyx"
            break;

#line 87 "jollyjack/jollyjack_cython.pyx"
          }

#line 87 "jollyjack/jollyjack_cython.pyx"
        }

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_sl, __pyx_t_6);

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr6__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr6__pyx_v_sl, __pyx_mstate_global->__pyx_n_u_stop); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_16);

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GOTREF(__pyx_t_7);

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GIVEREF(__pyx_t_6);

#line 87 "jollyjack/jollyjack_cython.pyx"
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6) != (0)) __PYX_ERR(0, 87, __pyx_L70_error);

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_GIVEREF(__pyx_t_16);

#line 87 "jollyjack/jollyjack_cython.pyx"
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_16) != (0)) __PYX_ERR(0, 87, __pyx_L70_error);

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_6 = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_16 = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_16 = __pyx_t_7; __Pyx_INCREF(__pyx_t_16);

#line 87 "jollyjack/jollyjack_cython.pyx"
        __pyx_t_10 = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        for (;;
#line 87 "jollyjack/jollyjack_cython.pyx"
) {

#line 87 "jollyjack/jollyjack_cython.pyx"
          if (__pyx_t_10 >= 2) break;

#line 87 "jollyjack/jollyjack_cython.pyx"
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 87 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_7 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_16, __pyx_t_10));

#line 87 "jollyjack/jollyjack_cython.pyx"
          #else

#line 87 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_16, __pyx_t_10);

#line 87 "jollyjack/jollyjack_cython.pyx"
          #endif

#line 87 "jollyjack/jollyjack_cython.pyx"
          ++__pyx_t_10;

#line 87 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
          __Pyx_GOTREF(__pyx_t_7);

#line 87 "jollyjack/jollyjack_cython.pyx"
          __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_x, __pyx_t_7);

#line 87 "jollyjack/jollyjack_cython.pyx"
          __pyx_t_7 = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_9, (PyObject*)__pyx_8genexpr6__pyx_v_x))) __PYX_ERR(0, 87, __pyx_L70_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
        }

#line 87 "jollyjack/jollyjack_cython.pyx"
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
      }

#line 87 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_sl); __pyx_8genexpr6__pyx_v_sl = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_x); __pyx_8genexpr6__pyx_v_x = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L77_exit_scope;

#line 87 "jollyjack/jollyjack_cython.pyx"
      __pyx_L70_error:;

#line 87 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_sl); __pyx_8genexpr6__pyx_v_sl = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_x); __pyx_8genexpr6__pyx_v_x = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L1_error;

#line 87 "jollyjack/jollyjack_cython.pyx"
      __pyx_L77_exit_scope:;

#line 87 "jollyjack/jollyjack_cython.pyx"
    } /* exit inner scope */

#line 87 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_20 = __pyx_convert_vector_from_py_int64_t(__pyx_t_9); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 87, __pyx_L1_error)

#line 87 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 87 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_ctarget_row_ranges = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_20);

    /* "jollyjack/jollyjack_cython.pyx":79
 *         ccolumn_names = [c.encode('utf8') for c in column_names]
 * 
 *     if row_ranges:             # <<<<<<<<<<<<<<
 *         for sl in row_ranges:
 *             if sl.start is None or sl.stop is None or sl.start < 0 or sl.stop < 0 or sl.stop <= sl.start:
*/

#line 79 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":90
 * 
 *     # Ensure that only one input is set
 *     assert (column_indices or column_names) and (not column_indices or not column_names), f"Either column_indices or column_names needs to be set"             # <<<<<<<<<<<<<<
 * 
 *     # Ensure the input is a 2D array, Fortran order
*/

#line 90 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 90 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 90 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_column_indices); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 90, __pyx_L1_error)

#line 90 "jollyjack/jollyjack_cython.pyx"
    if (!__pyx_t_5) {

#line 90 "jollyjack/jollyjack_cython.pyx"
    } else {

#line 90 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L79_next_and;

#line 90 "jollyjack/jollyjack_cython.pyx"
    }

#line 90 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_column_names); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 90, __pyx_L1_error)

#line 90 "jollyjack/jollyjack_cython.pyx"
    if (__pyx_t_5) {

#line 90 "jollyjack/jollyjack_cython.pyx"
    } else {

#line 90 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3 = __pyx_t_5;

#line 90 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L78_bool_binop_done;

#line 90 "jollyjack/jollyjack_cython.pyx"
    }

#line 90 "jollyjack/jollyjack_cython.pyx"
    __pyx_L79_next_and:;

#line 90 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_column_indices); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 90, __pyx_L1_error)

#line 90 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_21 = (!__pyx_t_5);

#line 90 "jollyjack/jollyjack_cython.pyx"
    if (!__pyx_t_21) {

#line 90 "jollyjack/jollyjack_cython.pyx"
    } else {

#line 90 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3 = __pyx_t_21;

#line 90 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L78_bool_binop_done;

#line 90 "jollyjack/jollyjack_cython.pyx"
    }

#line 90 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_21 = __Pyx_PyObject_IsTrue(__pyx_v_column_names); if (unlikely((__pyx_t_21 < 0))) __PYX_ERR(0, 90, __pyx_L1_error)

#line 90 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_5 = (!__pyx_t_21);

#line 90 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_3 = __pyx_t_5;

#line 90 "jollyjack/jollyjack_cython.pyx"
    __pyx_L78_bool_binop_done:;

#line 90 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_3)) {

#line 90 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_mstate_global->__pyx_kp_u_Either_column_indices_or_column, 0, 0);

#line 90 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 90, __pyx_L1_error)

#line 90 "jollyjack/jollyjack_cython.pyx"
    }

#line 90 "jollyjack/jollyjack_cython.pyx"
  }

#line 90 "jollyjack/jollyjack_cython.pyx"
  #else

#line 90 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 90, __pyx_L1_error)

#line 90 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":93
 * 
 *     # Ensure the input is a 2D array, Fortran order
 *     assert np_array.ndim == 2, f"Unexpected np_array.ndim, {np_array.ndim} != 2"             # <<<<<<<<<<<<<<
 *     assert np_array.strides[0] <= np_array.strides[1], f"Expected array in a Fortran (column-major) order"
 * 
*/

#line 93 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 93 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 93 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_3 = (__pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_np_array) == 2);

#line 93 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_3)) {

#line 93 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9 = __Pyx_PyUnicode_From_int(__pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_np_array), 0, ' ', 'd'); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 93, __pyx_L1_error)

#line 93 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_9);

#line 93 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_19[0] = __pyx_mstate_global->__pyx_kp_u_Unexpected_np_array_ndim;

#line 93 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_19[1] = __pyx_t_9;

#line 93 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_19[2] = __pyx_mstate_global->__pyx_kp_u_2;

#line 93 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_18 = __Pyx_PyUnicode_Join(__pyx_t_19, 3, 26 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9) + 5, 127);

#line 93 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 93, __pyx_L1_error)

#line 93 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_18);

#line 93 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 93 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_18, 0, 0);

#line 93 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 93 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 93, __pyx_L1_error)

#line 93 "jollyjack/jollyjack_cython.pyx"
    }

#line 93 "jollyjack/jollyjack_cython.pyx"
  }

#line 93 "jollyjack/jollyjack_cython.pyx"
  #else

#line 93 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 93, __pyx_L1_error)

#line 93 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":94
 *     # Ensure the input is a 2D array, Fortran order
 *     assert np_array.ndim == 2, f"Unexpected np_array.ndim, {np_array.ndim} != 2"
 *     assert np_array.strides[0] <= np_array.strides[1], f"Expected array in a Fortran (column-major) order"             # <<<<<<<<<<<<<<
 * 
 *     cdef int64_t cexpected_rows = np_array.shape[0]
*/

#line 94 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 94 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 94 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_3 = ((__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_np_array)[0]) <= (__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_np_array)[1]));

#line 94 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_3)) {

#line 94 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_mstate_global->__pyx_kp_u_Expected_array_in_a_Fortran_colu, 0, 0);

#line 94 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 94, __pyx_L1_error)

#line 94 "jollyjack/jollyjack_cython.pyx"
    }

#line 94 "jollyjack/jollyjack_cython.pyx"
  }

#line 94 "jollyjack/jollyjack_cython.pyx"
  #else

#line 94 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 94, __pyx_L1_error)

#line 94 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":96
 *     assert np_array.strides[0] <= np_array.strides[1], f"Expected array in a Fortran (column-major) order"
 * 
 *     cdef int64_t cexpected_rows = np_array.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     cdef CCacheOptions c_cache_options
*/

#line 96 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_cexpected_rows = (__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_np_array)[0]);

  /* "jollyjack/jollyjack_cython.pyx":99
 * 
 *     cdef CCacheOptions c_cache_options
 *     if cache_options is not None:             # <<<<<<<<<<<<<<
 *         c_cache_options = cache_options.unwrap()
 *     else:
*/

#line 99 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = (((PyObject *)__pyx_v_cache_options) != Py_None);

#line 99 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":100
 *     cdef CCacheOptions c_cache_options
 *     if cache_options is not None:
 *         c_cache_options = cache_options.unwrap()             # <<<<<<<<<<<<<<
 *     else:
 *         c_cache_options = CCacheOptions.LazyDefaults()
*/

#line 100 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_22 = ((struct __pyx_vtabstruct_7pyarrow_3lib_CacheOptions *)__pyx_v_cache_options->__pyx_vtab)->unwrap(__pyx_v_cache_options); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 100, __pyx_L1_error)

#line 100 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_c_cache_options = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_22);

    /* "jollyjack/jollyjack_cython.pyx":99
 * 
 *     cdef CCacheOptions c_cache_options
 *     if cache_options is not None:             # <<<<<<<<<<<<<<
 *         c_cache_options = cache_options.unwrap()
 *     else:
*/

#line 99 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L82;

#line 99 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":102
 *         c_cache_options = cache_options.unwrap()
 *     else:
 *         c_cache_options = CCacheOptions.LazyDefaults()             # <<<<<<<<<<<<<<
 * 
 *     cdef shared_ptr[CRandomAccessFile] rd_handle
*/

#line 102 "jollyjack/jollyjack_cython.pyx"
  /*else*/ {

#line 102 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_c_cache_options = arrow::io::CacheOptions::LazyDefaults();

#line 102 "jollyjack/jollyjack_cython.pyx"
  }

#line 102 "jollyjack/jollyjack_cython.pyx"
  __pyx_L82:;

  /* "jollyjack/jollyjack_cython.pyx":108
 *     cdef bool cuse_o_direct
 *     # Please note that the `JJ_READER_BACKEND` variable is experimental and may be changed or removed in future versions
 *     jj_reader_backend = os.environ.get("JJ_READER_BACKEND")             # <<<<<<<<<<<<<<
 *     if jj_reader_backend is None:
 *         get_reader(source, use_memory_map, &rd_handle)
*/

#line 108 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 108, __pyx_L1_error)

#line 108 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_16);

#line 108 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_environ); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 108, __pyx_L1_error)

#line 108 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 108 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

#line 108 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_9 = __pyx_t_7;

#line 108 "jollyjack/jollyjack_cython.pyx"
  __Pyx_INCREF(__pyx_t_9);

#line 108 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_8 = 0;

#line 108 "jollyjack/jollyjack_cython.pyx"
  {

#line 108 "jollyjack/jollyjack_cython.pyx"
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_n_u_JJ_READER_BACKEND};

#line 108 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_18 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 108 "jollyjack/jollyjack_cython.pyx"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 108 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 108 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 108, __pyx_L1_error)

#line 108 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_18);

#line 108 "jollyjack/jollyjack_cython.pyx"
  }

#line 108 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_jj_reader_backend = __pyx_t_18;

#line 108 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_18 = 0;

  /* "jollyjack/jollyjack_cython.pyx":109
 *     # Please note that the `JJ_READER_BACKEND` variable is experimental and may be changed or removed in future versions
 *     jj_reader_backend = os.environ.get("JJ_READER_BACKEND")
 *     if jj_reader_backend is None:             # <<<<<<<<<<<<<<
 *         get_reader(source, use_memory_map, &rd_handle)
 *     elif (jj_reader_backend == 'io_uring' or jj_reader_backend == 'io_uring_odirect'):
*/

#line 109 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = (__pyx_v_jj_reader_backend == Py_None);

#line 109 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_3) {

    /* "jollyjack/jollyjack_cython.pyx":110
 *     jj_reader_backend = os.environ.get("JJ_READER_BACKEND")
 *     if jj_reader_backend is None:
 *         get_reader(source, use_memory_map, &rd_handle)             # <<<<<<<<<<<<<<
 *     elif (jj_reader_backend == 'io_uring' or jj_reader_backend == 'io_uring_odirect'):
 *         pathstr = source.encode("utf-8")
*/

#line 110 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_use_memory_map); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)

#line 110 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_18 = __pyx_f_7pyarrow_3lib_get_reader(__pyx_v_source, __pyx_t_2, (&__pyx_v_rd_handle)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 110, __pyx_L1_error)

#line 110 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_18);

#line 110 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

    /* "jollyjack/jollyjack_cython.pyx":109
 *     # Please note that the `JJ_READER_BACKEND` variable is experimental and may be changed or removed in future versions
 *     jj_reader_backend = os.environ.get("JJ_READER_BACKEND")
 *     if jj_reader_backend is None:             # <<<<<<<<<<<<<<
 *         get_reader(source, use_memory_map, &rd_handle)
 *     elif (jj_reader_backend == 'io_uring' or jj_reader_backend == 'io_uring_odirect'):
*/

#line 109 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L83;

#line 109 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":111
 *     if jj_reader_backend is None:
 *         get_reader(source, use_memory_map, &rd_handle)
 *     elif (jj_reader_backend == 'io_uring' or jj_reader_backend == 'io_uring_odirect'):             # <<<<<<<<<<<<<<
 *         pathstr = source.encode("utf-8")
 *         cuse_o_direct = jj_reader_backend == 'io_uring_odirect'
*/

#line 111 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_jj_reader_backend, __pyx_mstate_global->__pyx_n_u_io_uring, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 111, __pyx_L1_error)

#line 111 "jollyjack/jollyjack_cython.pyx"
  if (!__pyx_t_5) {

#line 111 "jollyjack/jollyjack_cython.pyx"
  } else {

#line 111 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_3 = __pyx_t_5;

#line 111 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L84_bool_binop_done;

#line 111 "jollyjack/jollyjack_cython.pyx"
  }

#line 111 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_jj_reader_backend, __pyx_mstate_global->__pyx_n_u_io_uring_odirect, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 111, __pyx_L1_error)

#line 111 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __pyx_t_5;

#line 111 "jollyjack/jollyjack_cython.pyx"
  __pyx_L84_bool_binop_done:;

#line 111 "jollyjack/jollyjack_cython.pyx"
  if (likely(__pyx_t_3)) {

    /* "jollyjack/jollyjack_cython.pyx":112
 *         get_reader(source, use_memory_map, &rd_handle)
 *     elif (jj_reader_backend == 'io_uring' or jj_reader_backend == 'io_uring_odirect'):
 *         pathstr = source.encode("utf-8")             # <<<<<<<<<<<<<<
 *         cuse_o_direct = jj_reader_backend == 'io_uring_odirect'
 *         with nogil:
*/

#line 112 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_7 = __pyx_v_source;

#line 112 "jollyjack/jollyjack_cython.pyx"
    __Pyx_INCREF(__pyx_t_7);

#line 112 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_8 = 0;

#line 112 "jollyjack/jollyjack_cython.pyx"
    {

#line 112 "jollyjack/jollyjack_cython.pyx"
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_utf_8};

#line 112 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_18 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 112 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 112 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 112, __pyx_L1_error)

#line 112 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_18);

#line 112 "jollyjack/jollyjack_cython.pyx"
    }

#line 112 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_23 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_18); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 112, __pyx_L1_error)

#line 112 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 112 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_pathstr = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_23);

    /* "jollyjack/jollyjack_cython.pyx":113
 *     elif (jj_reader_backend == 'io_uring' or jj_reader_backend == 'io_uring_odirect'):
 *         pathstr = source.encode("utf-8")
 *         cuse_o_direct = jj_reader_backend == 'io_uring_odirect'             # <<<<<<<<<<<<<<
 *         with nogil:
 *             cjollyjack.ReadIntoMemoryIOUring (pathstr
*/

#line 113 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_18 = PyObject_RichCompare(__pyx_v_jj_reader_backend, __pyx_mstate_global->__pyx_n_u_io_uring_odirect, Py_EQ); __Pyx_XGOTREF(__pyx_t_18); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_18); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 113 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_cuse_o_direct = __pyx_t_2;

    /* "jollyjack/jollyjack_cython.pyx":114
 *         pathstr = source.encode("utf-8")
 *         cuse_o_direct = jj_reader_backend == 'io_uring_odirect'
 *         with nogil:             # <<<<<<<<<<<<<<
 *             cjollyjack.ReadIntoMemoryIOUring (pathstr
 *                 , c_metadata
*/

#line 114 "jollyjack/jollyjack_cython.pyx"
    {

#line 114 "jollyjack/jollyjack_cython.pyx"
        PyThreadState * _save;

#line 114 "jollyjack/jollyjack_cython.pyx"
        _save = PyEval_SaveThread();

#line 114 "jollyjack/jollyjack_cython.pyx"
        __Pyx_FastGIL_Remember();

#line 114 "jollyjack/jollyjack_cython.pyx"
        /*try:*/ {

          /* "jollyjack/jollyjack_cython.pyx":115
 *         cuse_o_direct = jj_reader_backend == 'io_uring_odirect'
 *         with nogil:
 *             cjollyjack.ReadIntoMemoryIOUring (pathstr             # <<<<<<<<<<<<<<
 *                 , c_metadata
 *                 , np_array.data
*/

#line 115 "jollyjack/jollyjack_cython.pyx"
          try {

#line 115 "jollyjack/jollyjack_cython.pyx"
            ReadIntoMemoryIOUring(__pyx_v_pathstr, __pyx_v_c_metadata, __pyx_f_5numpy_7ndarray_4data_data(__pyx_v_np_array), __pyx_v_cbuffer_size, __pyx_v_cstride0_size, __pyx_v_cstride1_size, __pyx_v_ccolumn_indices, __pyx_v_crow_group_indices, __pyx_v_ctarget_row_ranges, __pyx_v_ccolumn_names, __pyx_v_ctarget_column_indices, __pyx_v_cpre_buffer, __pyx_v_cuse_threads, __pyx_v_cuse_o_direct, __pyx_v_cexpected_rows, __pyx_v_c_cache_options);

#line 115 "jollyjack/jollyjack_cython.pyx"
          } catch(...) {
            PyGILState_STATE 
#line 115 "jollyjack/jollyjack_cython.pyx"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 115 "jollyjack/jollyjack_cython.pyx"
            __Pyx_CppExn2PyErr();

#line 115 "jollyjack/jollyjack_cython.pyx"
            __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 115 "jollyjack/jollyjack_cython.pyx"
            __PYX_ERR(0, 115, __pyx_L87_error)

#line 115 "jollyjack/jollyjack_cython.pyx"
          }

          /* "jollyjack/jollyjack_cython.pyx":131
 *                 , cexpected_rows
 *                 , c_cache_options)
 *             return             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError(f"Unsupported JJ_READER_BACKEND={jj_reader_backend}")
*/

#line 131 "jollyjack/jollyjack_cython.pyx"
          goto __pyx_L86_return;

#line 131 "jollyjack/jollyjack_cython.pyx"
        }

        /* "jollyjack/jollyjack_cython.pyx":114
 *         pathstr = source.encode("utf-8")
 *         cuse_o_direct = jj_reader_backend == 'io_uring_odirect'
 *         with nogil:             # <<<<<<<<<<<<<<
 *             cjollyjack.ReadIntoMemoryIOUring (pathstr
 *                 , c_metadata
*/

#line 114 "jollyjack/jollyjack_cython.pyx"
        /*finally:*/ {

#line 114 "jollyjack/jollyjack_cython.pyx"
          __pyx_L86_return: {

#line 114 "jollyjack/jollyjack_cython.pyx"
            __Pyx_FastGIL_Forget();

#line 114 "jollyjack/jollyjack_cython.pyx"
            PyEval_RestoreThread(_save);

#line 114 "jollyjack/jollyjack_cython.pyx"
            goto __pyx_L0;

#line 114 "jollyjack/jollyjack_cython.pyx"
          }

#line 114 "jollyjack/jollyjack_cython.pyx"
          __pyx_L87_error: {

#line 114 "jollyjack/jollyjack_cython.pyx"
            __Pyx_FastGIL_Forget();

#line 114 "jollyjack/jollyjack_cython.pyx"
            PyEval_RestoreThread(_save);

#line 114 "jollyjack/jollyjack_cython.pyx"
            goto __pyx_L1_error;

#line 114 "jollyjack/jollyjack_cython.pyx"
          }

#line 114 "jollyjack/jollyjack_cython.pyx"
        }

#line 114 "jollyjack/jollyjack_cython.pyx"
    }

    /* "jollyjack/jollyjack_cython.pyx":111
 *     if jj_reader_backend is None:
 *         get_reader(source, use_memory_map, &rd_handle)
 *     elif (jj_reader_backend == 'io_uring' or jj_reader_backend == 'io_uring_odirect'):             # <<<<<<<<<<<<<<
 *         pathstr = source.encode("utf-8")
 *         cuse_o_direct = jj_reader_backend == 'io_uring_odirect'
*/

#line 111 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":133
 *             return
 *     else:
 *         raise ValueError(f"Unsupported JJ_READER_BACKEND={jj_reader_backend}")             # <<<<<<<<<<<<<<
 * 
 *     with nogil:
*/

#line 133 "jollyjack/jollyjack_cython.pyx"
  /*else*/ {

#line 133 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_7 = NULL;

#line 133 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_v_jj_reader_backend, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_9);

#line 133 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_16 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Unsupported_JJ_READER_BACKEND, __pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_16);

#line 133 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 133 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_8 = 1;

#line 133 "jollyjack/jollyjack_cython.pyx"
    {

#line 133 "jollyjack/jollyjack_cython.pyx"
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_16};

#line 133 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_18 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 133 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 133 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

#line 133 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_18);

#line 133 "jollyjack/jollyjack_cython.pyx"
    }

#line 133 "jollyjack/jollyjack_cython.pyx"
    __Pyx_Raise(__pyx_t_18, 0, 0, 0);

#line 133 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

#line 133 "jollyjack/jollyjack_cython.pyx"
    __PYX_ERR(0, 133, __pyx_L1_error)

#line 133 "jollyjack/jollyjack_cython.pyx"
  }

#line 133 "jollyjack/jollyjack_cython.pyx"
  __pyx_L83:;

  /* "jollyjack/jollyjack_cython.pyx":135
 *         raise ValueError(f"Unsupported JJ_READER_BACKEND={jj_reader_backend}")
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         cjollyjack.ReadIntoMemory (rd_handle
 *             , c_metadata
*/

#line 135 "jollyjack/jollyjack_cython.pyx"
  {

#line 135 "jollyjack/jollyjack_cython.pyx"
      PyThreadState * _save;

#line 135 "jollyjack/jollyjack_cython.pyx"
      _save = PyEval_SaveThread();

#line 135 "jollyjack/jollyjack_cython.pyx"
      __Pyx_FastGIL_Remember();

#line 135 "jollyjack/jollyjack_cython.pyx"
      /*try:*/ {

        /* "jollyjack/jollyjack_cython.pyx":136
 * 
 *     with nogil:
 *         cjollyjack.ReadIntoMemory (rd_handle             # <<<<<<<<<<<<<<
 *             , c_metadata
 *             , np_array.data
*/

#line 136 "jollyjack/jollyjack_cython.pyx"
        try {

#line 136 "jollyjack/jollyjack_cython.pyx"
          ReadIntoMemory(__pyx_v_rd_handle, __pyx_v_c_metadata, __pyx_f_5numpy_7ndarray_4data_data(__pyx_v_np_array), __pyx_v_cbuffer_size, __pyx_v_cstride0_size, __pyx_v_cstride1_size, __pyx_v_ccolumn_indices, __pyx_v_crow_group_indices, __pyx_v_ctarget_row_ranges, __pyx_v_ccolumn_names, __pyx_v_ctarget_column_indices, __pyx_v_cpre_buffer, __pyx_v_cuse_threads, __pyx_v_cexpected_rows, __pyx_v_c_cache_options);

#line 136 "jollyjack/jollyjack_cython.pyx"
        } catch(...) {
          PyGILState_STATE 
#line 136 "jollyjack/jollyjack_cython.pyx"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 136 "jollyjack/jollyjack_cython.pyx"
          __Pyx_CppExn2PyErr();

#line 136 "jollyjack/jollyjack_cython.pyx"
          __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 136 "jollyjack/jollyjack_cython.pyx"
          __PYX_ERR(0, 136, __pyx_L90_error)

#line 136 "jollyjack/jollyjack_cython.pyx"
        }

        /* "jollyjack/jollyjack_cython.pyx":151
 *             , cexpected_rows
 *             , c_cache_options)
 *         return             # <<<<<<<<<<<<<<
 * 
 * cpdef void copy_to_torch_row_major (src_tensor, dst_tensor, row_indices):
*/

#line 151 "jollyjack/jollyjack_cython.pyx"
        goto __pyx_L89_return;

#line 151 "jollyjack/jollyjack_cython.pyx"
      }

      /* "jollyjack/jollyjack_cython.pyx":135
 *         raise ValueError(f"Unsupported JJ_READER_BACKEND={jj_reader_backend}")
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         cjollyjack.ReadIntoMemory (rd_handle
 *             , c_metadata
*/

#line 135 "jollyjack/jollyjack_cython.pyx"
      /*finally:*/ {

#line 135 "jollyjack/jollyjack_cython.pyx"
        __pyx_L89_return: {

#line 135 "jollyjack/jollyjack_cython.pyx"
          __Pyx_FastGIL_Forget();

#line 135 "jollyjack/jollyjack_cython.pyx"
          PyEval_RestoreThread(_save);

#line 135 "jollyjack/jollyjack_cython.pyx"
          goto __pyx_L0;

#line 135 "jollyjack/jollyjack_cython.pyx"
        }

#line 135 "jollyjack/jollyjack_cython.pyx"
        __pyx_L90_error: {

#line 135 "jollyjack/jollyjack_cython.pyx"
          __Pyx_FastGIL_Forget();

#line 135 "jollyjack/jollyjack_cython.pyx"
          PyEval_RestoreThread(_save);

#line 135 "jollyjack/jollyjack_cython.pyx"
          goto __pyx_L1_error;

#line 135 "jollyjack/jollyjack_cython.pyx"
        }

#line 135 "jollyjack/jollyjack_cython.pyx"
      }

#line 135 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":41
 *     return
 * 
 * cpdef void read_into_numpy (object source, FileMetaData metadata, cnp.ndarray np_array, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[int] crow_group_indices = row_group_indices
*/

#line 41 "jollyjack/jollyjack_cython.pyx"


#line 41 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_16);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_18);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.read_into_numpy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_v_sl);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_v_jj_reader_backend);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_item);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_item);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_c);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_item);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_item);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_c);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_sl);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_x);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 41 "jollyjack/jollyjack_cython.pyx"
}

#line 41 "jollyjack/jollyjack_cython.pyx"


#line 41 "jollyjack/jollyjack_cython.pyx"
/* Python wrapper */

#line 41 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_5read_into_numpy(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/

#line 41 "jollyjack/jollyjack_cython.pyx"
PyDoc_STRVAR(__pyx_doc_9jollyjack_16jollyjack_cython_4read_into_numpy, "read_into_numpy(source, FileMetaData metadata, ndarray np_array, row_group_indices, row_ranges=[], column_indices=[], column_names=[], pre_buffer=False, use_threads=True, use_memory_map=False, CacheOptions cache_options=None) -> void");
static PyMethodDef __pyx_mdef_9jollyjack_16jollyjack_cython_5read_into_numpy = 
#line 41 "jollyjack/jollyjack_cython.pyx"
{"read_into_numpy", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9jollyjack_16jollyjack_cython_5read_into_numpy, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9jollyjack_16jollyjack_cython_4read_into_numpy};

#line 41 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_5read_into_numpy(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_source = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_metadata = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyArrayObject *__pyx_v_np_array = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_row_group_indices = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_row_ranges = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_column_indices = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_column_names = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_pre_buffer = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_use_threads = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_use_memory_map = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;
  struct __pyx_obj_7pyarrow_3lib_CacheOptions *__pyx_v_cache_options = 0
#line 41 "jollyjack/jollyjack_cython.pyx"
;

#line 41 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 41 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;

#line 41 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 41 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("read_into_numpy (wrapper)", 0);

#line 41 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 41 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_ASSUME_SAFE_SIZE

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);

#line 41 "jollyjack/jollyjack_cython.pyx"
  #else

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 41 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);

#line 41 "jollyjack/jollyjack_cython.pyx"
  {

#line 41 "jollyjack/jollyjack_cython.pyx"
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_source,&__pyx_mstate_global->__pyx_n_u_metadata,&__pyx_mstate_global->__pyx_n_u_np_array,&__pyx_mstate_global->__pyx_n_u_row_group_indices,&__pyx_mstate_global->__pyx_n_u_row_ranges,&__pyx_mstate_global->__pyx_n_u_column_indices,&__pyx_mstate_global->__pyx_n_u_column_names,&__pyx_mstate_global->__pyx_n_u_pre_buffer,&__pyx_mstate_global->__pyx_n_u_use_threads,&__pyx_mstate_global->__pyx_n_u_use_memory_map,&__pyx_mstate_global->__pyx_n_u_cache_options,0};

#line 41 "jollyjack/jollyjack_cython.pyx"
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
    if (__pyx_kwds_len > 0) {

#line 41 "jollyjack/jollyjack_cython.pyx"
      switch (__pyx_nargs) {

#line 41 "jollyjack/jollyjack_cython.pyx"
        case 11:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case 10:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  9:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  8:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  7:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  6:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  5:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  4:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  3:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  2:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  1:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 41 "jollyjack/jollyjack_cython.pyx"
break;
        default: 
#line 41 "jollyjack/jollyjack_cython.pyx"
goto __pyx_L5_argtuple_error;

#line 41 "jollyjack/jollyjack_cython.pyx"
      }

#line 41 "jollyjack/jollyjack_cython.pyx"
      const Py_ssize_t kwd_pos_args = __pyx_nargs;

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "read_into_numpy", 0) < (0)) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[4]) values[4] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__4);

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[5]) values[5] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__5);

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[6]) values[6] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__6);

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)Py_False));

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)Py_True));

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[9]) values[9] = __Pyx_NewRef(((PyObject *)Py_False));

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[10]) values[10] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyarrow_3lib_CacheOptions *)Py_None));

#line 41 "jollyjack/jollyjack_cython.pyx"
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("read_into_numpy", 0, 4, 11, i); __PYX_ERR(0, 41, __pyx_L3_error) }

#line 41 "jollyjack/jollyjack_cython.pyx"
      }

#line 41 "jollyjack/jollyjack_cython.pyx"
    } else {

#line 41 "jollyjack/jollyjack_cython.pyx"
      switch (__pyx_nargs) {

#line 41 "jollyjack/jollyjack_cython.pyx"
        case 11:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case 10:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  9:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  8:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  7:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  6:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  5:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 41 "jollyjack/jollyjack_cython.pyx"
        case  4:

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 41 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
        break;
        default: 
#line 41 "jollyjack/jollyjack_cython.pyx"
goto __pyx_L5_argtuple_error;

#line 41 "jollyjack/jollyjack_cython.pyx"
      }

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[4]) values[4] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__4);

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[5]) values[5] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__5);

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[6]) values[6] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__6);

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)Py_False));

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)Py_True));

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[9]) values[9] = __Pyx_NewRef(((PyObject *)Py_False));

#line 41 "jollyjack/jollyjack_cython.pyx"
      if (!values[10]) values[10] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyarrow_3lib_CacheOptions *)Py_None));

#line 41 "jollyjack/jollyjack_cython.pyx"
    }

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_source = values[0];

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_metadata = ((struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *)values[1]);

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_np_array = ((PyArrayObject *)values[2]);

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_row_group_indices = values[3];

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_row_ranges = values[4];

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_column_indices = values[5];

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_column_names = values[6];

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_pre_buffer = values[7];

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_use_threads = values[8];

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_use_memory_map = values[9];

#line 41 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_cache_options = ((struct __pyx_obj_7pyarrow_3lib_CacheOptions *)values[10]);

#line 41 "jollyjack/jollyjack_cython.pyx"
  }

#line 41 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L6_skip;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L5_argtuple_error:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RaiseArgtupleInvalid("read_into_numpy", 0, 4, 11, __pyx_nargs); __PYX_ERR(0, 41, __pyx_L3_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L6_skip:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L3_error:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 41 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 41 "jollyjack/jollyjack_cython.pyx"
  }

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.read_into_numpy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 41 "jollyjack/jollyjack_cython.pyx"
  return NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_metadata), __pyx_mstate_global->__pyx_ptype_7pyarrow_8_parquet_FileMetaData, 1, "metadata", 0))) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_np_array), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "np_array", 0))) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cache_options), __pyx_mstate_global->__pyx_ptype_7pyarrow_3lib_CacheOptions, 1, "cache_options", 0))) __PYX_ERR(0, 41, __pyx_L1_error)
  __pyx_r = 
#line 41 "jollyjack/jollyjack_cython.pyx"
__pyx_pf_9jollyjack_16jollyjack_cython_4read_into_numpy(__pyx_self, __pyx_v_source, __pyx_v_metadata, __pyx_v_np_array, __pyx_v_row_group_indices, __pyx_v_row_ranges, __pyx_v_column_indices, __pyx_v_column_names, __pyx_v_pre_buffer, __pyx_v_use_threads, __pyx_v_use_memory_map, __pyx_v_cache_options);

#line 41 "jollyjack/jollyjack_cython.pyx"


#line 41 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 41 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 41 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 41 "jollyjack/jollyjack_cython.pyx"
  }

#line 41 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L7_cleaned_up;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 41 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 41 "jollyjack/jollyjack_cython.pyx"
  }

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L7_cleaned_up:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 41 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 41 "jollyjack/jollyjack_cython.pyx"
}

#line 41 "jollyjack/jollyjack_cython.pyx"


#line 41 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_4read_into_numpy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_source, struct __pyx_obj_7pyarrow_8_parquet_FileMetaData *__pyx_v_metadata, PyArrayObject *__pyx_v_np_array, PyObject *__pyx_v_row_group_indices, PyObject *__pyx_v_row_ranges, PyObject *__pyx_v_column_indices, PyObject *__pyx_v_column_names, PyObject *__pyx_v_pre_buffer, PyObject *__pyx_v_use_threads, PyObject *__pyx_v_use_memory_map, struct __pyx_obj_7pyarrow_3lib_CacheOptions *__pyx_v_cache_options) {

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 41 "jollyjack/jollyjack_cython.pyx"
  struct __pyx_opt_args_9jollyjack_16jollyjack_cython_read_into_numpy __pyx_t_1;

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("read_into_numpy", 0);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.__pyx_n = 7;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.row_ranges = __pyx_v_row_ranges;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.column_indices = __pyx_v_column_indices;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.column_names = __pyx_v_column_names;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.pre_buffer = __pyx_v_pre_buffer;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.use_threads = __pyx_v_use_threads;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.use_memory_map = __pyx_v_use_memory_map;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1.cache_options = __pyx_v_cache_options;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_f_9jollyjack_16jollyjack_cython_read_into_numpy(__pyx_v_source, __pyx_v_metadata, __pyx_v_np_array, __pyx_v_row_group_indices, 1, &__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = __pyx_t_2;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

#line 41 "jollyjack/jollyjack_cython.pyx"


#line 41 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.read_into_numpy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = NULL;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 41 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 41 "jollyjack/jollyjack_cython.pyx"
}

/* "jollyjack/jollyjack_cython.pyx":153
 *         return
 * 
 * cpdef void copy_to_torch_row_major (src_tensor, dst_tensor, row_indices):             # <<<<<<<<<<<<<<
 *     import torch
 * 
*/

#line 153 "jollyjack/jollyjack_cython.pyx"


#line 153 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_7copy_to_torch_row_major(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/

#line 153 "jollyjack/jollyjack_cython.pyx"
static void __pyx_f_9jollyjack_16jollyjack_cython_copy_to_torch_row_major(PyObject *__pyx_v_src_tensor, PyObject *__pyx_v_dst_tensor, PyObject *__pyx_v_row_indices, CYTHON_UNUSED int __pyx_skip_dispatch) {
  CYTHON_UNUSED PyObject *__pyx_v_torch = NULL
#line 153 "jollyjack/jollyjack_cython.pyx"
;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 153 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  size_t __pyx_t_4;

#line 153 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 153 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("copy_to_torch_row_major", 0);

  /* "jollyjack/jollyjack_cython.pyx":154
 * 
 * cpdef void copy_to_torch_row_major (src_tensor, dst_tensor, row_indices):
 *     import torch             # <<<<<<<<<<<<<<
 * 
 *     copy_to_numpy_row_major (src_array = src_tensor.numpy()
*/

#line 154 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_torch, 0, 0, NULL, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)

#line 154 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = __pyx_t_2;

#line 154 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 154 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_torch = __pyx_t_1;

#line 154 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = 0;

  /* "jollyjack/jollyjack_cython.pyx":156
 *     import torch
 * 
 *     copy_to_numpy_row_major (src_array = src_tensor.numpy()             # <<<<<<<<<<<<<<
 *         , dst_array = dst_tensor.numpy()
 *         , row_indices = row_indices
*/

#line 156 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __pyx_v_src_tensor;

#line 156 "jollyjack/jollyjack_cython.pyx"
  __Pyx_INCREF(__pyx_t_3);

#line 156 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_4 = 0;

#line 156 "jollyjack/jollyjack_cython.pyx"
  {

#line 156 "jollyjack/jollyjack_cython.pyx"
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};

#line 156 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_numpy, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 156 "jollyjack/jollyjack_cython.pyx"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 156 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)

#line 156 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 156 "jollyjack/jollyjack_cython.pyx"
  }

#line 156 "jollyjack/jollyjack_cython.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 156, __pyx_L1_error)

  /* "jollyjack/jollyjack_cython.pyx":157
 * 
 *     copy_to_numpy_row_major (src_array = src_tensor.numpy()
 *         , dst_array = dst_tensor.numpy()             # <<<<<<<<<<<<<<
 *         , row_indices = row_indices
 *     )
*/

#line 157 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __pyx_v_dst_tensor;

#line 157 "jollyjack/jollyjack_cython.pyx"
  __Pyx_INCREF(__pyx_t_5);

#line 157 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_4 = 0;

#line 157 "jollyjack/jollyjack_cython.pyx"
  {

#line 157 "jollyjack/jollyjack_cython.pyx"
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};

#line 157 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_numpy, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 157 "jollyjack/jollyjack_cython.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 157 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 157, __pyx_L1_error)

#line 157 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 157 "jollyjack/jollyjack_cython.pyx"
  }

#line 157 "jollyjack/jollyjack_cython.pyx"
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 157, __pyx_L1_error)

  /* "jollyjack/jollyjack_cython.pyx":156
 *     import torch
 * 
 *     copy_to_numpy_row_major (src_array = src_tensor.numpy()             # <<<<<<<<<<<<<<
 *         , dst_array = dst_tensor.numpy()
 *         , row_indices = row_indices
*/

#line 156 "jollyjack/jollyjack_cython.pyx"
  __pyx_f_9jollyjack_16jollyjack_cython_copy_to_numpy_row_major(((PyArrayObject *)__pyx_t_1), ((PyArrayObject *)__pyx_t_3), __pyx_v_row_indices, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)

#line 156 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 156 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "jollyjack/jollyjack_cython.pyx":153
 *         return
 * 
 * cpdef void copy_to_torch_row_major (src_tensor, dst_tensor, row_indices):             # <<<<<<<<<<<<<<
 *     import torch
 * 
*/

#line 153 "jollyjack/jollyjack_cython.pyx"


#line 153 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 153 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.copy_to_torch_row_major", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_v_torch);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 153 "jollyjack/jollyjack_cython.pyx"
}

#line 153 "jollyjack/jollyjack_cython.pyx"


#line 153 "jollyjack/jollyjack_cython.pyx"
/* Python wrapper */

#line 153 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_7copy_to_torch_row_major(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/

#line 153 "jollyjack/jollyjack_cython.pyx"
PyDoc_STRVAR(__pyx_doc_9jollyjack_16jollyjack_cython_6copy_to_torch_row_major, "copy_to_torch_row_major(src_tensor, dst_tensor, row_indices) -> void");
static PyMethodDef __pyx_mdef_9jollyjack_16jollyjack_cython_7copy_to_torch_row_major = 
#line 153 "jollyjack/jollyjack_cython.pyx"
{"copy_to_torch_row_major", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9jollyjack_16jollyjack_cython_7copy_to_torch_row_major, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9jollyjack_16jollyjack_cython_6copy_to_torch_row_major};

#line 153 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_7copy_to_torch_row_major(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_src_tensor = 0
#line 153 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_dst_tensor = 0
#line 153 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_row_indices = 0
#line 153 "jollyjack/jollyjack_cython.pyx"
;

#line 153 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 153 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;

#line 153 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 153 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;

#line 153 "jollyjack/jollyjack_cython.pyx"
  PyObject* values[3] = {0,0,0};

#line 153 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 153 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 153 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = 0;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("copy_to_torch_row_major (wrapper)", 0);

#line 153 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 153 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_ASSUME_SAFE_SIZE

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);

#line 153 "jollyjack/jollyjack_cython.pyx"
  #else

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 153 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);

#line 153 "jollyjack/jollyjack_cython.pyx"
  {

#line 153 "jollyjack/jollyjack_cython.pyx"
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_src_tensor,&__pyx_mstate_global->__pyx_n_u_dst_tensor,&__pyx_mstate_global->__pyx_n_u_row_indices,0};

#line 153 "jollyjack/jollyjack_cython.pyx"
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;

#line 153 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 153, __pyx_L3_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
    if (__pyx_kwds_len > 0) {

#line 153 "jollyjack/jollyjack_cython.pyx"
      switch (__pyx_nargs) {

#line 153 "jollyjack/jollyjack_cython.pyx"
        case  3:

#line 153 "jollyjack/jollyjack_cython.pyx"
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);

#line 153 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 153, __pyx_L3_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 153 "jollyjack/jollyjack_cython.pyx"
        case  2:

#line 153 "jollyjack/jollyjack_cython.pyx"
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);

#line 153 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 153, __pyx_L3_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 153 "jollyjack/jollyjack_cython.pyx"
        case  1:

#line 153 "jollyjack/jollyjack_cython.pyx"
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 153 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 153, __pyx_L3_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 153 "jollyjack/jollyjack_cython.pyx"
break;
        default: 
#line 153 "jollyjack/jollyjack_cython.pyx"
goto __pyx_L5_argtuple_error;

#line 153 "jollyjack/jollyjack_cython.pyx"
      }

#line 153 "jollyjack/jollyjack_cython.pyx"
      const Py_ssize_t kwd_pos_args = __pyx_nargs;

#line 153 "jollyjack/jollyjack_cython.pyx"
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "copy_to_torch_row_major", 0) < (0)) __PYX_ERR(0, 153, __pyx_L3_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {

#line 153 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("copy_to_torch_row_major", 1, 3, 3, i); __PYX_ERR(0, 153, __pyx_L3_error) }

#line 153 "jollyjack/jollyjack_cython.pyx"
      }

#line 153 "jollyjack/jollyjack_cython.pyx"
    } else if (unlikely(__pyx_nargs != 3)) {

#line 153 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L5_argtuple_error;

#line 153 "jollyjack/jollyjack_cython.pyx"
    } else {

#line 153 "jollyjack/jollyjack_cython.pyx"
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 153 "jollyjack/jollyjack_cython.pyx"
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 153, __pyx_L3_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);

#line 153 "jollyjack/jollyjack_cython.pyx"
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 153, __pyx_L3_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);

#line 153 "jollyjack/jollyjack_cython.pyx"
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 153, __pyx_L3_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
    }

#line 153 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_src_tensor = values[0];

#line 153 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_dst_tensor = values[1];

#line 153 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_row_indices = values[2];

#line 153 "jollyjack/jollyjack_cython.pyx"
  }

#line 153 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L6_skip;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_L5_argtuple_error:;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RaiseArgtupleInvalid("copy_to_torch_row_major", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 153, __pyx_L3_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_L6_skip:;

#line 153 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_L3_error:;

#line 153 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 153 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 153 "jollyjack/jollyjack_cython.pyx"
  }

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.copy_to_torch_row_major", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 153 "jollyjack/jollyjack_cython.pyx"
  return NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 153 "jollyjack/jollyjack_cython.pyx"
__pyx_pf_9jollyjack_16jollyjack_cython_6copy_to_torch_row_major(__pyx_self, __pyx_v_src_tensor, __pyx_v_dst_tensor, __pyx_v_row_indices);

#line 153 "jollyjack/jollyjack_cython.pyx"


#line 153 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 153 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 153 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 153 "jollyjack/jollyjack_cython.pyx"
  }

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 153 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 153 "jollyjack/jollyjack_cython.pyx"
}

#line 153 "jollyjack/jollyjack_cython.pyx"


#line 153 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_6copy_to_torch_row_major(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src_tensor, PyObject *__pyx_v_dst_tensor, PyObject *__pyx_v_row_indices) {

#line 153 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 153 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 153 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("copy_to_torch_row_major", 0);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_f_9jollyjack_16jollyjack_cython_copy_to_torch_row_major(__pyx_v_src_tensor, __pyx_v_dst_tensor, __pyx_v_row_indices, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 153, __pyx_L1_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = __pyx_t_1;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = 0;

#line 153 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

#line 153 "jollyjack/jollyjack_cython.pyx"


#line 153 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.copy_to_torch_row_major", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = NULL;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 153 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 153 "jollyjack/jollyjack_cython.pyx"
}

/* "jollyjack/jollyjack_cython.pyx":161
 *     )
 * 
 * cpdef void copy_to_numpy_row_major (cnp.ndarray src_array, cnp.ndarray dst_array, row_indices):             # <<<<<<<<<<<<<<
 * 
 *     if src_array.shape[0] == 0:
*/

#line 161 "jollyjack/jollyjack_cython.pyx"


#line 161 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_9copy_to_numpy_row_major(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/

#line 161 "jollyjack/jollyjack_cython.pyx"
static void __pyx_f_9jollyjack_16jollyjack_cython_copy_to_numpy_row_major(PyArrayObject *__pyx_v_src_array, PyArrayObject *__pyx_v_dst_array, PyObject *__pyx_v_row_indices, CYTHON_UNUSED int __pyx_skip_dispatch) {
  std::vector<int>  __pyx_v_crow_indices
#line 161 "jollyjack/jollyjack_cython.pyx"
;
  uint64_t __pyx_v_csrc_stride0
#line 161 "jollyjack/jollyjack_cython.pyx"
;
  uint64_t __pyx_v_csrc_stride1
#line 161 "jollyjack/jollyjack_cython.pyx"
;
  uint64_t __pyx_v_csrc_rows
#line 161 "jollyjack/jollyjack_cython.pyx"
;
  uint64_t __pyx_v_csrc_columns
#line 161 "jollyjack/jollyjack_cython.pyx"
;
  uint64_t __pyx_v_cdst_stride0
#line 161 "jollyjack/jollyjack_cython.pyx"
;
  uint64_t __pyx_v_cdst_stride1
#line 161 "jollyjack/jollyjack_cython.pyx"
;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 161 "jollyjack/jollyjack_cython.pyx"
  int __pyx_t_1;

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_3[3];

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_5[4];

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  Py_ssize_t __pyx_t_7;

#line 161 "jollyjack/jollyjack_cython.pyx"
  size_t __pyx_t_8;

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_9[5];

#line 161 "jollyjack/jollyjack_cython.pyx"
  std::vector<int>  __pyx_t_10;

#line 161 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 161 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("copy_to_numpy_row_major", 0);

  /* "jollyjack/jollyjack_cython.pyx":163
 * cpdef void copy_to_numpy_row_major (cnp.ndarray src_array, cnp.ndarray dst_array, row_indices):
 * 
 *     if src_array.shape[0] == 0:             # <<<<<<<<<<<<<<
 *         return # nothing to do
 * 
*/

#line 163 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = ((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_src_array)[0]) == 0);

#line 163 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_t_1) {

    /* "jollyjack/jollyjack_cython.pyx":164
 * 
 *     if src_array.shape[0] == 0:
 *         return # nothing to do             # <<<<<<<<<<<<<<
 * 
 *     assert src_array.ndim == 2, f"Unexpected src_array.ndim, {src_array.ndim} != 2"
*/

#line 164 "jollyjack/jollyjack_cython.pyx"
    goto __pyx_L0;

    /* "jollyjack/jollyjack_cython.pyx":163
 * cpdef void copy_to_numpy_row_major (cnp.ndarray src_array, cnp.ndarray dst_array, row_indices):
 * 
 *     if src_array.shape[0] == 0:             # <<<<<<<<<<<<<<
 *         return # nothing to do
 * 
*/

#line 163 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":166
 *         return # nothing to do
 * 
 *     assert src_array.ndim == 2, f"Unexpected src_array.ndim, {src_array.ndim} != 2"             # <<<<<<<<<<<<<<
 *     assert dst_array.ndim == 2, f"Unexpected dst_array.ndim, {dst_array.ndim} != 2"
 *     assert src_array.shape[0] <= dst_array.shape[0], f"src_array.shape[0] > dst_array.shape[0], {src_array.shape[0]} > {dst_array.shape[0]}"
*/

#line 166 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 166 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 166 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = (__pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_src_array) == 2);

#line 166 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 166 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_2 = __Pyx_PyUnicode_From_int(__pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_src_array), 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)

#line 166 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 166 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3[0] = __pyx_mstate_global->__pyx_kp_u_Unexpected_src_array_ndim;

#line 166 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3[1] = __pyx_t_2;

#line 166 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3[2] = __pyx_mstate_global->__pyx_kp_u_2;

#line 166 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, 27 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 5, 127);

#line 166 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)

#line 166 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 166 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 166 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_4, 0, 0);

#line 166 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 166 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 166, __pyx_L1_error)

#line 166 "jollyjack/jollyjack_cython.pyx"
    }

#line 166 "jollyjack/jollyjack_cython.pyx"
  }

#line 166 "jollyjack/jollyjack_cython.pyx"
  #else

#line 166 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 166, __pyx_L1_error)

#line 166 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":167
 * 
 *     assert src_array.ndim == 2, f"Unexpected src_array.ndim, {src_array.ndim} != 2"
 *     assert dst_array.ndim == 2, f"Unexpected dst_array.ndim, {dst_array.ndim} != 2"             # <<<<<<<<<<<<<<
 *     assert src_array.shape[0] <= dst_array.shape[0], f"src_array.shape[0] > dst_array.shape[0], {src_array.shape[0]} > {dst_array.shape[0]}"
 *     assert src_array.shape[1] == dst_array.shape[1], f"src_array.shape[1] != dst_array.shape[1], {src_array.shape[1]} != {dst_array.shape[1]}"
*/

#line 167 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 167 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 167 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = (__pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_dst_array) == 2);

#line 167 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 167 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_dst_array), 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 167 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3[0] = __pyx_mstate_global->__pyx_kp_u_Unexpected_dst_array_ndim;

#line 167 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3[1] = __pyx_t_4;

#line 167 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_3[2] = __pyx_mstate_global->__pyx_kp_u_2;

#line 167 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, 27 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 5, 127);

#line 167 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 167 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 167 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_2, 0, 0);

#line 167 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 167 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "jollyjack/jollyjack_cython.pyx"
    }

#line 167 "jollyjack/jollyjack_cython.pyx"
  }

#line 167 "jollyjack/jollyjack_cython.pyx"
  #else

#line 167 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":168
 *     assert src_array.ndim == 2, f"Unexpected src_array.ndim, {src_array.ndim} != 2"
 *     assert dst_array.ndim == 2, f"Unexpected dst_array.ndim, {dst_array.ndim} != 2"
 *     assert src_array.shape[0] <= dst_array.shape[0], f"src_array.shape[0] > dst_array.shape[0], {src_array.shape[0]} > {dst_array.shape[0]}"             # <<<<<<<<<<<<<<
 *     assert src_array.shape[1] == dst_array.shape[1], f"src_array.shape[1] != dst_array.shape[1], {src_array.shape[1]} != {dst_array.shape[1]}"
 *     assert (src_array.strides[0] <= src_array.strides[1]), f"Expected source array in a Fortran (column-major) order"
*/

#line 168 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 168 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 168 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = ((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_src_array)[0]) <= (__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_dst_array)[0]));

#line 168 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 168 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_2 = __Pyx_PyUnicode_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_src_array)[0]), 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 168 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyUnicode_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_dst_array)[0]), 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 168 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[0] = __pyx_mstate_global->__pyx_kp_u_src_array_shape_0_dst_array_shap;

#line 168 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[1] = __pyx_t_2;

#line 168 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[2] = __pyx_mstate_global->__pyx_kp_u__7;

#line 168 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[3] = __pyx_t_4;

#line 168 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_5, 4, 41 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4), 127);

#line 168 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 168 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 168 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 168 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_6, 0, 0);

#line 168 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 168 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "jollyjack/jollyjack_cython.pyx"
    }

#line 168 "jollyjack/jollyjack_cython.pyx"
  }

#line 168 "jollyjack/jollyjack_cython.pyx"
  #else

#line 168 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":169
 *     assert dst_array.ndim == 2, f"Unexpected dst_array.ndim, {dst_array.ndim} != 2"
 *     assert src_array.shape[0] <= dst_array.shape[0], f"src_array.shape[0] > dst_array.shape[0], {src_array.shape[0]} > {dst_array.shape[0]}"
 *     assert src_array.shape[1] == dst_array.shape[1], f"src_array.shape[1] != dst_array.shape[1], {src_array.shape[1]} != {dst_array.shape[1]}"             # <<<<<<<<<<<<<<
 *     assert (src_array.strides[0] <= src_array.strides[1]), f"Expected source array in a Fortran (column-major) order"
 *     assert (dst_array.strides[1] <= dst_array.strides[0]), f"Expected destination array in a C (row-major) order"
*/

#line 169 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 169 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 169 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = ((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_src_array)[1]) == (__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_dst_array)[1]));

#line 169 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 169 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyUnicode_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_src_array)[1]), 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 169, __pyx_L1_error)

#line 169 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 169 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyUnicode_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_dst_array)[1]), 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 169, __pyx_L1_error)

#line 169 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 169 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[0] = __pyx_mstate_global->__pyx_kp_u_src_array_shape_1_dst_array_shap;

#line 169 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[1] = __pyx_t_6;

#line 169 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[2] = __pyx_mstate_global->__pyx_kp_u__8;

#line 169 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[3] = __pyx_t_4;

#line 169 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_5, 4, 42 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 4 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4), 127);

#line 169 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)

#line 169 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 169 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 169 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 169 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_2, 0, 0);

#line 169 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 169 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 169, __pyx_L1_error)

#line 169 "jollyjack/jollyjack_cython.pyx"
    }

#line 169 "jollyjack/jollyjack_cython.pyx"
  }

#line 169 "jollyjack/jollyjack_cython.pyx"
  #else

#line 169 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 169, __pyx_L1_error)

#line 169 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":170
 *     assert src_array.shape[0] <= dst_array.shape[0], f"src_array.shape[0] > dst_array.shape[0], {src_array.shape[0]} > {dst_array.shape[0]}"
 *     assert src_array.shape[1] == dst_array.shape[1], f"src_array.shape[1] != dst_array.shape[1], {src_array.shape[1]} != {dst_array.shape[1]}"
 *     assert (src_array.strides[0] <= src_array.strides[1]), f"Expected source array in a Fortran (column-major) order"             # <<<<<<<<<<<<<<
 *     assert (dst_array.strides[1] <= dst_array.strides[0]), f"Expected destination array in a C (row-major) order"
 *     assert src_array.dtype == dst_array.dtype, f"Source and destination arrays have different datatypes, {src_array.dtype} != {dst_array.dtype}"
*/

#line 170 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 170 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 170 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = ((__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_src_array)[0]) <= (__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_src_array)[1]));

#line 170 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 170 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_mstate_global->__pyx_kp_u_Expected_source_array_in_a_Fortr, 0, 0);

#line 170 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 170, __pyx_L1_error)

#line 170 "jollyjack/jollyjack_cython.pyx"
    }

#line 170 "jollyjack/jollyjack_cython.pyx"
  }

#line 170 "jollyjack/jollyjack_cython.pyx"
  #else

#line 170 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 170, __pyx_L1_error)

#line 170 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":171
 *     assert src_array.shape[1] == dst_array.shape[1], f"src_array.shape[1] != dst_array.shape[1], {src_array.shape[1]} != {dst_array.shape[1]}"
 *     assert (src_array.strides[0] <= src_array.strides[1]), f"Expected source array in a Fortran (column-major) order"
 *     assert (dst_array.strides[1] <= dst_array.strides[0]), f"Expected destination array in a C (row-major) order"             # <<<<<<<<<<<<<<
 *     assert src_array.dtype == dst_array.dtype, f"Source and destination arrays have different datatypes, {src_array.dtype} != {dst_array.dtype}"
 *     assert len(row_indices) == src_array.shape[0], f"Unexpected len of row indices, {len(row_indices)} != {src_array.shape[0]}"
*/

#line 171 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 171 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 171 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = ((__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_dst_array)[1]) <= (__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_dst_array)[0]));

#line 171 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 171 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_mstate_global->__pyx_kp_u_Expected_destination_array_in_a, 0, 0);

#line 171 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "jollyjack/jollyjack_cython.pyx"
    }

#line 171 "jollyjack/jollyjack_cython.pyx"
  }

#line 171 "jollyjack/jollyjack_cython.pyx"
  #else

#line 171 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":172
 *     assert (src_array.strides[0] <= src_array.strides[1]), f"Expected source array in a Fortran (column-major) order"
 *     assert (dst_array.strides[1] <= dst_array.strides[0]), f"Expected destination array in a C (row-major) order"
 *     assert src_array.dtype == dst_array.dtype, f"Source and destination arrays have different datatypes, {src_array.dtype} != {dst_array.dtype}"             # <<<<<<<<<<<<<<
 *     assert len(row_indices) == src_array.shape[0], f"Unexpected len of row indices, {len(row_indices)} != {src_array.shape[0]}"
 *     assert min(row_indices) >= 0, f"Row index = {min(row_indices)} is not in the expected range [0, {dst_array.shape[0]})"
*/

#line 172 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 172 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 172 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_src_array), __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 172 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_dst_array), __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 172 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 172 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 172 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 172 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 172 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_src_array), __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 172 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 172 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_dst_array), __pyx_mstate_global->__pyx_n_u_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 172 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 172 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[0] = __pyx_mstate_global->__pyx_kp_u_Source_and_destination_arrays_ha;

#line 172 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[1] = __pyx_t_4;

#line 172 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[2] = __pyx_mstate_global->__pyx_kp_u__8;

#line 172 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[3] = __pyx_t_2;

#line 172 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_5, 4, 56 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 4 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));

#line 172 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_6, 0, 0);

#line 172 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 172 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
    }

#line 172 "jollyjack/jollyjack_cython.pyx"
  }

#line 172 "jollyjack/jollyjack_cython.pyx"
  #else

#line 172 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":173
 *     assert (dst_array.strides[1] <= dst_array.strides[0]), f"Expected destination array in a C (row-major) order"
 *     assert src_array.dtype == dst_array.dtype, f"Source and destination arrays have different datatypes, {src_array.dtype} != {dst_array.dtype}"
 *     assert len(row_indices) == src_array.shape[0], f"Unexpected len of row indices, {len(row_indices)} != {src_array.shape[0]}"             # <<<<<<<<<<<<<<
 *     assert min(row_indices) >= 0, f"Row index = {min(row_indices)} is not in the expected range [0, {dst_array.shape[0]})"
 *     assert max(row_indices) < dst_array.shape[0], f"Row index = {max(row_indices)} is not in the expected range [0, {dst_array.shape[0]})"
*/

#line 173 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 173 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 173 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_7 = PyObject_Length(__pyx_v_row_indices); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = (__pyx_t_7 == (__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_src_array)[0]));

#line 173 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 173 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_7 = PyObject_Length(__pyx_v_row_indices); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_7, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 173 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_2 = __Pyx_PyUnicode_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_src_array)[0]), 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 173 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[0] = __pyx_mstate_global->__pyx_kp_u_Unexpected_len_of_row_indices;

#line 173 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[1] = __pyx_t_6;

#line 173 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[2] = __pyx_mstate_global->__pyx_kp_u__8;

#line 173 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5[3] = __pyx_t_2;

#line 173 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_5, 4, 31 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 4 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2), 127);

#line 173 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 173 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 173 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 173 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_4, 0, 0);

#line 173 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 173 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "jollyjack/jollyjack_cython.pyx"
    }

#line 173 "jollyjack/jollyjack_cython.pyx"
  }

#line 173 "jollyjack/jollyjack_cython.pyx"
  #else

#line 173 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":174
 *     assert src_array.dtype == dst_array.dtype, f"Source and destination arrays have different datatypes, {src_array.dtype} != {dst_array.dtype}"
 *     assert len(row_indices) == src_array.shape[0], f"Unexpected len of row indices, {len(row_indices)} != {src_array.shape[0]}"
 *     assert min(row_indices) >= 0, f"Row index = {min(row_indices)} is not in the expected range [0, {dst_array.shape[0]})"             # <<<<<<<<<<<<<<
 *     assert max(row_indices) < dst_array.shape[0], f"Row index = {max(row_indices)} is not in the expected range [0, {dst_array.shape[0]})"
 * 
*/

#line 174 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 174 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 174 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_2 = NULL;

#line 174 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_8 = 1;

#line 174 "jollyjack/jollyjack_cython.pyx"
    {

#line 174 "jollyjack/jollyjack_cython.pyx"
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_row_indices};

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_min, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 174 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 174, __pyx_L1_error)

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 174 "jollyjack/jollyjack_cython.pyx"
    }

#line 174 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_mstate_global->__pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)

#line 174 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 174 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 174, __pyx_L1_error)

#line 174 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 174 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_v_row_indices); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 174, __pyx_L1_error)

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_2 = __Pyx_PyUnicode_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_dst_array)[0]), 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[0] = __pyx_mstate_global->__pyx_kp_u_Row_index;

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[1] = __pyx_t_4;

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[2] = __pyx_mstate_global->__pyx_kp_u_is_not_in_the_expected_range_0;

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[3] = __pyx_t_2;

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[4] = __pyx_mstate_global->__pyx_kp_u__9;

#line 174 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_9, 5, 12 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 34 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4));

#line 174 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 174, __pyx_L1_error)

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_6, 0, 0);

#line 174 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 174 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 174, __pyx_L1_error)

#line 174 "jollyjack/jollyjack_cython.pyx"
    }

#line 174 "jollyjack/jollyjack_cython.pyx"
  }

#line 174 "jollyjack/jollyjack_cython.pyx"
  #else

#line 174 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 174, __pyx_L1_error)

#line 174 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":175
 *     assert len(row_indices) == src_array.shape[0], f"Unexpected len of row indices, {len(row_indices)} != {src_array.shape[0]}"
 *     assert min(row_indices) >= 0, f"Row index = {min(row_indices)} is not in the expected range [0, {dst_array.shape[0]})"
 *     assert max(row_indices) < dst_array.shape[0], f"Row index = {max(row_indices)} is not in the expected range [0, {dst_array.shape[0]})"             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[int] crow_indices = row_indices
*/

#line 175 "jollyjack/jollyjack_cython.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 175 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(__pyx_assertions_enabled())) {

#line 175 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_2 = NULL;

#line 175 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_8 = 1;

#line 175 "jollyjack/jollyjack_cython.pyx"
    {

#line 175 "jollyjack/jollyjack_cython.pyx"
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_row_indices};

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_max, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 175 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 175 "jollyjack/jollyjack_cython.pyx"
    }

#line 175 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_2 = __Pyx_PyLong_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_dst_array)[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 175 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 175 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 175 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 175 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(!__pyx_t_1)) {

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_v_row_indices); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_4 = __Pyx_PyUnicode_From_npy_intp((__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_dst_array)[0]), 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[0] = __pyx_mstate_global->__pyx_kp_u_Row_index;

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[1] = __pyx_t_2;

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[2] = __pyx_mstate_global->__pyx_kp_u_is_not_in_the_expected_range_0;

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[3] = __pyx_t_4;

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_9[4] = __pyx_mstate_global->__pyx_kp_u__9;

#line 175 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_9, 5, 12 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 34 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));

#line 175 "jollyjack/jollyjack_cython.pyx"
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_AssertionError))), __pyx_t_6, 0, 0);

#line 175 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 175 "jollyjack/jollyjack_cython.pyx"
      __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
    }

#line 175 "jollyjack/jollyjack_cython.pyx"
  }

#line 175 "jollyjack/jollyjack_cython.pyx"
  #else

#line 175 "jollyjack/jollyjack_cython.pyx"
  if ((1)); else __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "jollyjack/jollyjack_cython.pyx"
  #endif

  /* "jollyjack/jollyjack_cython.pyx":177
 *     assert max(row_indices) < dst_array.shape[0], f"Row index = {max(row_indices)} is not in the expected range [0, {dst_array.shape[0]})"
 * 
 *     cdef vector[int] crow_indices = row_indices             # <<<<<<<<<<<<<<
 *     cdef uint64_t csrc_stride0 = src_array.strides[0]
 *     cdef uint64_t csrc_stride1 = src_array.strides[1]
*/

#line 177 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_10 = __pyx_convert_vector_from_py_int(__pyx_v_row_indices); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 177, __pyx_L1_error)

#line 177 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_crow_indices = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_10);

  /* "jollyjack/jollyjack_cython.pyx":178
 * 
 *     cdef vector[int] crow_indices = row_indices
 *     cdef uint64_t csrc_stride0 = src_array.strides[0]             # <<<<<<<<<<<<<<
 *     cdef uint64_t csrc_stride1 = src_array.strides[1]
 *     cdef uint64_t csrc_rows = src_array.shape[0]
*/

#line 178 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_csrc_stride0 = (__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_src_array)[0]);

  /* "jollyjack/jollyjack_cython.pyx":179
 *     cdef vector[int] crow_indices = row_indices
 *     cdef uint64_t csrc_stride0 = src_array.strides[0]
 *     cdef uint64_t csrc_stride1 = src_array.strides[1]             # <<<<<<<<<<<<<<
 *     cdef uint64_t csrc_rows = src_array.shape[0]
 *     cdef uint64_t csrc_columns = src_array.shape[1]
*/

#line 179 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_csrc_stride1 = (__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_src_array)[1]);

  /* "jollyjack/jollyjack_cython.pyx":180
 *     cdef uint64_t csrc_stride0 = src_array.strides[0]
 *     cdef uint64_t csrc_stride1 = src_array.strides[1]
 *     cdef uint64_t csrc_rows = src_array.shape[0]             # <<<<<<<<<<<<<<
 *     cdef uint64_t csrc_columns = src_array.shape[1]
 *     cdef uint64_t cdst_stride0 = dst_array.strides[0]
*/

#line 180 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_csrc_rows = (__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_src_array)[0]);

  /* "jollyjack/jollyjack_cython.pyx":181
 *     cdef uint64_t csrc_stride1 = src_array.strides[1]
 *     cdef uint64_t csrc_rows = src_array.shape[0]
 *     cdef uint64_t csrc_columns = src_array.shape[1]             # <<<<<<<<<<<<<<
 *     cdef uint64_t cdst_stride0 = dst_array.strides[0]
 *     cdef uint64_t cdst_stride1 = dst_array.strides[1]
*/

#line 181 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_csrc_columns = (__pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_src_array)[1]);

  /* "jollyjack/jollyjack_cython.pyx":182
 *     cdef uint64_t csrc_rows = src_array.shape[0]
 *     cdef uint64_t csrc_columns = src_array.shape[1]
 *     cdef uint64_t cdst_stride0 = dst_array.strides[0]             # <<<<<<<<<<<<<<
 *     cdef uint64_t cdst_stride1 = dst_array.strides[1]
 * 
*/

#line 182 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_cdst_stride0 = (__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_dst_array)[0]);

  /* "jollyjack/jollyjack_cython.pyx":183
 *     cdef uint64_t csrc_columns = src_array.shape[1]
 *     cdef uint64_t cdst_stride0 = dst_array.strides[0]
 *     cdef uint64_t cdst_stride1 = dst_array.strides[1]             # <<<<<<<<<<<<<<
 * 
 *     with nogil:
*/

#line 183 "jollyjack/jollyjack_cython.pyx"
  __pyx_v_cdst_stride1 = (__pyx_f_5numpy_7ndarray_7strides_strides(__pyx_v_dst_array)[1]);

  /* "jollyjack/jollyjack_cython.pyx":185
 *     cdef uint64_t cdst_stride1 = dst_array.strides[1]
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         cjollyjack.CopyToRowMajor (src_array.data
 *             , csrc_stride0
*/

#line 185 "jollyjack/jollyjack_cython.pyx"
  {

#line 185 "jollyjack/jollyjack_cython.pyx"
      PyThreadState * _save;

#line 185 "jollyjack/jollyjack_cython.pyx"
      _save = PyEval_SaveThread();

#line 185 "jollyjack/jollyjack_cython.pyx"
      __Pyx_FastGIL_Remember();

#line 185 "jollyjack/jollyjack_cython.pyx"
      /*try:*/ {

        /* "jollyjack/jollyjack_cython.pyx":186
 * 
 *     with nogil:
 *         cjollyjack.CopyToRowMajor (src_array.data             # <<<<<<<<<<<<<<
 *             , csrc_stride0
 *             , csrc_stride1
*/

#line 186 "jollyjack/jollyjack_cython.pyx"
        try {

#line 186 "jollyjack/jollyjack_cython.pyx"
          CopyToRowMajor(__pyx_f_5numpy_7ndarray_4data_data(__pyx_v_src_array), __pyx_v_csrc_stride0, __pyx_v_csrc_stride1, __pyx_v_csrc_rows, __pyx_v_csrc_columns, __pyx_f_5numpy_7ndarray_4data_data(__pyx_v_dst_array), __pyx_v_cdst_stride0, __pyx_v_cdst_stride1, __pyx_v_crow_indices);

#line 186 "jollyjack/jollyjack_cython.pyx"
        } catch(...) {
          PyGILState_STATE 
#line 186 "jollyjack/jollyjack_cython.pyx"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 186 "jollyjack/jollyjack_cython.pyx"
          __Pyx_CppExn2PyErr();

#line 186 "jollyjack/jollyjack_cython.pyx"
          __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 186 "jollyjack/jollyjack_cython.pyx"
          __PYX_ERR(0, 186, __pyx_L5_error)

#line 186 "jollyjack/jollyjack_cython.pyx"
        }

        /* "jollyjack/jollyjack_cython.pyx":195
 *             , cdst_stride1
 *             , crow_indices);
 *         return             # <<<<<<<<<<<<<<
*/

#line 195 "jollyjack/jollyjack_cython.pyx"
        goto __pyx_L4_return;

#line 195 "jollyjack/jollyjack_cython.pyx"
      }

      /* "jollyjack/jollyjack_cython.pyx":185
 *     cdef uint64_t cdst_stride1 = dst_array.strides[1]
 * 
 *     with nogil:             # <<<<<<<<<<<<<<
 *         cjollyjack.CopyToRowMajor (src_array.data
 *             , csrc_stride0
*/

#line 185 "jollyjack/jollyjack_cython.pyx"
      /*finally:*/ {

#line 185 "jollyjack/jollyjack_cython.pyx"
        __pyx_L4_return: {

#line 185 "jollyjack/jollyjack_cython.pyx"
          __Pyx_FastGIL_Forget();

#line 185 "jollyjack/jollyjack_cython.pyx"
          PyEval_RestoreThread(_save);

#line 185 "jollyjack/jollyjack_cython.pyx"
          goto __pyx_L0;

#line 185 "jollyjack/jollyjack_cython.pyx"
        }

#line 185 "jollyjack/jollyjack_cython.pyx"
        __pyx_L5_error: {

#line 185 "jollyjack/jollyjack_cython.pyx"
          __Pyx_FastGIL_Forget();

#line 185 "jollyjack/jollyjack_cython.pyx"
          PyEval_RestoreThread(_save);

#line 185 "jollyjack/jollyjack_cython.pyx"
          goto __pyx_L1_error;

#line 185 "jollyjack/jollyjack_cython.pyx"
        }

#line 185 "jollyjack/jollyjack_cython.pyx"
      }

#line 185 "jollyjack/jollyjack_cython.pyx"
  }

  /* "jollyjack/jollyjack_cython.pyx":161
 *     )
 * 
 * cpdef void copy_to_numpy_row_major (cnp.ndarray src_array, cnp.ndarray dst_array, row_indices):             # <<<<<<<<<<<<<<
 * 
 *     if src_array.shape[0] == 0:
*/

#line 161 "jollyjack/jollyjack_cython.pyx"


#line 161 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.copy_to_numpy_row_major", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 161 "jollyjack/jollyjack_cython.pyx"
}

#line 161 "jollyjack/jollyjack_cython.pyx"


#line 161 "jollyjack/jollyjack_cython.pyx"
/* Python wrapper */

#line 161 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_9copy_to_numpy_row_major(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/

#line 161 "jollyjack/jollyjack_cython.pyx"
PyDoc_STRVAR(__pyx_doc_9jollyjack_16jollyjack_cython_8copy_to_numpy_row_major, "copy_to_numpy_row_major(ndarray src_array, ndarray dst_array, row_indices) -> void");
static PyMethodDef __pyx_mdef_9jollyjack_16jollyjack_cython_9copy_to_numpy_row_major = 
#line 161 "jollyjack/jollyjack_cython.pyx"
{"copy_to_numpy_row_major", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9jollyjack_16jollyjack_cython_9copy_to_numpy_row_major, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9jollyjack_16jollyjack_cython_8copy_to_numpy_row_major};

#line 161 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pw_9jollyjack_16jollyjack_cython_9copy_to_numpy_row_major(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyArrayObject *__pyx_v_src_array = 0
#line 161 "jollyjack/jollyjack_cython.pyx"
;
  PyArrayObject *__pyx_v_dst_array = 0
#line 161 "jollyjack/jollyjack_cython.pyx"
;
  PyObject *__pyx_v_row_indices = 0
#line 161 "jollyjack/jollyjack_cython.pyx"
;

#line 161 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 161 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;

#line 161 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 161 "jollyjack/jollyjack_cython.pyx"
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject* values[3] = {0,0,0};

#line 161 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 161 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = 0;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("copy_to_numpy_row_major (wrapper)", 0);

#line 161 "jollyjack/jollyjack_cython.pyx"
  #if !CYTHON_METH_FASTCALL

#line 161 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_ASSUME_SAFE_SIZE

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);

#line 161 "jollyjack/jollyjack_cython.pyx"
  #else

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 161 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);

#line 161 "jollyjack/jollyjack_cython.pyx"
  {

#line 161 "jollyjack/jollyjack_cython.pyx"
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_src_array,&__pyx_mstate_global->__pyx_n_u_dst_array,&__pyx_mstate_global->__pyx_n_u_row_indices,0};

#line 161 "jollyjack/jollyjack_cython.pyx"
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;

#line 161 "jollyjack/jollyjack_cython.pyx"
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 161, __pyx_L3_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
    if (__pyx_kwds_len > 0) {

#line 161 "jollyjack/jollyjack_cython.pyx"
      switch (__pyx_nargs) {

#line 161 "jollyjack/jollyjack_cython.pyx"
        case  3:

#line 161 "jollyjack/jollyjack_cython.pyx"
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);

#line 161 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 161, __pyx_L3_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 161 "jollyjack/jollyjack_cython.pyx"
        case  2:

#line 161 "jollyjack/jollyjack_cython.pyx"
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);

#line 161 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 161, __pyx_L3_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;

#line 161 "jollyjack/jollyjack_cython.pyx"
        case  1:

#line 161 "jollyjack/jollyjack_cython.pyx"
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 161 "jollyjack/jollyjack_cython.pyx"
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 161, __pyx_L3_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 161 "jollyjack/jollyjack_cython.pyx"
break;
        default: 
#line 161 "jollyjack/jollyjack_cython.pyx"
goto __pyx_L5_argtuple_error;

#line 161 "jollyjack/jollyjack_cython.pyx"
      }

#line 161 "jollyjack/jollyjack_cython.pyx"
      const Py_ssize_t kwd_pos_args = __pyx_nargs;

#line 161 "jollyjack/jollyjack_cython.pyx"
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "copy_to_numpy_row_major", 0) < (0)) __PYX_ERR(0, 161, __pyx_L3_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {

#line 161 "jollyjack/jollyjack_cython.pyx"
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("copy_to_numpy_row_major", 1, 3, 3, i); __PYX_ERR(0, 161, __pyx_L3_error) }

#line 161 "jollyjack/jollyjack_cython.pyx"
      }

#line 161 "jollyjack/jollyjack_cython.pyx"
    } else if (unlikely(__pyx_nargs != 3)) {

#line 161 "jollyjack/jollyjack_cython.pyx"
      goto __pyx_L5_argtuple_error;

#line 161 "jollyjack/jollyjack_cython.pyx"
    } else {

#line 161 "jollyjack/jollyjack_cython.pyx"
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);

#line 161 "jollyjack/jollyjack_cython.pyx"
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 161, __pyx_L3_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);

#line 161 "jollyjack/jollyjack_cython.pyx"
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 161, __pyx_L3_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);

#line 161 "jollyjack/jollyjack_cython.pyx"
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 161, __pyx_L3_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
    }

#line 161 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_src_array = ((PyArrayObject *)values[0]);

#line 161 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_dst_array = ((PyArrayObject *)values[1]);

#line 161 "jollyjack/jollyjack_cython.pyx"
    __pyx_v_row_indices = values[2];

#line 161 "jollyjack/jollyjack_cython.pyx"
  }

#line 161 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L6_skip;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L5_argtuple_error:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RaiseArgtupleInvalid("copy_to_numpy_row_major", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 161, __pyx_L3_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L6_skip:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L3_error:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 161 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 161 "jollyjack/jollyjack_cython.pyx"
  }

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.copy_to_numpy_row_major", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 161 "jollyjack/jollyjack_cython.pyx"
  return NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_src_array), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "src_array", 0))) __PYX_ERR(0, 161, __pyx_L1_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dst_array), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "dst_array", 0))) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_r = 
#line 161 "jollyjack/jollyjack_cython.pyx"
__pyx_pf_9jollyjack_16jollyjack_cython_8copy_to_numpy_row_major(__pyx_self, __pyx_v_src_array, __pyx_v_dst_array, __pyx_v_row_indices);

#line 161 "jollyjack/jollyjack_cython.pyx"


#line 161 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 161 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 161 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 161 "jollyjack/jollyjack_cython.pyx"
  }

#line 161 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L7_cleaned_up;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {

#line 161 "jollyjack/jollyjack_cython.pyx"
    Py_XDECREF(values[__pyx_temp]);

#line 161 "jollyjack/jollyjack_cython.pyx"
  }

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L7_cleaned_up:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 161 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 161 "jollyjack/jollyjack_cython.pyx"
}

#line 161 "jollyjack/jollyjack_cython.pyx"


#line 161 "jollyjack/jollyjack_cython.pyx"
static PyObject *__pyx_pf_9jollyjack_16jollyjack_cython_8copy_to_numpy_row_major(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_src_array, PyArrayObject *__pyx_v_dst_array, PyObject *__pyx_v_row_indices) {

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_r = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyDeclarations

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  int __pyx_lineno = 0;

#line 161 "jollyjack/jollyjack_cython.pyx"
  const char *__pyx_filename = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  int __pyx_clineno = 0;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannySetupContext("copy_to_numpy_row_major", 0);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_f_9jollyjack_16jollyjack_cython_copy_to_numpy_row_major(__pyx_v_src_array, __pyx_v_dst_array, __pyx_v_row_indices, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 161, __pyx_L1_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = __pyx_t_1;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = 0;

#line 161 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

#line 161 "jollyjack/jollyjack_cython.pyx"


#line 161 "jollyjack/jollyjack_cython.pyx"
  /* function exit code */

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_AddTraceback("jollyjack.jollyjack_cython.copy_to_numpy_row_major", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_r = NULL;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 161 "jollyjack/jollyjack_cython.pyx"
  return __pyx_r;

#line 161 "jollyjack/jollyjack_cython.pyx"
}
/* #### Code section: module_exttypes ### */

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyLongObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyLongObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyLongObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyLongObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(7, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyComplexObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyComplexObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(8, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_5numpy_dtype = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "dtype",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArray_Descr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArray_Descr),
  #else
  sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArray_Descr),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 229, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_flatiter = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "flatiter",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayIterObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayIterObject),
  #else
  sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayIterObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 274, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_broadcast = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "broadcast",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayMultiIterObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayMultiIterObject),
  #else
  sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayMultiIterObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 278, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_ndarray = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "ndarray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayObject),
  #else
  sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 317, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_generic = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "generic",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_generic) __PYX_ERR(2, 826, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_number = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "number",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_number) __PYX_ERR(2, 828, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_integer = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "integer",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_integer) __PYX_ERR(2, 830, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "signedinteger",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 832, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "unsignedinteger",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 834, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_inexact = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "inexact",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 836, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_floating = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "floating",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_floating) __PYX_ERR(2, 838, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "complexfloating",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 840, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_flexible = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "flexible",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 842, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_character = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "character",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_character) __PYX_ERR(2, 844, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_ufunc = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "ufunc",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyUFuncObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyUFuncObject),
  #else
  sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyUFuncObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 908, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("datetime"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_8datetime_date = __Pyx_ImportType_3_2_4(__pyx_t_1, "datetime", "date",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyDateTime_Date), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_Date),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyDateTime_Date), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_Date),
  #else
  sizeof(PyDateTime_Date), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_Date),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_8datetime_date) __PYX_ERR(4, 43, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7cpython_8datetime_time = __Pyx_ImportType_3_2_4(__pyx_t_1, "datetime", "time",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyDateTime_Time), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_Time),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyDateTime_Time), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_Time),
  #else
  sizeof(PyDateTime_Time), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_Time),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_8datetime_time) __PYX_ERR(4, 56, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7cpython_8datetime_datetime = __Pyx_ImportType_3_2_4(__pyx_t_1, "datetime", "datetime",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyDateTime_DateTime), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_DateTime),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyDateTime_DateTime), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_DateTime),
  #else
  sizeof(PyDateTime_DateTime), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_DateTime),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_8datetime_datetime) __PYX_ERR(4, 82, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7cpython_8datetime_timedelta = __Pyx_ImportType_3_2_4(__pyx_t_1, "datetime", "timedelta",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyDateTime_Delta), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_Delta),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyDateTime_Delta), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_Delta),
  #else
  sizeof(PyDateTime_Delta), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_Delta),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_8datetime_timedelta) __PYX_ERR(4, 120, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7cpython_8datetime_tzinfo = __Pyx_ImportType_3_2_4(__pyx_t_1, "datetime", "tzinfo",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyDateTime_TZInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_TZInfo),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyDateTime_TZInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_TZInfo),
  #else
  sizeof(PyDateTime_TZInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyDateTime_TZInfo),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_8datetime_tzinfo) __PYX_ERR(4, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyarrow.lib"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib__Weakrefable = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "_Weakrefable",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib__Weakrefable), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__Weakrefable),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib__Weakrefable), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__Weakrefable),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib__Weakrefable), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__Weakrefable),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib__Weakrefable) __PYX_ERR(9, 37, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_IpcWriteOptions = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "IpcWriteOptions",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_IpcWriteOptions), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_IpcWriteOptions),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_IpcWriteOptions), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_IpcWriteOptions),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_IpcWriteOptions), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_IpcWriteOptions),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_IpcWriteOptions) __PYX_ERR(9, 41, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_IpcReadOptions = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "IpcReadOptions",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_IpcReadOptions), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_IpcReadOptions),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_IpcReadOptions), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_IpcReadOptions),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_IpcReadOptions), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_IpcReadOptions),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_IpcReadOptions) __PYX_ERR(9, 49, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Message = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Message",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Message), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Message),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Message), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Message),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Message), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Message),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Message) __PYX_ERR(9, 60, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_MemoryPool = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "MemoryPool",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_MemoryPool), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MemoryPool),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_MemoryPool), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MemoryPool),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_MemoryPool), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MemoryPool),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MemoryPool) __PYX_ERR(9, 65, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_MemoryPool = (struct __pyx_vtabstruct_7pyarrow_3lib_MemoryPool*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MemoryPool); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_MemoryPool)) __PYX_ERR(9, 65, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_DataType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "DataType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_DataType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DataType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_DataType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DataType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_DataType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DataType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DataType) __PYX_ERR(9, 78, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_DataType = (struct __pyx_vtabstruct_7pyarrow_3lib_DataType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DataType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_DataType)) __PYX_ERR(9, 78, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "ListType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListType) __PYX_ERR(9, 88, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_ListType = (struct __pyx_vtabstruct_7pyarrow_3lib_ListType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_ListType)) __PYX_ERR(9, 88, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "LargeListType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListType) __PYX_ERR(9, 93, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_LargeListType = (struct __pyx_vtabstruct_7pyarrow_3lib_LargeListType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_LargeListType)) __PYX_ERR(9, 93, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListViewType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "ListViewType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListViewType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListViewType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListViewType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListViewType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListViewType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListViewType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListViewType) __PYX_ERR(9, 98, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_ListViewType = (struct __pyx_vtabstruct_7pyarrow_3lib_ListViewType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListViewType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_ListViewType)) __PYX_ERR(9, 98, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListViewType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "LargeListViewType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListViewType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListViewType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListViewType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListViewType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListViewType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListViewType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListViewType) __PYX_ERR(9, 103, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_LargeListViewType = (struct __pyx_vtabstruct_7pyarrow_3lib_LargeListViewType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListViewType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_LargeListViewType)) __PYX_ERR(9, 103, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_MapType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "MapType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_MapType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MapType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_MapType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MapType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_MapType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MapType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MapType) __PYX_ERR(9, 108, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_MapType = (struct __pyx_vtabstruct_7pyarrow_3lib_MapType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MapType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_MapType)) __PYX_ERR(9, 108, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeListType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "FixedSizeListType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeListType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeListType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeListType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeListType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeListType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeListType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeListType) __PYX_ERR(9, 113, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_FixedSizeListType = (struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeListType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeListType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_FixedSizeListType)) __PYX_ERR(9, 113, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_StructType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "StructType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_StructType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StructType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_StructType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StructType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_StructType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StructType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StructType) __PYX_ERR(9, 118, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_StructType = (struct __pyx_vtabstruct_7pyarrow_3lib_StructType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StructType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_StructType)) __PYX_ERR(9, 118, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_DictionaryMemo = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "DictionaryMemo",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_DictionaryMemo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DictionaryMemo),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_DictionaryMemo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DictionaryMemo),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_DictionaryMemo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DictionaryMemo),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DictionaryMemo) __PYX_ERR(9, 125, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_DictionaryType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "DictionaryType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_DictionaryType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DictionaryType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_DictionaryType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DictionaryType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_DictionaryType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DictionaryType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DictionaryType) __PYX_ERR(9, 133, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_DictionaryType = (struct __pyx_vtabstruct_7pyarrow_3lib_DictionaryType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DictionaryType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_DictionaryType)) __PYX_ERR(9, 133, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_TimestampType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "TimestampType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_TimestampType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_TimestampType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_TimestampType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_TimestampType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_TimestampType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_TimestampType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_TimestampType) __PYX_ERR(9, 138, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_TimestampType = (struct __pyx_vtabstruct_7pyarrow_3lib_TimestampType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_TimestampType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_TimestampType)) __PYX_ERR(9, 138, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Time32Type = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Time32Type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Time32Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Time32Type),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Time32Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Time32Type),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Time32Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Time32Type),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Time32Type) __PYX_ERR(9, 143, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Time32Type = (struct __pyx_vtabstruct_7pyarrow_3lib_Time32Type*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Time32Type); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Time32Type)) __PYX_ERR(9, 143, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Time64Type = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Time64Type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Time64Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Time64Type),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Time64Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Time64Type),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Time64Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Time64Type),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Time64Type) __PYX_ERR(9, 148, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Time64Type = (struct __pyx_vtabstruct_7pyarrow_3lib_Time64Type*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Time64Type); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Time64Type)) __PYX_ERR(9, 148, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_DurationType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "DurationType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_DurationType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DurationType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_DurationType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DurationType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_DurationType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DurationType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DurationType) __PYX_ERR(9, 153, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_DurationType = (struct __pyx_vtabstruct_7pyarrow_3lib_DurationType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DurationType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_DurationType)) __PYX_ERR(9, 153, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "FixedSizeBinaryType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryType) __PYX_ERR(9, 158, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_FixedSizeBinaryType = (struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_FixedSizeBinaryType)) __PYX_ERR(9, 158, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal32Type = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Decimal32Type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal32Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal32Type),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal32Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal32Type),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal32Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal32Type),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal32Type) __PYX_ERR(9, 163, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Decimal32Type = (struct __pyx_vtabstruct_7pyarrow_3lib_Decimal32Type*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal32Type); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Decimal32Type)) __PYX_ERR(9, 163, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal64Type = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Decimal64Type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal64Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal64Type),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal64Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal64Type),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal64Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal64Type),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal64Type) __PYX_ERR(9, 168, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Decimal64Type = (struct __pyx_vtabstruct_7pyarrow_3lib_Decimal64Type*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal64Type); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Decimal64Type)) __PYX_ERR(9, 168, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal128Type = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Decimal128Type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal128Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal128Type),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal128Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal128Type),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal128Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal128Type),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal128Type) __PYX_ERR(9, 173, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Decimal128Type = (struct __pyx_vtabstruct_7pyarrow_3lib_Decimal128Type*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal128Type); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Decimal128Type)) __PYX_ERR(9, 173, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal256Type = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Decimal256Type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal256Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal256Type),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal256Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal256Type),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal256Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal256Type),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal256Type) __PYX_ERR(9, 178, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Decimal256Type = (struct __pyx_vtabstruct_7pyarrow_3lib_Decimal256Type*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal256Type); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Decimal256Type)) __PYX_ERR(9, 178, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_RunEndEncodedType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "RunEndEncodedType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_RunEndEncodedType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_RunEndEncodedType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_RunEndEncodedType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_RunEndEncodedType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_RunEndEncodedType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_RunEndEncodedType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_RunEndEncodedType) __PYX_ERR(9, 183, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_RunEndEncodedType = (struct __pyx_vtabstruct_7pyarrow_3lib_RunEndEncodedType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_RunEndEncodedType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_RunEndEncodedType)) __PYX_ERR(9, 183, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_BaseExtensionType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "BaseExtensionType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_BaseExtensionType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BaseExtensionType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_BaseExtensionType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BaseExtensionType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_BaseExtensionType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BaseExtensionType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BaseExtensionType) __PYX_ERR(9, 188, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_BaseExtensionType = (struct __pyx_vtabstruct_7pyarrow_3lib_BaseExtensionType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BaseExtensionType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_BaseExtensionType)) __PYX_ERR(9, 188, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_ExtensionType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "ExtensionType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_ExtensionType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ExtensionType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_ExtensionType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ExtensionType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_ExtensionType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ExtensionType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ExtensionType) __PYX_ERR(9, 193, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_ExtensionType = (struct __pyx_vtabstruct_7pyarrow_3lib_ExtensionType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ExtensionType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_ExtensionType)) __PYX_ERR(9, 193, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedShapeTensorType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "FixedShapeTensorType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedShapeTensorType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedShapeTensorType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedShapeTensorType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedShapeTensorType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedShapeTensorType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedShapeTensorType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedShapeTensorType) __PYX_ERR(9, 198, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_FixedShapeTensorType = (struct __pyx_vtabstruct_7pyarrow_3lib_FixedShapeTensorType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedShapeTensorType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_FixedShapeTensorType)) __PYX_ERR(9, 198, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Bool8Type = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Bool8Type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Bool8Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Bool8Type),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Bool8Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Bool8Type),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Bool8Type), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Bool8Type),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Bool8Type) __PYX_ERR(9, 202, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Bool8Type = (struct __pyx_vtabstruct_7pyarrow_3lib_Bool8Type*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Bool8Type); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Bool8Type)) __PYX_ERR(9, 202, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_OpaqueType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "OpaqueType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_OpaqueType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_OpaqueType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_OpaqueType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_OpaqueType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_OpaqueType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_OpaqueType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_OpaqueType) __PYX_ERR(9, 206, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_OpaqueType = (struct __pyx_vtabstruct_7pyarrow_3lib_OpaqueType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_OpaqueType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_OpaqueType)) __PYX_ERR(9, 206, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_UuidType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "UuidType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_UuidType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UuidType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_UuidType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UuidType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_UuidType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UuidType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UuidType) __PYX_ERR(9, 210, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_UuidType = (struct __pyx_vtabstruct_7pyarrow_3lib_UuidType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UuidType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_UuidType)) __PYX_ERR(9, 210, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_JsonType = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "JsonType",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_JsonType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_JsonType),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_JsonType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_JsonType),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_JsonType), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_JsonType),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_JsonType) __PYX_ERR(9, 214, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_JsonType = (struct __pyx_vtabstruct_7pyarrow_3lib_JsonType*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_JsonType); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_JsonType)) __PYX_ERR(9, 214, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib__Metadata = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "_Metadata",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib__Metadata), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__Metadata),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib__Metadata), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__Metadata),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib__Metadata), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__Metadata),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib__Metadata) __PYX_ERR(9, 219, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_KeyValueMetadata = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "KeyValueMetadata",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_KeyValueMetadata),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_KeyValueMetadata) __PYX_ERR(9, 225, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_KeyValueMetadata = (struct __pyx_vtabstruct_7pyarrow_3lib_KeyValueMetadata*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_KeyValueMetadata); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_KeyValueMetadata)) __PYX_ERR(9, 225, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Field = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Field",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Field), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Field),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Field), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Field),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Field), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Field),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Field) __PYX_ERR(9, 237, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Field = (struct __pyx_vtabstruct_7pyarrow_3lib_Field*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Field); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Field)) __PYX_ERR(9, 237, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Schema = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Schema",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Schema), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Schema),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Schema), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Schema),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Schema), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Schema),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Schema) __PYX_ERR(9, 248, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Schema = (struct __pyx_vtabstruct_7pyarrow_3lib_Schema*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Schema); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Schema)) __PYX_ERR(9, 248, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Scalar = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Scalar",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Scalar), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Scalar),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Scalar), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Scalar),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Scalar), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Scalar),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Scalar) __PYX_ERR(9, 257, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Scalar = (struct __pyx_vtabstruct_7pyarrow_3lib_Scalar*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Scalar); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Scalar)) __PYX_ERR(9, 257, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_ArrayStatistics = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "ArrayStatistics",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_ArrayStatistics), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ArrayStatistics),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_ArrayStatistics), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ArrayStatistics),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_ArrayStatistics), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ArrayStatistics),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ArrayStatistics) __PYX_ERR(9, 269, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_ArrayStatistics = (struct __pyx_vtabstruct_7pyarrow_3lib_ArrayStatistics*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ArrayStatistics); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_ArrayStatistics)) __PYX_ERR(9, 269, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib__PandasConvertible = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "_PandasConvertible",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib__PandasConvertible), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__PandasConvertible),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib__PandasConvertible), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__PandasConvertible),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib__PandasConvertible), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__PandasConvertible),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib__PandasConvertible) __PYX_ERR(9, 277, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Array) __PYX_ERR(9, 281, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Array = (struct __pyx_vtabstruct_7pyarrow_3lib_Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Array)) __PYX_ERR(9, 281, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Tensor = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Tensor",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Tensor), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Tensor),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Tensor), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Tensor),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Tensor), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Tensor),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Tensor) __PYX_ERR(9, 297, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Tensor = (struct __pyx_vtabstruct_7pyarrow_3lib_Tensor*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Tensor); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Tensor)) __PYX_ERR(9, 297, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCSRMatrix = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "SparseCSRMatrix",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCSRMatrix), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCSRMatrix),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCSRMatrix), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCSRMatrix),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCSRMatrix), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCSRMatrix),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCSRMatrix) __PYX_ERR(9, 310, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_SparseCSRMatrix = (struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSRMatrix*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCSRMatrix); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_SparseCSRMatrix)) __PYX_ERR(9, 310, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCSCMatrix = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "SparseCSCMatrix",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCSCMatrix), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCSCMatrix),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCSCMatrix), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCSCMatrix),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCSCMatrix), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCSCMatrix),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCSCMatrix) __PYX_ERR(9, 321, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_SparseCSCMatrix = (struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSCMatrix*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCSCMatrix); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_SparseCSCMatrix)) __PYX_ERR(9, 321, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCOOTensor = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "SparseCOOTensor",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCOOTensor), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCOOTensor),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCOOTensor), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCOOTensor),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCOOTensor), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCOOTensor),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCOOTensor) __PYX_ERR(9, 332, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_SparseCOOTensor = (struct __pyx_vtabstruct_7pyarrow_3lib_SparseCOOTensor*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCOOTensor); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_SparseCOOTensor)) __PYX_ERR(9, 332, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCSFTensor = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "SparseCSFTensor",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCSFTensor), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCSFTensor),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCSFTensor), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCSFTensor),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_SparseCSFTensor), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_SparseCSFTensor),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCSFTensor) __PYX_ERR(9, 343, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_SparseCSFTensor = (struct __pyx_vtabstruct_7pyarrow_3lib_SparseCSFTensor*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_SparseCSFTensor); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_SparseCSFTensor)) __PYX_ERR(9, 343, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_NullArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "NullArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_NullArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_NullArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_NullArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_NullArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_NullArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_NullArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_NullArray) __PYX_ERR(9, 354, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_NullArray = (struct __pyx_vtabstruct_7pyarrow_3lib_NullArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_NullArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_NullArray)) __PYX_ERR(9, 354, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_BooleanArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "BooleanArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_BooleanArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BooleanArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_BooleanArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BooleanArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_BooleanArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BooleanArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BooleanArray) __PYX_ERR(9, 358, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_BooleanArray = (struct __pyx_vtabstruct_7pyarrow_3lib_BooleanArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BooleanArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_BooleanArray)) __PYX_ERR(9, 358, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_NumericArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "NumericArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_NumericArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_NumericArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_NumericArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_NumericArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_NumericArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_NumericArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_NumericArray) __PYX_ERR(9, 362, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_NumericArray = (struct __pyx_vtabstruct_7pyarrow_3lib_NumericArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_NumericArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_NumericArray)) __PYX_ERR(9, 362, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_IntegerArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "IntegerArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_IntegerArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_IntegerArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_IntegerArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_IntegerArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_IntegerArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_IntegerArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_IntegerArray) __PYX_ERR(9, 366, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_IntegerArray = (struct __pyx_vtabstruct_7pyarrow_3lib_IntegerArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_IntegerArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_IntegerArray)) __PYX_ERR(9, 366, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_FloatingPointArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "FloatingPointArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_FloatingPointArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FloatingPointArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_FloatingPointArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FloatingPointArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_FloatingPointArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FloatingPointArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FloatingPointArray) __PYX_ERR(9, 370, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_FloatingPointArray = (struct __pyx_vtabstruct_7pyarrow_3lib_FloatingPointArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FloatingPointArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_FloatingPointArray)) __PYX_ERR(9, 370, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int8Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Int8Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int8Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int8Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int8Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int8Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int8Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int8Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int8Array) __PYX_ERR(9, 374, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Int8Array = (struct __pyx_vtabstruct_7pyarrow_3lib_Int8Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int8Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Int8Array)) __PYX_ERR(9, 374, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt8Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "UInt8Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt8Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt8Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt8Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt8Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt8Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt8Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt8Array) __PYX_ERR(9, 378, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_UInt8Array = (struct __pyx_vtabstruct_7pyarrow_3lib_UInt8Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt8Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_UInt8Array)) __PYX_ERR(9, 378, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int16Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Int16Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int16Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int16Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int16Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int16Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int16Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int16Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int16Array) __PYX_ERR(9, 382, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Int16Array = (struct __pyx_vtabstruct_7pyarrow_3lib_Int16Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int16Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Int16Array)) __PYX_ERR(9, 382, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt16Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "UInt16Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt16Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt16Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt16Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt16Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt16Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt16Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt16Array) __PYX_ERR(9, 386, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_UInt16Array = (struct __pyx_vtabstruct_7pyarrow_3lib_UInt16Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt16Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_UInt16Array)) __PYX_ERR(9, 386, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int32Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Int32Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int32Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int32Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int32Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int32Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int32Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int32Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int32Array) __PYX_ERR(9, 390, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Int32Array = (struct __pyx_vtabstruct_7pyarrow_3lib_Int32Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int32Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Int32Array)) __PYX_ERR(9, 390, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt32Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "UInt32Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt32Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt32Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt32Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt32Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt32Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt32Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt32Array) __PYX_ERR(9, 394, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_UInt32Array = (struct __pyx_vtabstruct_7pyarrow_3lib_UInt32Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt32Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_UInt32Array)) __PYX_ERR(9, 394, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int64Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Int64Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int64Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int64Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int64Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int64Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Int64Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Int64Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int64Array) __PYX_ERR(9, 398, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Int64Array = (struct __pyx_vtabstruct_7pyarrow_3lib_Int64Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Int64Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Int64Array)) __PYX_ERR(9, 398, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt64Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "UInt64Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt64Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt64Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt64Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt64Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_UInt64Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UInt64Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt64Array) __PYX_ERR(9, 402, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_UInt64Array = (struct __pyx_vtabstruct_7pyarrow_3lib_UInt64Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UInt64Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_UInt64Array)) __PYX_ERR(9, 402, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_HalfFloatArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "HalfFloatArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_HalfFloatArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_HalfFloatArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_HalfFloatArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_HalfFloatArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_HalfFloatArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_HalfFloatArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_HalfFloatArray) __PYX_ERR(9, 406, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_HalfFloatArray = (struct __pyx_vtabstruct_7pyarrow_3lib_HalfFloatArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_HalfFloatArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_HalfFloatArray)) __PYX_ERR(9, 406, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_FloatArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "FloatArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_FloatArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FloatArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_FloatArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FloatArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_FloatArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FloatArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FloatArray) __PYX_ERR(9, 410, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_FloatArray = (struct __pyx_vtabstruct_7pyarrow_3lib_FloatArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FloatArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_FloatArray)) __PYX_ERR(9, 410, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_DoubleArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "DoubleArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_DoubleArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DoubleArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_DoubleArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DoubleArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_DoubleArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DoubleArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DoubleArray) __PYX_ERR(9, 414, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_DoubleArray = (struct __pyx_vtabstruct_7pyarrow_3lib_DoubleArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DoubleArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_DoubleArray)) __PYX_ERR(9, 414, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "FixedSizeBinaryArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeBinaryArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryArray) __PYX_ERR(9, 418, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_FixedSizeBinaryArray = (struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeBinaryArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeBinaryArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_FixedSizeBinaryArray)) __PYX_ERR(9, 418, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal32Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Decimal32Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal32Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal32Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal32Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal32Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal32Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal32Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal32Array) __PYX_ERR(9, 422, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Decimal32Array = (struct __pyx_vtabstruct_7pyarrow_3lib_Decimal32Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal32Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Decimal32Array)) __PYX_ERR(9, 422, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal64Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Decimal64Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal64Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal64Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal64Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal64Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal64Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal64Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal64Array) __PYX_ERR(9, 426, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Decimal64Array = (struct __pyx_vtabstruct_7pyarrow_3lib_Decimal64Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal64Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Decimal64Array)) __PYX_ERR(9, 426, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal128Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Decimal128Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal128Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal128Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal128Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal128Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal128Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal128Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal128Array) __PYX_ERR(9, 430, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Decimal128Array = (struct __pyx_vtabstruct_7pyarrow_3lib_Decimal128Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal128Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Decimal128Array)) __PYX_ERR(9, 430, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal256Array = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Decimal256Array",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal256Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal256Array),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal256Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal256Array),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Decimal256Array), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Decimal256Array),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal256Array) __PYX_ERR(9, 434, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Decimal256Array = (struct __pyx_vtabstruct_7pyarrow_3lib_Decimal256Array*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Decimal256Array); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Decimal256Array)) __PYX_ERR(9, 434, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_StructArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "StructArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_StructArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StructArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_StructArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StructArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_StructArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StructArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StructArray) __PYX_ERR(9, 438, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_StructArray = (struct __pyx_vtabstruct_7pyarrow_3lib_StructArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StructArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_StructArray)) __PYX_ERR(9, 438, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_BaseListArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "BaseListArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_BaseListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BaseListArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_BaseListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BaseListArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_BaseListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BaseListArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BaseListArray) __PYX_ERR(9, 442, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_BaseListArray = (struct __pyx_vtabstruct_7pyarrow_3lib_BaseListArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BaseListArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_BaseListArray)) __PYX_ERR(9, 442, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "ListArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListArray) __PYX_ERR(9, 446, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_ListArray = (struct __pyx_vtabstruct_7pyarrow_3lib_ListArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_ListArray)) __PYX_ERR(9, 446, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "LargeListArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListArray) __PYX_ERR(9, 450, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_LargeListArray = (struct __pyx_vtabstruct_7pyarrow_3lib_LargeListArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_LargeListArray)) __PYX_ERR(9, 450, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListViewArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "ListViewArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListViewArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListViewArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_ListViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ListViewArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListViewArray) __PYX_ERR(9, 454, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_ListViewArray = (struct __pyx_vtabstruct_7pyarrow_3lib_ListViewArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ListViewArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_ListViewArray)) __PYX_ERR(9, 454, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListViewArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "LargeListViewArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListViewArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListViewArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_LargeListViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_LargeListViewArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListViewArray) __PYX_ERR(9, 458, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_LargeListViewArray = (struct __pyx_vtabstruct_7pyarrow_3lib_LargeListViewArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_LargeListViewArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_LargeListViewArray)) __PYX_ERR(9, 458, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_MapArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "MapArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_MapArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MapArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_MapArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MapArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_MapArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MapArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MapArray) __PYX_ERR(9, 462, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_MapArray = (struct __pyx_vtabstruct_7pyarrow_3lib_MapArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MapArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_MapArray)) __PYX_ERR(9, 462, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeListArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "FixedSizeListArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeListArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeListArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_FixedSizeListArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_FixedSizeListArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeListArray) __PYX_ERR(9, 466, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_FixedSizeListArray = (struct __pyx_vtabstruct_7pyarrow_3lib_FixedSizeListArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_FixedSizeListArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_FixedSizeListArray)) __PYX_ERR(9, 466, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_UnionArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "UnionArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_UnionArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UnionArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_UnionArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UnionArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_UnionArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_UnionArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UnionArray) __PYX_ERR(9, 470, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_UnionArray = (struct __pyx_vtabstruct_7pyarrow_3lib_UnionArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_UnionArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_UnionArray)) __PYX_ERR(9, 470, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_StringArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "StringArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_StringArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StringArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_StringArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StringArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_StringArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StringArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StringArray) __PYX_ERR(9, 474, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_StringArray = (struct __pyx_vtabstruct_7pyarrow_3lib_StringArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StringArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_StringArray)) __PYX_ERR(9, 474, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_BinaryArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "BinaryArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_BinaryArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BinaryArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_BinaryArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BinaryArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_BinaryArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BinaryArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BinaryArray) __PYX_ERR(9, 478, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_BinaryArray = (struct __pyx_vtabstruct_7pyarrow_3lib_BinaryArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BinaryArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_BinaryArray)) __PYX_ERR(9, 478, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_StringViewArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "StringViewArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_StringViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StringViewArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_StringViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StringViewArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_StringViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StringViewArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StringViewArray) __PYX_ERR(9, 482, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_StringViewArray = (struct __pyx_vtabstruct_7pyarrow_3lib_StringViewArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StringViewArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_StringViewArray)) __PYX_ERR(9, 482, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_BinaryViewArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "BinaryViewArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_BinaryViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BinaryViewArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_BinaryViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BinaryViewArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_BinaryViewArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BinaryViewArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BinaryViewArray) __PYX_ERR(9, 486, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_BinaryViewArray = (struct __pyx_vtabstruct_7pyarrow_3lib_BinaryViewArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BinaryViewArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_BinaryViewArray)) __PYX_ERR(9, 486, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_DictionaryArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "DictionaryArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_DictionaryArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DictionaryArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_DictionaryArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DictionaryArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_DictionaryArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_DictionaryArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DictionaryArray) __PYX_ERR(9, 490, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_DictionaryArray = (struct __pyx_vtabstruct_7pyarrow_3lib_DictionaryArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_DictionaryArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_DictionaryArray)) __PYX_ERR(9, 490, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_ExtensionArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "ExtensionArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_ExtensionArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ExtensionArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_ExtensionArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ExtensionArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_ExtensionArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ExtensionArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ExtensionArray) __PYX_ERR(9, 495, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_ExtensionArray = (struct __pyx_vtabstruct_7pyarrow_3lib_ExtensionArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ExtensionArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_ExtensionArray)) __PYX_ERR(9, 495, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_MonthDayNanoIntervalArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "MonthDayNanoIntervalArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_MonthDayNanoIntervalArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MonthDayNanoIntervalArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_MonthDayNanoIntervalArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MonthDayNanoIntervalArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_MonthDayNanoIntervalArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MonthDayNanoIntervalArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MonthDayNanoIntervalArray) __PYX_ERR(9, 499, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_MonthDayNanoIntervalArray = (struct __pyx_vtabstruct_7pyarrow_3lib_MonthDayNanoIntervalArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MonthDayNanoIntervalArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_MonthDayNanoIntervalArray)) __PYX_ERR(9, 499, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_ChunkedArray = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "ChunkedArray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_ChunkedArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ChunkedArray),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_ChunkedArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ChunkedArray),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_ChunkedArray), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ChunkedArray),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ChunkedArray) __PYX_ERR(9, 507, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_ChunkedArray = (struct __pyx_vtabstruct_7pyarrow_3lib_ChunkedArray*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ChunkedArray); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_ChunkedArray)) __PYX_ERR(9, 507, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib__Tabular = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "_Tabular",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib__Tabular), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__Tabular),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib__Tabular), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__Tabular),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib__Tabular), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__Tabular),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib__Tabular) __PYX_ERR(9, 522, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib__Tabular = (struct __pyx_vtabstruct_7pyarrow_3lib__Tabular*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib__Tabular); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib__Tabular)) __PYX_ERR(9, 522, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Table = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Table",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Table), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Table),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Table), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Table),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Table), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Table),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Table) __PYX_ERR(9, 526, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Table = (struct __pyx_vtabstruct_7pyarrow_3lib_Table*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Table); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Table)) __PYX_ERR(9, 526, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_RecordBatch = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "RecordBatch",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_RecordBatch), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_RecordBatch),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_RecordBatch), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_RecordBatch),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_RecordBatch), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_RecordBatch),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_RecordBatch) __PYX_ERR(9, 536, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_RecordBatch = (struct __pyx_vtabstruct_7pyarrow_3lib_RecordBatch*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_RecordBatch); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_RecordBatch)) __PYX_ERR(9, 536, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Device = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Device",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Device), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Device),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Device), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Device),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Device), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Device),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Device) __PYX_ERR(9, 545, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Device = (struct __pyx_vtabstruct_7pyarrow_3lib_Device*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Device); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Device)) __PYX_ERR(9, 545, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_MemoryManager = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "MemoryManager",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_MemoryManager), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MemoryManager),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_MemoryManager), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MemoryManager),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_MemoryManager), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_MemoryManager),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MemoryManager) __PYX_ERR(9, 557, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_MemoryManager = (struct __pyx_vtabstruct_7pyarrow_3lib_MemoryManager*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_MemoryManager); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_MemoryManager)) __PYX_ERR(9, 557, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Buffer = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Buffer",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Buffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Buffer),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Buffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Buffer),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Buffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Buffer),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Buffer) __PYX_ERR(9, 569, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Buffer = (struct __pyx_vtabstruct_7pyarrow_3lib_Buffer*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Buffer); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Buffer)) __PYX_ERR(9, 569, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_ResizableBuffer = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "ResizableBuffer",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_ResizableBuffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ResizableBuffer),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_ResizableBuffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ResizableBuffer),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_ResizableBuffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_ResizableBuffer),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ResizableBuffer) __PYX_ERR(9, 579, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_ResizableBuffer = (struct __pyx_vtabstruct_7pyarrow_3lib_ResizableBuffer*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_ResizableBuffer); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_ResizableBuffer)) __PYX_ERR(9, 579, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_NativeFile = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "NativeFile",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_NativeFile), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_NativeFile),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_NativeFile), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_NativeFile),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_NativeFile), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_NativeFile),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_NativeFile) __PYX_ERR(9, 584, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_NativeFile = (struct __pyx_vtabstruct_7pyarrow_3lib_NativeFile*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_NativeFile); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_NativeFile)) __PYX_ERR(9, 584, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_BufferedInputStream = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "BufferedInputStream",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_BufferedInputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BufferedInputStream),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_BufferedInputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BufferedInputStream),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_BufferedInputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BufferedInputStream),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BufferedInputStream) __PYX_ERR(9, 608, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_BufferedInputStream = (struct __pyx_vtabstruct_7pyarrow_3lib_BufferedInputStream*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BufferedInputStream); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_BufferedInputStream)) __PYX_ERR(9, 608, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_BufferedOutputStream = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "BufferedOutputStream",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_BufferedOutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BufferedOutputStream),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_BufferedOutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BufferedOutputStream),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_BufferedOutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_BufferedOutputStream),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BufferedOutputStream) __PYX_ERR(9, 612, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_BufferedOutputStream = (struct __pyx_vtabstruct_7pyarrow_3lib_BufferedOutputStream*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_BufferedOutputStream); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_BufferedOutputStream)) __PYX_ERR(9, 612, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_CompressedInputStream = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "CompressedInputStream",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_CompressedInputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_CompressedInputStream),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_CompressedInputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_CompressedInputStream),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_CompressedInputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_CompressedInputStream),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_CompressedInputStream) __PYX_ERR(9, 616, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_CompressedInputStream = (struct __pyx_vtabstruct_7pyarrow_3lib_CompressedInputStream*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_CompressedInputStream); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_CompressedInputStream)) __PYX_ERR(9, 616, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_CompressedOutputStream = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "CompressedOutputStream",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_CompressedOutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_CompressedOutputStream),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_CompressedOutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_CompressedOutputStream),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_CompressedOutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_CompressedOutputStream),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_CompressedOutputStream) __PYX_ERR(9, 620, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_CompressedOutputStream = (struct __pyx_vtabstruct_7pyarrow_3lib_CompressedOutputStream*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_CompressedOutputStream); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_CompressedOutputStream)) __PYX_ERR(9, 620, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib__CRecordBatchWriter = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "_CRecordBatchWriter",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib__CRecordBatchWriter), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__CRecordBatchWriter),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib__CRecordBatchWriter), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__CRecordBatchWriter),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib__CRecordBatchWriter), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib__CRecordBatchWriter),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib__CRecordBatchWriter) __PYX_ERR(9, 624, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_RecordBatchReader = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "RecordBatchReader",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_RecordBatchReader), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_RecordBatchReader),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_RecordBatchReader), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_RecordBatchReader),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_RecordBatchReader), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_RecordBatchReader),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_RecordBatchReader) __PYX_ERR(9, 629, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_CacheOptions = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "CacheOptions",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_CacheOptions), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_CacheOptions),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_CacheOptions), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_CacheOptions),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_CacheOptions), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_CacheOptions),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_CacheOptions) __PYX_ERR(9, 634, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_CacheOptions = (struct __pyx_vtabstruct_7pyarrow_3lib_CacheOptions*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_CacheOptions); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_CacheOptions)) __PYX_ERR(9, 634, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_Codec = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "Codec",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_Codec), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Codec),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_Codec), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Codec),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_Codec), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_Codec),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Codec) __PYX_ERR(9, 646, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_Codec = (struct __pyx_vtabstruct_7pyarrow_3lib_Codec*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_Codec); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_Codec)) __PYX_ERR(9, 646, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_3lib_StopToken = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow.lib", "StopToken",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_3lib_StopToken), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StopToken),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_3lib_StopToken), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StopToken),
  #else
  sizeof(struct __pyx_obj_7pyarrow_3lib_StopToken), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_3lib_StopToken),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StopToken) __PYX_ERR(9, 654, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_3lib_StopToken = (struct __pyx_vtabstruct_7pyarrow_3lib_StopToken*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_3lib_StopToken); if (unlikely(!__pyx_vtabptr_7pyarrow_3lib_StopToken)) __PYX_ERR(9, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyarrow._parquet"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_FileEncryptionProperties = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow._parquet", "FileEncryptionProperties",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties),
  #else
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_FileEncryptionProperties),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_FileEncryptionProperties) __PYX_ERR(5, 25, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_8_parquet_FileEncryptionProperties = (struct __pyx_vtabstruct_7pyarrow_8_parquet_FileEncryptionProperties*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_FileEncryptionProperties); if (unlikely(!__pyx_vtabptr_7pyarrow_8_parquet_FileEncryptionProperties)) __PYX_ERR(5, 25, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_ParquetSchema = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow._parquet", "ParquetSchema",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_ParquetSchema), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_ParquetSchema),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_ParquetSchema), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_ParquetSchema),
  #else
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_ParquetSchema), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_ParquetSchema),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_ParquetSchema) __PYX_ERR(5, 78, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_FileMetaData = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow._parquet", "FileMetaData",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData),
  #else
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_FileMetaData),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_FileMetaData) __PYX_ERR(5, 83, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_8_parquet_FileMetaData = (struct __pyx_vtabstruct_7pyarrow_8_parquet_FileMetaData*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_FileMetaData); if (unlikely(!__pyx_vtabptr_7pyarrow_8_parquet_FileMetaData)) __PYX_ERR(5, 83, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_RowGroupMetaData = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow._parquet", "RowGroupMetaData",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData),
  #else
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_RowGroupMetaData),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_RowGroupMetaData) __PYX_ERR(5, 93, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_8_parquet_RowGroupMetaData = (struct __pyx_vtabstruct_7pyarrow_8_parquet_RowGroupMetaData*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_RowGroupMetaData); if (unlikely(!__pyx_vtabptr_7pyarrow_8_parquet_RowGroupMetaData)) __PYX_ERR(5, 93, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_ColumnChunkMetaData = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow._parquet", "ColumnChunkMetaData",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData),
  #else
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_ColumnChunkMetaData),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_ColumnChunkMetaData) __PYX_ERR(5, 109, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_8_parquet_ColumnChunkMetaData = (struct __pyx_vtabstruct_7pyarrow_8_parquet_ColumnChunkMetaData*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_ColumnChunkMetaData); if (unlikely(!__pyx_vtabptr_7pyarrow_8_parquet_ColumnChunkMetaData)) __PYX_ERR(5, 109, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_Statistics = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow._parquet", "Statistics",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_Statistics), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_Statistics),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_Statistics), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_Statistics),
  #else
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_Statistics), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_Statistics),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_Statistics) __PYX_ERR(5, 120, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_8_parquet_Statistics = (struct __pyx_vtabstruct_7pyarrow_8_parquet_Statistics*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_Statistics); if (unlikely(!__pyx_vtabptr_7pyarrow_8_parquet_Statistics)) __PYX_ERR(5, 120, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_GeoStatistics = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow._parquet", "GeoStatistics",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics),
  #else
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_GeoStatistics),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_GeoStatistics) __PYX_ERR(5, 130, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_8_parquet_GeoStatistics = (struct __pyx_vtabstruct_7pyarrow_8_parquet_GeoStatistics*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_GeoStatistics); if (unlikely(!__pyx_vtabptr_7pyarrow_8_parquet_GeoStatistics)) __PYX_ERR(5, 130, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_FileDecryptionProperties = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyarrow._parquet", "FileDecryptionProperties",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties),
  #else
  sizeof(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyarrow_8_parquet_FileDecryptionProperties),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_FileDecryptionProperties) __PYX_ERR(5, 140, __pyx_L1_error)
  __pyx_vtabptr_7pyarrow_8_parquet_FileDecryptionProperties = (struct __pyx_vtabstruct_7pyarrow_8_parquet_FileDecryptionProperties*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyarrow_8_parquet_FileDecryptionProperties); if (unlikely(!__pyx_vtabptr_7pyarrow_8_parquet_FileDecryptionProperties)) __PYX_ERR(5, 140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  {
    __pyx_t_1 = PyImport_ImportModule("pyarrow.lib"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_import_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_PyObject_PyObject_bool_std_share);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_import_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_import_name = __pyx_import_signature + 79;
    void (**const __pyx_import_pointers[])(void) = {(void (**)(void))&__pyx_f_7pyarrow_3lib_get_reader, (void (**)(void)) NULL};
    void (**const *__pyx_import_pointer)(void) = __pyx_import_pointers;
    const char *__pyx_import_current_signature = __pyx_import_signature;
    while (*__pyx_import_pointer) {
      if (__Pyx_ImportFunction_3_2_4(__pyx_t_1, __pyx_import_name, *__pyx_import_pointer, __pyx_import_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_import_pointer;
      __pyx_import_name = strchr(__pyx_import_name, '\0') + 1;
      __pyx_import_signature = strchr(__pyx_import_signature, '\0') + 1;
      if (*__pyx_import_signature != '\0') __pyx_import_current_signature = __pyx_import_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_jollyjack_cython(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_jollyjack_cython},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "jollyjack_cython",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_jollyjack_cython(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_jollyjack_cython(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_jollyjack_cython(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'jollyjack_cython' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "jollyjack_cython" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_jollyjack_cython", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_jollyjack__jollyjack_cython) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "jollyjack.jollyjack_cython")) {
      if (unlikely((PyDict_SetItemString(modules, "jollyjack.jollyjack_cython", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  (void)__Pyx_modinit_type_init_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/

  /* "jollyjack/jollyjack_cython.pyx":3
 * # distutils: include_dirs = .
 * 
 * import os             # <<<<<<<<<<<<<<
 * import cython
 * import pyarrow.parquet as pq
*/

#line 3 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_os, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)

#line 3 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 3 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 3 "jollyjack/jollyjack_cython.pyx"
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_os, __pyx_t_2) < (0)) __PYX_ERR(0, 3, __pyx_L1_error)

#line 3 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "jollyjack/jollyjack_cython.pyx":5
 * import os
 * import cython
 * import pyarrow.parquet as pq             # <<<<<<<<<<<<<<
 * cimport numpy as cnp
 * 
*/

#line 5 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyarrow_parquet, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)

#line 5 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_parquet);

#line 5 "jollyjack/jollyjack_cython.pyx"
  Py_DECREF(__pyx_t_1);

#line 5 "jollyjack/jollyjack_cython.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)

#line 5 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_1 = __pyx_t_3;

#line 5 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 5 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "jollyjack/jollyjack_cython.pyx"
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pq, __pyx_t_2) < (0)) __PYX_ERR(0, 5, __pyx_L1_error)

#line 5 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "jollyjack/jollyjack_cython.pyx":17
 * from pyarrow._parquet cimport *
 * from pyarrow.lib cimport (CacheOptions, NativeFile, get_reader)
 * from collections.abc import Iterable             # <<<<<<<<<<<<<<
 * 
 * def is_iterable_of_iterables(obj):
*/

#line 17 "jollyjack/jollyjack_cython.pyx"
  {

#line 17 "jollyjack/jollyjack_cython.pyx"
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_Iterable};

#line 17 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_collections_abc, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)

#line 17 "jollyjack/jollyjack_cython.pyx"
  }

#line 17 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 17 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 17 "jollyjack/jollyjack_cython.pyx"
  {

#line 17 "jollyjack/jollyjack_cython.pyx"
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_Iterable};

#line 17 "jollyjack/jollyjack_cython.pyx"
    __pyx_t_4 = 0; {

#line 17 "jollyjack/jollyjack_cython.pyx"
      __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_4]); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 17, __pyx_L1_error)

#line 17 "jollyjack/jollyjack_cython.pyx"
      __Pyx_GOTREF(__pyx_t_5);

#line 17 "jollyjack/jollyjack_cython.pyx"
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_4], __pyx_t_5) < (0)) __PYX_ERR(0, 17, __pyx_L1_error)

#line 17 "jollyjack/jollyjack_cython.pyx"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 17 "jollyjack/jollyjack_cython.pyx"
    }

#line 17 "jollyjack/jollyjack_cython.pyx"
  }

#line 17 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "jollyjack/jollyjack_cython.pyx":19
 * from collections.abc import Iterable
 * 
 * def is_iterable_of_iterables(obj):             # <<<<<<<<<<<<<<
 *     return isinstance(obj, Iterable) and isinstance(obj[0], Iterable) and not isinstance(obj[0], str)
 * 
*/

#line 19 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9jollyjack_16jollyjack_cython_1is_iterable_of_iterables, 0, __pyx_mstate_global->__pyx_n_u_is_iterable_of_iterables, NULL, __pyx_mstate_global->__pyx_n_u_jollyjack_jollyjack_cython, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 19 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000

#line 19 "jollyjack/jollyjack_cython.pyx"
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);

#line 19 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 19 "jollyjack/jollyjack_cython.pyx"
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_is_iterable_of_iterables, __pyx_t_2) < (0)) __PYX_ERR(0, 19, __pyx_L1_error)

#line 19 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "jollyjack/jollyjack_cython.pyx":22
 *     return isinstance(obj, Iterable) and isinstance(obj[0], Iterable) and not isinstance(obj[0], str)
 * 
 * cpdef void read_into_torch (object source, FileMetaData metadata, tensor, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):             # <<<<<<<<<<<<<<
 * 
 *     import torch
*/

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_mstate_global->__pyx_k_ = __pyx_t_2;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_mstate_global->__pyx_k__2 = __pyx_t_2;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_mstate_global->__pyx_k__3 = __pyx_t_2;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_7 = PyTuple_Pack(7, __pyx_t_2, __pyx_t_5, __pyx_t_6, Py_False, Py_True, Py_False, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_9jollyjack_16jollyjack_cython_3read_into_torch, 0, __pyx_mstate_global->__pyx_n_u_read_into_torch, NULL, __pyx_mstate_global->__pyx_n_u_jollyjack_jollyjack_cython, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 22 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000

#line 22 "jollyjack/jollyjack_cython.pyx"
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_6);

#line 22 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_6, __pyx_t_7);

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 22 "jollyjack/jollyjack_cython.pyx"
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_read_into_torch, __pyx_t_6) < (0)) __PYX_ERR(0, 22, __pyx_L1_error)

#line 22 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "jollyjack/jollyjack_cython.pyx":41
 *     return
 * 
 * cpdef void read_into_numpy (object source, FileMetaData metadata, cnp.ndarray np_array, row_group_indices, row_ranges = [], column_indices = [], column_names = [], pre_buffer = False, use_threads = True, use_memory_map = False, CacheOptions cache_options = None):             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[int] crow_group_indices = row_group_indices
*/

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_mstate_global->__pyx_k__4 = __pyx_t_6;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GIVEREF(__pyx_t_6);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_6 = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_mstate_global->__pyx_k__5 = __pyx_t_6;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GIVEREF(__pyx_t_6);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_6 = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_mstate_global->__pyx_k__6 = __pyx_t_6;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GIVEREF(__pyx_t_6);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_6 = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_2 = PyTuple_Pack(7, __pyx_t_6, __pyx_t_7, __pyx_t_5, Py_False, Py_True, Py_False, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9jollyjack_16jollyjack_cython_5read_into_numpy, 0, __pyx_mstate_global->__pyx_n_u_read_into_numpy, NULL, __pyx_mstate_global->__pyx_n_u_jollyjack_jollyjack_cython, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 41 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000

#line 41 "jollyjack/jollyjack_cython.pyx"
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);

#line 41 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_t_2);

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 41 "jollyjack/jollyjack_cython.pyx"
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_read_into_numpy, __pyx_t_5) < (0)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "jollyjack/jollyjack_cython.pyx":153
 *         return
 * 
 * cpdef void copy_to_torch_row_major (src_tensor, dst_tensor, row_indices):             # <<<<<<<<<<<<<<
 *     import torch
 * 
*/

#line 153 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9jollyjack_16jollyjack_cython_7copy_to_torch_row_major, 0, __pyx_mstate_global->__pyx_n_u_copy_to_torch_row_major, NULL, __pyx_mstate_global->__pyx_n_u_jollyjack_jollyjack_cython, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 153, __pyx_L1_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 153 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000

#line 153 "jollyjack/jollyjack_cython.pyx"
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);

#line 153 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 153 "jollyjack/jollyjack_cython.pyx"
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_copy_to_torch_row_major, __pyx_t_5) < (0)) __PYX_ERR(0, 153, __pyx_L1_error)

#line 153 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "jollyjack/jollyjack_cython.pyx":161
 *     )
 * 
 * cpdef void copy_to_numpy_row_major (cnp.ndarray src_array, cnp.ndarray dst_array, row_indices):             # <<<<<<<<<<<<<<
 * 
 *     if src_array.shape[0] == 0:
*/

#line 161 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9jollyjack_16jollyjack_cython_9copy_to_numpy_row_major, 0, __pyx_mstate_global->__pyx_n_u_copy_to_numpy_row_major, NULL, __pyx_mstate_global->__pyx_n_u_jollyjack_jollyjack_cython, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 161, __pyx_L1_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 161 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000

#line 161 "jollyjack/jollyjack_cython.pyx"
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);

#line 161 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 161 "jollyjack/jollyjack_cython.pyx"
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_copy_to_numpy_row_major, __pyx_t_5) < (0)) __PYX_ERR(0, 161, __pyx_L1_error)

#line 161 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "jollyjack/jollyjack_cython.pyx":1
 * # distutils: include_dirs = .             # <<<<<<<<<<<<<<
 * 
 * import os
*/

#line 1 "jollyjack/jollyjack_cython.pyx"
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "jollyjack/jollyjack_cython.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 1 "jollyjack/jollyjack_cython.pyx"
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_5) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "jollyjack/jollyjack_cython.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 1 "jollyjack/jollyjack_cython.pyx"


#line 1 "jollyjack/jollyjack_cython.pyx"
  /*--- Wrapped vars code ---*/

#line 1 "jollyjack/jollyjack_cython.pyx"


#line 1 "jollyjack/jollyjack_cython.pyx"
  goto __pyx_L0;

#line 1 "jollyjack/jollyjack_cython.pyx"
  __pyx_L1_error:;

#line 1 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "jollyjack/jollyjack_cython.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 1 "jollyjack/jollyjack_cython.pyx"
  if (__pyx_m) {

#line 1 "jollyjack/jollyjack_cython.pyx"
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {

#line 1 "jollyjack/jollyjack_cython.pyx"
      __Pyx_AddTraceback("init jollyjack.jollyjack_cython", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "jollyjack/jollyjack_cython.pyx"
    }

#line 1 "jollyjack/jollyjack_cython.pyx"
    #if !CYTHON_USE_MODULE_STATE

#line 1 "jollyjack/jollyjack_cython.pyx"
    Py_CLEAR(__pyx_m);

#line 1 "jollyjack/jollyjack_cython.pyx"
    #else

#line 1 "jollyjack/jollyjack_cython.pyx"
    Py_DECREF(__pyx_m);

#line 1 "jollyjack/jollyjack_cython.pyx"
    if (pystate_addmodule_run) {

#line 1 "jollyjack/jollyjack_cython.pyx"
      PyObject *tp, *value, *tb;

#line 1 "jollyjack/jollyjack_cython.pyx"
      PyErr_Fetch(&tp, &value, &tb);

#line 1 "jollyjack/jollyjack_cython.pyx"
      PyState_RemoveModule(&__pyx_moduledef);

#line 1 "jollyjack/jollyjack_cython.pyx"
      PyErr_Restore(tp, value, tb);

#line 1 "jollyjack/jollyjack_cython.pyx"
    }

#line 1 "jollyjack/jollyjack_cython.pyx"
    #endif

#line 1 "jollyjack/jollyjack_cython.pyx"
  } else if (!PyErr_Occurred()) {

#line 1 "jollyjack/jollyjack_cython.pyx"
    PyErr_SetString(PyExc_ImportError, "init jollyjack.jollyjack_cython");

#line 1 "jollyjack/jollyjack_cython.pyx"
  }

#line 1 "jollyjack/jollyjack_cython.pyx"
  __pyx_L0:;

#line 1 "jollyjack/jollyjack_cython.pyx"
  __Pyx_RefNannyFinishContext();

#line 1 "jollyjack/jollyjack_cython.pyx"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1 "jollyjack/jollyjack_cython.pyx"
  return (__pyx_m != NULL) ? 0 : -1;

#line 1 "jollyjack/jollyjack_cython.pyx"
  #else

#line 1 "jollyjack/jollyjack_cython.pyx"
  return __pyx_m;

#line 1 "jollyjack/jollyjack_cython.pyx"
  #endif

#line 1 "jollyjack/jollyjack_cython.pyx"
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_min); if (!__pyx_builtin_min) __PYX_ERR(0, 174, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_max); if (!__pyx_builtin_max) __PYX_ERR(0, 175, __pyx_L1_error)

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);
  __Pyx_RefNannyFinishContext();
  return 0;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 10; } index[] = {{17},{5},{53},{48},{51},{55},{179},{18},{12},{11},{56},{27},{31},{26},{27},{30},{1},{1},{3},{4},{1},{8},{54},{22},{19},{34},{30},{39},{34},{41},{42},{5},{8},{17},{20},{18},{13},{18},{15},{14},{12},{23},{23},{9},{10},{5},{6},{7},{6},{8},{3},{8},{16},{13},{24},{5},{26},{4},{8},{3},{8},{3},{10},{8},{8},{14},{5},{3},{2},{7},{3},{2},{10},{15},{14},{12},{15},{15},{17},{11},{10},{12},{10},{6},{9},{10},{5},{4},{4},{6},{8},{5},{14},{11},{4},{6},{89},{26},{48},{97},{967},{545}};
    #if (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (2204 bytes) */
const char* const cstring = "BZh91AY&SY\345\023\037\243\000\005^\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\300@@@@@@@@@@@@\000@\000`\007_S\336\246\275\327mn\235\333eS\270\352wnx\037O!\251\242*\030\322\037\250OM\2056\243L\247\251\3514\361M\246G\244\3054\315#\324\365\032\031==SA\241\247\243D\362\032\231\251\355&\210mOQ\352= \r\021\031\023L 4\312\036\"zS\304i\031\244\000h\000\006\200\000\323@\000h\000\032\000z\215\000h@4\202D\362F@\003@\000m \304\001\240\000\320\000\001\240\364\200\000\320\032\r\0004\004\231B\236i5=OS\323!4\3654\321\246&\203F\200\006\201\220\000\000\0004\000\000\001\246\200\002\2512a14\302mL\2320\002`F\t\202d\300\000L\230M0\230\2004`\206\004a0\230M0\224\324 h\206\242m1\240\204\365O\325\007\265\031\251\246\240h4\001\240\000\017Ph\000\001\352\000\006\215\014\232zmSH\314\023R\232\324\030f\333G_\334\03491\372\376\3757z\234\035F\352\235a\270\237\325\202<\377C\366\325\324o\316\033\327\364)\320\331\333\243\355\245M\335_\215Z\337.\024\326PR\255_u\\6\030]qz\273\274%^*\252\275g\216\314\0300b\305=\357\233\245\217d\376$\222Yt\206\220\022\004\204\231\220\223d\2377\035?L\231\306\343\346\207\035\017\323\316\346s\202Zf\252\027\002\342a\222\0059\030\300\200\216\355\252t{\300u\220 \350\247)\234mC6B\002 \373e\0106\340\020aDT\002\245ph\214@\007\214\013\031\210\210\024\035\275\004U\271\240HIPN\357x-Z\305x|\334\301Z\224\252N\010w\312\257\273\254D\256\363V2\2502\334<\374\004\237$\350\360E\005Q\024Hq<C\217\366\367\036\244f\2157\233Q\003L\271\217B\333\177\331\006>\r\006\275\322\0254\242\231\014{-\263\3238\360\201\355BAE\225\014\356%\345M\247\n\024\006\022K,2\271\207UF\023\000I\276M\255\211\214_\312i\220C\272\202\260T[\211L\225L\240\3650Bk\264tp\266\351\276\331NK\010\344\266\373\004\232\023\2453\357\362\311\000\322\35282D\200[\277V\020\312\252\360\302\212\374\014&)\024\010\233zN\001\324lu\260\276\302f\033G\217'\t\3571*\311\216o\343P\275_\034&|,0M\266\200\022\335\211\250\265\366-\003\n\256\253S\030@R0\266\327_G\374R\001\245\034\017o\232\020\031\220\022&\361\340\2436<m\347\204""\301\320\205\014\350\343\325\351\222\235\361d?\244!=\024\024\207\003\373J\230\231\343\2144FS\241\243\306\024\222bd\261(6\373\253\345\370\311\371\272\275!<'\215o\242=\031\305\267vI:g\266s,?<\334\375\203^?F\266\3622\334\013SBcdF\261i\203\010\342HL\035\311\364}qBP\206X\370%\257\213\331\270/\363\366\305XJ\020B\032\257\354B\330\244\360\020\224\"\240\231\204\206\002\003AtFDw_\r\311\244\370\247\302\266\360vF\307\2767n)\020\322f\275'\2254\026+\366\373\214\030\230,i8\316\013AFu<\0003\030wS\345\342\212\004\335H\305ol\232\346N\007\362\204n\242\262O\367\263\361\010A=r\236u\245\034\220\347j:\374C\024(\304\016\230\2231\004\343\272\213\206\301$o\220\2137#\220\360\326\304\341\207.{Cn\222y\010\217\244qw>\035c\373\362\261c\312\030,\324D\354\265>\340\341\3444\276\263\213\013\303Q\305\354}\356\274\305\026\213+2\274;h\005+h\373&\014l\tdtd\364\215I\350\332*u\262rU0\310\000m\313\006\270\263K:k4\365\031\234\232F\315\264O<\272\022\211\243\251\213\250/2\260\024)\030xO\264}\305d\026\354\220\305\312\323(y\311\014a\"\230\231\0108\231q\267B\244M\314\2163\226%\342\244\300\223<\325\n\355`0\022\326\212\320P\347\r\226\007`\367\\W\344}\306\203\205\3302\0102R\020\026\202U\305\241R\311\032&\202\014\313g\020\245\222\300\254\224\263\033\302\350\274\001\213\204\272\007j\010\304(\342\204\210\002\010\264H\005\026\021\222B\026p\315]!H\343JJPK\265\223\316\262#$\324\005\241R\241\323C\301\253;Bt\002\214$$\020\3432\"\264J\331\345O\025\234R\270\032\216@=P\004Q\021\n\302\324\020N\352\001J\224\030e4\\\256u\210\035\023$\230#\030T\004\323>\206\251\rm\362\253_c`\013\247*dPk\313\245\014\206e%\027\264C\n\2712\250\023\202z\004\027\025t\r\352\344,\261\206\034\200-\212\026\034\242\273\360\220Z\213\005\261\225\234\253SDB\344\205\014=+\211\232\374\306\211EU\255Jh\220A\031\215\024\000\361\001\311^\357\206FZ%C\211!\214\203P\002\312!A%w&)\021\2601p*\004\000\322\352\224\211\204\257c \320JH\346X\0067`M\003\2004F\225\225C\2448)D\322\"\315\021\025\344\216\025FE\001\250\020\"{\262\332\210:\250^\004\006\324\346\340\2579\3040f\265""\030\024\305\253\221\234\366O\201\266\207LvB\271.\273\033\232\266\253?\030\226[\201\241$\301A\014R\313\210,\002\220\344WYE\367\341\342\206\031\013\364\316\017+q\013\227i\247b3\217\306nh\314i9\241\235;\270\r\244\013\316\337z\241\203(PH2AFUSL\0001\010X%\221\nk\232\305\2331\206Jj\211\031t\016%\355m\221\246\332\0255\003\023\214\225\217\345\226n\233;u\2251\3526\301\030\252\362\026aFiB\315@\3301}\240\241TR\026\200\300\271\342\240Kz\252\020jx\305c0p\020\205\224\250\303i&\324\272\365\304\266E\201,S\313\267\276\034\252\0368\351\231\334\204\026\321\005@\232\304e\231\021TX\200\001\236\366\305\340g\307p\324\204\326A5\242{\351\315DER)\320t\320\tT\316HQ\202<6\n\301|3s\234[\247\315S\263\324\346mh\354\n\320\262\203\366\240rsX\000\301e\320\272n\030\024\301\334\205\016\326REagT\031\317\202G\335\234H\3405=\2140\337\330\331*X\307\202D\323\034\374\301\236\365\014ap\346n\203ex>\001\201\317\275|\034\373K\313\314FC[A\236\323\343\255_\014%\353\373=\010\350k\323?~\356J\241\247{\203V\276\023\340\"+\006p\214b5\215c9\366\001\223,/\206\366\010\321<A\356\034\036\263\017\276\206\036xr\336\365\262\331 \257&\022\017\302fYB.\320\367\204\377gx*\231c\213\201\251\351\206\030\026\366\017\247^W\023\224\237\356\020I\"\374\300\246.\222\016)\276\002 \357\375+ u\353\354x\000\202<{\204\336\231.*\010<\276@\221\0020\204\n\350\265G \004\001\222\023\200\273\363\240|\371\354\2774\2564\215%\353\326\231!\354\033\350:\333SY\233c\373\202\033`,%\034\020\"1\021J\376Y\221;k\237}\331\\\177'\302\032\014\346\214\322|\370\276\00431\304\362\2238c\221\250F\216F\333 \364\303$\2038\323\247e\003e\221\0209\342T\215\245\245\006D\320\347w\275\r\007d\0358t\016H\327\\\341K\205\355\265\027\002\300\202\304\010\304@\3316\\\251\332\244\330E\324\177r\205^\2631\036\200\337\r\231\\\177\204*1>b\363\035:I\214\273\365zz\370\327\212\010z9\370\021BP\362\016\244\275X\251\331\023\332\032\315r\016\325\253C\230B\346\031@(\030\217\n\374\016Q\251ZX\217\035\351\243\t\215\3617\231yC\375F\002G)\021\t\004\204\216\360\217\240\203\034`\277\254\037\200\311F+""\026\301\r\346\212\020\315\333\211:`@\310\010A\004\214\237\253\210\004\217J\002cS\336\343e;m\320\007\323\205Wt\335\344\\\026X\333Xw\035S\227\264\333\20212i\202$\232@2\010\000R\301\361Y\335\300\270\352./\374]\311\024\341BC\224L~\214";
    PyObject *data = __Pyx_DecompressString(cstring, 2204, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (1990 bytes) */
const char* const cstring = "x\332\215T\317_\323\330\026\247\241BuP\037\240\010\210\316\355\350\033\220\221\216\025\364!o\324)P\024|\343\360C\361\327h\346&\271\265\251m\222\346&@\364\251,Yf\231e\226Yf\331e\227]\336e\226\374\t\376\t\357\334\004\n\370c\336\360)=\247\367\236\237\337\363\275\347\335\265\367H\267-\244\227\220\244\333\232BQ\3666\272^T\25521\221\254W\355\232&\252\232\242\312\204\"\275}\242\341\032\374\326\010\001{KG\022A\224X\305-\203\310\026Q\0206M\354 UC\030-\350\246eb\r\215%\216\0235\\\321\315+\020I!f\333^!\324R5l\251\272v\330w\016\215\231\372\346\327]\250n\2332\371\273\231\036\352\026AV\031[h\316\261\312\220E\245\220\264\252J\304\304\026\251:\210Z\246\nqMn\244\241\345\342\362\304\324\364\024\302\232\202LR\201\214\024Q[\222\253\230R\016\003@e\253U(\031Y\216Ah\016-\226\220\243\3331\036\034\016\003\354\016;\000\226\032\007\210+h\024k\232n\305\315\212\340\256j\257G\221\242\232\220D\335 \334{\001W)\311\255\222\272\r\260$\001M\202\025\264\252oB\247\n\331B\267c\035\232}\r\341\326\366\220\320\276\202#Ee\014A\025\265T\"&\321,\244`\310\314k\276\212\036k\244\215?\265\304\330>\007\203\256\035\271\252B\345\320\257\231\344\346,8r\255\031\337r\244\246\374\331\025\265\r\003f\004wKK\342j\2610_\\\025g\013s\017\212\017\347o\347\356\242;\2103\357\nV\024Q\343\343J\306\010\351$;\001N\263k0/^\315\336\025\237;\034\032\316>\r(\032\345_\340\016|\330\300UUI@j\237\312\272f\251\257m\335\246\373'|\206\020\273]v\002\352\213kWQE\257V\235\n\226\337\374\334\326D9fO\316p\266\342\2749Q\326M\222\253\331\300\205\244\204\022V\253\311\310\324\032o\366\260\231]\303V\371\013\213\003\234h\031\033\344\305\265\227\200\304\301@\366\017\257\242\317\r\363/9^\237[\346\301\322\266J\023\323\213\300f,U\311\027X\213\342\262\263\005\377\363@x\361!\331\262VI\tSG\223U=\007u\302.P5Be,\227\211\250\033\234KT\256\302\021l\001\021\036\230L$@\002&P\345\214\205\313\034\226\344\243{\342\360\216\220u\303\021-]\214\201\020\201Fb\374.\367\217-P\313\007\307\355n\270b\021\215\302\021\247+\321d]\201\357\r\325\324\265\222n\002\222\242X\2625Y\024_""\023\250B\264MxG\373R\324\223\367$\252Tlw\004\272\272\007\211\250\227\332:\005\245F\333\023\316}>\3537\304\241\"\024\007\315\303\367V\215X\230\277\241Z\374[Wl\210\006\177\274U\370\336{\014\320k\334\322kHm\320\270s]\252\350\324\300&\274i\313\320\r\243n\230D\224l\376,\r\007\234\364\315\334\336\255(\002\273\304\r+.\024\376\3526\256&\341\371\022\000\210\367\301<\370\031\203\330N\270?\005~pH\215\211\r\255\300\032\332+\0274\005F\017\344M\226i\233`\\I\300\247\026\006\206Z\304\240\026P!>\022\341\n\206\223$\265)\021k\244\246\233\016\200c\360_V\231\327E\201\202\323\360\000a\203-;\277K|\203\242\361\261\003\025\036\265\256W\201\323\22623\003\3045\211\"\032\226\371\013\212\241\230\231Q\365\231\231U\330hz\255 Cat\001\336\014\354\207\361+\0350\356\030\010bn\247\242\357zv\323#\336o\301px\263\321\033\r \277\313\307\321\340\020\334\244\277\333\231qW\\\342\315\370\217\340:\037\316\205\365F\272\261\324TZ\227[\025\366\364\031{\366\007\373\343%{\211\031\226XG\205U>~\352\350\370UX\023@\254\t\n\027\212\340p\341\010\313\235 \226;-.\254N\273s\227\347=\343.y8:;\020\235=\357\335\364{\243\213\337G\347\006\243\363#\321\3200W\006\207\270>|\376SW\307\261.\326\261\305\003\245\036\362x\017\005\314\005\026L.L\341\036\017{\257\363\025\027\257:g\323 f\323s\351\335\364x\220\335M\377\340\217\005ca6\314G\207\325s\356\226W\217\240\367B\224\276\340\255$f?\206\251\360L\270\032~h\256\2652ly\235\255?aO\236\263\347\320\250\310D\336\350\247\023\035\3072\333\216\333\355\326\243\314Y\367)\367\354\336\376\350*\336\270_`\377\274\021\256D\231a/\027\034\013@\311\372\027\203\315\020G\307\007\2754\033\231\n\013\311m\326\233\362\352~\247?\345o\005un\325\353_\366\261O\203\313A\231[\237\2102\247w\250\233e\375\227\374I\277\034\224\300\361\021\273\365[k\245%\263\225U\016o\352\031\357\374\231\360\234\213\347\202\304\205$\224\271(\013*\027\252P\021\342d\274\276\367\356\274w\302\317\372w\302\336(3\350\245\274^o\024\262\216AF\022\376\322\260\233\005^\306H\360$,D\307\007\\\205\r_\217Q\212m\317x+\236\344w\373\365 \023\246\303\373""\215\302_\224\354\336\361\373\375\373\301\002xO\261\177=h\345[s\255:T\263\222z\314\213z,\254s\261.\274\342\342\325>I\010\027D(\t_\026\227\347\345\003H=;\013n6\3529\275c\272\347\274N\357\006\300w\335_\017&\301d2\224\032\335\r\263\331\327\234lJ\255\256\226\304\226W\331\352c\366\030f\007\364\374\223\375)1I\216\3761\340V\374\224\177!\300A}7\0164\340nz\004\246`\006\003\001\r\263G-2\027\342\266\323\376B\360S\370\250q\266\201!\305`\253\257\225\347|<\271\363\2737\351iA\021\350\"6\347\232V\353.{\376\202\337\364\354L\2737\275~o\216}\377\357F\276q\037\240\205\2274\355N{yo\316\243{\004\234\n\353\007\214\353\345~\231\355\017\356\023 #\214l\302\337\014\360nf\324\177\037\026>\365t\034\033\362\004o,\306\033\010\333\315\272\207 <\030\366\001\334\323\354\374OA>(D'F \372\273\340\022\233\270\333\240\315,\277\375\321\353\006\232\245\000S68\036\\\002B\236<\025\rL\004+\001\216\372\372\243>\340F,\377\306\247\347\344n\346\324N\305K\261\241[\215\336Fv\027\006\336\367\003\020*\037\235:\035\235\032pq,\377\337's|;\305\347\351\270].v\251w\311\203\367p|\027\360\271\345.\300\2575\206n7V\032j\263\036\375\305Y\026\220\227\375\241\344\261\262\0330_x\230\260~\014f\324Y}\223m&\233b\226\223jV\270\317\305}a\221\213EaI\370F\224x\211Io\330\033\010a2\323a\316[px\227\232\347~\363\340\007bIx\300\305\003\341?\020\245gg\221\277~\210\321\357/\006O\303\225\2204\362\337<\3469\357AF\325\337b?\253L\205r)\243\037\330\207dg\036\251nr\007\357\274\207i:A\0270\321b\223\217\330#`\363\323x\345\202#\024i1\353-{\373\216\275\373\357a\007\352\347\240\231\376\260\300n(L!\214hL\003K\233\331\033\007f\263P\303\006p@\ts\315T\263\277Y`\305\304\346\003\224\3601U\344\225\024\205\005.\026\204{\302n\372\262\017\014\316\002R{\213\363\220zq/\026NNc.\177\335\266M\230[\300\325#\2548\007\213\364\020A\376\007\r\373\025\001";
    PyObject *data = __Pyx_DecompressString(cstring, 1990, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (3404 bytes) */
const char* const bytes = "{0} out of bounds != 2Either column_indices or column_names needs to be setExpected array in a Fortran (column-major) orderExpected destination array in a C (row-major) orderExpected source array in a Fortran (column-major) orderNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.Requested to read Row index = Row range 'Source and destination arrays have different datatypes, Unexpected dst_array.ndim, Unexpected len of row indices, Unexpected np_array.ndim, Unexpected src_array.ndim, Unsupported JJ_READER_BACKEND=.? >  != )add_note columns, but the number of columns in numpy array is ' is not a valid range' is not contiguous is not in the expected range [0, jollyjack/jollyjack_cython.pyxnumpy._core.multiarray failed to importnumpy._core.umath failed to importsrc_array.shape[0] > dst_array.shape[0], src_array.shape[1] != dst_array.shape[1], utf-8IterableJJ_READER_BACKEND__Pyx_PyDict_NextRefasyncio.coroutinescache_optionscline_in_tracebackcollections.abccolumn_indicescolumn_namescopy_to_numpy_row_majorcopy_to_torch_row_majordst_arraydst_tensordtypeencodeenvironformat__func__getio_uringio_uring_odirect_is_coroutineis_iterable_of_iterablesitemsjollyjack.jollyjack_cythonkeys__main__maxmetadatamin__module____name__np_arraynum_row_groupsnumpyobjosparquetpoppqpre_bufferpyarrow.parquet__pyx_vtable____qualname__read_into_numpyread_into_torchrow_group_indicesrow_indicesrow_ranges__set_name__setdefaultsourcesrc_arraysrc_tensorstartstepstoptensor__test__torchuse_memory_mapuse_threadsutf8valuesPyObject *(PyObject *, bool, std::shared_ptr< arrow::io::RandomAccessFile>  *)\000get_reader\200\001\330\013\014\340\004\034\230M\250\032\2606\270\021\330\026 \240\006\240a\330\030\031\200\001\330\004\013\210:\220Q\220e\230:\240T\250\032\2601\260C\260q\270\004\270J\300d\310$\310j\320XY\320Y\\\320\\]\320]a\320ab\320\000j\320j\177\360\000\000@\002S\002\360\000\000S""\002d\002\360\000\000d\002y\002\360\000\000y\002P\003\360\000\000P\003t\003\360\000\000t\003u\003\340\013\014\340\004\024\220J\230a\330\025\026\330\025\033\2306\240\021\330\036\037\330\027\030\330\033\034\330\031\032\330\027\030\330\030\031\330\033\034\330\032\033\360\006\000\005\006\320\000x\360\000\000y\001N\002\360\000\000N\002a\002\360\000\000a\002r\002\360\000\000r\002G\003\360\000\000G\003^\003\360\000\000^\003B\004\360\000\000B\004C\004\340\004*\250!\340\004\"\240(\250(\260!\2601\330\004\"\240(\250(\260!\2601\330\004\027\220x\230q\330\004\034\230A\330\004\035\230Q\340\004\"\240(\250&\260\001\260\024\260R\260~\300S\310\010\320PV\320VW\320WZ\320Z\\\320\\_\320_a\320ab\360\n\000\005\010\200y\220\007\220q\330\010\025\220X\230Q\340\004\007\200\177\220d\230*\240A\320%5\260Q\330\010\032\230.\250\005\250Q\330\010!\240\036\250w\260a\330\t\030\230\004\320\0344\260A\260Q\330\010\032\230!\2304\230q\240\003\2404\240x\250q\330\010!\240\021\240$\240a\240s\250$\250h\260a\330\t\n\330\010\017\210s\220!\320\023#\2403\240h\250f\260A\260T\3209M\310Q\310c\320QR\360\000\000S\001Y\002\360\000\000Y\002Z\002\360\000\000Z\002b\002\360\000\000b\002h\002\360\000\000h\002i\002\360\000\000i\002j\002\330\010\032\230!\340\004\007\200}\220D\230\n\240!\240>\260\021\330\010\030\230\001\230\021\230'\240\021\240(\250$\250e\260<\270u\300A\330\010!\240\034\250W\260A\330\t\026\220d\320\0322\260!\2601\330\010\030\230\001\230\024\230Q\230b\240\007\240q\250\010\260\004\260H\270A\330\010!\240\021\240$\240a\240s\250$\250h\260a\330\t\n\330\010\017\210s\220!\220>\240\023\240H\250F\260!\2604\3207K\3101\310C\310q\360\000\000Q\001U\002\360\000\000U\002V\002\360\000\000V\002^\002\360\000\000^\002d\002\360\000\000d\002e\002\360\000\000e\002f\002\330\010\030\230\001\230\021\230'\240\021\240(\250$\250e\2601\340\004\007\200q\330\010\014\210F\220!\330\014\017\210r\220\027\230\003\2305\240\003\2402\240V\2503\250e\2603\260b\270\007\270r\300\022\3003\300b\310\006\310b\320PR\320RU\320UW\320W]\320]`\320`b\320bc\330\020""\026\220j\240\001\240\035\250a\250q\340\014\017\210r\220\026\220w\230e\2404\240r\250\026\250s\260!\330\020\026\220j\240\001\240\035\250a\250q\340\010\035\230Q\230b\240\004\240F\250+\260T\270\025\270a\270r\300\030\310\022\3101\360\006\000\005\r\210O\2303\230n\250E\260\024\260_\300C\300t\310?\320Z[\360\006\000\005\014\2108\2206\230\023\230C\320\037;\2701\270H\300A\330\004\013\2108\2208\2301\230C\230s\240(\250(\260!\2604\260q\340\004\"\240(\250&\260\001\260\021\360\006\000\005\010\200~\220W\230A\330\010\032\230-\240w\250a\340\010'\240}\260A\360\014\000\005\031\230\002\230(\240$\240a\240q\330\004\007\320\007\031\230\023\230A\330\010\022\220!\2208\320\033+\2501\250A\330\n\034\230C\230{\250#\320-?\270s\300!\330\010\022\220&\230\007\230q\240\001\330\010\030\320\030*\250#\250Q\330\r\016\330\026-\250Q\250a\330\022\023\330\022\032\230!\330\022\023\330\022\023\330\022\023\330\022\023\330\022\023\330\022\023\330\022\023\330\022\023\330\022\023\330\022\023\330\022\023\330\022\023\330\022\023\330\014\r\340\010\016\210j\230\001\320\0319\270\021\270!\340\t\n\330\022\"\240!\2401\330\016\017\330\016\026\220a\330\016\017\330\016\017\330\016\017\330\016\017\330\016\017\330\016\017\330\016\017\330\016\017\330\016\017\330\016\017\330\016\017\330\016\017\330\010\t\200\001\340\004\007\200y\220\006\220a\220s\230#\230Q\330\010\t\340\004\013\2109\220F\230#\230S\320 =\270Q\270i\300q\330\004\013\2109\220F\230#\230S\320 =\270Q\270i\300q\330\004\013\2109\220F\230!\2303\230c\240\031\250&\260\001\260\024\3205`\320`a\320aj\320jp\320pq\320qw\320wx\360\000\000y\001B\002\360\000\000B\002H\002\360\000\000H\002I\002\360\000\000I\002J\002\330\004\013\2109\220F\230!\2303\230c\240\031\250&\260\001\260\024\3205a\320ab\320bk\320kq\320qr\320ry\320yz\360\000\000{\001D\002\360\000\000D\002J\002\360\000\000J\002K\002\360\000\000K\002L\002\330\004\014\210I\220X\230Q\230c\240\023\240I\250X\260Q\260e\2701\330\004\014\210I\220X\230Q\230c\240\023\240I\250X\260Q\260e\2701\330\004\013\2109\220G\2303\230i\240x\320/i\320i""j\320js\320s~\320~\177\360\000\000@\002I\002\360\000\000I\002J\002\330\004\013\2103\210a\210}\230C\230y\250\006\250a\250t\3203T\320TU\320UX\320XY\320Yj\320jk\320kt\320tz\320z{\320{|\330\004\013\2103\210a\210}\230C\230s\240.\260\001\260\023\260A\3205d\320de\320en\320nt\320tu\320uv\330\004\013\2103\210a\210}\230B\230i\240v\250Q\250d\260.\300\001\300\023\300A\320Et\320tu\320u~\360\000\000\177\001E\002\360\000\000E\002F\002\360\000\000F\002G\002\340\004$\240A\330\004!\240\031\250(\260!\2601\330\004!\240\031\250(\260!\2601\330\004\036\230i\240v\250Q\250a\330\004!\240\031\250&\260\001\260\021\330\004!\240\031\250(\260!\2601\330\004!\240\031\250(\260!\2601\340\t\n\330\022\"\240!\2409\250A\330\016\017\330\016\017\330\016\017\330\016\017\330\016\027\220q\330\016\017\330\016\017\330\016\017\330\010\t";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 96; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 32) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 96; i < 102; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 102; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 96;
      for (Py_ssize_t i=0; i<6; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int8_t const cint_constants_1[] = {0,1};
    for (int i = 0; i < 2; i++) {
      numbertab[i] = PyLong_FromLong(cint_constants_1[i - 0]);
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<2; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 4;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 4;
    unsigned int flags : 10;
    unsigned int first_line : 8;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 19};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_obj};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_jollyjack_jollyjack_cython_pyx, __pyx_mstate->__pyx_n_u_is_iterable_of_iterables, __pyx_mstate->__pyx_kp_b_iso88591_Qe_T_1Cq_Jd_jXYY_aab, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {11, 0, 0, 11, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 22};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_source, __pyx_mstate->__pyx_n_u_metadata, __pyx_mstate->__pyx_n_u_tensor, __pyx_mstate->__pyx_n_u_row_group_indices, __pyx_mstate->__pyx_n_u_row_ranges, __pyx_mstate->__pyx_n_u_column_indices, __pyx_mstate->__pyx_n_u_column_names, __pyx_mstate->__pyx_n_u_pre_buffer, __pyx_mstate->__pyx_n_u_use_threads, __pyx_mstate->__pyx_n_u_use_memory_map, __pyx_mstate->__pyx_n_u_cache_options};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_jollyjack_jollyjack_cython_pyx, __pyx_mstate->__pyx_n_u_read_into_torch, __pyx_mstate->__pyx_kp_b_iso88591_jj_S_S_d_d_y_y_P_P_t_t_u_Ja_6, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {11, 0, 0, 11, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 41};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_source, __pyx_mstate->__pyx_n_u_metadata, __pyx_mstate->__pyx_n_u_np_array, __pyx_mstate->__pyx_n_u_row_group_indices, __pyx_mstate->__pyx_n_u_row_ranges, __pyx_mstate->__pyx_n_u_column_indices, __pyx_mstate->__pyx_n_u_column_names, __pyx_mstate->__pyx_n_u_pre_buffer, __pyx_mstate->__pyx_n_u_use_threads, __pyx_mstate->__pyx_n_u_use_memory_map, __pyx_mstate->__pyx_n_u_cache_options};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_jollyjack_jollyjack_cython_pyx, __pyx_mstate->__pyx_n_u_read_into_numpy, __pyx_mstate->__pyx_kp_b_iso88591_x_y_N_N_a_a_r_r_G_G_B_B_C_1_1_x, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 153};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_src_tensor, __pyx_mstate->__pyx_n_u_dst_tensor, __pyx_mstate->__pyx_n_u_row_indices};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_jollyjack_jollyjack_cython_pyx, __pyx_mstate->__pyx_n_u_copy_to_torch_row_major, __pyx_mstate->__pyx_kp_b_iso88591_M_6_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 161};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_src_array, __pyx_mstate->__pyx_n_u_dst_array, __pyx_mstate->__pyx_n_u_row_indices};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_jollyjack_jollyjack_cython_pyx, __pyx_mstate->__pyx_n_u_copy_to_numpy_row_major, __pyx_mstate->__pyx_kp_b_iso88591_y_as_Q_9F_S_Qiq_9F_S_Qiq_9F_3c, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* NumpyImportArray.init */
  /*
   * Cython has automatically inserted a call to _import_array since
   * you didn't include one when you cimported numpy. To disable this
   * add the line
   *   <void>numpy._import_array
   */
  #ifdef NPY_FEATURE_VERSION
  #ifndef NO_IMPORT_ARRAY
  if (unlikely(_import_array() == -1)) {
      PyErr_SetString(PyExc_ImportError, "numpy.core.multiarray failed to import "
      "(auto-generated because you didn't call 'numpy.import_array()' after cimporting numpy; "
      "use '<void>numpy._import_array' to disable if you are certain you don't need it).");
  }
  #endif
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

#line 1 "cython_utility"

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* CIntToDigits (used by CIntToPyUnicode) */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp = PySequence_Repeat(padding, uoffset - prepend_sign);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyUnicode_Concat(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyUnicode_Concat(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* COrdinalToPyUnicode (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value) {
    return value <= 1114111;
}
static PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t ulength, char padding_char) {
    Py_ssize_t padding_length = ulength - 1;
    if (likely((padding_length <= 250) && (value < 0xD800 || value > 0xDFFF))) {
        char chars[256];
        if (value <= 255) {
            memset(chars, padding_char, (size_t) padding_length);
            chars[ulength-1] = (char) value;
            return PyUnicode_DecodeLatin1(chars, ulength, NULL);
        }
        char *cpos = chars + sizeof(chars);
        if (value < 0x800) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xc0 | (value & 0x1f));
        } else if (value < 0x10000) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xe0 | (value & 0x0f));
        } else {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xf0 | (value & 0x07));
        }
        cpos -= padding_length;
        memset(cpos, padding_char, (size_t) padding_length);
        return PyUnicode_DecodeUTF8(cpos, chars + sizeof(chars) - cpos, NULL);
    }
    if (value <= 127 && CYTHON_USE_UNICODE_INTERNALS) {
        const char chars[1] = {(char) value};
        return __Pyx_PyUnicode_BuildFromAscii(ulength, chars, 1, 0, padding_char);
    }
    {
        PyObject *uchar, *padding_uchar, *padding, *result;
        padding_uchar = PyUnicode_FromOrdinal(padding_char);
        if (unlikely(!padding_uchar)) return NULL;
        padding = PySequence_Repeat(padding_uchar, padding_length);
        Py_DECREF(padding_uchar);
        if (unlikely(!padding)) return NULL;
        uchar = PyUnicode_FromOrdinal(value);
        if (unlikely(!uchar)) {
            Py_DECREF(padding);
            return NULL;
        }
        result = PyUnicode_Concat(padding, uchar);
        Py_DECREF(padding);
        Py_DECREF(uchar);
        return result;
    }
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (Py_ssize_t) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_npy_intp(npy_intp value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_intp neg_one = (npy_intp) -1, const_zero = (npy_intp) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (npy_intp) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_npy_intp(npy_intp value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(npy_intp)*3+2];
    char *dpos, *end = digits + sizeof(npy_intp)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    npy_intp remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_intp neg_one = (npy_intp) -1, const_zero = (npy_intp) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (npy_intp) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (npy_intp) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (npy_intp) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
        Py_INCREF(values[i]);
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* IterFinish (used by dict_iter) */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by dict_iter) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* RaiseNeedMoreValuesToUnpack (used by UnpackTuple2) */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack (used by UnpackItemEndCheck) */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck (used by UnpackTuple2) */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError (used by UnpackTupleError) */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError (used by UnpackTuple2) */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else {
      Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(t);
 #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely(size < 0)) return;
 #endif
      if (size < index) {
        __Pyx_RaiseNeedMoreValuesError(size);
      } else {
        __Pyx_RaiseTooManyValuesError(index);
      }
    }
}

/* UnpackTuple2 (used by dict_iter) */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
        PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple) {
    if (likely(is_tuple || PyTuple_Check(tuple))) {
        Py_ssize_t size;
        if (has_known_size) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        size = __Pyx_PyTuple_GET_SIZE(tuple);
        if (likely(size == 2)) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        if (size >= 0) {
            __Pyx_UnpackTupleError(tuple, 2);
        }
        return -1;
    } else {
        return __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple);
    }
}
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
    value1 = __Pyx_PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = __Pyx_PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#else
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
#if !CYTHON_AVOID_BORROWED_REFS
static CYTHON_INLINE int __Pyx_dict_iter_next_source_is_dict(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem) {
    PyObject *key, *value;
    if (unlikely(orig_length != PyDict_Size(iter_obj))) {
        PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
        return -1;
    }
    if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
        return 0;
    }
    if (pitem) {
        PyObject* tuple = PyTuple_New(2);
        if (unlikely(!tuple)) {
            return -1;
        }
        Py_INCREF(key);
        Py_INCREF(value);
        #if CYTHON_ASSUME_SAFE_MACROS
        PyTuple_SET_ITEM(tuple, 0, key);
        PyTuple_SET_ITEM(tuple, 1, value);
        #else
        if (unlikely(PyTuple_SetItem(tuple, 0, key) < 0)) {
            Py_DECREF(value);
            Py_DECREF(tuple);
            return -1;
        }
        if (unlikely(PyTuple_SetItem(tuple, 1, value) < 0)) {
            Py_DECREF(tuple);
            return -1;
        }
        #endif
        *pitem = tuple;
    } else {
        if (pkey) {
            Py_INCREF(key);
            *pkey = key;
        }
        if (pvalue) {
            Py_INCREF(value);
            *pvalue = value;
        }
    }
    return 1;
}
#endif
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_AVOID_BORROWED_REFS
    if (source_is_dict) {
        int result;
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_BEGIN_CRITICAL_SECTION(iter_obj);
#endif
        result = __Pyx_dict_iter_next_source_is_dict(iter_obj, orig_length, ppos, pkey, pvalue, pitem);
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_END_CRITICAL_SECTION();
#endif
        return result;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(tuple_size < 0)) return -1;
        #endif
        if (unlikely(pos >= tuple_size)) return 0;
        *ppos = pos + 1;
        #if CYTHON_ASSUME_SAFE_MACROS
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        #else
        next_item = PyTuple_GetItem(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        #endif
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t list_size = __Pyx_PyList_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(list_size < 0)) return -1;
        #endif
        if (unlikely(pos >= list_size)) return 0;
        *ppos = pos + 1;
        next_item = __Pyx_PyList_GetItemRef(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (int) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* PxdImportShared (used by FunctionImport) */
#ifndef __PYX_HAVE_RT_ImportFromPxd_3_2_4
#define __PYX_HAVE_RT_ImportFromPxd_3_2_4
static int __Pyx_ImportFromPxd_3_2_4(PyObject *module, const char *name, void **p, const char *sig, const char *what) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, "__pyx_capi__");
    if (!d)
        goto bad;
#if (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030d0000) || (!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030d0000)
    PyDict_GetItemStringRef(d, name, &cobj);
#else
    cobj = PyDict_GetItemString(d, name);
    Py_XINCREF(cobj);
#endif
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C %.8s %.200s",
                PyModule_GetName(module), what, name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C %.8s %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             what, PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(d);
    Py_XDECREF(cobj);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_3_2_4
#define __PYX_HAVE_RT_ImportFunction_3_2_4
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    int result = __Pyx_ImportFromPxd_3_2_4(module, funcname, &tmp.p, sig, "function");
    if (result == 0) {
        *f = tmp.fp;
    }
    return result;
}
#endif

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__10);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType (used by FetchCommonType) */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* CallTypeTraverse (used by CythonFunctionShared) */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return ::std::complex< long double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return x + y*(__pyx_t_long_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      __pyx_t_long_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsl(b.real) >= fabsl(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                long double r = b.imag / b.real;
                long double s = (long double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_long_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            long double r = b.real / b.imag;
            long double s = (long double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_long_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            long double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_long_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtl(z.real*z.real + z.imag*z.imag);
          #else
            return hypotl(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
            __pyx_t_long_double_complex z;
            long double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    long double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_long__double(a, a);
                    case 3:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, a);
                    case 4:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powl(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2l(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_long__double(a);
                theta = atan2l(a.imag, a.real);
            }
            lnr = logl(r);
            z_r = expl(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosl(z_theta);
            z.imag = z_r * sinl(z_theta);
            return z;
        }
    #endif
#endif

/* PyObjectVectorCallKwBuilder (used by CIntToPy) */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyLong_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_npy_intp(npy_intp value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_intp neg_one = (npy_intp) -1, const_zero = (npy_intp) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_intp) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(npy_intp) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(npy_intp) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_intp) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(npy_intp) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(npy_intp),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(npy_intp));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__11);
    }
    goto done;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
